(***********************************************************************
 * File: Actions.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $Author: jdcampo $
 *******************************************************************
 * This file specifies the actions (state transformers)
 * 
 ******************************************************************)

Require Import Addressing OS.

Inductive action : Set :=
(** action not modifying observable state *)
| Silent :  action            
(** OS read action *)
| Read   :  vadd -> action                
(** Hypervisor read action *)
| Read_hyper : vadd -> action             
(** OS write action *)
| Write  :  vadd -> value -> action       
(** Hypervisor write action *)
| Write_hyper  :  vadd -> value -> action 
(** new mapping for untrusted OS*)
| New_untrusted :  os_ident -> vadd -> padd  -> action
(** Hypervisor new mapping action *)
| New_hyper   :  vadd -> madd -> action
(** delete mapping for untrusted OS *)
| Del_untrusted  :  os_ident -> vadd -> action
(** delete mapping for trusted OS *)
(* | Del_trusted  :  vadd -> action           *)
(** Hypervisor delete own mapping *)
| Del_hyper : vadd -> action
(** Hypervisor switch between 2 OS *)
| Switch : os_ident -> action
(** Local switch within an OS *)
(* | Lswitch_trusted : padd -> action          *)
(** Local switch within an OS *)
| Lswitch_untrusted : os_ident -> padd -> action
(** hypercall *)
| Hcall : Hyperv_call -> action
(** return control from OS to Hypervisor *)
| Ret_ctrl : action
(** used by the Hypervisor to give control back to an Untrusted OS after an hypercall *)
| Chmod : action
(** classify a page, register it*)
(* | Page_pin_trusted : padd -> ptype -> action  *)
(** un-classify a page, deregister it*)
(* | Page_unpin_trusted : padd -> action   *)
(** classify a page, register it*)
| Page_pin_untrusted : os_ident -> padd -> ptype -> action
(** un-classify a page, deregister it*)
| Page_unpin_untrusted : os_ident -> padd -> action
.

