(***********************************************************************
 * File: Dict_bound.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies a bounded dictionary
 * 
 *************************************************************************)
Module Type DICT_BOUND.

Parameter dict : Set.

Parameter key : Set.
Parameter info : Set.
Parameter eq_key : forall (k k':key), {k = k'} + {k <> k'}.

Parameter empty : dict.
Parameter set : dict -> key -> info -> dict*option (key*info).
Parameter get : dict -> key -> option info.
Parameter is_empty : dict -> Prop.
Parameter is_full : dict -> Prop.
Parameter remove : dict -> key -> dict.
Parameter filter : dict -> (key -> bool) -> dict*option (key*info).

(* Dict behavior *)
Axiom is_empty_empty : is_empty empty.
Axiom not_empty_set : forall (d:dict) (k:key) (inf:info),
  ~ is_empty (fst (set d k inf)).

Axiom get_empty : forall (k:key), get empty k = None.
Axiom get_set : forall (d:dict) (k:key) (inf:info),
  Some inf = get (fst (set d k inf)) k.
Axiom get_set_distinct : forall (d:dict) (k k':key) (inf inf':info),
  let (d',entry) := set d k inf in
    k <> k' -> (entry = None \/ (exists k'', exists inf'', entry = Some (k'',inf'') /\ k' <> k'')) ->
    Some inf' = get d k' ->
    Some inf' = get d' k'.

Axiom get_set_distinct_not_new : forall (d:dict) (k k':key) (inf inf':info),
  let (d',entry) := set d k inf in
    k <> k' -> 
    Some inf' = get d' k' ->
    Some inf' = get d k'.

Axiom is_full_dec : forall (d:dict), {is_full d}+{~(is_full d)}.

Axiom set_entry_behavior_not_full : forall (d:dict) (k:key) (inf:info),
  let (d',entry) := set d k inf in
    ~ is_full d -> entry = None.

Axiom set_entry_behavior_update : forall (d:dict) (k:key) (inf:info),
  let (d',entry) := set d k inf in
    get d k <> None -> entry = None.

Axiom set_entry_behavior_full : forall (d:dict) (k:key) (inf:info),
  let (d',entry) := set d k inf in
    is_full d -> get d k = None ->
      exists k', exists inf',
        entry = Some (k',inf') /\
        get d k' = Some inf' /\
        get d' k' = None.


Axiom get_remove : forall (d:dict) (k:key),
  get (remove d k) k = None.

Axiom get_remove_distinct : forall (d:dict) (k k':key) (inf:info),
  k <> k' -> get d k = Some inf -> get (remove d k') k = Some inf.

Axiom get_remove_old : forall (d:dict) (k k':key) (inf:info),
  get (remove d k') k = Some inf -> get d k = Some inf.

Axiom get_filter_p : forall d p k,
  p k = true ->
  get (fst (filter d p)) k = None.

Axiom get_filter_notp : forall d p k,
  p k = false ->
  get (fst (filter d p)) k = get d k.

Axiom get_filter_some_entry : forall d p k i,
  snd (filter d p) = Some (k,i) ->
  get d k = Some i.

Axiom get_filter_some_entry_p : forall d p k i,
  snd (filter d p) = Some (k,i) -> p k = true.

Axiom get_filter_no_entry : forall d p k i,
  snd (filter d p) = None ->
  get d k = Some i ->
  p k = false.


End DICT_BOUND.





