Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_NONCACHEABLE (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.

Lemma read_valid_syn :
  forall s pt_addr va pg,
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    aos_page_of_vadd_smem s va = Some pg ->
    is_rw_page pg ->
    non_cacheable_synonyms (st_hypervisor s) (smem_mm (page_smem_smem (get (st_memory s) va pt_addr))).
Proof. 
  intros.
  STAC.destruct_valid_state.
  unfold non_cacheable_synonyms; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  assert (madd_type (smem_mm (st_memory s)) pt_addr0 = Some PT).
  rewrite STAC.read_same_madd_type with (pt_addr':=pt_addr) (va':=va) (ma':=m) (pg':=pg); crush.
  unfold madd_type; crush.
  rewrite H7; crush.
  rewrite <- STAC.read_no_pt_change with (ma':=m) (pg':=pg) in H6; crush.
  assert (madd_type (smem_mm (st_memory s)) pt_addr' = Some PT).
  rewrite STAC.read_same_madd_type with (pt_addr':=pt_addr) (va':=va) (ma':=m) (pg':=pg); crush.
  unfold madd_type; crush.
  rewrite H10; crush.
  rewrite <- STAC.read_no_pt_change with (ma':=m) (pg':=pg)in H9; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr0 va0 = Some ma);
    try unfold madd_of_vadd, madd_page_table; crush.
  rewrite H7; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma);
    try unfold madd_of_vadd, madd_page_table; crush.
  rewrite H10; crush.

  destruct (snd(add_cache (smem_c (st_memory s)) (va,m) pg)) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush.
  symmetry in H7, H10.
  destruct (vadd_eq va' v); crush;
    STAC.synonym_cached.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=pg) (pg':=x6) in H12; crush.
  apply (H_syn pt_addr0 pt_page mp va0 ma pg0 pt_addr' pt_page' mp' va'); crush.
  rewrite get_mem_unchanged with (ma:=m) (pg:=pg) in H12; crush.
  apply (H_syn pt_addr0 pt_page mp va0 ma pg0 pt_addr' pt_page' mp' va'); crush.
Qed.  

Lemma write_valid_syn :
  forall s pt_addr va pg val,
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    aos_page_of_vadd_smem s va = Some pg ->
    is_rw_page pg ->
    non_cacheable_synonyms (st_hypervisor s) (smem_mm (set (st_memory s) va pt_addr val)).
Proof. 
  intros.
  STAC.destruct_valid_state.
  unfold non_cacheable_synonyms; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  assert (madd_type (smem_mm (st_memory s)) pt_addr0 = Some PT).
  rewrite <- STAC.write_same_madd_type with (pt_addr:=pt_addr) (va:=va) (val:=val); crush.
  unfold madd_type; crush.
  rewrite H7; crush.
  rewrite <- STAC.write_no_pt_change with (pg':=pg) in H6; 
    unfold aos_page_of_vadd_smem, page_of_vadd_smem; crush.
  assert (madd_type (smem_mm (st_memory s)) pt_addr' = Some PT).
  rewrite <- STAC.write_same_madd_type with (pt_addr:=pt_addr) (va:=va) (val:=val); crush.
  unfold madd_type; crush.
  rewrite H10; crush.
  rewrite <- STAC.write_no_pt_change with (pg':=pg) in H9; 
    unfold aos_page_of_vadd_smem, page_of_vadd_smem; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr0 va0 = Some ma);
    try unfold madd_of_vadd, madd_page_table; crush.
  rewrite H7; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma);
    try unfold madd_of_vadd, madd_page_table; crush.
  rewrite H10; crush.

  destruct (snd (add_cache (smem_c (st_memory s)) (va,m) 
     (Page (RW_c (Some val))
           (page_owned_by pg)
           (page_cacheable pg)))) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush.
  symmetry in H7, H10.
  destruct (vadd_eq va' v); crush;
    STAC.synonym_cached.
  rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0) 
    (pg:=Page (RW_c (Some val))
           (page_owned_by pg)
           (page_cacheable pg)) 
    (pg':=x6) in H12; crush.
  apply (H_syn pt_addr0 pt_page mp va0 ma pg0 pt_addr' pt_page' mp' va'); crush.
  assert (is_cacheable (Page (RW_c (Some val))
                             (page_owned_by pg)
                             (page_cacheable pg))); crush.
  STAC.some_entry_add_cache.
  apply (H_syn pt_addr0 pt_page mp va0 m pg pt_addr' pt_page' mp' va'); crush.
  unfold page_of_vadd_madd_smem in *; crush.
  symmetry in Heqval0.
  symmetry in H7, H10.
  destruct (vadd_eq va va'); crush.
  STAC.synonym_cached.
  assert (va' <> va); crush.
  STAC.synonym_cached.
  
  destruct (madd_eq m ma); crush.
  destruct set_new_mmem with (st_memory s) pt_addr va ma val pg; crush.
  unfold page_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va ma pg; crush.
  apply (H_syn pt_addr0 pt_page mp va0 ma x5 pt_addr' pt_page' mp' va'); crush.
  apply (H_syn pt_addr0 pt_page mp va0 ma pg pt_addr' pt_page' mp' va'); crush.
  rewrite set_mem_unchanged2 with (ma:=m) (pg:=Page (RW_c (Some val))
                                                    (page_owned_by pg)
                                                    (page_cacheable pg)) in H12; crush.
  apply (H_syn pt_addr0 pt_page mp va0 ma pg0 pt_addr' pt_page' mp' va'); crush.
Qed.

Lemma new_non_cacheable_synonyms : 
  forall s osi' pt_addr' va' ma',
    valid_state s ->
    madd_type (smem_mm (st_memory s)) ma' <> None ->
    osi_current_page_table s osi' = Some pt_addr' ->
    non_cacheable_synonyms (st_hypervisor s) (smem_mm (ACT.new s osi' pt_addr' va' ma')).
Proof.
  Local Opaque ACT.new.
  unfold non_cacheable_synonyms; crush.
  STAC.destruct_valid_state.

  assert (madd_of_vadd (smem_mm (ACT.new s osi' pt_addr' va' ma')) pt_addr va =
    Some ma);
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H6; crush.
  assert (madd_of_vadd (smem_mm (ACT.new s osi' pt_addr' va' ma')) pt_addr'0 va'0 =
    Some ma);
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H9; crush.

  assert (va' <> va \/ pt_addr <> pt_addr' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); intros;
  [|assert (va' <> va'0 \/ pt_addr'0 <> pt_addr' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr'0 va'0 = Some ma)]; intros.

  rewrite STAC.new_same_madd_of_vadd in H15; crush;
    unfold madd_type; crush.
  rewrite H6; crush.
  rewrite STAC.new_same_madd_of_vadd in H16; crush;
    unfold madd_type; crush.
  rewrite H9; crush.

  destruct STAC.new_get_mem_complete_semantics with s osi' pt_addr' va' ma' ma
    as [?|[?|[?|[?|?]]]]; crush; unfold is_cacheable in *; crush.

  destruct (vadd_eq va' va); crush;
    [assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr'0 va'0 = Some ma) |
     assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma)]; crush;
    try (assert (va'0 <> va); [crush; fail | STAC.synonym_cached]);
    try (assert (va <> va'); [crush; fail | STAC.synonym_cached]).
  destruct (vadd_eq va' va); crush;
    [assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr'0 va'0 = Some ma') |
     assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma')]; crush;
    try (assert (va'0 <> va); [crush; fail | STAC.synonym_cached]);
    try (assert (va <> va'); [crush; fail | STAC.synonym_cached]).
  destruct H_valid_pt_non_cacheable with pt_addr' x5;
    unfold is_pt_page; crush.

  destruct (madd_eq ma' ma); crush.
  destruct H20; crush.
  apply H20; unfold shared_page; crush.
  destruct (vadd_eq va' va); crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr'0 va'0 = Some ma); crush.
  unfold madd_of_vadd, madd_page_table in H23; crush.
  destruct (page_content p0) eqn:?; crush.
  exists x1; exists x0; exists p0; exists p1; exists va'0;
    unfold osi_pa_to_pg; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  unfold madd_of_vadd, madd_page_table in H23; crush.
  destruct (page_content p0) eqn:?; crush.
  exists x4; exists x3; exists p0; exists p1; exists va;
    unfold osi_pa_to_pg; crush.

  rewrite H22 in *; clear H22.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma') eqn:?; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma).
  destruct (vadd_eq va' va); crush.
  destruct (madd_eq pt_addr' pt_addr); crush.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  rewrite STAC.new_get_mem_pt_addr
    with (pt_page:=pt_page0) (pt:=pt) in H5; crush.
  injection H6; intros; subst mp; clear H6.
  rewrite <- PT.get_set in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr'0 va'0 = Some ma).
  destruct (vadd_eq va' va'0); crush.
  destruct (madd_eq pt_addr' pt_addr'0); crush.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  unfold osi_current_page_table in *; crush.
  rewrite STAC.new_get_mem_pt_addr
    with (pt_page:=pt_page0) (pt:=pt) in H8; crush.
  injection H9; intros; subst mp'; clear H9.
  rewrite <- PT.get_set in *; crush.

  unfold madd_of_vadd, madd_page_table in H22, H23; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (page_content p3) eqn:?; crush.
  symmetry in Heqc, H11.
  STAC.syn_non_cacheable.  
Qed.  
  

Lemma non_cachable_synonyms_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_state s ->
  non_cacheable_synonyms (st_hypervisor s') (smem_mm (st_memory s')).
Proof.
  Local Opaque ACT.new new_page.
  Local Hint Resolve read_valid_syn write_valid_syn
    new_non_cacheable_synonyms.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; crush.

  (* Delete *)
  unfold non_cacheable_synonyms; crush.
  assert (madd_of_vadd (smem_mm (delete (st_memory s) x1 v)) pt_addr va =
    Some ma); crush.
  unfold madd_of_vadd, madd_page_table; crush;
    rewrite H11; crush.
  assert (madd_of_vadd (smem_mm (delete (st_memory s) x1 v)) pt_addr' va' =
    Some ma); crush.
  unfold madd_of_vadd, madd_page_table; crush;
    rewrite H14; crush.
  assert (ACT.pre_del o v s).
  unfold ACT.pre_del; crush;
    unfold ACT.pre_all_del; crush.

  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  apply STAC.del_madd_of_vadd_old with
    (s':=(State (st_active_os s)
      (OSS.set (st_oss s) o
        (OS None (os_current_page x0)))
      (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
      (ACT.delete_smem o x1 v s)))
    (osi':=o) (va':=v); crush.
  unfold ACT.post_del; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma); crush.
  apply STAC.del_madd_of_vadd_old with
    (s':=(State (st_active_os s)
      (OSS.set (st_oss s) o
        (OS None (os_current_page x0)))
      (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
      (ACT.delete_smem o x1 v s)))
    (osi':=o) (va':=v); crush.
  unfold ACT.post_del; crush.
  unfold madd_of_vadd, madd_page_table in H23, H24; crush.
  destruct (page_content p1) eqn:?; crush.
  destruct (page_content p2) eqn:?; crush.
  symmetry in Heqc.
  assert (exists pg, Some pg = get_mem (smem_mm (st_memory s)) ma); crush.
  destruct H_valid_virtmap with (osi:=x6) (pt_addr:=pt_addr) (pt_page:=p2)
    (pt:=p) (va:=va) (ma:=ma); crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO x6); crush.
  unfold madd_owner in *; crush.
  STAC.syn_non_cacheable.    

  edestruct STAC.del_get_mem_same_page_but_value; crush.
  unfold ACT.post_del; crush.
  unfold same_page_but_value in *; crush.

  (* Del hyper *)
  unfold non_cacheable_synonyms; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x0 v) eqn:?; crush.
  assert (madd_of_vadd m pt_addr va = Some ma).
  destruct (get_cache (smem_c (st_memory s)) (v,x1)) eqn:?; crush.
  rewrite delete_mem_some_entry with (mm':=m) (ma:=x1) (pg:=p) in H7; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  destruct (madd_eq x1 pt_addr); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H7]; crush.
  destruct H_valid_cache with v pt_addr pt_page; crush;
    unfold is_rw_page in *; crush.
  rewrite H8; crush.
  rewrite delete_mem_no_entry with (mm':=m) (ma:=x1) in H7; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H8; crush.
  assert (madd_of_vadd m pt_addr' va' = Some ma).
  destruct (get_cache (smem_c (st_memory s)) (v,x1)) eqn:?; crush.
  rewrite delete_mem_some_entry with (mm':=m) (ma:=x1) (pg:=p) in H10; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  destruct (madd_eq x1 pt_addr'); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H10]; crush.
  destruct H_valid_cache with v pt_addr' pt_page'; crush;
    unfold is_rw_page in *; crush.
  rewrite H11; crush.
  rewrite delete_mem_no_entry with (mm':=m) (ma:=x1) in H10; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H11; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma).
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  clear H18.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  destruct (madd_eq x0 pt_addr); crush.
  rewrite Heqc; crush.
  rewrite <- get_add_mem in *; crush.
  destruct (vadd_eq v va); crush.
  rewrite PT.get_remove in *; crush.
  apply PT.get_remove_old with (k':=v); crush.
  rewrite get_add_mem_distinct in Heqopt_res2; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va' = Some ma).
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  clear H17.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  destruct (madd_eq x0 pt_addr'); crush.
  rewrite Heqc; crush.
  rewrite <- get_add_mem in *; crush.
  destruct (vadd_eq v va'); crush.
  rewrite PT.get_remove in *; crush.
  apply PT.get_remove_old with (k':=v); crush.
  rewrite get_add_mem_distinct in Heqopt_res3; crush.
  clear H17 H18.

  assert (get_mem m ma = Some pg); crush.
  destruct (get_cache (smem_c (st_memory s)) (v,x1)) eqn:?; crush.
  rewrite delete_mem_some_entry with (mm':=m) (ma:=x1) (pg:=p) in H13; crush.
  destruct (madd_eq x1 ma); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H13]; crush.
  destruct (vadd_eq v va); crush.
  assert (va' <> va); crush.
  STAC.synonym_cached.
  assert (va <> v); crush.
  STAC.synonym_cached.
  rewrite delete_mem_no_entry with (mm':=m) (ma:=x1) in H13; crush.
  assert (get_mem (smem_mm (st_memory s)) ma = Some pg); crush.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  destruct (madd_eq x0 ma); crush.
  rewrite <- get_add_mem in *; crush.
  unfold is_cacheable in *; crush.
  destruct H_valid_pt_non_cacheable with ma p; crush.
  unfold is_pt_page; crush.
  rewrite get_add_mem_distinct in H17; crush.
  symmetry in H18.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p2) eqn:?; crush.
  destruct (page_content p3) eqn:?; crush.
  symmetry in Heqc.
  STAC.syn_non_cacheable.
  unfold remove_page_table_entry in *; crush.

  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold osi_current_page_table, madd_of_vadd, madd_page_table in *; crush.
  rewrite H20 in *; crush.

  (* Switch *)
  rewrite ctxt_switch_mem in *; crush.

  (* Lswitch *)
  rewrite ctxt_switch_mem in *; crush.

  (* Page_pin untrusted *)
  unfold non_cacheable_synonyms; crush.
  destruct (madd_eq x pt_addr); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H17]; crush.
  Local Transparent new_page.
  crush.
  injection H18; intros; subst mp.
  rewrite PT.get_empty in *; crush.
  Local Opaque new_page.
  destruct (madd_eq x pt_addr'); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H20]; crush.
  Local Transparent new_page.
  crush.
  injection H21; intros; subst mp'.
  rewrite PT.get_empty in *; crush.
  Local Opaque new_page.
  assert (osi_pa_to_ma s x9 x10 = Some pt_addr); crush.
  unfold osi_pa_to_ma; crush.
  destruct (os_ident_eq o x9); crush.
  rewrite <- HYP.get_set in *; crush.    
  destruct (padd_eq p x10); crush.
  rewrite <- P2M.get_set in *; crush.
  rewrite P2M.get_set_distinct in H26; crush.
  rewrite HYP.get_set_distinct in H15; crush.
  assert (osi_pa_to_ma s x6 x7 = Some pt_addr'); crush.
  unfold osi_pa_to_ma; crush.
  destruct (os_ident_eq o x6); crush.
  rewrite <- HYP.get_set in *; crush.    
  destruct (padd_eq p x7); crush.
  rewrite <- P2M.get_set in *; crush.
  rewrite P2M.get_set_distinct in H25; crush.
  rewrite HYP.get_set_distinct in H16; crush.
  destruct (madd_eq x ma); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H23]; unfold is_cacheable in *; crush.
  symmetry in H4, H17, H18, H20.
  STAC.syn_non_cacheable.  
  symmetry in H23, H17, H18, H20.
  STAC.syn_non_cacheable.  

  (* Page unpin *)
  unfold non_cacheable_synonyms; crush.
  destruct (madd_eq ma x0); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H20]; unfold is_cacheable in *; crush.
  Local Transparent new_page.
  crush.  
  Local Opaque new_page.
  destruct (madd_eq x0 pt_addr); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H14]; crush.
  Local Transparent new_page.
  crush.
  Local Opaque new_page.
  destruct (madd_eq x0 pt_addr'); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H17]; crush.
  Local Transparent new_page.
  crush.
  Local Opaque new_page.
  assert (osi_pa_to_ma s x7 x8 = Some pt_addr); crush.
  unfold osi_pa_to_ma; crush.
  destruct (os_ident_eq o x7); crush.
  rewrite <- HYP.get_set in *; crush.    
  destruct (padd_eq p x8); crush.
  rewrite P2M.get_remove in *; crush.
  apply P2M.get_remove_old with (k':=p); crush.
  rewrite HYP.get_set_distinct in H12; crush.
  assert (osi_pa_to_ma s x4 x5 = Some pt_addr'); crush.
  unfold osi_pa_to_ma; crush.
  destruct (os_ident_eq o x4); crush.
  rewrite <- HYP.get_set in *; crush.    
  destruct (padd_eq p x5); crush.
  rewrite P2M.get_remove in *; crush.
  apply P2M.get_remove_old with (k':=p); crush.
  rewrite HYP.get_set_distinct in H13; crush.
  symmetry in H14, H15, H17, H20.
  STAC.syn_non_cacheable.  
Qed.

End EXEC_INV_NONCACHEABLE.
