Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VALID_PT_NON_CACHEABLE (C:CACHE PG)
       (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
       (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
       (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM. 
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.

Lemma read_pt_non_cacheable : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_pt_non_cacheable (smem_mm (page_smem_smem
      (get (st_memory s) va pt_addr))).
Proof.
  unfold page_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  unfold valid_pt_non_cacheable, is_cacheable; crush.
  erewrite <- STAC.read_no_pt_change in H3; crush.
  destruct H_valid_pt_non_cacheable with pt_addr0 pt_page; crush.
  erewrite STAC.read_same_madd_type; crush.
  unfold madd_type, is_pt_page in *; crush.
  rewrite H4; crush.
Qed.

Lemma write_pt_non_cacheable : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page) (val:value),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_pt_non_cacheable (smem_mm (set (st_memory s) va pt_addr val)).
Proof.
  unfold page_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  unfold valid_pt_non_cacheable, is_cacheable; crush.
  erewrite <- STAC.write_no_pt_change in H3; crush.
  destruct H_valid_pt_non_cacheable with pt_addr0 pt_page; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem; crush.
  rewrite <- STAC.write_same_madd_type with (pt_addr:=pt_addr) (va:=va) 
    (val:=val); crush.
  unfold madd_type, is_pt_page in *; crush.
  rewrite H4; crush.
Qed.

Lemma new_pt_non_cacheable : 
  forall (s:state) (osi:os_ident) (pt_addr:madd)
    (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_type (smem_mm (st_memory s)) ma <> None ->

    ((os_accessible va /\ exists pa, osi_pa_to_ma s osi pa = Some ma) \/
     (~ os_accessible va /\ madd_owner (smem_mm (st_memory s)) ma = Hyp_PO)) ->
      
    valid_pt_non_cacheable (smem_mm (ACT.new s osi pt_addr va ma)).
Proof.
  Local Opaque ACT.new.
  intros.
  STAC.destruct_valid_state.
  unfold valid_pt_non_cacheable, is_pt_page; crush.
  destruct STAC.new_get_mem_complete_semantics 
    with s osi pt_addr va ma pt_addr0 as [? | [? | [? | [? | ?]]]]; crush;
      unfold page_of_vadd_madd_smem, is_cacheable in *; crush;
        repeat
          match goal with
            | [H: get_cache (smem_c (st_memory s)) (?va, ?ma) = Some ?pg|-_]=>
              symmetry in H
            |[H1: Some ?pg = get_mem (smem_mm (st_memory ?s)) ?ma,
              H2: Some ?pg' = get_cache (smem_c (st_memory ?s)) (?va,?ma),
              H3: page_owned_by ?pg = Os_PO ?osi |-
                page_owned_by ?pg' = Os_PO ?osi] =>
            destruct H_valid_cache with va ma pg'; crush; fail
            | [H: PT_c ?pt = page_content ?pg|-_]=> symmetry in H
            | [H1: Some ?pg = get_cache (smem_c (st_memory ?s)) (?va,?ma),
               H2: Some ?pg' = get_mem (smem_mm (st_memory ?s)) ?ma,
               H3: page_content ?pg' = PT_c ?pt,
               H4: valid_cache ?s |-_] =>
              destruct H4 with va ma pg; crush; 
                unfold is_rw_page in *; crush
            | [H1: Some ?pg = get_cache (smem_c (st_memory ?s)) (?va,?ma),
               H3: page_content ?pg = PT_c ?pt,
               H4: valid_cache ?s |-_] =>
              destruct H4 with va ma pg; crush; 
                unfold is_rw_page in *; crush
            | [H: Some ?pg = get_mem ?mem ?ma |-_] => symmetry in H
            | [H1: get_mem ?mem ?ma = Some ?pg,
               H2: Some ?ma = P2M.get ?p2m ?pt_padd
               |- ex (fun pt_addr' => ex (fun pt_page' => 
                  Some pt_addr' = P2M.get ?p2m ?pt_padd /\ 
                  Some pt_page' = get_mem ?mem pt_addr' /\ _))] =>
              exists ma, pg; unfold is_pt_page; crush
          end.
  destruct H_valid_pt_non_cacheable with ma pt_page; 
    unfold is_pt_page; crush.
  destruct H_valid_pt_non_cacheable with pt_addr x0; 
    unfold is_pt_page; crush.
  rewrite H9 in H3; crush.
  destruct H_valid_pt_non_cacheable with pt_addr0 pt_page; 
    unfold is_pt_page; crush.
Qed.        


Lemma valid_pt_non_cacheable_inv :forall (s s' : state) (a : action),
    ACT.valid_exec s a s' ->
    valid_state s ->
    valid_pt_non_cacheable (smem_mm (st_memory s')).
Proof. 
  Local Opaque new_page.
  Local Hint Resolve read_pt_non_cacheable write_pt_non_cacheable
        new_pt_non_cacheable.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; crush; 
    unfold valid_pt_non_cacheable in *; crush.

  (* Delete *)
  destruct H_valid_os_idents with o; crush.
  STAC.get_osi_current_pt_page o; crush.
  generalize H4; unfold osi_current_page_table; crush.
  assert (madd_type (smem_mm (st_memory s)) pt_addr = Some PT).
  erewrite <- STAC.del_same_madd_type with (s':=
    (State (st_active_os s) 
           (OSS.set (st_oss s) o (OS None (os_current_page x0)))
           (st_exec_mode s) (st_aos_activity s)
           (st_hypervisor s) (ACT.delete_smem o x1 v s))); 
    crush.
  unfold madd_type, is_pt_page in *; crush.
  rewrite H8; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  unfold madd_type in H11; crush.
  destruct (page_content p) eqn:?; crush.
  edestruct STAC.del_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  crush.
  destruct H16; crush.
  destruct H_valid_pt_non_cacheable with pt_addr p; 
    unfold is_pt_page, is_cacheable in *; crush.

  (* Del hyper *)
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  generalize H; unfold osi_current_page_table; crush.
  assert (madd_type (smem_mm (st_memory s)) pt_addr = Some PT).
  rewrite <- STAC.del_hyper_same_madd_type with (pt_addr':=x0) (va':=v); 
    crush.
  unfold madd_type, is_pt_page in *; crush.
  rewrite H5; crush.
  unfold madd_type in H7; crush.
  destruct (page_content p) eqn:?; crush.
  edestruct STAC.del_hyper_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  unfold ACT.post_del_hyper; crush.
  crush.
  destruct H12; crush.
  unfold is_cacheable in *; crush.
  destruct H_valid_pt_non_cacheable with pt_addr p; 
    unfold is_pt_page; crush.

  (* Switch *)
  rewrite ctxt_switch_mem in *; crush.
  destruct H_valid_pt_non_cacheable with pt_addr pt_page; crush.

  (* LSwitch *)
  rewrite ctxt_switch_mem in *; crush.
  destruct H_valid_pt_non_cacheable with pt_addr pt_page; crush.

  (* Page pin *)
  destruct (madd_eq x pt_addr); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in *]; crush.
  Local Transparent new_page.
  unfold is_pt_page, is_cacheable in *; crush.
  Local Opaque new_page.
  destruct H_valid_pt_non_cacheable with pt_addr pt_page; crush.

  (* Page unpin *)
  destruct (madd_eq x0 pt_addr); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in *]; crush.
  Local Transparent new_page.
  unfold is_pt_page, is_cacheable in *; crush.
  Local Opaque new_page.
  destruct H_valid_pt_non_cacheable with pt_addr pt_page; crush.
Qed.

End EXEC_INV_VALID_PT_NON_CACHEABLE.