Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VA_HAS_PA (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.

Lemma read_va_has_pa : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    va_has_valid_pa 
      (smem_mm (page_smem_smem (get (st_memory s) va pt_addr))) 
      (st_hypervisor s).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  STAC.destruct_valid_state.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some m).
  unfold madd_of_vadd_smem in *; crush.

  unfold va_has_valid_pa; crush.
  rewrite <- STAC.read_no_pt_change with (ma':=m) (pg':=pg) in H6; 
    unfold is_rw_page; crush.
  assert (madd_type (smem_mm (page_smem_smem (get (st_memory s) va pt_addr)))
            pt_addr0 = Some PT).
  unfold madd_type; crush; rewrite <- H7; crush.
  rewrite <- H10.
  apply STAC.read_same_madd_type with (ma':=m) (pg':=pg); 
    unfold is_rw_page; crush.
Qed.

Lemma write_va_has_pa :
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page) (val:value),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    va_has_valid_pa (smem_mm (set (st_memory s) va pt_addr val)) 
      (st_hypervisor s).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  STAC.destruct_valid_state.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some m).
  unfold madd_of_vadd_smem in *; crush.

  unfold va_has_valid_pa; intros.
  rewrite <- STAC.write_no_pt_change with (pg':=pg) in H6; 
    unfold is_rw_page, aos_page_of_vadd_smem, page_of_vadd_smem; crush.
  assert (madd_type (smem_mm (set (st_memory s) va pt_addr val))
                    pt_addr0 = Some PT).
  unfold madd_type; crush; rewrite <- H7; crush.
  rewrite <- H10.
  symmetry.
  apply STAC.write_same_madd_type; crush.
Qed.

Lemma new_va_has_pa : 
  forall (s:state) (osi:os_ident) (pt_addr:madd) (va:vadd) (ma:madd),
    valid_state s ->

    ((exists pa, osi_pa_to_ma s osi pa = Some ma) \/ 
     ~ os_accessible va) ->

    osi_current_page_table s osi = Some pt_addr ->
    madd_type (smem_mm (st_memory s)) ma <> None ->
    va_has_valid_pa (smem_mm (ACT.new s osi pt_addr va ma)) 
                    (st_hypervisor s).
Proof.
  Local Opaque ACT.new.
  intros.
  STAC.destruct_valid_state.
  unfold va_has_valid_pa; crush.
  
  destruct STAC.new_get_mem_complete_semantics 
    with s osi pt_addr va ma pt_addr0 as [?|[?|[?|[? | ?]]]]; crush;
  unfold page_of_vadd_madd_smem in *; crush;
  repeat (match goal with 
            | [H: Some ?pg = get_cache ?c (?va,?ma) |-_] =>
              symmetry in H
            | [H: PT_c ?pt = page_content ?pg|-_]=> symmetry in H
            | [H1: get_cache (smem_c (st_memory ?s)) (?va,?ma) = Some ?pg,
               H2: page_content ?pg = PT_c ?pt,
               H3: valid_cache ?s |-_] =>
              destruct H3 with va ma pg; crush; 
                unfold is_rw_page in *; crush
          end).
  injection H6; intros; subst pt; clear H6.
  destruct (vadd_eq va0 va); crush;
    [rewrite <- PT.get_set in * |
     rewrite PT.get_set_distinct in H7]; crush.
  destruct H0; crush; unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi0); crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi); crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.
  unfold osi_current_page_table in *; crush.
  rewrite H5 in H6; injection H6; intros; subst osi0; crush.
  rewrite H12 in *; crush.
Qed.  

Lemma del_va_has_pa : 
  forall (s:state) (osi:os_ident) (pt_addr:madd) (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    va_has_valid_pa (smem_mm (ACT.delete_smem osi pt_addr va s)) 
      (st_hypervisor s).
Proof.
  intros.
  STAC.destruct_valid_state.
  unfold va_has_valid_pa; crush.
  
  destruct (eq_cache_key (va0, pt_addr0) (va, pt_addr)); crush.
  (* (va0,pt_addr0) = (va,pt_addr) *)
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr va) eqn:?;
    crush.
  assert (get_mem m pt_addr = Some pt_page).
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?; crush;
    [rewrite (delete_mem_some_entry (st_memory s) m pt_addr va ma p) in H4 |
      rewrite (delete_mem_no_entry (st_memory s) m pt_addr va ma) in H4]; crush.
  destruct (madd_eq ma pt_addr); crush;
    [rewrite <- get_add_mem in * |
      rewrite get_add_mem_distinct in H4]; crush.
  destruct H_valid_cache with va pt_addr pt_page; crush;
    unfold is_rw_page in *; crush.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite <- get_add_mem in *; crush.
  injection H5; intros; subst pt; crush.
  rewrite PT.get_remove in *; crush.
  unfold remove_page_table_entry in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H13 in *; crush.

  (* (va0, pt_addr0) <> (va,pt_addr) *)
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr va) eqn:?;
    crush.
  assert (get_mem m pt_addr0 = Some pt_page); crush.
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?; crush.
  rewrite (delete_mem_some_entry (st_memory s) m pt_addr va ma p) in H4; crush.
  destruct (madd_eq ma pt_addr0); crush;
    [rewrite <- get_add_mem in * | rewrite get_add_mem_distinct in H4]; crush.
  destruct H_valid_cache with va pt_addr0 pt_page; crush;
    unfold is_rw_page in *; crush.
  rewrite (delete_mem_no_entry (st_memory s) m pt_addr va ma) in H4; crush.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; crush.
  destruct (madd_eq pt_addr pt_addr0); crush;
    [rewrite <- get_add_mem in * | rewrite get_add_mem_distinct in H8]; crush.
  injection H5; intros; subst pt; clear H5.
  assert (PT.get p0 va0 = Some ma0); crush.
  apply PT.get_remove_old with (k':=va); crush.
  unfold remove_page_table_entry in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H13 in *; crush.
Qed.  

Lemma va_has_valid_pa_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_state s ->
  va_has_valid_pa (smem_mm (st_memory s')) (st_hypervisor s').
Proof.
  Local Opaque new_page.
  Local Hint Resolve read_va_has_pa write_va_has_pa new_va_has_pa 
    del_va_has_pa.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; crush; unfold va_has_valid_pa in *; crush.

  generalize (del_va_has_pa s o x1 v x2); intros.
  unfold va_has_valid_pa in H13.
  apply H13 with (osi:=osi) (pt_addr:=pt_addr) (pt_page:=pt_page) 
                 (pt:=pt) (va:=va); crush.

  generalize (del_va_has_pa s (st_active_os s) x0 v x1); intros.
  unfold va_has_valid_pa in H10.
  apply H10 with (osi:=osi) (pt_addr:=pt_addr) (pt_page:=pt_page) 
                 (pt:=pt) (va:=va); crush.

  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.

  (* Page_pin_untrusted *)
  destruct (os_ident_eq osi o); crush;
    [rewrite <- HYP.get_set in * |
     rewrite HYP.get_set_distinct in H14]; crush.
  destruct (madd_eq pt_addr x); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H16]; crush.
  Local Transparent new_page.
  crush.
  Local Opaque new_page.
  injection H17; crush; rewrite PT.get_empty in *; crush.
  destruct (madd_eq x ma); crush;
    [exists p; rewrite <- P2M.get_set|]; crush.
  destruct H_va_has_pa with o x0 pt_addr pt_page pt va ma; crush.
  destruct (padd_eq p x5); crush;
    [rewrite <- P2M.get_set in * | rewrite P2M.get_set_distinct in H15];
    crush.
  exists x6; rewrite P2M.get_set_distinct; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (madd_eq pt_addr x); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H16]; crush.
  Local Transparent new_page.
  crush.
  Local Opaque new_page.  
  injection H17; crush; rewrite PT.get_empty in *; crush.

  (* Page unpin *)
  destruct (os_ident_eq osi o); crush;
    [rewrite <- HYP.get_set in * |
     rewrite HYP.get_set_distinct in H11]; crush.
  destruct (madd_eq pt_addr x0); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H13]; crush.
  Local Transparent new_page.
  crush.
  Local Opaque new_page.  
  assert (P2M.get x x1 = Some pt_addr); crush.
  destruct (padd_eq x1 p); crush;
    [rewrite P2M.get_remove in * |
     rewrite P2M.get_remove_old with (k':=p) (inf:=pt_addr)]; crush.
  destruct (madd_eq x3 ma); crush.
  unfold osi_pa_to_ma in *; crush.
  destruct H10 with x1 pt_page va pt; crush.
  unfold osi_pa_to_pg; crush.
  destruct H_va_has_pa with o x pt_addr pt_page pt va ma; crush.
  exists x4; crush;
  rewrite P2M.get_remove_distinct with (inf:=ma); crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (madd_eq pt_addr x0); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in H13]; crush.
  Local Transparent new_page.
  crush.
  Local Opaque new_page.  
Qed.

End EXEC_INV_VA_HAS_PA.