Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VALID_CACHE (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.



Lemma read_valid_cache_inv : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_cache (State (st_active_os s) (st_oss s) (st_exec_mode s)
                       (st_aos_activity s) (st_hypervisor s)
                       (page_smem_smem (get (st_memory s) va pt_addr))).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some m).
  unfold madd_of_vadd_smem in *; crush.
  STAC.destruct_valid_state; apply H_valid_tlb; crush.

  unfold valid_cache; intros.
  assert (page_of_vadd_madd_smem (page_smem_smem (get (st_memory s) va pt_addr)) va0 ma = Some pg_cache); [unfold page_of_vadd_madd_smem; crush|simpl in *].
  rewrite get_new_cache with (ma:=m) (pg:=pg) (val:=x) in H4; auto.
  2: rewrite get_page_result with (pg:=pg); crush;
    unfold page_of_vadd_smem; crush.
  destruct (eq_cache_key (va,m) (va0,ma)); [injection e; intros; subst; clear e|].
  (* (va,m) = (va0,ma) *)
  assert (is_cacheable pg).
  destruct (page_cacheable pg) eqn:?; auto.
  assert (~is_cacheable pg); [unfold is_cacheable; rewrite Heqb; auto| clear Heqb].
  rewrite <- get_add_cache_non_cacheable with (k:=(va0,ma)) (pg:=pg) in H4;
    [|unfold page_of_vadd_madd_smem in *; crush].
  unfold page_of_vadd_madd_smem in H1; crush.
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma pg; crush.
  rewrite <- get_add_cache_cacheable in H4; auto.
  injection H4; intros; subst; clear H4.
  unfold is_rw_page; crush.
  unfold page_of_vadd_madd_smem in H1; crush.
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma pg; crush.
  exists x0; crush.
  rewrite get_mem_unchanged with (ma:=ma) (pg:=pg); crush.
  specialize (add_cache_entry_behavior_update (smem_c (st_memory s)) (va0,ma) pg);
    crush.
  apply H11; crush.
  exists pg; crush.  
  destruct (add_cache (smem_c (st_memory s)) (va0,ma) pg) eqn:?; crush.
  destruct o.
  destruct p; destruct c0; crush.
  destruct (madd_eq m ma); crush.
  (* m = ma *)
  assert (get_cache (smem_c (st_memory s)) (v,ma) = Some p).
  apply STAC.cache_deleted_entry_was_there with (va:=va0) (ma:=ma) (pg:=pg); crush.
  rewrite Heqp; crush.    
  destruct (vadd_eq va0 v); crush.
  STAC.destruct_valid_state. 
  STAC.synonym_cached.
  (* m <> ma *)
  rewrite get_cache_entry2 with (va':=v) (ma:=ma) (ma':=m) (pg:=pg) (pg':=p);
    crush.
  rewrite Heqp; crush.
  (* No entry was removed *)
  rewrite get_mem_unchanged with (ma:=ma) (pg:=pg); crush.
  rewrite Heqp; crush.
  (* (v,m) <> (va,ma) *)
  assert (get_cache (smem_c (st_memory s)) (va0,ma) = Some pg_cache).
  specialize (get_add_cache_distinct_not_new (smem_c (st_memory s)) (va,m) (va0,ma) 
                                             pg pg_cache); crush.
  crush; STAC.destruct_valid_state; destruct H_valid_cache with va0 ma pg_cache; crush.
  exists x0; crush.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=pg) (pg':=p);
    crush.
  STAC.simpl_some_cache_entry.
  destruct STAC.no_cache_alias with s va0 v ma pg_cache x2; crush.
  rewrite get_mem_unchanged with (ma:=m) (pg:=pg);
    crush.
Qed.

Lemma write_valid_cache_inv : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page) (val:value),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_cache (State (st_active_os s) (st_oss s) (st_exec_mode s)
                       (st_aos_activity s) (st_hypervisor s)
                       (set (st_memory s) va pt_addr val)).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some m).
  unfold madd_of_vadd_smem in *; crush.
  STAC.destruct_valid_state; apply H_valid_tlb; crush.
  unfold valid_cache; intros.
  assert (page_of_vadd_madd_smem (set (st_memory s) va pt_addr val) va0 ma = Some pg_cache); [unfold page_of_vadd_madd_smem; crush|simpl in *].
  rewrite set_new_cache with (ma:=m) (pg:=pg) (old_val:=x) in H4; auto.
  2: unfold page_of_vadd_smem; crush.
  destruct (eq_cache_key (va,m) (va0,ma)); [injection e; intros; subst; clear e|].
  (* (va,m) = (va0,ma) *)
  assert (is_cacheable pg).
  destruct (page_cacheable pg) eqn:?; auto.
  rewrite <- get_add_cache_non_cacheable in H4; crush.
  unfold page_of_vadd_madd_smem in H1; crush.
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma pg; crush.
  unfold is_cacheable in *; crush.
  rewrite <- get_add_cache_cacheable in H4; auto.
  injection H4; intros; subst.
  unfold is_rw_page; crush.
  unfold page_of_vadd_madd_smem in H1; crush.
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma pg; crush.
  destruct set_new_mmem with (st_memory s) pt_addr va0 ma val pg; crush.
  unfold page_of_vadd_smem, page_of_vadd_madd_smem; crush.
  exists x1; crush.
  unfold same_page_type in *; crush.
  rewrite H2 in *; crush.
  destruct (page_content x1); crush.
  destruct set_new_mmem with (st_memory s) pt_addr va0 ma val pg; crush.
  unfold page_of_vadd_smem, page_of_vadd_madd_smem; crush.
  exists x0; crush.
  unfold same_page_type in *; crush.
  rewrite H2 in *; crush.
  destruct (page_content x0); crush.
  (* (v,m) <> (va,ma) *)
  assert (Some pg_cache =  get_cache (smem_c (st_memory s)) (va0,ma)).
  specialize (get_add_cache_distinct_not_new (smem_c (st_memory s)) (va,m) (va0,ma) 
                                             (Page (RW_c (Some val))
                                                   (page_owned_by pg) 
                                                   (page_cacheable pg)) pg_cache); 
    intros.
  match goal with 
    |[H:context[add_cache ?c ?k ?new_pg] |-_] =>
     destruct (add_cache c k new_pg) eqn:?
  end; crush.
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma pg_cache; crush.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0) 
                                            (pg:=Page (RW_c (Some val))
                                                       (page_owned_by pg)
                                                       (page_cacheable pg))
                                            (pg':=p); crush.
  symmetry in H6.
  destruct (vadd_eq va va0); crush.
  STAC.synonym_cached.
  STAC.simpl_some_cache_entry.
  destruct STAC.no_cache_alias with s va0 v m0 pg_cache x2; crush.
  rewrite set_mem_unchanged2 with (ma:=m) (pg:=Page (RW_c (Some val))
                                                       (page_owned_by pg)
                                                       (page_cacheable pg)); crush.
  symmetry in H6.
  destruct (vadd_eq va va0); crush.
  STAC.synonym_cached.
Qed.

Lemma new_valid_cache_inv : 
  forall (s:state) (oss:OSS.dict) (osi:os_ident) (pt_addr:madd) (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_type (smem_mm (st_memory s)) ma <> None ->
    valid_cache (State (st_active_os s) oss (st_exec_mode s)
                       (st_aos_activity s) (st_hypervisor s)
                       (ACT.new s osi pt_addr va ma)).
Proof.
  unfold valid_cache; intros; simpl in *.
  destruct (shared_page_dec s va ma) eqn:?.
  rewrite new_cache_shared in H2.
  destruct (madd_eq ma ma0); subst.
  rewrite get_remove_ma_cache_eq in H2; try discriminate.
  rewrite get_remove_ma_cache_ne in H2; auto.
  assert (get_cache (smem_c (st_memory s)) (va0,ma0) = Some pg_cache).
  unfold remove_new_va_cache in *; crush.
  apply get_remove_cache_old with (k':=(va,m)); crush. 
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma0 pg_cache; crush.
  
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr va) eqn:?; crush.
  (* Some entry *)
  destruct (madd_eq m ma0); crush.
  destruct (get_cache (smem_c (st_memory s)) (va,ma0)) eqn:?; crush.
  destruct (vadd_eq va0 va); crush.
  (* va0 = va *)
  exists p; crush.
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma0 = Some p); [|crush].
  apply STAC.new_get_mem_shared_distinct_ma_some_entry 
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
  (* va0 <> va *)
  assert (va <> va0); crush.
  STAC.synonym_cached.
  (* No entry *)
  exists x; crush.
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma0 = Some x); [|crush].
  rewrite STAC.new_get_mem_shared_distinct_ma_no_entry
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
  (* Mapped to other *)
  exists x; crush.
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.  
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma0 = Some x); [|crush].
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
  (* Not mapped *)
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.  
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma0 = Some x); [|crush].
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
  (* not shared *)
  rewrite new_cache_not_shared in H2.
  assert (get_cache (smem_c (st_memory s)) (va0,ma0) = Some pg_cache).
  unfold remove_new_va_cache in *; crush.
  destruct (eq_cache_key (va,m) (va0,ma0)); crush.
  rewrite get_remove_cache in *; crush.
  apply get_remove_cache_old with (va,m); crush.
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma0 pg_cache; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr va) eqn:?; crush.
  (* Some entry *)
  destruct (madd_eq m ma0); crush.
  destruct (get_cache (smem_c (st_memory s)) (va,ma0)) eqn:?; crush.
  destruct (vadd_eq va va0); crush.
  (* va = va0 *)
  exists pg_cache; crush.
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.  
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va0 ma)) ma0 = Some pg_cache); [|crush].
  apply STAC.new_get_mem_not_shared_distinct_ma_some_entry 
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
  (* va <> va0 *)
  STAC.synonym_cached.
  (* No entry *)
  exists x; crush.
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.  
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma0 = Some x); [|crush].
  rewrite STAC.new_get_mem_not_shared_distinct_ma_no_entry
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
  destruct (vadd_eq va va0); crush.
  destruct n; crush.
  unfold shared_page; crush.
  destruct H_valid_cache_va_ma with va0 ma pg_cache; crush.
  exists x3; exists x2; exists x4; exists x5; exists va0; crush.
  (* Mapped to other *)
  exists x; crush.
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.  
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma0 = Some x); [|crush].
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
  (* Not mapped *)
  exists x; crush.
  unfold osi_current_page_table in *; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi.  
  assert (get_mem (smem_mm (ACT.new s osi pt_addr va ma)) ma0 = Some x); [|crush].
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped
    with (pt_page':=pt_page) (pt':=pt); crush.
  unfold madd_type in H1; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.
  unfold is_rw_page in *; crush.
Qed.  

Lemma del_valid_cache_inv : 
  forall (s:state) (oss:OSS.dict) (osi:os_ident) (pt_addr:madd) (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    valid_cache (State (st_active_os s) oss (st_exec_mode s)
                       (st_aos_activity s) (st_hypervisor s)
                       (ACT.delete_smem osi pt_addr va s)).
Proof.
  unfold valid_cache; intros; simpl in *.
  rewrite  delete_cache with (ma:=ma) in H2; auto.
  assert (get_cache (smem_c (st_memory s)) (va0,ma0) = Some pg_cache).
  destruct (eq_cache_key (va,ma) (va0,ma0)); crush.
  rewrite get_remove_cache in *; crush.
  rewrite get_remove_cache_old with (k':=(va,ma)) (pg:=pg_cache); crush.
  STAC.destruct_valid_state; destruct H_valid_cache with va0 ma0 pg_cache; crush.

  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H15 in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr va) eqn:?; crush.
  rename m into mm'.
  assert (ma0 <> pt_addr); crush.
  unfold is_rw_page in *; crush.
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=ma) (pg:=p); crush. 
  destruct (madd_eq ma ma0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  exists p; crush.
  destruct H_valid_cache with va ma0 p; crush.
  destruct H_valid_cache with va ma0 p; crush.
  destruct (vadd_eq va va0); crush.
  destruct STAC.no_cache_alias with s va va0 ma0 p pg_cache; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H15 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H15; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=ma); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H15 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H15; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H15 in *; crush.
Qed.  

Lemma valid_cache_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_state s ->
  valid_cache s'.
Proof.
  Local Hint Resolve read_valid_cache_inv write_valid_cache_inv 
        new_valid_cache_inv del_valid_cache_inv.

  STAC.action_induction a; crush;
    STAC.destruct_action_sem; crush;
    try (unfold valid_cache, valid_state in *; crush; fail).

  (* Switch *)
  STAC.destruct_valid_state; crush.
  unfold valid_cache; intros; simpl in *.
  rewrite (ctxt_switch_cache (st_memory s)) in H2.
  rewrite (ctxt_switch_mem (st_memory s)); crush.
  
  (* Switch *)
  STAC.destruct_valid_state; crush.
  unfold valid_cache; intros; simpl in *.
  rewrite (ctxt_switch_cache (st_memory s)) in H5.
  rewrite (ctxt_switch_mem (st_memory s)); crush.
  
  
  (* Page_pin_untrusted *)
  STAC.destruct_valid_state; crush.
  unfold valid_cache; intros; simpl in *.
  rewrite get_add_mem_distinct; [|crush].
  destruct H_valid_cache with va ma pg_cache; crush.
  destruct H_valid_cache with va ma pg_cache; crush.
  unfold is_rw_page in *; crush.
  STAC.destruct_valid_state; crush.
  unfold valid_cache; intros; simpl in *.
  rewrite get_add_mem_distinct; [|crush].
  destruct H_valid_cache with va ma pg_cache; crush.
  destruct H_valid_cache with va ma pg_cache; crush.
  unfold is_rw_page in *; crush.
  
  (* Page_unpin *)
  STAC.destruct_valid_state.
  unfold valid_cache; intros; simpl in *.
  destruct H_valid_cache with va ma pg_cache; crush.
  rewrite get_add_mem_distinct; [|crush].
  exists x1; crush.  
  unfold osi_pa_to_ma in *; crush.
  destruct H_valid_cache_va_ma with va x3 pg_cache; crush.
  destruct H10 with x4 x5 va x6; crush.
  assert (o = x0); crush.
  assert (madd_owner (smem_mm (st_memory s)) x3 = Os_PO o); crush.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO x0); crush.
  destruct H_valid_virtmap with x0 i0 x5 x6 va x3; 
    unfold madd_owner in *; crush.
  unfold os_accessible in *.
  destruct (ctxt_vadd_accessible ctxt va); 
    [rewrite H22 in H20 |
     rewrite H23 in H20]; unfold os_accessible in *; crush.
Qed.                      

End EXEC_INV_VALID_CACHE.