Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VALID_CACHE_VA_MA (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).


  Import PG.
  Import C.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.


Lemma read_valid_cache_va_ma :
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_cache_va_ma (State (st_active_os s) (st_oss s) (st_exec_mode s)
                       (st_aos_activity s) (st_hypervisor s)
                       (page_smem_smem (get (st_memory s) va pt_addr))).
Proof.
  intros.
  STAC.destruct_valid_state; unfold valid_cache_va_ma; crush.
  unfold page_of_vadd_smem in *; crush.
  unfold is_rw_page in *; crush.
  rewrite get_new_cache with (ma:=m) (pg:=pg) (val:=x) in H3; crush;
    try (apply get_page_result; unfold page_of_vadd_smem; crush; fail).
  destruct (eq_cache_key (va,m) (va0,ma)); crush.
  (* (va,m) = (va0,ma) *)
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va0 = Some ma); crush.
  unfold madd_of_vadd_smem in *; crush.
  apply H_valid_tlb; unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  exists (st_active_os s); exists (os_current_page i); exists p0; exists p1;
    unfold osi_pa_to_pg; crush.
  rewrite <- STAC.read_no_pt_change with (ma':=ma) (pg':=pg); crush;
    unfold is_rw_page, madd_type; crush.
  rewrite Heqc; auto.
  (* (v,m) <> (va,ma) *)
  generalize (get_add_cache_distinct_not_new
    (smem_c (st_memory s)) (va,m) (va0,ma) pg pg_cache); crush.
  destruct H_valid_cache_va_ma with va0 ma pg_cache; crush.
  exists x0; exists x1; exists x2; exists x3; unfold osi_pa_to_pg in *; crush.
  rewrite <- STAC.read_no_pt_change with (ma':=m) (pg':=pg); crush;
    unfold is_rw_page, madd_type; crush.
  rewrite H6; auto.
Qed.

Lemma write_valid_cache_va_ma : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page) (val:value),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_cache_va_ma (State (st_active_os s) (st_oss s) (st_exec_mode s)
      (st_aos_activity s) (st_hypervisor s)
      (set (st_memory s) va pt_addr val)).
Proof.
  intros.
  STAC.destruct_valid_state; unfold valid_cache_va_ma; crush.
  unfold page_of_vadd_smem in *; crush.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=m) (pg:=pg) (old_val:=x) in H3; crush;
    try (unfold page_of_vadd_smem; crush; fail).
  destruct (eq_cache_key (va,m) (va0,ma)); crush.
  (* (va,m) = (va0,ma) *)
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va0 = Some ma); crush.
  unfold madd_of_vadd_smem in *; crush.
  apply H_valid_tlb; unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  exists (st_active_os s); exists (os_current_page i); exists p0; exists p1;
    unfold osi_pa_to_pg; crush.
  rewrite <- STAC.write_no_pt_change with (pg':=pg); crush;
    unfold is_rw_page, madd_type, aos_current_page_table,
      osi_current_page_table; crush.
  unfold aos_page_of_vadd_smem, aos_current_page_table, osi_current_page_table,
    page_of_vadd_smem; crush.
  rewrite Heqc; auto.
  (* (v,m) <> (va,ma) *)
  assert (get_cache (smem_c (st_memory s)) (va0,ma) = Some pg_cache).
  generalize (get_add_cache_distinct_not_new
    (smem_c (st_memory s)) (va,m) (va0,ma)
    (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))
    pg_cache); intros.
  destruct (add_cache (smem_c (st_memory s)) (va,m)
  (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))) eqn:?; crush.
  destruct H_valid_cache_va_ma with va0 ma pg_cache; crush.
  exists x0; exists x1; exists x2; exists x3; unfold osi_pa_to_pg in *; crush.
  rewrite <- STAC.write_no_pt_change with (pg':=pg); crush;
    unfold is_rw_page, madd_type; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem; crush.
  rewrite H6; auto.
Qed.

Lemma new_valid_cache_va_ma : 
  forall (s:state) (oss:OSS.dict) (osi:os_ident) (pt_addr:madd) (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_type (smem_mm (st_memory s)) ma <> None ->
    valid_cache_va_ma (State (st_active_os s) oss (st_exec_mode s)
      (st_aos_activity s) (st_hypervisor s)
      (ACT.new s osi pt_addr va ma)).
Proof.
  Local Opaque ACT.new.
  intros.
  STAC.destruct_valid_state; unfold valid_cache_va_ma; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma) eqn:?; crush.

  assert (get_cache (smem_c (st_memory s)) (va0,ma0) = Some pg_cache).
  Local Transparent ACT.new.
  crush; [ rewrite new_cache_shared in H2; destruct (madd_eq ma0 ma) |
           rewrite new_cache_not_shared in H2]; crush;
    try rewrite get_remove_ma_cache_eq in *; 
    try rewrite get_remove_ma_cache_ne in *; crush;
    unfold remove_new_va_cache in *; crush;
    destruct (eq_cache_key (va,m) (va0,ma0)); crush;
    try rewrite get_remove_cache in *; crush;
    apply get_remove_cache_old with (k':=(va,m)); crush.
  Local Opaque ACT.new.
  destruct H_valid_cache_va_ma with va0 ma0 pg_cache; crush.
  unfold osi_pa_to_pg in H4; crush.
  destruct (madd_eq i0 pt_addr); crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  exists x; exists x0;
    exists (Page (PT_c (PT.set pt va ma)) (page_owned_by x1)
      (page_cacheable x1)); exists (PT.set pt va ma);
    unfold osi_pa_to_pg in *; crush.
  symmetry.
  apply STAC.new_get_mem_pt_addr; crush.
  destruct (vadd_eq va va0); crush;
    [rewrite <- PT.get_set |
     rewrite PT.get_set_distinct]; crush.
  destruct (madd_eq ma ma0); crush.
  Local Transparent ACT.new.
  crush; [ rewrite new_cache_shared in H2 |
           rewrite new_cache_not_shared in H2]; crush;
    try rewrite get_remove_ma_cache_ne in *; crush;
      unfold remove_new_va_cache in *; crush;
      unfold madd_of_vadd, madd_page_table in *; crush;
      rewrite H5 in *; crush;
      try rewrite get_remove_cache in *; crush.
  Local Opaque ACT.new.
  exists x; exists x0; exists x1; exists x2; unfold osi_pa_to_pg; crush.
  rewrite STAC.new_get_mem_distinct_pt; crush.
  unfold madd_type; crush.
  rewrite H5; crush.
Qed.

Lemma del_valid_cache_va_ma : 
  forall (s:state) (oss:OSS.dict) (osi:os_ident) (pt_addr:madd) (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    valid_cache_va_ma (State (st_active_os s) oss (st_exec_mode s)
                       (st_aos_activity s) (st_hypervisor s)
                       (ACT.delete_smem osi pt_addr va s)).
Proof.
  intros.
  STAC.destruct_valid_state; unfold valid_cache_va_ma; crush.
  assert (get_cache (smem_c (st_memory s)) (va0,ma0) = Some pg_cache).
  rewrite delete_cache with (ma:=ma) in H2; crush.
  destruct (eq_cache_key (va,ma) (va0,ma0)); crush;
    [rewrite get_remove_cache in * |
     apply get_remove_cache_old with (k':=(va,ma))]; crush.
  destruct H_valid_cache_va_ma with va0 ma0 pg_cache; crush.
  unfold osi_pa_to_pg in H4; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  destruct (madd_eq i0 pt_addr); crush.
  rewrite H5 in H12; crush.
  injection H12; intros; subst x2; clear H12.
  exists x; exists x0; exists 
    (Page (PT_c (PT.remove pt va))
      (page_owned_by pt_page)
      (page_cacheable pt_page)); 
  exists (PT.remove pt va); unfold osi_pa_to_pg; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr va) eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=ma) (pg:=p); crush. 
  destruct (madd_eq ma pt_addr); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  destruct H_valid_cache with va pt_addr p; crush.
  unfold is_rw_page in *; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H5 in *; crush.
  rewrite <- get_add_mem; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=ma); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H5 in *; crush.
  rewrite <- get_add_mem; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H5 in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H5 in *; crush.
  destruct (vadd_eq va0 va); crush;
    [|rewrite PT.get_remove_distinct with (inf:=ma0)]; crush.
  rewrite delete_cache with (ma:=ma) in H2; crush.
  rewrite get_remove_cache in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H5; crush.
  exists x; exists x0; exists x1; exists x2; unfold osi_pa_to_pg; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) pt_addr va) eqn:?; crush.
  rename m into mm'.
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=ma) (pg:=p); crush. 
  destruct (madd_eq ma i0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  destruct H_valid_cache with va i0 p; crush.
  unfold is_rw_page in *; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H12 in *; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=ma); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H12 in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H12 in *; crush.
Qed.



Lemma valid_cache_va_ma_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_state s ->
  valid_cache_va_ma s'.
Proof.
  Local Hint Resolve read_valid_cache_va_ma write_valid_cache_va_ma
    new_valid_cache_va_ma del_valid_cache_va_ma.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; unfold valid_cache_va_ma; crush.

  (* Switch *)
  rewrite ctxt_switch_cache in *; crush.
  destruct H_valid_cache_va_ma with va ma pg_cache; crush.
  exists x, x0, x1, x2; unfold osi_pa_to_pg in *; crush.
  rewrite ctxt_switch_mem; crush.

  (* LSwitch *)
  rewrite ctxt_switch_cache in *; crush.
  destruct H_valid_cache_va_ma with va ma pg_cache; crush.
  exists x0, x1, x2, x3; unfold osi_pa_to_pg in *; crush.
  rewrite ctxt_switch_mem; crush.

  (* Pin *)
  destruct H_valid_cache_va_ma with va ma pg_cache; crush.
  exists x5, x6, x7, x8; unfold osi_pa_to_pg in *; crush.
  destruct (os_ident_eq o x5); crush;
    [rewrite <- HYP.get_set | rewrite HYP.get_set_distinct]; crush.
  destruct (padd_eq p x6); crush;
    [rewrite <- P2M.get_set | rewrite P2M.get_set_distinct]; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (madd_eq x i0); crush;
    rewrite get_add_mem_distinct; crush.
  destruct (madd_eq x i0); crush;
    rewrite get_add_mem_distinct; crush.
  destruct H_valid_cache_va_ma with va ma pg_cache; crush.
  exists x5, x6, x7, x8; unfold osi_pa_to_pg in *; crush.
  destruct (os_ident_eq o x5); crush;
    [rewrite <- HYP.get_set | rewrite HYP.get_set_distinct]; crush.
  destruct (padd_eq p x6); crush;
    [rewrite <- P2M.get_set | rewrite P2M.get_set_distinct]; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (madd_eq x i0); crush;
    rewrite get_add_mem_distinct; crush.
  destruct (madd_eq x i0); crush;
    rewrite get_add_mem_distinct; crush.

  (* Unpin *)
  destruct H_valid_cache_va_ma with va ma pg_cache; crush.
  exists x1, x4, x5, x6; unfold osi_pa_to_pg in *; crush.
  destruct (os_ident_eq o x1); crush;
    [rewrite <- HYP.get_set | rewrite HYP.get_set_distinct]; crush.
  destruct (padd_eq p x4); crush;
    [rewrite P2M.get_remove |
     rewrite P2M.get_remove_distinct with (inf:=i0)]; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite (H9 x5 x6) in H14; crush.
  rewrite PT.get_empty in *; crush.
  destruct (madd_eq x0 i0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  unfold osi_pa_to_ma in *; crush.
  elim n; crush.
  destruct (madd_eq x0 i0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO x1); crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO o); crush.
Qed.  

End EXEC_INV_VALID_CACHE_VA_MA.