Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VALID_CURRENT_PAGE (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.

Lemma read_valid_currpg : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_current_page (st_hypervisor s) (smem_mm (page_smem_smem
      (get (st_memory s) va pt_addr))) (st_oss s).
Proof.
  unfold page_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  unfold valid_current_page; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  exists pt_addr0, pt_page; crush.
  erewrite <- STAC.read_no_pt_change; crush.
  unfold madd_type; crush; rewrite H8; crush.
  unfold is_pt_page; crush.
Qed.

Lemma write_valid_currpg : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page) (val:value),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_current_page (st_hypervisor s) (smem_mm (set (st_memory s) va pt_addr val)) (st_oss s).
Proof.
  unfold page_of_vadd_smem in *; crush.
  STAC.destruct_valid_state.
  unfold valid_current_page; crush.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  exists pt_addr0, pt_page; crush.
  erewrite <- STAC.write_no_pt_change; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem; crush.
  unfold madd_type; crush; rewrite H8; crush.
  unfold is_pt_page; crush.
Qed.

Lemma new_valid_currpg : 
  forall (s:state) (osi:os_ident) (pt_addr:madd)
    (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_type (smem_mm (st_memory s)) ma <> None ->

    ((os_accessible va /\ exists pa, osi_pa_to_ma s osi pa = Some ma) \/
     (~ os_accessible va /\ madd_owner (smem_mm (st_memory s)) ma = Hyp_PO)) ->
      
    valid_current_page (st_hypervisor s) (smem_mm (ACT.new s osi pt_addr va ma)) (st_oss s).
Proof.
  Local Opaque ACT.new.
  intros.
  STAC.destruct_valid_state.
  unfold valid_current_page; crush.
  destruct H_valid_os_idents with osi0; crush.
  STAC.get_osi_current_pt_page osi0; crush.
  generalize H0; unfold osi_current_page_table; crush.
  destruct STAC.new_get_mem_complete_semantics with s osi pt_addr va ma pt_addr0
    as [? | [? | [? | [? | ?]]]]; crush;
      unfold page_of_vadd_madd_smem in *; crush;
        repeat
          match goal with
            | [H: get_cache (smem_c (st_memory s)) (?va, ?ma) = Some ?pg|-_]=>
              symmetry in H
            |[H1: Some ?pg = get_mem (smem_mm (st_memory ?s)) ?ma,
              H2: Some ?pg' = get_cache (smem_c (st_memory ?s)) (?va,?ma),
              H3: page_owned_by ?pg = Os_PO ?osi |-
                page_owned_by ?pg' = Os_PO ?osi] =>
            destruct H_valid_cache with va ma pg'; crush; fail
            | [H: PT_c ?pt = page_content ?pg|-_]=> symmetry in H
            | [H1: Some ?pg = get_cache (smem_c (st_memory ?s)) (?va,?ma),
               H2: Some ?pg' = get_mem (smem_mm (st_memory ?s)) ?ma,
               H3: page_content ?pg' = PT_c ?pt,
               H4: valid_cache ?s |-_] =>
              destruct H4 with va ma pg; crush; 
                unfold is_rw_page in *; crush
            | [H: Some ?pg = get_mem ?mem ?ma |-_] => symmetry in H
            | [H1: get_mem ?mem ?ma = Some ?pg,
               H2: Some ?ma = P2M.get ?p2m ?pt_padd
               |- ex (fun pt_addr' => ex (fun pt_page' => 
                  Some pt_addr' = P2M.get ?p2m ?pt_padd /\ 
                  Some pt_page' = get_mem ?mem pt_addr' /\ _))] =>
              exists ma, pg; unfold is_pt_page; crush
          end.

  exists pt_addr0, pt_page;
    unfold is_pt_page; crush.
Qed.        

Lemma valid_current_page_inv  : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_state s ->
  valid_current_page (st_hypervisor s') (smem_mm (st_memory s')) (st_oss s').
Proof. 
  Local Opaque new_page.
  Local Hint Resolve read_valid_currpg write_valid_currpg 
        new_valid_currpg.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; unfold valid_current_page in *; crush.
  
  (* New *)
  destruct (os_ident_eq o osi); crush;
    [rewrite <- OSS.get_set in * |
     rewrite OSS.get_set_distinct in *]; crush;
    apply new_valid_currpg; crush.

  (* Delete *)
  destruct H_valid_os_idents with osi; crush.
  destruct H_valid_currpg with osi x x3; crush.
  exists x4.
  edestruct STAC.del_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  exists x6; crush.
  destruct (os_ident_eq o osi); crush;
    [rewrite <- OSS.get_set in * |
     rewrite OSS.get_set_distinct in *]; crush.
  destruct H15; crush.
  unfold same_page_type in H17.
  unfold is_pt_page in *; crush.
  rewrite H12 in *; crush.
  destruct (page_content x6); crush.
  destruct H15; crush.

  (* Del hyper *)
  destruct H_valid_os_idents with osi; crush.
  destruct H_valid_currpg with osi x x2; crush.
  exists x3.
  edestruct STAC.del_hyper_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  unfold ACT.post_del_hyper; crush.
  exists x5; crush.
  destruct H11; crush.
  unfold same_page_type in H13.
  unfold is_pt_page in *; crush.
  rewrite H8 in *; crush.
  destruct (page_content x5); crush.
  destruct H11; crush.

  (* Switch *)
  destruct H_valid_currpg with osi system p2m; crush.
  exists x, x0; rewrite ctxt_switch_mem; crush.

  (* LSwitch *)
  destruct (os_ident_eq o osi); crush;
    [rewrite <- OSS.get_set in * |
     rewrite OSS.get_set_distinct in *]; crush.
  unfold madd_type in H4; crush.
  destruct (page_content p0) eqn:?; crush.
  exists x, p0; rewrite ctxt_switch_mem; 
    unfold is_pt_page, osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  destruct H_valid_os_idents with osi; crush.
  destruct H_valid_currpg with osi x0 x1; crush.
  exists x2, x3; rewrite ctxt_switch_mem; crush.

  (* Hcall *)
  destruct (os_ident_eq (st_active_os s) osi); crush;
    [rewrite <- OSS.get_set in * |
     rewrite OSS.get_set_distinct in *]; crush.

  (* Page pin *)
  destruct H_valid_os_idents with osi; crush.
  destruct H_valid_currpg with osi x5 x6; crush.
  exists x7, x8; crush.
  destruct (os_ident_eq o osi); crush;
    [rewrite <- OSS.get_set in *; rewrite <- HYP.get_set in * |
     rewrite OSS.get_set_distinct in *; rewrite HYP.get_set_distinct in *];
    crush.
  rewrite P2M.get_set_distinct; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (madd_eq x x7); crush.
  rewrite get_add_mem_distinct; crush.
  
  (* Page unpin *)
  destruct H_valid_os_idents with osi; crush.
  destruct H_valid_currpg with osi x1 x4; crush.
  exists x5, x6; crush.
  destruct (os_ident_eq o osi); crush;
    [rewrite <- OSS.get_set in *; rewrite <- HYP.get_set in * |
     rewrite OSS.get_set_distinct in *; 
       rewrite HYP.get_set_distinct in *]; 
    crush.
  rewrite P2M.get_remove_distinct with (inf:=x5); crush.
  destruct (os_ident_eq o osi); crush;
    [rewrite <- OSS.get_set in *; rewrite <- HYP.get_set in * |
     rewrite OSS.get_set_distinct in *; 
       rewrite HYP.get_set_distinct in *]; 
    crush.
  destruct (madd_eq x0 x5); crush.
  elim H6.
  apply H_injective_hyper with osi x4 x5; crush.
  rewrite get_add_mem_distinct; crush.
  destruct (madd_eq x0 x5); crush.
  assert (madd_owner (smem_mm (st_memory s)) x5 = Os_PO o); crush;
    unfold madd_owner in *; crush.
  rewrite get_add_mem_distinct; crush.
Qed.  

End EXEC_INV_VALID_CURRENT_PAGE.