Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VALID_HYPER (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM. 
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.
  
Lemma read_valid_hypervisor : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_hypervisor (st_hypervisor s)(smem_mm (page_smem_smem
      (get (st_memory s) va pt_addr))).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  STAC.destruct_valid_state.
  unfold valid_hypervisor; crush.
  destruct (snd(add_cache (smem_c (st_memory s)) (va,m) pg)) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush.
  (* m0 = ma *)
  unfold madd_owner; crush;
    rewrite get_cache_entry1 with (va':=v) (ma:=m) (pg:=pg) (pg':=x1) in *;
      crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  destruct H_valid_cache with v ma x1; crush.
  (* m0 <> ma *)
  unfold madd_owner; crush;
  rewrite get_cache_entry2
    with (va':=v) (ma:=m) (ma':=m0) (pg:=pg) (pg':=x1) in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  rewrite get_mem_unchanged with (ma:=m) (pg:=pg); crush.
Qed.

Lemma write_valid_hypervisor : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page) (val:value),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_hypervisor (st_hypervisor s) (smem_mm (set (st_memory s) va pt_addr val)).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  STAC.destruct_valid_state.
  unfold valid_hypervisor; crush.
  destruct (madd_eq m ma); crush.
  destruct set_new_mmem with (st_memory s) pt_addr va ma val pg;
    unfold page_of_vadd_smem; crush.
  unfold madd_owner; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va ma pg; crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m)
    (Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))).
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush.
  destruct H_valid_cache with v ma x1; crush.
  unfold madd_owner; crush;
    rewrite set_cache_entry1 with (va':=v) (ma:=m)
      (pg:=(Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg)))
      (pg':=x1) in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  unfold madd_owner; crush;
    rewrite set_mem_unchanged1 with (va':=v) (pg':=x1) (ma:=m) (ma':=m0)
      (pg:=(Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))) in *;
      crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  unfold madd_owner; crush;
    rewrite set_mem_unchanged2 with (ma:=m)
      (pg:=(Page (RW_c (Some val)) (page_owned_by pg) (page_cacheable pg))) in *;
      crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
Qed.

Lemma new_valid_hypervisor : 
  forall (s:state) (osi:os_ident) (pt_addr:madd)
    (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_type (smem_mm (st_memory s)) ma <> None ->

    ((os_accessible va /\ exists pa, osi_pa_to_ma s osi pa = Some ma) \/
     (~ os_accessible va /\ madd_owner (smem_mm (st_memory s)) ma = Hyp_PO)) ->
      
    valid_hypervisor (st_hypervisor s) (smem_mm (ACT.new s osi pt_addr va ma)).
Proof.
  Local Opaque ACT.new.
  intros.
  STAC.destruct_valid_state.
  unfold valid_hypervisor; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma0 = Os_PO osi0); crush;
    unfold madd_owner in H5; crush.
  destruct STAC.new_get_mem_complete_semantics with s osi pt_addr va ma ma0
    as [? | [? | [? | [? | ?]]]]; crush;
      unfold madd_owner; crush;
      unfold page_of_vadd_madd_smem in *; crush;
        repeat
          match goal with
            | [H: get_cache (smem_c (st_memory s)) (?va, ?ma) = Some ?pg|-_]=>
              symmetry in H
            |[H1: Some ?pg = get_mem (smem_mm (st_memory ?s)) ?ma,
              H2: Some ?pg' = get_cache (smem_c (st_memory ?s)) (?va,?ma),
              H3: page_owned_by ?pg = Os_PO ?osi |-
                page_owned_by ?pg' = Os_PO ?osi] =>
            destruct H_valid_cache with va ma pg'; crush; fail
          end.
Qed.        

Lemma  valid_hypervisor_inv :forall (s s' : state) (a : action),
    ACT.valid_exec s a s' ->
    valid_state s ->
    valid_hypervisor (st_hypervisor s') (smem_mm (st_memory s')).
Proof. 
  Local Opaque new_page.
  Local Hint Resolve read_valid_hypervisor write_valid_hypervisor
    new_valid_hypervisor.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; unfold valid_hypervisor in *; crush.

  (* Delete *)
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in H9; crush.
  unfold madd_owner; crush;
    edestruct STAC.del_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  destruct H11; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  destruct H11; crush.

  (* Del hyper *)
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in H6; crush.
  unfold madd_owner; crush;
    edestruct STAC.del_hyper_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  unfold ACT.post_del_hyper; crush.
  destruct H8; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  unfold ACT.post_del_hyper; crush.
  destruct H8; crush.

  (* Switch *)
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in H4; crush.
  unfold madd_owner; crush; rewrite ctxt_switch_mem in *; crush.
  (* LSwitch *)
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in H7; crush.
  unfold madd_owner; crush; rewrite ctxt_switch_mem in *; crush.

  (* Page pin *)
  destruct (os_ident_eq o osi); crush;
    [rewrite <- HYP.get_set in * |
     rewrite HYP.get_set_distinct in H14]; crush.
  destruct (padd_eq p pa); crush;
    [rewrite <- P2M.get_set in * |
     rewrite P2M.get_set_distinct in H15]; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold madd_owner; crush.
  Local Transparent new_page.
  rewrite <- get_add_mem in *; crush.
  Local Opaque new_page.
  rewrite <- get_add_mem in *; crush.
  destruct (madd_eq x ma); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  rewrite get_add_mem_distinct in *; crush.
  rewrite get_add_mem_distinct in *; crush.
  destruct (madd_eq x ma); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  rewrite get_add_mem_distinct in *; crush.
  rewrite get_add_mem_distinct in *; crush.
  
  (* Page unpin *)
  destruct (os_ident_eq o osi); crush;
    [rewrite <- HYP.get_set in * |
     rewrite HYP.get_set_distinct in H11]; crush.
  assert (P2M.get x pa = Some ma); crush.
  destruct (padd_eq p pa); crush;
    [rewrite P2M.get_remove in * |
     apply P2M.get_remove_old with (k':=p)]; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  destruct (madd_eq x0 ma); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in *]; crush.
  destruct (padd_eq p pa); crush.
  rewrite P2M.get_remove in *; crush.
  elim n.
  apply H_injective_hyper with osi x ma; crush.
  destruct (madd_eq x0 ma); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in *]; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  destruct (madd_eq x0 ma); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in *]; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) x3 = Os_PO o); crush.
  apply H_valid_hyper with (p2m:=x) (pa:=p); crush.
  unfold madd_owner in *; crush.
  destruct (madd_eq x0 ma); crush;
    [rewrite <- get_add_mem in * |
     rewrite get_add_mem_distinct in *]; crush.
Qed.  

End EXEC_INV_VALID_HYPER.