Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VALID_TLB (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.


Lemma read_valid_tlb : 
  forall s pt_addr pt_addr' v va ma pg,
    valid_state s ->
    Some ma = get_tlb (smem_t (page_smem_smem (get (st_memory s) v pt_addr))) va ->
    aos_current_page_table s = Some pt_addr ->
    aos_current_page_table 
      (State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s)
             (st_hypervisor s) (page_smem_smem (get (st_memory s) v pt_addr))) = 
    Some pt_addr' ->
    aos_page_of_vadd_smem s v = Some pg ->

    madd_of_vadd (smem_mm (page_smem_smem (get (st_memory s) v pt_addr))) pt_addr' va = 
      Some ma.
Proof.
  intros.
  STAC.destruct_valid_state; crush.
  rewrite <- STAC.read_same_madd_of_vadd; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  rewrite get_new_tlb with (ma:=m) in H0; crush.
  destruct (vadd_eq v va); crush.
  rewrite <- get_add_tlb in H0; crush.
  unfold madd_of_vadd_smem in *; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  assert (Some ma = get_tlb (smem_t (st_memory s)) va); crush.
  generalize (get_add_tlb_distinct_not_new (smem_t (st_memory s)) v va m ma);
    crush.
Qed.  

Lemma write_valid_tlb : 
  forall s pt_addr pt_addr' v va ma pg val,
    valid_state s ->
    Some ma = get_tlb (smem_t (set (st_memory s) v pt_addr val)) va ->
    aos_current_page_table s = Some pt_addr ->
    aos_current_page_table 
      (State (st_active_os s) (st_oss s) (st_exec_mode s) (st_aos_activity s)
             (st_hypervisor s) (set (st_memory s) v pt_addr val)) = 
    Some pt_addr' ->
    aos_page_of_vadd_smem s v = Some pg ->
    is_rw_page pg ->

    madd_of_vadd (smem_mm (set (st_memory s) v pt_addr val)) pt_addr' va = 
      Some ma.
Proof.
  intros.
  STAC.destruct_valid_state.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in H3; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr v = Some m).
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  generalize H1 H2; clear H2.
  unfold aos_current_page_table, osi_current_page_table; crush.
  generalize H5; intros.
  unfold madd_of_vadd, madd_page_table in H5; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite <- STAC.write_same_madd_of_vadd with (pt_page:=p0) (old_pg:=pg); crush.
  rewrite set_new_tlb with (ma:=m) in H0; crush.
  destruct (vadd_eq v va); crush.
  rewrite <- get_add_tlb in H0; crush.
  assert (Some ma = get_tlb (smem_t (st_memory s)) va); crush.
  generalize (get_add_tlb_distinct_not_new (smem_t (st_memory s)) v va m ma);
    crush.
  unfold aos_page_of_vadd_smem; crush.
  unfold page_of_vadd_smem; crush.
  unfold is_pt_page; crush.
Qed.  

Lemma new_shared_valid_tlb :
  forall s pt_addr pt_addr' va va' ma ma',
    valid_state s ->
    madd_type (smem_mm (st_memory s)) ma' <> None ->
    aos_current_page_table s = Some pt_addr ->
    Some ma = get_tlb (smem_t (new_shared (st_memory s) pt_addr' va' ma')) va ->
    (pt_addr <> pt_addr' \/ va <> va') ->
    madd_of_vadd (smem_mm (new_shared (st_memory s) pt_addr' va' ma')) pt_addr va = Some ma.
Proof.
  intros.
  STAC.destruct_valid_state.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  rewrite H_oss in Heqopt_res; crush.
  destruct (madd_eq pt_addr pt_addr'); crush.
  (* pt_addr = pt_addr' *)
  destruct H3; crush.
  rewrite STAC.new_get_mem_shared_pt_addr with (pt_page:=pt_page) (pt:=pt); crush.
  rewrite PT.get_set_distinct; crush.
  rewrite new_tlb_shared in *; crush.
  assert (get_tlb (smem_t (st_memory s)) va = Some ma); crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va = Some ma); crush.
  apply H_valid_tlb; crush.  
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H7 in *; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma') eqn:?; crush.
  (* pt_addr <> pt_addr' *)
  rewrite new_tlb_shared in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite STAC.new_get_mem_shared_distinct_pt; crush.
  rewrite H7; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  apply H_valid_tlb; crush.  
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H7 in *; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma') eqn:?; crush.
  unfold madd_type; crush.
  rewrite H7; crush.
Qed.  

Lemma new_not_shared_valid_tlb :
  forall s pt_addr pt_addr' va va' ma ma',
    valid_state s ->
    madd_type (smem_mm (st_memory s)) ma' <> None ->
    aos_current_page_table s = Some pt_addr ->
    Some ma = get_tlb (smem_t (new_not_shared (st_memory s) pt_addr' va' ma')) va ->
    (pt_addr <> pt_addr' \/ va <> va') ->
    madd_of_vadd (smem_mm (new_not_shared (st_memory s) pt_addr' va' ma')) pt_addr va = Some ma.
Proof.
  intros.
  STAC.destruct_valid_state.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite new_tlb_not_shared in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  rewrite H_oss in Heqopt_res; crush.
  destruct (madd_eq pt_addr pt_addr'); crush.
  (* pt_addr = pt_addr' *)
  destruct H3; crush.
  rewrite STAC.new_get_mem_not_shared_pt_addr with (pt_page:=pt_page) (pt:=pt); crush.
  rewrite PT.get_set_distinct; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va = Some ma); crush.
  apply H_valid_tlb; crush.  
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H7 in *; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma') eqn:?; crush.
  (* pt_addr <> pt_addr' *)
  rewrite STAC.new_get_mem_not_shared_distinct_pt; crush.
  rewrite H7; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  apply H_valid_tlb; crush.  
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H7 in *; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma') eqn:?; crush.
  unfold madd_type; crush.
  rewrite H7; crush.
Qed.

Lemma valid_tlb_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_state s ->
  valid_tlb s'.
Proof.
  Local Hint Resolve read_valid_tlb write_valid_tlb
    new_shared_valid_tlb new_not_shared_valid_tlb.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; unfold valid_tlb; crush.

  (* New untrusted shared page *)
  assert (get_tlb (smem_t (new_shared (st_memory s) x0 v x2)) va = Some ma); crush.
  apply get_remove_tlb_old with v; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite <- OSS.get_set in *; crush.
  assert (v <> va); crush.
  rewrite get_remove_tlb in *; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite OSS.get_set_distinct in *; crush.
  assert (x0 <> pt_addr); crush.
  rewrite new_tlb_shared in *; crush.
  unfold aos_current_page_table, osi_current_page_table in H10, H3; crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO o); crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO (st_active_os s)); crush.
  (* New untrusted shared page *)
  assert (get_tlb (smem_t (new_not_shared (st_memory s) x0 v x2)) va = Some ma); crush.
  apply get_remove_tlb_old with v; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite <- OSS.get_set in *; crush.
  assert (v <> va); crush.
  rewrite get_remove_tlb in *; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite OSS.get_set_distinct in *; crush.
  assert (x0 <> pt_addr); crush.
  rewrite new_tlb_not_shared in *; crush.
  unfold aos_current_page_table, osi_current_page_table in H10, H3; crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO o); crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO (st_active_os s)); crush.

  (* New hyper shared page *)
  assert (get_tlb (smem_t (new_shared (st_memory s) x v m)) va = Some ma); crush.
  apply get_remove_tlb_old with v; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  assert (v <> va); crush.
  rewrite get_remove_tlb in *; crush.
  assert (get_tlb (smem_t (new_not_shared (st_memory s) x v m)) va = Some ma); crush.
  apply get_remove_tlb_old with v; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  assert (v <> va); crush.
  rewrite get_remove_tlb in *; crush.

  (* Del untrusted *)
  destruct (vadd_eq v va); crush.
  rewrite get_remove_tlb in *; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite <- OSS.get_set in *; crush.
  destruct (madd_eq x1 pt_addr ); crush.
  (* x1 = pt_addr *)
  unfold madd_of_vadd, madd_page_table; crush.
  generalize (STAC.del_get_mem_pt_page s
             (State (st_active_os s) (OSS.set (st_oss s) (st_active_os s) 
                                              {|os_hcall:=None; 
                                                os_current_page:=os_current_page x0|})
                    (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                    (ACT.delete_smem (st_active_os s) pt_addr v s))
             pt_addr); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite H10 with (pt_page:=p0) (pt:=p1) (osi':=(st_active_os s)) (va':=v); crush.
  rewrite PT.get_remove_distinct with (inf:=ma); crush.
  rewrite delete_tlb in *; crush.
  assert (get_tlb (smem_t (st_memory s)) va = Some ma); crush.
  apply get_remove_tlb_old with v; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  exists pt_addr; exists x2; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  unfold ACT.post_del; crush.
  (* x1 <> pt_addr *)
  unfold madd_of_vadd, madd_page_table; crush.
  generalize (STAC.del_get_mem_distinct_pt s
             (State (st_active_os s) (OSS.set (st_oss s) (st_active_os s) 
                                              {|os_hcall:=None; 
                                                os_current_page:=os_current_page x0|})
                    (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                    (ACT.delete_smem (st_active_os s) x1 v s))); crush.
  rewrite H10 with (osi':=(st_active_os s)) (pt_addr':=x1) (va':=v); crush.
  rewrite delete_tlb in *; crush.
  assert (get_tlb (smem_t (st_memory s)) va = Some ma); crush.
  apply get_remove_tlb_old with v; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  unfold madd_type; crush.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table in H9; crush.
  rewrite H14; crush.

  rewrite delete_tlb in *; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite OSS.get_set_distinct in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  generalize (STAC.del_get_mem_distinct_pt s
             (State (st_active_os s) (OSS.set (st_oss s) o
                                              {|os_hcall:=None; 
                                                os_current_page:=os_current_page x0|})
                    (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                    (ACT.delete_smem o x1 v s))); crush.
  rewrite H10 with (osi':=o) (pt_addr':=x1) (va':=v); crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  unfold madd_type; crush.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table in H9; crush.
  rewrite H14; crush.
  clear H8 H10.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) x1 = Os_PO o); crush.
  assert (madd_owner (smem_mm (st_memory s)) x1 = Os_PO (st_active_os s)); crush.
  
  (* Del hyper *)
  destruct (vadd_eq v va); crush.
  rewrite get_remove_tlb in *; crush.
  assert (aos_current_page_table s = Some pt_addr); crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x0 v) eqn:?; crush.
  rename m into mm'.
  unfold remove_page_table_entry in *; crush.
  destruct (page_content p) eqn:?; try discriminate.
  injection Heqo; intros; clear Heqo.

  destruct (get_cache (smem_c (st_memory s)) (v,x1)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x1) (pg:=p1); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite get_add_mem_distinct; crush.  
  destruct (madd_eq pt_addr x0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  rewrite PT.get_remove_distinct with (inf:=ma); crush.
  rewrite delete_tlb in *; crush.
  assert (get_tlb (smem_t (st_memory s)) va = Some ma); crush. 
  apply get_remove_tlb_old with v; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) x0 va = Some ma); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  rewrite delete_tlb in *; crush.
  assert (get_tlb (smem_t (st_memory s)) va = Some ma); crush. 
  apply get_remove_tlb_old with v; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  destruct H_valid_cache with v pt_addr p1; unfold is_rw_page in *; crush.
  STAC.get_aos_data; STAC.get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table in H6; crush.
  unfold remove_page_table_entry; crush.
  rewrite Heqc; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x1); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  destruct (madd_eq pt_addr x0); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  rewrite PT.get_remove_distinct with (inf:=ma); crush.
  rewrite delete_tlb in *; crush.
  assert (get_tlb (smem_t (st_memory s)) va = Some ma); crush. 
  apply get_remove_tlb_old with v; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) x0 va = Some ma); crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite Heqc in *; crush.
  rewrite delete_tlb in *; crush.
  assert (get_tlb (smem_t (st_memory s)) va = Some ma); crush. 
  apply get_remove_tlb_old with v; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  unfold remove_page_table_entry; crush.
  rewrite Heqc; crush.
  unfold remove_page_table_entry in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  (* Switch *)
  rewrite ctxt_switch_tlb in H2; rewrite get_empty_tlb in H2; crush.

  (* Lswitch_untrusted *)
  rewrite ctxt_switch_tlb in H5; rewrite get_empty_tlb in H5; crush.

  (* Hcall *)
  apply H_valid_tlb; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite <- OSS.get_set in *; crush.

  (* Page pin untrusted *)
  match goal with
      |[H:aos_current_page_table ?s' = Some ?pt_addr |- _] =>
       let Hlemma := fresh "H" in
         generalize (STAC.page_pin_madd_of_vadd s s'); 
           intros Hlemma; crush; erewrite Hlemma; crush
  end.
  unfold ACT.pre_pin; crush.
  exists x3; crush.
  unfold ACT.post_pin; crush.
  exists x; crush.
  clear H16.
  apply H_valid_tlb; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct (os_ident_eq o (st_active_os s)); crush.
  rewrite <- OSS.get_set in *; crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (padd_eq p (os_current_page x1)); crush.
  rewrite <- P2M.get_set in *; crush.
  STAC.get_aos_data; STAC.get_aos_current_pt_page; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite P2M.get_set_distinct in *; crush.
  rewrite OSS.get_set_distinct in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  
  match goal with
      |[H:aos_current_page_table ?s' = Some ?pt_addr |- _] =>
       let Hlemma := fresh "H" in
         generalize (STAC.page_pin_madd_of_vadd s s'); 
           intros Hlemma; crush; erewrite Hlemma; crush
  end.
  unfold ACT.pre_pin; crush.
  exists x3; crush.
  unfold ACT.post_pin; crush.
  exists x; crush.
  clear H16.
  apply H_valid_tlb; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct (os_ident_eq o (st_active_os s)); crush.
  rewrite <- OSS.get_set in *; crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (padd_eq p (os_current_page x1)); crush.
  rewrite <- P2M.get_set in *; crush.
  STAC.get_aos_data; STAC.get_aos_current_pt_page; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite P2M.get_set_distinct in *; crush.
  rewrite OSS.get_set_distinct in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  
  (* Page unpin untrusted *)
  assert (aos_current_page_table s = Some pt_addr); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct (os_ident_eq o (st_active_os s)); crush.
  rewrite <- OSS.get_set in *; crush.
  rewrite <- HYP.get_set in *; crush.
  apply P2M.get_remove_old with p; crush.
  rewrite OSS.get_set_distinct in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  destruct (os_ident_eq o (st_active_os s)); crush.
  match goal with
      |[H:aos_current_page_table ?s' = Some ?pt_addr |- _] =>
       let Hlemma := fresh "H" in
         generalize (STAC.page_unpin_madd_of_vadd_distinct_pt_padd s s'); 
           intros Hlemma; crush
  end.
  rewrite H14 with (osi:=st_active_os s) (pt_padd:=os_current_page x2) 
                   (osi':=st_active_os s) (ma:=ma) (pa':=p); crush.
  unfold ACT.pre_unpin; crush.
  exists x2; crush.
  exists x3; crush.
  unfold ACT.post_unpin; crush.
  exists x; crush.
  exists x0; crush.
  unfold osi_pa_to_ma; crush.  
  unfold aos_current_page_table, osi_current_page_table in H13; crush.

  unfold aos_current_page_table, osi_current_page_table in H13; crush.
  match goal with
      |[H:aos_current_page_table ?s' = Some ?pt_addr |- _] =>
       let Hlemma := fresh "H" in
         generalize (STAC.page_unpin_madd_of_vadd_distinct_osi s s'); 
           intros Hlemma; crush
  end.
  
  rewrite H14 with (osi:=st_active_os s) (pt_padd:=os_current_page i)
                   (osi':=o) (ma:=ma) (pa':=p); crush.
  unfold ACT.pre_unpin; crush.
  exists x2; crush.
  exists x3; crush.
  unfold ACT.post_unpin; crush.
  exists x; crush.
  exists x0; crush.
  unfold osi_pa_to_ma; crush.  
  apply H_valid_tlb; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
Qed.


End EXEC_INV_VALID_TLB.