Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Module EXEC_INV_VIRTMAP (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST)
  (EI:EXEC_INV_INIT C SMEM ST ACT ST_VALID).

  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import EI.

Lemma read_valid_virtual_mapping : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_virtual_mapping (smem_mm (page_smem_smem
      (get (st_memory s) va pt_addr))) (st_hypervisor s).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  STAC.destruct_valid_state.
  unfold valid_virtual_mapping; crush.
  rewrite <- STAC.read_no_pt_change with (ma':=m) (pg':=pg) in H4; crush;
    unfold is_rw_page; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr0 va0 = Some ma).
  unfold madd_of_vadd, madd_page_table; crush;
    rewrite <- H6; crush.
  unfold madd_of_vadd, madd_page_table in H9; crush.  
  destruct (page_content pt_page) eqn:?; crush.
  injection H6; intros; subst p0; clear H6 H7.
  destruct H_valid_virtmap with osi pt_addr0 pt_page pt va0 ma; crush.
  destruct (snd (add_cache (smem_c (st_memory s)) (va,m) pg)) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush;
    [rewrite get_cache_entry1 with (va':=v) (ma:=m) (pg:=pg) (pg':=x4) |
    rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=pg) (pg':=x4)];
      crush.
  exists x4; crush;
    destruct H_valid_cache with v ma x4; crush.
  rewrite get_mem_unchanged with (ma:=m) (pg:=pg); crush.
  rewrite STAC.read_same_madd_type with (pt_addr':=pt_addr) (va':=va)
    (ma':=m) (pg':=pg); crush;
    unfold is_rw_page; crush.
  unfold madd_type; crush; rewrite <- H6; crush.
Qed.

Lemma write_valid_virtual_mapping : 
  forall (s:state) (pt_addr:madd) (va:vadd) (pg:page) (val:value),
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (st_memory s) pt_addr va = Some pg ->
    is_rw_page pg ->
    valid_virtual_mapping (smem_mm (set (st_memory s) va pt_addr val))
      (st_hypervisor s).
Proof.
  unfold page_of_vadd_smem, is_rw_page in *; crush.
  STAC.destruct_valid_state.
  unfold valid_virtual_mapping; crush.
  rewrite <- STAC.write_no_pt_change with (pg':=pg) in H4; crush;
    unfold is_rw_page, aos_page_of_vadd_smem, page_of_vadd_smem; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr0 va0 = Some ma).
  unfold madd_of_vadd, madd_page_table; crush;
    rewrite <- H6; crush.
  unfold madd_of_vadd, madd_page_table in H9; crush.  
  destruct (page_content pt_page) eqn:?; crush.
  injection H6; intros; subst p0; clear H6 H7.
  destruct H_valid_virtmap with osi pt_addr0 pt_page pt va0 ma; crush.
  destruct (madd_eq m ma); crush.
  destruct set_new_mmem with (st_memory s) pt_addr va ma val pg; crush;
    unfold page_of_vadd_smem; crush.
  exists x3; crush;
    unfold page_of_vadd_madd_smem in *; crush;
    try destruct H_valid_cache with va ma pg; crush.
  
  destruct (snd (add_cache (smem_c (st_memory s)) (va,m) pg)) eqn:?.
  STAC.destruct_add_cache (smem_c (st_memory s)).
  STAC.simpl_some_cache_entry.
  destruct (madd_eq m0 ma); crush;
    [rewrite set_cache_entry1 with (va':=v) (ma:=m) (pg:=pg) (pg':=x4) |
     rewrite set_mem_unchanged1 with (va':=v) (ma:=m) (ma':=m0)
       (pg:=pg) (pg':=x4)]; crush.
  exists x4; crush;
    try destruct H_valid_cache with v ma x4; crush.
  rewrite set_mem_unchanged2 with (ma:=m) (pg:=pg); crush.
  rewrite <- STAC.write_same_madd_type
    with (pt_addr:=pt_addr) (va:=va) (val:=val); crush.
  unfold madd_type; crush; rewrite <- H6; crush.
Qed.

Lemma new_valid_virtual_mapping : 
  forall (s:state) (osi:os_ident) (pt_addr:madd)
    (va:vadd) (ma:madd),
    valid_state s ->
    osi_current_page_table s osi = Some pt_addr ->
    madd_type (smem_mm (st_memory s)) ma <> None ->

    ((os_accessible va /\ exists pa, osi_pa_to_ma s osi pa = Some ma) \/
     (~ os_accessible va /\ madd_owner (smem_mm (st_memory s)) ma = Hyp_PO)) ->
      
    valid_virtual_mapping (smem_mm (ACT.new s osi pt_addr va ma))
      (st_hypervisor s).
Proof.
  Local Opaque ACT.new.
  intros.
  STAC.destruct_valid_state.
  unfold valid_virtual_mapping; crush.

  destruct (eq_cache_key (va0, pt_addr0) (va, pt_addr)); crush;
    [|assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr0 va0 = Some ma0)].
  (* (va0,pt_addr0) = (va,pt_addr) *)
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  rewrite STAC.new_get_mem_pt_addr with (pt_page:=pt_page0) (pt:=pt0) in H4;
    crush.
  injection H6; intros; subst pt; clear H6.
  rewrite <- PT.get_set in *; crush.
  destruct H2; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma0 = Os_PO osi); crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma0 = Os_PO osi0); crush.
  rewrite H7 in H6; crush.
  injection H6; intros; subst osi0; clear H6.

  destruct STAC.new_get_mem_complete_semantics with s osi pt_addr va ma0 ma0
    as [? | [? | [? | [? | ?]]]]; crush;
  unfold osi_current_page_table; crush;
  repeat(match goal with
    | [H: get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma = _ |-
      context[get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma]] =>
    rewrite H
    | [H: _ = get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma |-
      context[get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma]] =>
    rewrite <- H
    | [|- ex (fun pg => Some pg = Some ?pg2 /\ _)] =>
      exists pg2
  end; crush);
  try (unfold page_of_vadd_madd_smem in *; crush;
    [destruct H_valid_cache with x ma0 x4|]; crush;
    unfold madd_owner in *; crush; fail).
  unfold madd_owner in H7; crush.
  exists p; crush.

  destruct STAC.new_get_mem_complete_semantics with s osi pt_addr va ma0 ma0
    as [? | [? | [? | [? | ?]]]]; crush;
  unfold osi_current_page_table; crush;
  repeat(match goal with
    | [H: get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma = _ |-
      context[get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma]] =>
    rewrite H
    | [H: _ = get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma |-
      context[get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma]] =>
    rewrite <- H
    | [|- ex (fun pg => Some pg = Some ?pg2 /\ _)] =>
      exists pg2
  end; crush);
  try (unfold page_of_vadd_madd_smem in *; crush;
    [destruct H_valid_cache with x3 ma0 x4|]; crush;
    unfold madd_owner in *; crush; fail).
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  exists p; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s)) ma); crush.

  (* (va0,pt_addr0) <> (va,pt_addr) *)
  rewrite <- STAC.new_same_madd_of_vadd
    with (osi':=osi) (pt_addr':=pt_addr) (va':=va) (ma':=ma);
      unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H6; crush.
  destruct (vadd_eq va0 va); crush.
  unfold madd_type; crush.
  rewrite <- H6; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct H_valid_virtmap with osi0 pt_addr0 p0 p1 va0 ma0; crush.

  destruct STAC.new_get_mem_complete_semantics with s osi pt_addr va ma pt_addr0
    as [? | [? | [? | [? | ?]]]]; crush.
  destruct H_valid_cache with va pt_addr0 pt_page;
    unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with x1 ma x2; crush.
  unfold page_of_vadd_madd_smem in *; crush;  
    try (destruct H_valid_cache with x1 ma pt_page;
      unfold is_rw_page in *; crush; fail).

  destruct STAC.new_get_mem_complete_semantics with s osi pt_addr va ma ma0
    as [? | [? | [? | [? | ?]]]]; crush;
  repeat(match goal with
    | [H: get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma = _ |-
      context[get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma]] =>
    rewrite H
    | [H: _ = get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma |-
      context[get_mem (smem_mm (ACT.new _ _ _ _ _)) ?ma]] =>
    rewrite <- H
    | [|- ex (fun pg => Some pg = Some ?pg2 /\ _)] =>
      exists pg2
  end; crush);
  try (destruct H_valid_cache with va ma0 x2; crush; fail);
  unfold page_of_vadd_madd_smem in *; crush;
    try (destruct H_valid_cache with x2 ma x3; crush; fail).
Qed.


Lemma valid_virtual_mapping_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_state s ->
  valid_virtual_mapping (smem_mm (st_memory s')) (st_hypervisor s').
Proof.
  Local Opaque new_page.
  Local Hint Resolve read_valid_virtual_mapping write_valid_virtual_mapping
    new_valid_virtual_mapping.

  STAC.action_induction a; intros; STAC.destruct_action_sem;
    STAC.destruct_valid_state; unfold valid_virtual_mapping in *; crush.

  (* Delete *)
  assert (madd_of_vadd (smem_mm (delete (st_memory s) x1 v)) pt_addr va = Some ma).
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H10; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  eapply STAC.del_madd_of_vadd_old with (osi':=o) (va':=v) ; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  crush.
  unfold madd_of_vadd, madd_page_table in H14; crush.
  destruct (page_content p0) eqn:?; crush.
  assert (page_owned_by p0 = Os_PO osi); crush. 
  edestruct STAC.del_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  crush.
  destruct H16; crush.

  destruct H_valid_virtmap with osi pt_addr p0 p1 va ma; crush.
  edestruct STAC.del_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold ACT.post_del; crush.
  crush.
  rewrite H19; crush.
  destruct H20; crush.
  exists x5; crush.

  (* Delete hyper *)
  assert (madd_of_vadd (smem_mm (delete (st_memory s) x0 v)) pt_addr va = Some ma).
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H7; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma); crush.
  eapply STAC.del_hyper_madd_of_vadd_old with (va':=v) ; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  unfold ACT.post_del_hyper; crush.
  crush.
  unfold madd_of_vadd, madd_page_table in H11; crush.
  destruct (page_content p0) eqn:?; crush.
  assert (page_owned_by p0 = Os_PO osi); crush. 
  edestruct STAC.del_hyper_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  unfold ACT.post_del_hyper; crush.
  crush.
  destruct H13; crush.

  destruct H_valid_virtmap with osi pt_addr p0 p1 va ma; crush.
  edestruct STAC.del_hyper_get_mem_same_page_but_value; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  unfold ACT.post_del_hyper; crush.
  crush.
  rewrite H16; crush.
  destruct H17; crush.
  exists x4; crush.

  (* Switch *)
  rewrite ctxt_switch_mem in *; crush.
  rewrite ctxt_switch_mem in *; crush.

  (* Page pin *)
  assert (get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page); crush.
  destruct (madd_eq pt_addr x); crush;
    [rewrite <- get_add_mem in * | rewrite get_add_mem_distinct in H15]; crush.
  Local Transparent new_page.
  crush.
  injection H17; intros; subst pt; rewrite PT.get_empty in *; crush.
  Local Opaque new_page.
  destruct (madd_eq x ma); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  destruct H_valid_virtmap with osi pt_addr pt_page pt va ma; crush.
  destruct (os_ident_eq o osi); crush;
    [rewrite <- HYP.get_set in *|
     rewrite HYP.get_set_distinct in H14]; crush.
  destruct (padd_eq p x6); crush;
    [rewrite <- P2M.get_set in *|
     rewrite P2M.get_set_distinct in H19]; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; crush.
  rewrite H22 in *; crush.
  destruct (os_ident_eq o osi); crush;
    [rewrite <- HYP.get_set in *|
     rewrite HYP.get_set_distinct in H14]; crush.
  destruct (padd_eq p x6); crush;
    [rewrite <- P2M.get_set in *|
     rewrite P2M.get_set_distinct in H19]; crush.

  (* Page unpin *)
  assert (osi_pa_to_ma s osi x4 = Some pt_addr); crush.
  unfold osi_pa_to_ma; crush.
  destruct (os_ident_eq o osi); crush;
    [rewrite <- HYP.get_set in *|
     rewrite HYP.get_set_distinct in H11]; crush.
  destruct (padd_eq p x4); crush.
  rewrite P2M.get_remove in *; crush.
  apply P2M.get_remove_old with (k':=p); crush.
  clear H11; unfold osi_pa_to_ma in H17; crush.
  destruct (madd_eq pt_addr x0); crush;
    [rewrite <- get_add_mem in * | rewrite get_add_mem_distinct in H12]; crush.
  Local Transparent new_page.
  crush.
  Local Opaque new_page.
  destruct (madd_eq x0 ma); crush;
    [rewrite <- get_add_mem | rewrite get_add_mem_distinct]; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO o); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  unfold osi_pa_to_ma in *; crush.
  destruct H_valid_virtmap with osi pt_addr pt_page pt va x3; crush.
  unfold madd_owner in H11; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; crush.
  unfold madd_owner; crush.
  rewrite H19 in H11; crush.
  rewrite H11 in H18; injection H18; crush; clear H18.
  unfold osi_pa_to_ma in *; crush.
  destruct H10 with (pt_padd:=x4) (pt_page:=pt_page) (mp:=pt) (va:=va); 
    unfold osi_pa_to_pg; crush.
Qed.

End EXEC_INV_VIRTMAP.