(**
 * $Revision: 12919 $
 * $Date: 2016-09-22 11:57:12 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Valid state definition.
 * 
 *************************************************************************)

(**********************)
(**********************)
Require Export State State_valid Tactics State_tactics Exec_invariant_init.

Require Import Exec_inv_inj_hyper Exec_inv_valid_cache Exec_inv_valid_tlb
        Exec_inv_non_cacheable_syn Exec_inv_valid_cache_va_ma
        Exec_inv_valid_virtmap Exec_inv_va_has_pa Exec_inv_valid_hyper
        Exec_inv_valid_current_page Exec_inv_pt_non_cacheable.


Module EXEC_INV (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST).

Import ST.
Import ST_VALID.
Declare Module EI : EXEC_INV_INIT C SMEM ST ACT ST_VALID.
Import EI.

Module EIH := EXEC_INV_INJ_HYPER C SMEM ST ACT ST_VALID EI. Import EIH. (* DONE *)
Module EVC := EXEC_INV_VALID_CACHE C SMEM ST ACT ST_VALID EI. Import EVC. (* DONE *)
Module EVT := EXEC_INV_VALID_TLB C SMEM ST ACT ST_VALID EI. Import EVT. (* DONE *)
Module ECS := EXEC_INV_NONCACHEABLE C SMEM ST ACT ST_VALID EI. Import ECS. (* DONE *)
Module ECVM := EXEC_INV_VALID_CACHE_VA_MA C SMEM ST ACT ST_VALID EI. Import ECVM. (* DONE *)
Module EVV := EXEC_INV_VIRTMAP C SMEM ST ACT ST_VALID EI. Import EVV. (* DONE *)
Module EVP := EXEC_INV_VA_HAS_PA C SMEM ST ACT ST_VALID EI. Import EVP. (* DONE *)
Module EVH := EXEC_INV_VALID_HYPER C SMEM ST ACT ST_VALID EI. Import EVH.
Module EVCP := EXEC_INV_VALID_CURRENT_PAGE C SMEM ST ACT ST_VALID EI. Import EVCP.
Module EPTNC := EXEC_INV_VALID_PT_NON_CACHEABLE C SMEM ST ACT ST_VALID EI.
  Import EPTNC.
Theorem valid_state_invariant :
  forall (s s' : state) (a : action),
    valid_state s ->
    ACT.valid_exec s a s' ->
    valid_state s'.
Proof.
  intros; STAC.destruct_valid_state; unfold valid_state; crush.

  eapply valid_running_os_not_hypercall_inv; eauto.
  eapply valid_hyper_exec_mode_inv; eauto.
  eapply valid_os_exec_mode_inv; eauto.
  eapply valid_os_idents_inv; eauto.
  eapply valid_hypervisor_inv; eauto.
  eapply valid_pt_non_cacheable_inv; eauto.
  eapply non_cachable_synonyms_inv; eauto.
  eapply valid_virtual_mapping_inv; eauto.
  eapply valid_current_page_inv; eauto.
  eapply injective_hyper_mappings_inv; eauto.
  eapply va_has_valid_pa_inv; eauto.
  eapply valid_cache_inv; eauto.
  eapply valid_cache_va_ma_inv; eauto.
  eapply valid_tlb_inv; eauto.
Qed.

End EXEC_INV.