Require Import State State_valid Tactics State_tactics.

Module Type EXEC_INV_INIT (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM) (ST:STATE C SMEM)
  (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST).

Import PG.
Import C.
Import T.
Import MM.
Import SMEM.
Import ST.
Import ST_VALID.
Module STAC := STAC C SMEM ST ACT ST_VALID.
Import STAC.

Lemma valid_running_os_not_hypercall_inv : forall (s s':state) (a:action),
  ACT.valid_exec s a s' -> 
  running_os_not_hypercall (st_active_os s) (st_aos_activity s) (st_oss s) ->
  running_os_not_hypercall (st_active_os s') (st_aos_activity s') (st_oss s').
Proof.

  STAC.action_induction a; intros;STAC.destruct_action_sem.
  (* Chmod *)
  unfold hyper_running in H.
  unfold running_os_not_hypercall in *.
  intros osi H2 H3.
  unfold osi_hypercall in H1.
  crush.
  
Qed.


Lemma valid_hyper_exec_mode_inv : forall (s s':state) (a:action),
  ACT.valid_exec s a s' -> valid_hyper_exec_mode (st_aos_activity s) (st_exec_mode s) ->
  valid_hyper_exec_mode (st_aos_activity s') (st_exec_mode s').
Proof.
  STAC.action_induction a; intros;
    STAC.destruct_action_sem.
Qed.

Lemma valid_os_exec_mode_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_os_exec_mode (st_active_os s) (st_aos_activity s) (st_exec_mode s) (st_oss s) ->
  valid_os_exec_mode (st_active_os s') (st_aos_activity s') (st_exec_mode s') (st_oss s').
Proof.
  STAC.action_induction a; intros;
    STAC.destruct_action_sem.
Qed.


Lemma valid_os_idents_inv : forall (s s' : state) (a : action),
  ACT.valid_exec s a s' ->
  valid_os_idents (st_oss s) (st_hypervisor s) ->
  valid_os_idents (st_oss s') (st_hypervisor s').
Proof.
  STAC.action_induction a; intros; STAC.destruct_action_sem.
  
  (* new untrusted *)
  unfold valid_os_idents; crush.
  case (os_ident_eq o osi).
  intro.
  subst.
  rewrite <- OSS.get_set.
  exists (OS None (os_current_page x1)).
  split.
  trivial.
  unfold valid_os_idents in H0.
  destruct (H0 osi).
  crush.
  intro.
  rewrite OSS.get_set_distinct.
  unfold valid_os_idents in H0.
  apply H0.
  assumption.

(* del untrusted *)
  unfold valid_os_idents; crush.
  case (os_ident_eq o osi).
  intro.
  subst.
  rewrite <- OSS.get_set.
  exists (OS None (os_current_page x0)).
  crush.
  unfold valid_os_idents in H0.
  destruct (H0 osi).
  crush.
  intro.
  rewrite OSS.get_set_distinct.
  unfold valid_os_idents.
  crush.
  assumption.

  (* Lswitch untrusted *)
  unfold valid_os_idents; crush.
  case (os_ident_eq o osi).
  intro.
  subst.
  rewrite <- OSS.get_set.
  exists (OS None p).
  crush.
  unfold valid_os_idents in H0.
  destruct (H0 osi).
  destruct H5.
  assumption.
  intro.
  rewrite OSS.get_set_distinct.
  crush.
  assumption.

  (* Hcall *)
   unfold valid_os_idents in *.
   crush.
   destruct (H0 osi).
   crush.
   case (os_ident_eq (st_active_os s) osi).
   intro.
   exists (OS (Some h) (os_current_page x)).
   replace osi with (st_active_os s).
   rewrite <- (OSS.get_set).
   crush.
   intro.
   rewrite (OSS.get_set_distinct).
   crush.
   assumption.

   (* page pin untrusted *)
   unfold valid_os_idents; crush.
   case (os_ident_eq o osi).
   intro.
   subst.
   rewrite <- (OSS.get_set).
   exists (OS None (os_current_page x1)).
   crush.
   unfold valid_os_idents in H0.
   destruct (H0 osi).
   destruct H12.
   rewrite <- (HYP.get_set).
   crush.
   intro.
   rewrite (OSS.get_set_distinct).
   rewrite (HYP.get_set_distinct).
   crush.
   assumption.
   assumption.

   unfold valid_os_idents; crush.
   case (os_ident_eq o osi).
   intro.
   subst.
   rewrite <- (OSS.get_set).
   exists (OS None (os_current_page x2)).
   crush.
   unfold valid_os_idents in H0.
   destruct (H0 osi).
   elim H11.
   intros.
   destruct H13.
   rewrite <- (HYP.get_set).
   crush.
   intro.
   rewrite (OSS.get_set_distinct).
   unfold valid_os_idents in H0.
   destruct (H0 osi).
   crush.
   rewrite (HYP.get_set_distinct).
   crush.
   assumption.
   assumption.
Qed.

End EXEC_INV_INIT.