(***********************************************************************
 * File: OS_Isolation.v
 * $Revision $
 * $Date: 2016-09-22 11:57:12 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy $
 ***********************************************************************
 * Proof of os isolation.
 **********************************************************************) 
Require Import State State_valid State_equivalence State_tactics.

Set Undo 20000.

Module OS_ISOLATION (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM)
  (ST:STATE C SMEM) (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST).

Import PG.
Import C.
Import T.
Import MM.
Import SMEM.
Import ST.
Import ST_VALID.

Module STAC := STAC C SMEM ST ACT ST_VALID.
Module STEQ := STEQ C SMEM ST ACT ST_VALID.
Import STEQ.

  Inductive osi_action (osi:os_ident) (s:state) : action -> Prop  :=
  | osi_active : forall (a:action), st_active_os s = osi -> st_aos_activity s = OSA_running ->
    osi_action osi s a
  | new_normal_osi  : forall va pa, osi_action osi s (New_untrusted osi va pa)
  | del_untrusted_osi  : forall va, osi_action osi s (Del_untrusted osi va)
  | lswitch_untrusted_osi  : forall pa, osi_action osi s (Lswitch_untrusted osi pa)
  | pin_untrusted_osi  : forall pa ptype, osi_action osi s (Page_pin_untrusted osi pa ptype)
  | unpin_untrusted_osi  : forall pa, osi_action osi s (Page_unpin_untrusted osi pa)
  | chmod_osi : st_active_os s = osi -> osi_action osi s Chmod
  | switch_from_osi : forall o, st_active_os s = osi -> osi_action osi s (Switch o)
  | switch_to_osi : osi_action osi s (Switch osi).

Lemma shared_page_eq : forall s1 s2 osi va pa ma1 ma2,
  valid_state s1 -> valid_state s2 ->
  osi_equivalent s1 s2 osi ->
  osi_pa_to_ma s1 osi pa = Some ma1 ->
  shared_page s1 va ma1 ->
  osi_pa_to_ma s2 osi pa = Some ma2 ->
  shared_page s2 va ma2.
Proof.
  unfold shared_page; crush.
  rename x into pt_padd.
  rename x0 into osi'.
  rename x1 into pt_page1.
  rename x2 into mp1.
  rename x3 into va1.

  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi).
  STAC.destruct_valid_state_H H.
  unfold valid_hypervisor in H_valid_hyper.
  unfold osi_pa_to_ma in H2; crush.

  assert (osi' = osi); crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi'); crush.
  STAC.destruct_valid_state_H H.
  unfold valid_virtual_mapping in H_valid_virtmap.
  unfold osi_pa_to_pg in H3; crush.
  rename i0 into pt_addr.
  destruct H_valid_virtmap with osi' pt_addr pt_page1 mp1 va1 ma1; crush.
  unfold valid_hypervisor in H_valid_hyper.
  rewrite <- H_valid_hyper with osi' i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold madd_owner; crush.
  remember (ctxt_vadd_accessible ctxt va1); unfold os_accessible in *.
  destruct b; auto.
  unfold madd_owner in H8.
  rewrite <- Heqval in H8.
  rewrite H11 in H8; congruence.

  exists pt_padd.
  exists osi.
  unfold osi_equivalent in H1.
  unfold osi_equivalent in *; crush.
  destruct H10; crush.
  unfold osi_pt_pages_eq in H10.
  destruct H10 with pt_padd pt_page1 mp1; crush.
  unfold madd_owner in H8; crush.
  exists x. exists x0.  exists va1; crush.
  destruct H15 with va1 pa ma1; crush.
Qed.

Lemma no_va_mapped_eq :
  forall s s' pa ma1 ma2 osi,
    osi_pt_pages_eq s' s osi ->
    osi_pa_to_ma s osi pa = Some ma1 ->
    osi_pa_to_ma s' osi pa = Some ma2 ->
    ACT.no_va_mapped_to_ma osi ma1 s ->
    ACT.no_va_mapped_to_ma osi ma2 s'.
Proof.    
  unfold ACT.no_va_mapped_to_ma; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  destruct H with pt_padd pt_page mp; crush.
  unfold osi_pa_to_pg; crush.
  destruct H8 with va pa ma2; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_pg in *; crush.
  elim H2 with pt_padd x va x0; crush.
  unfold osi_pa_to_ma in *; crush.
Qed.

Local Hint Resolve no_va_mapped_eq.

Lemma read_os_isolation_not_osi :
  forall (s s':state) (osi:os_ident) (va:vadd),
    valid_state s ->
    ACT.pre_all_read va s ->
    ACT.post_all_read va s s' ->
    osi_equivalent s s' osi.
Proof.
  unfold ACT.pre_all_read, ACT.post_all_read in *; crush;
  STAC.unify_read_post; crush.
  unfold page_of_vadd_smem in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr0 va = Some m); crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.

  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in *; crush.
  exists pt_page1; crush.
  unfold page_of_vadd_smem in H0; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m) x).
  destruct p; crush.
  destruct o; crush.
  destruct p. destruct c0.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=x) (pg':=p); crush.
  STAC.apply_absurd_cache_get_entry_pt.
  rewrite get_mem_unchanged with (ma:=m) (pg:=x); crush.
  exists mp1; crush.
  unfold page_of_vadd_smem in H0; crush.
  unfold  is_rw_page in *; crush.
  rewrite <- STAC.read_same_mem with (ma':=m) (pt_addr:=i0) (pg:=x) (val:=x1) in *; crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s); exists (os_current_page i1); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H5; crush.

  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in *; crush.
  exists pt_page1; crush.
  unfold page_of_vadd_smem in H0; crush.
  remember (add_cache (smem_c (st_memory s)) (va,m) x).
  destruct p; crush.
  destruct o; crush.
  destruct p. destruct c0.
  assert (m0 <> i0); crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=m) (ma':=i0) (pg:=x) (pg':=p) in H5; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v i0 pt_page1; crush.
  symmetry; apply STAC.cache_deleted_entry_was_there with va m x; crush.
  unfold is_rw_page in *; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=x) (pg':=p) in H5; crush.
  rewrite get_mem_unchanged with (ma:=m) (pg:=x) in H5; crush.
  exists mp1; crush.
  unfold page_of_vadd_smem in H0; crush.
  unfold  is_rw_page in *; crush.
  rewrite <- STAC.read_same_mem with (ma':=m) (pt_addr:=i0) (pg:=x) (val:=x1) in *; crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s); exists (os_current_page i1); crush.
  rewrite STAC.read_same_madd_of_vadd with (va':=va) (pt_addr':=pt_addr0); crush.
  unfold madd_of_vadd_smem, madd_of_vadd, madd_page_table in *; crush.
  rewrite <- H6; crush.
  rewrite <- H6; crush.

  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper; crush.
  destruct (os_ident_eq (st_active_os s) osi); crush.

  unfold osi_hyper_eq; crush.
  rewrite <- STAC.read_no_va_mapped_same_mem with (osi:=osi) (pa:=pa); crush.

  unfold osi_hyper_eq; crush.
  rewrite STAC.read_no_va_mapped_same_mem with (osi:=osi)
    (pa:=pa) (va':=va) (pt_addr':=pt_addr0); crush.
Qed.  

Lemma write_os_isolation_not_osi :
  forall (s s':state) (osi:os_ident) (va:vadd) (val:value),
    valid_state s ->
    ((st_active_os s) <> osi \/ ~ os_accessible va) ->
    ACT.pre_all_write va val s ->
    ACT.post_all_write va val s s' ->
    osi_equivalent s s' osi.
Proof.  
  Local Hint Resolve STAC.madd_of_vadd_smem_then_mmem.

  unfold ACT.pre_all_write, ACT.post_all_write in *; crush;
  STAC.unify_write_post; crush.
  unfold page_of_vadd_smem in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr0 va = Some m); crush;
    clear Heqopt_res.
  assert (page_owned_by x <> Os_PO osi); crush.
  unfold madd_of_vadd, madd_page_table in H2; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct H_valid_virtmap with (st_active_os s) pt_addr0 p0 p1 va m; crush.
  rewrite <- H_valid_hyper with (st_active_os s) i0 (os_current_page i) pt_addr0;
    crush.
  unfold madd_owner; crush.
  destruct H0; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va m x; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  rewrite H7 in *; crush.
  rewrite H8 in *; crush.
  rewrite H8 in *; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va m x; crush.
  
  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in *; crush.
  exists pt_page1; crush.
  apply STAC.set_mem_unchanged_pt_page with (ma:=m) (old_pg:=x); crush.
  unfold is_pt_page; crush.
  exists mp1; crush.
  rewrite STAC.write_smem_mem_neq with (ma':=m) (pg:=x) (pt_addr:=i0);
    crush.

  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H7; crush.
  unfold osi_pa_to_ma in H9; crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi i pa m in H5; crush.
  unfold madd_owner in *; crush.
  unfold page_of_vadd_madd_smem in H1; crush;
    try destruct H_valid_cache with va m x; crush.
  unfold page_of_vadd_madd_smem in H1; crush;
    try destruct H_valid_cache with va m x; crush.

  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  assert (get_mem (smem_mm (st_memory s)) i0 = Some pt_page1).
  eapply STAC.set_mem_old_pt_page; crush.
  unfold is_pt_page; crush.
  exists pt_page1; crush.
  exists mp1; crush.
  rewrite STAC.write_smem_mem_neq with (ma':=m) (pg:=x) (pt_addr:=i0) in H12;
    crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H8; crush.
  STAC.destruct_valid_state.
  rewrite <- H_valid_hyper with osi i pa m in H5; crush.
  unfold madd_owner in H5; crush.
  unfold page_of_vadd_madd_smem in H1; crush;
    destruct H_valid_cache with va m x; crush.
  unfold page_of_vadd_madd_smem in H1; crush;
    destruct H_valid_cache with va m x; crush.

  unfold osi_equivalent, osi_equivalent_smem, osi_equivalent_hyper; crush.
  destruct (os_ident_eq osi (st_active_os s)); crush.

  unfold osi_hyper_eq; crush.
  rewrite <- STAC.write_no_va_mapped_same_mem with (osi:=osi) (pa:=pa); crush.

  unfold osi_hyper_eq; crush.
  rewrite STAC.write_no_va_mapped_same_mem with (osi:=osi)
    (pa:=pa) (va':=va) (pt_addr':=pt_addr0) (val:=val); crush.
Qed.

Lemma del_untrusted_os_isolation_not_osi:
  forall (s s':state) (osi o:os_ident) (va:vadd),
    valid_state s ->
    o <> osi ->
    ACT.pre_del o va s ->
    ACT.post_del o va s s' ->
    osi_equivalent s s' osi.
Proof.  
  intros.

  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.

  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H3; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s' osi pt_padd = Some i0); crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  generalize H1.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  destruct (madd_eq i0 x); crush.
  (* i0 = x *)
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO o); crush.
  unfold ACT.post_del in *; crush.
  destruct H_valid_os_idents with o; crush.
  STAC.get_osi_current_pt_page o; crush.
  unfold osi_current_page_table in *; crush.
  (* i0 <> x *)
  exists pt_page1; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' i0 = Some pt_page1); crush.
  unfold remove_page_table_entry in *; crush.
  unfold madd_of_vadd, madd_page_table in H12; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct; crush.
  destruct (get_cache (smem_c (st_memory s)) (va,x0)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x0) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va i0 p; crush.
  unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x0); crush.
  unfold remove_page_table_entry in *; crush.
  unfold madd_of_vadd, madd_page_table in H12; crush.
  destruct (page_content p0) eqn:?; crush.
  exists mp1; crush.
  erewrite <- STAC.del_osi_pa_to_ma; crush.
  erewrite <- STAC.del_page_of_vadd_madd_smem_ne_osi; crush.
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state; crush.

  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H4; crush.
  assert (osi_pa_to_ma s' osi pt_padd = Some i0); crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); crush.
  erewrite STAC.del_osi_pa_to_ma; crush.
  generalize H1.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  destruct (madd_eq i0 x); crush.
  (* i0 = x *)
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO osi); crush.
  destruct H_valid_os_idents with o; crush.
  STAC.get_osi_current_pt_page o; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_owner in *; crush.
  (* i0 <> x *)
  exists pt_page1; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' i0 = Some pt_page1).
  destruct (get_cache (smem_c (st_memory s)) (va,x0)) eqn:?.
  assert (madd_type (smem_mm (delete (st_memory s) x1 va)) i0 =
    madd_type (smem_mm (st_memory s)) i0).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type in *; crush.
  rewrite <- H5 in *; crush.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x0) (pg:=p) in *; crush.
  rewrite get_add_mem_distinct in H4; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va i0 p; unfold is_rw_page in *; crush.
  rewrite H17 in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x0) in *; crush.

  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H13; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct in *; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H13; crush.
  destruct (page_content p0) eqn:?; crush.
  exists mp1; crush.
  erewrite STAC.del_osi_pa_to_ma; crush.
  erewrite STAC.del_page_of_vadd_madd_smem_ne_osi; crush.
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state; crush.
  unfold ACT.post_del in *; crush.

  unfold osi_equivalent; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  destruct (os_ident_eq osi (st_active_os s)); crush. 
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  rewrite OSS.get_set_distinct; crush.
  unfold osi_equivalent_smem; crush.

  unfold osi_equivalent_hyper; crush.
  unfold osi_hyper_eq; crush.
  assert (osi_pa_to_ma s' osi pa = Some ma1); crush.
  unfold ACT.pre_del, ACT.post_del in *; crush.
  unfold ACT.pre_del, ACT.post_del in *; crush.
  erewrite <- STAC.del_no_va_mapped_same_mem; crush.
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO o); crush.
  unfold osi_current_page_table in *; crush.

  unfold osi_hyper_eq; crush.
  assert (osi_pa_to_ma s osi pa = Some ma1); crush.
  unfold ACT.pre_del, ACT.post_del in *; crush.
  unfold ACT.pre_del, ACT.post_del in *; crush.
  erewrite STAC.del_no_va_mapped_same_mem; crush. 
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO o); crush.
  unfold osi_current_page_table in *; crush.
Qed.

Lemma del_hyper_os_isolation_not_osi:
  forall (s s':state) (osi:os_ident) (va:vadd),
    valid_state s ->
    ACT.pre_del_hyper va s ->
    ACT.post_del_hyper va s s' ->
    osi_equivalent s s' osi.
Proof.  
  intros.
  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.
  
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H2; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s' osi pt_padd = Some i0); crush.
  erewrite <- STAC.del_hyper_osi_pa_to_ma; crush.
  generalize H0.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  destruct (madd_eq i0 pt_addr); crush.
  (* i0 = pt_addr *)
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi); crush.
  exists (Page (PT_c (PT.remove pt va)) (page_owned_by pt_page1) (page_cacheable pt_page1));
    crush.
  unfold osi_pa_to_pg in *; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H10 in *; crush.
  injection H3; intros; subst mp1; clear H3.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' x = Some (Page (PT_c (PT.remove pt va))
    (page_owned_by pt_page1) (page_cacheable pt_page1))); crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H10 in *; crush.
  rewrite <- MM.get_add_mem; crush.
  destruct (get_cache (smem_c (st_memory s)) (va,x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va x p; unfold is_rw_page in *; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H10; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H10; crush.
  unfold remove_page_table_entry in *; crush.
  rewrite H10 in *; crush.
  exists (PT.remove pt va); crush.
  exists ma1; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H10 in *; crush.
  injection H3; intros; subst mp1; clear H3.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  rewrite PT.get_remove_distinct with (inf:=ma1); crush.
  destruct H_valid_virtmap with (st_active_os s) x pt_page1 pt va x2; crush;
    unfold madd_owner in *; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper in *; crush.
  erewrite <- STAC.del_hyper_page_of_vadd_madd_smem_ne_osi; crush.
  unfold madd_type, is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem in H13; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  rewrite H17; crush.
  rewrite H14; crush.
  (* i0 <> pt_addr *)
  exists pt_page1; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del_hyper in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' i0 = Some  pt_page1).
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H10 in *; crush.
  rewrite MM.get_add_mem_distinct; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del in *; crush.
  destruct (get_cache (smem_c (st_memory s)) (va,x2)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p); crush.
  rewrite get_add_mem_distinct; crush.
  destruct H_valid_cache with va i0 p; unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2); crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H10 in *; crush.
  exists mp1; crush.
  erewrite <- STAC.del_hyper_osi_pa_to_ma; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold osi_pa_to_ma in *; crush.
  erewrite <- STAC.del_hyper_page_of_vadd_madd_smem_ne_osi; crush.
  unfold madd_type; crush.
  unfold page_of_vadd_madd_smem in H13; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  rewrite H18; crush.
  unfold is_rw_page in *; crush.
  rewrite H14; crush.

  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H3; crush.
  assert (osi_pa_to_ma s' osi pt_padd = Some i0); crush.
  unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some i0); crush.
  erewrite STAC.del_hyper_osi_pa_to_ma; crush.
  generalize H0.
  unfold ACT.pre_del_hyper, ACT.pre_all_del; crush.
  destruct (madd_eq i0 x); crush.
  (* i0 = x *)
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) x = Os_PO osi); crush.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H10; crush.
  rewrite H15 in *; crush.
  exists pt_page; crush.
  unfold osi_pa_to_pg; crush.
  exists pt; crush.
  exists ma1; crush.
  unfold ACT.post_del_hyper in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x1 va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' x = Some pt_page1).
  unfold ACT.pre_del_hyper, ACT.pre_all_del in *; crush.
  destruct (get_cache (smem_c (st_memory s)) (va,x3)) eqn:?.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x3) (pg:=p) in H3; crush.
  rewrite get_add_mem_distinct in H3; crush.
  destruct H_valid_cache with va x p; unfold is_rw_page in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x3) in H3; crush.

  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H15 in *; crush.
  rewrite <- MM.get_add_mem in H17; crush.
  injection H4; intros; subst mp1; clear H4.
  rewrite PT.get_remove_old with (k':=va) (inf:=ma1); crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H15 in *; crush.
  unfold ACT.post_del_hyper in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold ACT.post_del_hyper in *; crush.
  erewrite STAC.del_hyper_page_of_vadd_madd_smem_ne_osi; crush.
  assert (madd_type (smem_mm (st_memory s')) ma1 = Some RW).
  unfold page_of_vadd_madd_smem in H17; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H15 in *; crush.
  rewrite delete_cache with (ma:=x0) in Heqval; crush.
  destruct (eq_cache_key (va,x0) (va0,ma1)); crush.
  rewrite get_remove_cache in Heqval; crush.
  destruct (get_cache (smem_c (st_memory s)) (va0,ma1)) eqn:?; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in Heqval; crush.
  rewrite STAC.del_hyper_same_madd_type; crush.
  unfold madd_type; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  rewrite H22; crush.
  rewrite get_remove_cache_old with (k':=(va,x0)) (pg:=pg) in Heqo; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H15; crush.
  unfold madd_type; crush.
  unfold is_rw_page in *; crush.
  rewrite H18; crush.
  unfold ACT.post_del_hyper in *; crush.
  erewrite <- STAC.del_hyper_same_madd_type; crush.
  (* i0 <> x *)
  exists pt_page1; crush.
  unfold osi_pa_to_pg; crush.
  unfold ACT.post_del_hyper in *; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del in *; crush.
  destruct (remove_page_table_entry (smem_mm (st_memory s)) x va) eqn:?; crush.
  rename m into mm'.
  assert (get_mem mm' i0 = Some pt_page1).
  destruct (get_cache (smem_c (st_memory s)) (va,x2)) eqn:?.
  assert (madd_type (smem_mm (delete (st_memory s) x va)) i0 =
    madd_type (smem_mm (st_memory s)) i0).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type in *; crush.
  rewrite <- H4 in *; crush.
  rewrite delete_mem_some_entry with (mm':=mm') (ma:=x2) (pg:=p) in H3; crush.
  rewrite get_add_mem_distinct in H3; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with va i0 p; unfold is_rw_page in *; crush.
  rewrite H14 in *; crush.
  rewrite delete_mem_no_entry with (mm':=mm') (ma:=x2) in H3; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H12; crush.
  destruct (page_content p) eqn:?; crush.
  rewrite MM.get_add_mem_distinct in *; crush.
  unfold remove_page_table_entry in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in H12; crush.
  destruct (page_content p0) eqn:?; crush.
  exists mp1; crush.
  erewrite STAC.del_hyper_osi_pa_to_ma; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold ACT.post_del_hyper in *; crush.
  erewrite STAC.del_hyper_page_of_vadd_madd_smem_ne_osi; crush.
  assert (madd_type (smem_mm (st_memory s')) ma1 = Some RW).
  2:unfold ACT.post_del_hyper in *; crush;
    erewrite  STAC.del_hyper_same_madd_type in H16; crush.
  unfold page_of_vadd_madd_smem in H13; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper in *; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  rewrite delete_cache with (ma:=x2) in Heqval; crush.
  destruct (eq_cache_key (va,x2) (va0,ma1)); crush.
  rewrite get_remove_cache in Heqval; crush.
  destruct (get_cache (smem_c (st_memory s)) (va0,ma1)) eqn:?; crush.
  rewrite get_remove_cache_distinct with (pg:=p) in Heqval; crush.
  rewrite STAC.del_hyper_same_madd_type; crush.
  unfold madd_type; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  rewrite H18; crush.
  rewrite get_remove_cache_old with (k':=(va,x2)) (pg:=pg) in Heqo; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite Heqc; crush.
  unfold madd_type; crush.
  unfold is_rw_page in *; crush.
  rewrite H14; crush.

  unfold osi_equivalent; crush.
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper in *; crush.
  destruct (os_ident_eq osi (st_active_os s)); crush. 
  unfold ACT.pre_del_hyper, ACT.pre_all_del, ACT.post_del_hyper in *; crush.
  unfold osi_equivalent_smem; crush.

  unfold osi_equivalent_hyper; crush.
  unfold osi_hyper_eq; crush.
  assert (osi_pa_to_ma s' osi pa = Some ma1); crush.
  unfold ACT.pre_del_hyper, ACT.post_del_hyper in *; crush.
  unfold ACT.pre_del_hyper, ACT.post_del_hyper in *; crush.
  erewrite <- STAC.del_no_va_mapped_same_mem; crush. 
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  unfold osi_current_page_table, is_rw_page in *; crush.

  unfold osi_hyper_eq; crush.
  assert (osi_pa_to_ma s osi pa = Some ma1); crush.
  unfold ACT.pre_del_hyper, ACT.post_del_hyper in *; crush.
  unfold ACT.pre_del_hyper, ACT.post_del_hyper in *; crush.
  erewrite STAC.del_no_va_mapped_same_mem; crush. 
  assert (madd_type (smem_mm (delete (st_memory s) x va)) x =
    madd_type (smem_mm (st_memory s)) x).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type, is_rw_page in *; crush.
  rewrite H4 in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H15 in *; crush.
Qed.

Lemma new_untrusted_os_isolation_not_osi:
  forall (s s':state) (osi o:os_ident) (va:vadd) (pa:padd),
    valid_state s ->
    o <> osi ->
    ACT.pre_new o va pa s ->
    ACT.post_all_new o va pa s s' ->
    osi_equivalent s s' osi.
Proof.  
  Opaque ACT.new.
  unfold ACT.pre_new, ACT.post_new, ACT.post_all_new in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) x2 = Os_PO o); crush.
  unfold osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state.
  rewrite H_valid_hyper with o i pa x2; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with o; crush.
  STAC.get_osi_current_pt_page o; crush.
  unfold osi_current_page_table in *; crush.

  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.

  unfold osi_pt_pages_eq, osi_pa_to_pg in *; crush.
  exists pt_page1; crush.
  rewrite STAC.new_get_mem_distinct_pt; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  unfold madd_type; crush.
  rewrite <- H13; crush.
  exists mp1; crush.
  rewrite STAC.new_page_ne; crush.
  unfold osi_pa_to_ma in H17; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  unfold is_rw_page in *; crush.
  exists o; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H14; crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; exists i0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite <- H13; crush.

  unfold osi_pt_pages_eq, osi_pa_to_pg; crush.
  exists pt_page1; crush.
  rewrite STAC.new_get_mem_distinct_pt in H13; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type with (osi:=o) (pt_addr':=x0) (va':=va) (ma':=x2); 
    crush.
  unfold madd_type; crush.
  rewrite <- H16; crush.
  assert (madd_owner (smem_mm (st_memory s)) x2 = Os_PO o); crush.
  unfold madd_owner in *; crush.

  exists mp1; crush.
  rewrite STAC.new_page_ne in H19; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists o; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H14; crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; exists i0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite STAC.new_get_mem_distinct_pt in H13; crush.
  rewrite <- H16; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO o); crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type with (osi:=o) (pt_addr':=x0) (va':=va) (ma':=x2); crush.
  unfold madd_type; crush.
  rewrite <- H16; crush.
  unfold madd_owner in *; crush.

  unfold osi_equivalent, osi_equivalent_smem; crush.
  destruct (os_ident_eq osi (st_active_os s)); crush.
  rewrite OSS.get_set_distinct; crush.

  unfold osi_equivalent_hyper; crush.
  unfold osi_hyper_eq; crush.
  erewrite <- STAC.new_no_va_mapped_same_mem with (osi:=osi) (pa:=pa0); crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO o); crush.
  unfold madd_owner in *; crush.

  unfold osi_hyper_eq; crush.
  erewrite STAC.new_no_va_mapped_same_mem; crush. 
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO o); crush.
  unfold madd_owner in *; crush.
Qed.  


Lemma new_hyper_os_isolation_not_osi:
  forall (s s':state) (osi:os_ident) (va:vadd) (ma:madd),
    valid_state s ->
    ACT.pre_new_hyper va ma s ->
    ACT.post_new_hyper va ma s s' ->
    osi_equivalent s s' osi.
Proof.  
  unfold ACT.pre_new_hyper, ACT.post_new_hyper in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with (st_active_os s); crush.
  STAC.get_osi_current_pt_page (st_active_os s); crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.

  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.
   
  unfold osi_pt_pages_eq, osi_pa_to_pg in *; crush.
  destruct (madd_eq i0 x); crush.
  exists (Page (PT_c (PT.set pt va ma))
    (page_owned_by pt_page1) (page_cacheable pt_page1)); crush.
  erewrite STAC.new_get_mem_pt_addr; crush.
  unfold madd_owner in *; crush.
  exists (PT.set pt va ma); crush.
  exists ma1; crush.
  rewrite PT.get_set_distinct; crush. 
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold osi_pa_to_ma in *; crush.
  unfold madd_owner in H13; crush.
  destruct H_valid_virtmap with osi x pt_page1 mp1 va0 ma1; crush.
  rewrite <- H_valid_hyper with osi i pt_padd x; crush.
  unfold madd_owner; crush.
  rewrite STAC.new_page_ne; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H10; crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; exists x; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.

  (* i0 <> x *)
  exists pt_page1; crush.
  rewrite STAC.new_get_mem_distinct_pt; crush.
  assert (madd_owner (smem_mm (st_memory s)) i0 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold madd_type; crush.
  rewrite <- H9; crush.
  exists mp1; crush.
  rewrite STAC.new_page_ne; crush.
  unfold osi_pa_to_ma in H13; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  unfold is_rw_page in *; crush.
  exists (st_active_os s); exists (os_current_page x0); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H10; crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; exists i0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite <- H9; crush.

  unfold osi_pt_pages_eq, osi_pa_to_pg in *; crush.
  destruct (madd_eq i0 x); crush.
  erewrite STAC.new_get_mem_pt_addr in H9; crush.
  2: unfold madd_owner in *; crush.
  exists pt; crush.
  exists ma1; crush.
  unfold osi_pa_to_ma in *; crush.
  injection H12; intros; subst mp1; clear H12.
  destruct (vadd_eq va va0); crush;
    [rewrite <- PT.get_set in H9 | rewrite PT.get_set_distinct in H9]; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_type (smem_mm (ACT.new s (st_active_os s) x va ma)) ma1 = Some RW).
  unfold page_of_vadd_madd_smem in H14; crush.
  Transparent ACT.new.
  unfold ACT.new in Heqval.
  Opaque ACT.new.
  crush.
  rewrite new_cache_shared in Heqval; crush.
  destruct (madd_eq ma ma1); crush.
  rewrite get_remove_ma_cache_eq in Heqval; crush.
  rewrite get_remove_ma_cache_ne in Heqval; crush.
  unfold remove_new_va_cache in Heqval; crush.
  destruct (eq_cache_key (va,m) (va0,ma1)); crush.
  rewrite get_remove_cache in Heqval; crush.
  destruct (get_cache (smem_c (st_memory s)) (va0,ma1)) eqn:?; crush.
  rewrite get_remove_cache_distinct with (pg:=p0) in Heqval; crush.
  erewrite STAC.new_shared_same_madd_type; crush.
  unfold madd_type; crush.
  destruct H_valid_cache with va0 ma1 p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H18; crush.
  unfold madd_owner in *; crush.
  
  rewrite get_remove_cache_old with (k':=(va,m)) (pg:=pg) in Heqo; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.new_shared_same_madd_type; crush.
  unfold madd_type; crush.
  rewrite H18; crush.
  unfold madd_owner in *; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.new_shared_same_madd_type; crush.
  unfold madd_type; crush.
  rewrite H18; crush.
  unfold madd_owner in *; crush.
  (* not shared *)
  rewrite new_cache_not_shared in Heqval; crush.
  unfold remove_new_va_cache in Heqval; crush.
  destruct (eq_cache_key (va,m) (va0,ma1)); crush.
  rewrite get_remove_cache in Heqval; crush.
  destruct (get_cache (smem_c (st_memory s)) (va0,ma1)) eqn:?; crush.
  rewrite get_remove_cache_distinct with (pg:=p0) in Heqval; crush.
  erewrite STAC.new_not_shared_same_madd_type; crush.
  unfold madd_type; crush.
  destruct H_valid_cache with va0 ma1 p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H18; crush.
  unfold madd_owner in *; crush.
  rewrite get_remove_cache_old with (k':=(va,m)) (pg:=pg) in Heqo; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.new_not_shared_same_madd_type; crush.
  unfold madd_type; crush.
  rewrite H18; crush.
  unfold madd_owner in *; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  erewrite STAC.new_not_shared_same_madd_type; crush.
  unfold madd_type; crush.
  rewrite H18; crush.
  unfold madd_owner in *; crush.
  unfold madd_type; crush.
  unfold is_rw_page in *; crush.
  rewrite H15; crush.
  rewrite STAC.new_page_ne in H14; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  erewrite STAC.new_same_madd_type in H16; crush.
  unfold madd_type in H16; crush.
  rewrite H10 in *; crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H10; crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; exists x; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite H10; crush.
  injection H12; intros; subst mp1; crush.
  rewrite PT.get_set_distinct in H9; crush.
  rewrite <- PT.get_set in H9; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.

  (* i0 <> x *)  
  exists pt_page1; crush.
  rewrite STAC.new_get_mem_distinct_pt in H9; crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type with (osi:=st_active_os s) (pt_addr':=x) (va':=va) (ma':=ma); crush.
  unfold madd_type; crush.
  rewrite <- H12; crush.
  unfold madd_owner in *; crush.
  exists mp1; crush.
  erewrite STAC.new_page_ne in H15; crush.
  4: unfold madd_owner in *; crush.
  unfold osi_pa_to_ma in H14; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  assert (madd_type (smem_mm (ACT.new s (st_active_os s) ma1 va ma)) ma1 = Some RW).
  unfold page_of_vadd_madd_smem in H15; crush.
  Transparent ACT.new.
  unfold ACT.new in Heqval.
  Opaque ACT.new.
  crush.
  rewrite new_cache_shared in Heqval; crush.
  destruct (madd_eq ma ma1); crush.
  rewrite get_remove_ma_cache_eq in Heqval; crush.
  rewrite get_remove_ma_cache_ne in Heqval; crush.
  unfold remove_new_va_cache in Heqval; crush.
  destruct (eq_cache_key (va,m) (va0,ma1)); crush.
  rewrite get_remove_cache in Heqval; crush.
  destruct (get_cache (smem_c (st_memory s)) (va0,ma1)) eqn:?; crush.
  rewrite get_remove_cache_distinct with (pg:=p0) in Heqval; crush.
  erewrite STAC.new_shared_same_madd_type; crush.
  unfold madd_type; crush.
  destruct H_valid_cache with va0 ma1 p0; crush.
  unfold is_rw_page in *; crush.
  unfold madd_owner in *; crush.
  rewrite get_remove_cache_old with (k':=(va,m)) (pg:=pg) in Heqo; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  (* not shared *)
  rewrite new_cache_not_shared in Heqval; crush.
  unfold remove_new_va_cache in Heqval; crush.
  destruct (eq_cache_key (va,m) (va0,ma1)); crush.
  rewrite get_remove_cache in Heqval; crush.
  destruct (get_cache (smem_c (st_memory s)) (va0,ma1)) eqn:?; crush.
  rewrite get_remove_cache_distinct with (pg:=p0) in Heqval; crush.
  erewrite STAC.new_not_shared_same_madd_type; crush.
  unfold madd_type; crush.
  destruct H_valid_cache with va0 ma1 p0; crush.
  unfold is_rw_page in *; crush.
  unfold madd_owner in *; crush.
  rewrite get_remove_cache_old with (k':=(va,m)) (pg:=pg) in Heqo; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.
  destruct H_valid_cache with va0 ma1 pg; crush.
  unfold is_rw_page in *; crush.

  unfold madd_type; crush.
  unfold is_rw_page in *; crush.
  rewrite H16; crush.
  erewrite STAC.new_same_madd_type in H17; crush.
  unfold madd_type in H17; crush.
  rewrite H10 in H17; crush.
  unfold madd_owner in *; crush.
  exists (st_active_os s); exists (os_current_page x0); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H10; crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; exists i0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite STAC.new_get_mem_distinct_pt in H9; crush.
  rewrite <- H12; crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type with (osi:=st_active_os s) (pt_addr':=x) (va':=va) (ma':=ma); crush.
  unfold madd_type; crush.
  rewrite <- H12; crush.
  unfold madd_owner in *; crush.

  unfold osi_equivalent, osi_equivalent_smem, osi_equivalent_hyper; crush.
  destruct (os_ident_eq osi (st_active_os s)); crush.

  unfold osi_equivalent_hyper; crush.
  unfold osi_hyper_eq; crush.
  erewrite <- STAC.new_no_va_mapped_same_mem with (osi:=osi) (pa:=pa); crush.
  unfold is_rw_page in *; crush.
  unfold madd_owner in *; crush.

  unfold osi_hyper_eq; crush.
  erewrite STAC.new_no_va_mapped_same_mem; crush.
  assert (madd_type (smem_mm (ACT.new s (st_active_os s) ma1 va ma)) ma1 = Some PT); crush.
  erewrite STAC.new_same_madd_type; crush.
  unfold madd_type; crush.
  rewrite H10; crush.
  unfold madd_owner in *; crush.
  unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H12 in *; crush.
  unfold madd_owner in *; crush.
Qed.  

Lemma pin_os_isolation_not_osi:
  forall (s s':state) (osi o:os_ident) (pa:padd) (t:ptype),
    valid_state s ->
    o <> osi ->
    ACT.pre_pin o pa t s ->
    ACT.post_pin o pa t s s' ->
    osi_equivalent s s' osi.
Proof.
  Opaque new_page.
  Local Hint Resolve STAC.page_pin_osi_pa_to_ma
    STAC.page_pin_osi_pa_to_ma_distinct_osi
    STAC.page_pin_osi_pa_to_ma_distinct_osi
    STAC.page_pin_madd_of_vadd STAC.page_pin_madd_of_vadd
    STAC.page_pin_page_of_vadd_madd_smem
    STAC.page_pin_page_of_vadd_madd_smem_old.
  intros.
  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.
  
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H3; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s osi pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s' osi pt_padd = Some pt_addr); crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  destruct (madd_eq x pt_addr); crush;
    rewrite MM.get_add_mem_distinct; crush.
  exists mp1; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  unfold ACT.pre_pin, ACT.post_pin,
    osi_pa_to_ma, page_of_vadd_madd_smem in *; crush.
  destruct (madd_eq x ma1); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  STAC.destruct_valid_state.
  rewrite H_valid_hyper with osi i pa0 ma1; crush.
  unfold madd_owner in H2; crush.
  rewrite MM.get_add_mem_distinct; crush.

  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H4; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s' osi pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some pt_addr); crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  rewrite get_add_mem_distinct in H4; crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi); crush.
  STAC.destruct_valid_state; crush.
  unfold madd_owner in *; crush.
  exists mp1; crush.
  unfold ACT.post_pin, osi_pa_to_ma in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  unfold ACT.pre_pin, ACT.post_pin,
    osi_pa_to_ma, page_of_vadd_madd_smem in *; crush.
  rewrite get_add_mem_distinct in H10; crush.
  STAC.destruct_valid_state; crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  rewrite HYP.get_set_distinct in *; crush.
  unfold madd_owner in H24; crush.

  unfold osi_equivalent; crush.
  unfold ACT.post_pin in *; crush.
  destruct (os_ident_eq (st_active_os s) osi); crush.

  unfold ACT.pre_pin, ACT.post_pin in *; crush;
    rewrite OSS.get_set_distinct; crush.

  unfold osi_equivalent_smem; crush.
  
  unfold osi_equivalent_hyper; crush.
  unfold osi_hyper_eq; crush.
  exists ma1; crush.
  erewrite <- STAC.pin_same_get_mem; crush.
  unfold osi_hyper_eq; crush.
  exists ma1; crush.
  erewrite STAC.pin_same_get_mem; crush.
Qed.

Lemma unpin_os_isolation_not_osi:
  forall (s s':state) (osi o:os_ident) (pa:padd),
    valid_state s ->
    o <> osi ->
    ACT.pre_unpin o pa s ->
    ACT.post_unpin o pa s s' ->
    osi_equivalent s s' osi.
Proof.
  Local Hint Resolve STAC.page_unpin_osi_pa_to_ma_distinct_osi
    STAC.page_unpin_osi_pa_to_ma_distinct_pa
    STAC.page_unpin_osi_pa_to_ma_old
    STAC.page_unpin_osi_pa_to_pg_distinct_osi
    STAC.page_unpin_osi_pa_to_pg_distinct_pa
    STAC.page_unpin_osi_pa_to_pg_old
    STAC.page_unpin_madd_of_vadd_distinct_pt_padd
    STAC.page_unpin_madd_of_vadd_distinct_osi
    STAC.page_unpin_madd_of_vadd_old
    STAC.page_unpin_page_of_vadd_madd_smem
    STAC.page_unpin_page_of_vadd_madd_smem_old.
  intros.
  match goal with
      |[|-osi_equivalent ?s ?s' ?osi] =>
       assert (osi_pt_pages_eq s s' osi);
         [|assert (osi_pt_pages_eq s' s osi)]
                                           
  end.

  unfold ACT.pre_unpin in *.
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H3; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s osi pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s' osi pt_padd = Some pt_addr); crush.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_pa_to_ma in *; crush.
  rewrite HYP.get_set_distinct; crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  destruct (madd_eq x2 pt_addr); crush.
  (* x2 = pt_addr *)
  STAC.destruct_valid_state.
  rewrite HYP.get_set_distinct in *; crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO o); crush.
  rewrite MM.get_add_mem_distinct; crush.
  exists mp1; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin,
    osi_pa_to_ma, page_of_vadd_madd_smem in *; crush.
  destruct (madd_eq x2 ma1); crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO o); crush.
  rewrite MM.get_add_mem_distinct; crush.

  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg in H4; crush.
  rename i0 into pt_addr.
  assert (osi_pa_to_ma s' osi pt_padd = Some pt_addr); unfold osi_pa_to_ma; crush.
  assert (osi_pa_to_ma s osi pt_padd = Some pt_addr); crush.
  exists pt_page1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  rewrite get_add_mem_distinct in H4; crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO o); crush.
  exists mp1; crush.
  unfold ACT.post_unpin, osi_pa_to_ma in *; crush.
  rewrite HYP.get_set_distinct in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin,
    osi_pa_to_ma, page_of_vadd_madd_smem in *; crush.
  rewrite get_add_mem_distinct in H10; crush.
  rewrite HYP.get_set_distinct in *; crush.
  STAC.destruct_valid_state.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma1 = Os_PO o); crush.
  

  unfold osi_equivalent; crush.
  unfold ACT.post_unpin in *; crush.
  destruct (os_ident_eq (st_active_os s) osi); crush.

  unfold ACT.pre_unpin, ACT.post_unpin in *; crush;
    rewrite OSS.get_set_distinct; crush.

  unfold osi_equivalent_smem; crush.

  unfold osi_equivalent_hyper; crush.
  unfold osi_hyper_eq; crush.
  exists ma1; crush.
  erewrite <- STAC.unpin_no_va_mapped_same_mem
    with (osi':= o); crush.
  unfold osi_hyper_eq; crush.
  exists ma1; crush.
  erewrite STAC.unpin_no_va_mapped_same_mem with (osi':=o); crush.
Qed.  

Lemma ctxt_switch_eq_smem :
  forall s s' osi,
    st_hypervisor s' = st_hypervisor s ->
    st_memory s' = ctxt_switch (st_memory s) ->
    osi_equivalent_smem s s' osi.
Proof.                              
  unfold osi_equivalent_smem, osi_pt_pages_eq,
    osi_pa_to_ma, osi_pa_to_pg, madd_of_vadd, madd_page_table,
    page_of_vadd_madd_smem; simpl; crush;
    try rewrite H in *; try rewrite H0 in *; crush;
    try rewrite ctxt_switch_mem in *; try rewrite ctxt_switch_cache in *;
    try rewrite ctxt_switch_tlb in *; intros; repeat split; intros;
    repeat eexists; crush.
Qed.  

Lemma ctxt_switch_eq_hyper :
  forall s s' osi,
    st_hypervisor s' = st_hypervisor s ->
    st_memory s' = ctxt_switch (st_memory s) ->
    osi_equivalent_hyper s s' osi.
Proof.                              
  unfold osi_equivalent_hyper, osi_hyper_eq,
    osi_pa_to_ma, osi_pa_to_pg, madd_of_vadd, madd_page_table,
    page_of_vadd_madd_smem; simpl; crush;
    try rewrite H in *; try rewrite H0 in *; crush;
    try rewrite ctxt_switch_mem in *; try rewrite ctxt_switch_cache in *;
    try rewrite ctxt_switch_tlb in *; intros; repeat split; intros;
    repeat eexists; crush.
  eapply no_va_mapped_eq; crush.
  apply ctxt_switch_eq_smem; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H; crush.
  eapply no_va_mapped_eq; crush.
  apply ctxt_switch_eq_smem with (s:=s); crush.
  unfold osi_pa_to_ma; crush.
  rewrite H; crush.
  unfold osi_pa_to_ma; crush.
Qed.  


Theorem os_isolation_not_osi :
  forall (s s':state) (osi:os_ident) (a:action),
    ACT.valid_exec s a s' ->
    valid_state s ->
    ~ osi_action osi s a ->
    STEQ.osi_equivalent s s' osi.
Proof.
  intros.
  Local Hint Resolve osi_equivalent_refl
    read_os_isolation_not_osi
    write_os_isolation_not_osi
    new_untrusted_os_isolation_not_osi
    new_hyper_os_isolation_not_osi
    del_untrusted_os_isolation_not_osi
    del_hyper_os_isolation_not_osi
    pin_os_isolation_not_osi
    unpin_os_isolation_not_osi.
  Local Hint Constructors osi_action.

  STAC.action_induction a;
    unfold ACT.valid_exec in *; crush;
    unfold ACT.post_silent, ACT.pre_read, ACT.pre_read_hyper,
      ACT.post_read, ACT.post_read_hyper, ACT.pre_write, ACT.pre_write_hyper,
      ACT.post_write, ACT.post_write_hyper in *;
    crush; try (elim H1; crush; fail).

  (* Write *)
  eapply write_os_isolation_not_osi; crush.

  (* Write hyper *)
  eapply write_os_isolation_not_osi; crush.
  right; unfold ACT.pre_all_write in *; crush.
  unfold aos_page_owner in *; crush.
  unfold aos_page_of_vadd_smem, osi_page_of_vadd_smem in *; crush.
  unfold page_of_vadd_smem in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) m v = Some m0).
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct H_valid_virtmap with (st_active_os s) m p0 p1 v m0; crush. 
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with v m0 x; crush.

  (* New untrusted *)
  eapply new_untrusted_os_isolation_not_osi; crush; crush.

  (* Del *)
  eapply del_untrusted_os_isolation_not_osi; crush; crush.
    
  (* Switch *)
  unfold ACT.pre_switch, ACT.post_switch in *; crush.
  unfold osi_equivalent; crush.
  destruct (os_ident_eq osi (st_active_os s)); right; crush.
  apply ctxt_switch_eq_smem; crush.
  apply ctxt_switch_eq_hyper; crush.

  (* Lswitch *)
  unfold ACT.pre_lswitch, ACT.post_lswitch in *; crush.
  unfold osi_equivalent; crush.
  destruct (os_ident_eq (st_active_os s) osi); crush.
  destruct (os_ident_eq (st_active_os s) o); crush.
  destruct (os_ident_eq (st_active_os s) osi); crush.
  elim H1; crush.
  rewrite OSS.get_set_distinct; crush.
  rewrite OSS.get_set_distinct; crush.
  apply ctxt_switch_eq_smem; crush.
  apply ctxt_switch_eq_hyper; crush.
  
  (* Hcall *)
  unfold ACT.pre_hcall, ACT.post_hcall in *; crush.
  unfold osi_equivalent,
    osi_equivalent_smem, osi_pt_pages_eq,
    osi_pa_to_ma, osi_pa_to_pg, madd_of_vadd, madd_page_table,
    page_of_vadd_madd_smem; simpl; intros; repeat split; intros;
    repeat eexists; crush.
  destruct (os_ident_eq osi (st_active_os s)); right; crush.
  destruct (os_ident_eq osi (st_active_os s)); crush.
  elim H1; crush.
  rewrite OSS.get_set_distinct; crush.

  (* Ret ctrl *)
  unfold ACT.pre_ret_ctrl, ACT.post_ret_ctrl in *; crush.
  unfold osi_equivalent,
    osi_equivalent_smem, osi_pt_pages_eq,
    osi_pa_to_ma, osi_pa_to_pg, madd_of_vadd, madd_page_table,
    page_of_vadd_madd_smem; simpl; intros; repeat split; intros;
    repeat eexists; crush.
  destruct (os_ident_eq osi (st_active_os s)); right; crush.

  (* Chmod *)
  unfold ACT.pre_chmod, ACT.post_chmod in *; crush.
  unfold osi_equivalent,
    osi_equivalent_smem, osi_pt_pages_eq,
    osi_pa_to_ma, osi_pa_to_pg, madd_of_vadd, madd_page_table,
    page_of_vadd_madd_smem; simpl; intros; repeat split; intros;
    repeat eexists; crush.
  destruct (os_ident_eq osi (st_active_os s)); right; crush.

  (* Page pin *)
  eapply pin_os_isolation_not_osi; crush; crush.

  (* Page unpin *)
  eapply unpin_os_isolation_not_osi; crush; crush.
Qed.  


(* Two step execution *)
Lemma write_os_isolation_pt_pages : 
  forall s1 s2 s1' s2' va val,
    osi_equivalent s1 s2 (st_active_os s1) ->
    valid_state s1 -> valid_state s2 ->
    st_active_os s1 = st_active_os s2 ->
    ACT.valid_exec s1 (Write va val) s1' ->
    ACT.valid_exec s2 (Write va val) s2' ->
    osi_pt_pages_eq s1' s2' (st_active_os s1).
Proof.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write, osi_equivalent in *; crush.
  clear H13.
  repeat STAC.unify_write_post; crush.
  unfold osi_equivalent_smem in *; crush.
  
  unfold osi_pt_pages_eq; crush.
  assert (osi_pa_to_pg s1 (st_active_os s1) pt_padd = Some pt_page1).
  clear dependent s2.
  unfold osi_pa_to_pg in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) i0 = Os_PO (st_active_os s1)).
  STAC.destruct_valid_state_H H0; crush.
  unfold madd_owner in H; crush.
  unfold page_of_vadd_smem in H10; crush.
  rewrite STAC.set_mem_unchanged_pt_page
    with (ma:=m) (pt_page':=p) (old_pg:=x0) in H12; crush.
  assert (madd_type (smem_mm (st_memory s1)) i0 = Some PT); crush.
  rewrite <- STAC.write_same_madd_type with (va:=va) (pt_addr:=pt_addr1) (val:=val); crush.
  unfold madd_type; crush.
  rewrite <- H17; crush.
  unfold madd_type, is_pt_page in *; crush.
  destruct (page_content p) eqn:?; crush.
  clear H12.
  destruct H5 with pt_padd pt_page1 mp1; crush.
  rename x1 into pt_page2, x2 into mp2.
  exists pt_page2; crush.
  rewrite H2 in *; clear dependent s1.
  unfold osi_pa_to_pg in *; crush.
  unfold page_of_vadd_smem in H8; crush.
  apply STAC.set_mem_unchanged_pt_page with (ma:=m) (old_pg:=x); crush.
  unfold is_pt_page; crush.
  unfold osi_pa_to_ma; crush.
  exists mp2; crush.
  destruct H20 with va0 pa ma1; crush; clear H20.
  unfold osi_pa_to_ma; crush.
  unfold page_of_vadd_smem in H10; crush.
  
  assert (exists pa', osi_pa_to_ma s1 (st_active_os s1) pa' = Some m).
  clear dependent s2.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) pt_addr1 va = Some m).
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  destruct H_va_has_pa with (st_active_os s1) p2m pt_addr1 pt_page pt va m; crush.
  unfold madd_of_vadd, madd_page_table in H; crush.
  rewrite H4 in *; crush.
  exists x2; crush.
  unfold osi_pa_to_ma; crush.
  destruct H20 as [pa' ?].
  unfold page_of_vadd_smem in H8; crush.
  assert (osi_pa_to_ma s2 (st_active_os s1) pa' = Some m0); crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) pt_addr1 va = Some m); crush.
  unfold madd_of_vadd, madd_page_table in H28; crush.
  destruct (page_content p0) eqn:?; crush.
  unfold aos_current_page_table, osi_current_page_table in Heqopt_pt_addr0; crush.
  destruct H5 with (os_current_page i0) p0 p1; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  destruct H31 with va pa' m; crush.
  rewrite H2 in *; clear dependent s1.
  assert (madd_of_vadd (smem_mm (st_memory s2)) pt_addr0 va = Some m0); crush.
  unfold aos_current_page_table, osi_current_page_table in *;  crush.
  unfold madd_of_vadd, madd_page_table in H; crush.
  destruct (page_content p2) eqn:?; crush.
  unfold osi_pa_to_pg in H29; crush.

  destruct (padd_eq pa' pa); crush.
  (* pa' = pa *)
  unfold osi_pa_to_pg in H18; crush.
  assert (page_of_vadd_madd_smem (st_memory s2) va0 x1 = Some x).
  unfold osi_pa_to_pg in H12; crush.
  rewrite STAC.page_of_vadd_madd_smem_same_madd
    with (va2:=va) (pt_addr1:=i2) (pt_addr2:=pt_addr0); crush.
  exists (st_active_os s1); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold  aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s2); exists (os_current_page i3); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H19; crush.
  rewrite H25 with x0 in H28; crush.
  unfold osi_pa_to_pg in H12; crush.
  rewrite STAC.write_smem_mem_eq with (pt_addr':=i2) (old_pg:=x); crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.write_smem_mem_eq with (pt_addr':=i0) (old_pg:=x) in H26; crush.
  exists (st_active_os s1); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H17; crush.
  rewrite STAC.page_of_vadd_madd_smem_same_madd
    with (va2:=va) (pt_addr1:=i0) (pt_addr2:=pt_addr1); crush.
  exists (st_active_os s1); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s1); exists (os_current_page i5); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H17; crush.
  exists (st_active_os s1); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H19; crush.
  rewrite STAC.page_of_vadd_madd_smem_same_madd
    with (va2:=va) (pt_addr1:=i2) (pt_addr2:=pt_addr0); crush.
  exists (st_active_os s1); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s1); exists (os_current_page i5); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H19; crush.
  rewrite STAC.page_of_vadd_madd_smem_same_madd
    with (va2:=va) (pt_addr1:=i0) (pt_addr2:=pt_addr1); crush.
  unfold osi_pa_to_ma in *; crush.
  exists (st_active_os s1); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s1); exists (os_current_page i4); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H17; crush.
  unfold osi_pa_to_ma in *; crush.

  (* pa' <> pa *)
  unfold osi_pa_to_pg in H18; crush.
  rewrite STAC.write_smem_mem_neq with (ma':=m) (pg:=x0) (pt_addr:=i0) in H26; crush.
  assert (page_of_vadd_madd_smem (st_memory s2) va0 x1 = Some pg).
  apply H25; crush.
  rewrite H2 in *; clear dependent s1.
  unfold osi_pa_to_pg in H12; crush.
  rewrite STAC.write_smem_mem_neq with (ma':=m0) (pg:=x) (pt_addr:=i2); crush.
  exists (st_active_os s2); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H19; crush.
  unfold osi_pa_to_ma in *; crush.
  apply n.
  STAC.destruct_valid_state.
  rewrite <- H_injective_hyper with (st_active_os s2) i1 pa pa' m0; crush.
  exists (st_active_os s1); exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H17; crush.
  clear dependent s2.
  unfold osi_pa_to_ma in *; crush.
  apply n.
  STAC.destruct_valid_state.
  rewrite <- H_injective_hyper with (st_active_os s1) i pa pa' m; crush.
Qed.

Lemma write_os_isolation : 
  forall s1 s2 s1' s2' osi va val,
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Write va val) s1' ->
    ACT.valid_exec s2 (Write va val) s2' ->
    osi_equivalent s1' s2' osi.
Proof.
  intros.
  destruct (os_ident_eq (st_active_os s1) osi); crush.
  unfold osi_equivalent; crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write, osi_equivalent in *; crush;
  repeat STAC.unify_write_post; crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write, osi_equivalent in *; crush;
  repeat STAC.unify_write_post; crush.

  generalize H; unfold osi_equivalent,
    osi_equivalent_smem; crush.
  eapply write_os_isolation_pt_pages; crush.
  destruct H4; crush.
  destruct H4; crush.
  rewrite <- H9.
  eapply write_os_isolation_pt_pages with (s1:=s2) (s2:=s1); crush.
  apply osi_equivalent_symm; auto.
  rewrite H9; crush.

  unfold ACT.valid_exec in *; crush.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper in *; crush.
  unfold osi_hyper_eq; crush.
  assert (ACT.no_va_mapped_to_ma (st_active_os s1) ma1 s1).
  eapply STAC.write_no_va_mapped_old
    with (s:=s1) (ma:=ma1) (va:=va) (val:=val); crush.
  unfold ACT.pre_write in *; crush.
  destruct H8 with pa ma1 pg; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush. 
  erewrite STAC.write_no_va_mapped_same_mem; crush.
  destruct H; crush.
  rewrite <- H19.
  exists x; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush. 
  unfold osi_pa_to_ma in *; crush.
  eapply STAC.write_no_va_mapped_old with (s:=s2) (ma:=x) (va:=va) (val:=val); crush.
  unfold ACT.pre_write in *; crush.
  rewrite H19; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush. 
  erewrite <- STAC.write_no_va_mapped_same_mem; crush.
  unfold osi_hyper_eq; crush.
  assert (ACT.no_va_mapped_to_ma (st_active_os s1) ma1 s2).
  destruct H; crush.
  rewrite <- H15.
  eapply STAC.write_no_va_mapped_old with (s:=s2) (ma:=ma1) (va:=va) (val:=val); crush.
  unfold ACT.pre_write in *; crush.
  rewrite H15; crush.
  destruct H9 with pa ma1 pg; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush.  
  erewrite STAC.write_no_va_mapped_same_mem; crush.
  exists x; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush. 
  eapply STAC.write_no_va_mapped_old with (s:=s1) (ma:=x) (va:=va) (val:=val); crush.
  unfold ACT.pre_write in *; crush.
  unfold ACT.pre_write, ACT.pre_all_write, ACT.post_write,
    ACT.post_all_write in *; crush.
  repeat STAC.unify_write_post; crush. 
  erewrite <- STAC.write_no_va_mapped_same_mem; crush.

  unfold ACT.valid_exec in *; crush;
    unfold ACT.pre_write, ACT.post_write in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_symm.
  eapply write_os_isolation_not_osi; crush.
  apply osi_equivalent_trans with s2; crush.
  eapply write_os_isolation_not_osi; crush.
  destruct H; crush.
  destruct H; crush.
Qed.

Lemma write_hyper_os_isolation :
  forall s1 s2 s1' s2' osi va val,
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Write_hyper va val) s1' ->
    ACT.valid_exec s2 (Write_hyper va val) s2' ->
    osi_equivalent s1' s2' osi.
Proof.    
  unfold ACT.valid_exec; crush;
    unfold ACT.pre_write_hyper in *; crush.
  assert (~ os_accessible va); crush.
  unfold ACT.pre_all_write in *; crush.
  clear dependent s2.
  unfold aos_page_owner, aos_page_of_vadd_smem,
    page_of_vadd_smem, aos_current_page_table, osi_current_page_table in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) i va = Some m); crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in H1; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  destruct H_valid_virtmap with (st_active_os s1) i p0 p1 va m; crush. 
  rewrite <- H_valid_hyper with (st_active_os s1) i1 (os_current_page i0) i; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H_valid_cache with va m x0; crush.

  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_symm; crush.
  apply osi_equivalent_trans with s2; crush.
Qed.  

Lemma new_os_isolation_pt_pages_page_of_vadd : 
  forall s1 s2 osi pt_addr1' pt_addr2' va'
         va pa ma1 ma2 pg,
    valid_state s1 -> valid_state s2 ->
    osi_equivalent s1 s2 osi ->
    osi_current_page_table s1 osi = Some pt_addr1' ->
    osi_current_page_table s2 osi = Some pt_addr2' ->
    osi_pa_to_ma s1 osi pa = Some ma1 ->
    osi_pa_to_ma s2 osi pa = Some ma2 ->

    (exists pt_padd, exists pt_addr1,
       osi_pa_to_ma s1 osi pt_padd = Some pt_addr1 /\
       madd_of_vadd (smem_mm  (st_memory s1)) pt_addr1 va = Some ma1) \/
    va' = va ->

    is_rw_page pg ->
    page_of_vadd_madd_smem (ACT.new s1 osi pt_addr1' va' ma1) va ma1 = Some pg ->
    page_of_vadd_madd_smem (ACT.new s2 osi pt_addr2' va' ma2) va ma2 = Some pg.
Proof.
  unfold osi_equivalent, osi_equivalent_smem, osi_equivalent_hyper; crush.
  STAC.destruct_valid_state_H H.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  clear H_valid_currpg.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.

  destruct (padd_eq pa (os_current_page x)); crush.
  (* pa is the current page *)
  unfold osi_pa_to_ma in *; crush.
  rewrite STAC.new_page_pt_addr with (pt_page:=pt_page) (pt:=pt) in H8; crush.
  unfold is_rw_page in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va pt_addr1' p; crush.
  unfold is_rw_page in *; crush.
  (* pa is not the current pt page *)
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  assert (same_page_type p pg = true).
  eapply STAC.new_same_madd_type_page with (s:=s1) (pt_addr':=pt_addr1'); crush.
  unfold madd_type; crush.
  rewrite H19; crush.
  destruct (shared_page_dec s1 va' ma1); crush.
  (* shared page *)
  destruct (remove_ma_cache (smem_c (st_memory s1)) ma1) eqn:?; crush.
  destruct o.
  destruct p0; destruct c0.
  edestruct get_remove_ma_cache_some_entry
    with (c:=smem_c (st_memory s1)) (ma:=ma1); crush.
  rewrite Heqp0; crush.
  destruct H_valid_cache_va_ma with v ma1 p0; crush.
  assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO x2); crush.
  unfold osi_pa_to_pg in *; crush.
  destruct H_valid_virtmap with x2 i0 x4 x5 v m; crush.
  assert (madd_owner (smem_mm (st_memory s1)) i0 = Os_PO x2); crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt v) eqn:?; unfold os_accessible in *; auto.
  rewrite H30 in *; crush.
  rewrite H31 in *; crush.
  rewrite H29 in H30; crush; injection H30; crush.
  rewrite STAC.new_page_shared_entry_updated_page
    with (v:=v) (ma':=m) (pg':=p0) in H8; crush.
  destruct H10 with x3 x4 x5; crush.
  destruct H32 with v pa m; crush.
  unfold osi_pa_to_ma; crush.
  destruct (remove_ma_cache (smem_c (st_memory s2)) ma2) eqn:?.
  destruct o; crush.
  destruct p1. destruct c1.
  edestruct get_remove_ma_cache_some_entry
    with (c:=smem_c (st_memory s2)) (ma:=ma2); crush.
  rewrite Heqp1; crush.
  crush.
  rewrite STAC.new_page_shared_entry_updated_page
    with (v:=v0) (ma':=m0) (pg':=p1); crush.
  assert (page_of_vadd_madd_smem (st_memory s2) v m0 = Some p0).
  unfold osi_pa_to_ma in *; crush.
  apply H35; crush.
  unfold page_of_vadd_madd_smem; crush.
  unfold page_of_vadd_madd_smem in H36; crush.
  destruct (vadd_eq v v0); crush.
  destruct STAC.no_cache_alias with s2 v v0 m0 p0 p1; crush.
  destruct (vadd_eq v v0); crush.
  unfold osi_pa_to_pg in *; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s2)) i v = Some x8); crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H31; crush.
  STAC.synonym_cached.
  unfold madd_type; crush.
  rewrite H23; crush.
  eapply shared_page_eq with (s1:=s1) (osi:=x2); crush.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp1; crush.
  rewrite STAC.new_page_shared_no_entry_updated_page
    with (pg:=p0); crush.
  unfold madd_type; crush.
  rewrite H23; crush.
  eapply shared_page_eq with (s1:=s1) (osi:=x2); crush.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp1; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval0; crush.
  rewrite Heqp1; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold page_of_vadd_madd_smem in H35 at 2; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval1; crush.
  rewrite Heqp1; crush.
  apply H35; crush.
  unfold page_of_vadd_madd_smem; crush.
  unfold madd_type; crush.
  rewrite H19; crush.
  rewrite Heqp0; crush.
  (* No entry removed in s1 *)
  rewrite STAC.new_page_shared_no_entry_updated_page
    with (pg:=p) in H8; crush.
  destruct (remove_ma_cache (smem_c (st_memory s2)) ma2) eqn:?.
  destruct o; crush.
  destruct p0. destruct c1.
  edestruct get_remove_ma_cache_some_entry
    with (c:=smem_c (st_memory s2)) (ma:=ma2); crush.
  rewrite Heqp1; crush.
  crush.
  rewrite STAC.new_page_shared_entry_updated_page
    with (v:=v) (ma':=m) (pg':=p0); crush.
  destruct H_valid_cache_va_ma0 with v m p0; crush.
  destruct H13 with x3 x4 x5; crush.
  assert (madd_owner (smem_mm (st_memory s2)) m = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s2)) m = Os_PO x2); crush.
  unfold osi_pa_to_pg in *; crush.
  destruct H_valid_virtmap0 with x2 i0 x4 x5 v m; crush.
  assert (madd_owner (smem_mm (st_memory s2)) i0 = Os_PO x2); crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt v) eqn:?; unfold os_accessible in *; auto.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  rewrite H31 in *; crush.
  destruct H30 with v pa m; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (page_of_vadd_madd_smem (st_memory s1) v ma1 = Some p); crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval0; crush.
  rewrite Heqp0; crush.
  assert (page_of_vadd_madd_smem (st_memory s1) v ma1 = Some p0); crush.
  apply H33; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache0 with v m p0; crush.
  unfold madd_type; crush.
  rewrite H23; crush.
  eapply shared_page_eq with (s1:=s1) (osi:=osi); crush.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp1; crush.
  (* No entry removed from s2 *)
  rewrite STAC.new_page_shared_no_entry_updated_page
    with (pg:=p); crush.
  unfold madd_type; crush.
  rewrite H23; crush.
  eapply shared_page_eq with (s1:=s1) (osi:=osi); crush.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite Heqp1; crush.
  destruct ACT.no_va_mapped_to_ma_dec with osi ma1 s1; crush.
  destruct H11 with pa ma1 p; crush.
  unfold osi_pa_to_ma; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval0; crush.
  rewrite Heqp1; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct v; crush.
  destruct H10 with x2 x3 x5; crush.
  destruct H29 with x4 pa ma1; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval0; crush.
  rewrite Heqp1; crush.
  unfold page_of_vadd_madd_smem in H32 at 2; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval1; crush.
  rewrite Heqp1; crush.
  apply H32; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval2; crush.
  rewrite Heqp0; crush.
  unfold madd_type; crush; rewrite H19; crush.
  rewrite Heqp0; crush.
  unfold page_of_vadd_madd_smem; crush;
    rewrite get_remove_ma_cache_no_entry in Heqval0; crush;
    rewrite Heqp0; crush.
  (* Not shared page *)
  assert (va = va'); crush.
  destruct H6; crush.
  destruct (vadd_eq va' va); crush.
  elim n0.
  unfold shared_page.
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p1) eqn:?; crush.
  exists x2; exists osi; exists p1; exists p2; exists va; crush.
  unfold osi_pa_to_pg; crush.
  rewrite STAC.new_page_not_shared_updated_page in H8; crush.
  rewrite STAC.new_page_not_shared_updated_page; crush.
  destruct ACT.no_va_mapped_to_ma_dec with osi ma1 s1; crush.
  destruct H11 with pa ma1 p; crush.
  unfold is_rw_page in *; crush.
  unfold same_page_type in H22; rewrite H7 in H22; simpl in *.
  destruct (page_content p); crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache_va_ma0 with va' ma2 p0; crush.
  destruct H26 with x3 x4 va' x5; crush.
  assert (madd_owner (smem_mm (st_memory s2)) ma2 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s2)) ma2 = Os_PO x2); crush.
  unfold osi_pa_to_pg in *; crush.
  destruct H_valid_virtmap0 with x2 i0 x4 x5 va' ma2; crush.
  assert (madd_owner (smem_mm (st_memory s2)) i0 = Os_PO x2); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H32 in *; crush.
  unfold page_of_vadd_madd_smem in H8; crush.
  destruct H_valid_cache_va_ma with va' ma1 pg; crush.
  destruct n1 with x3 x4 va' x5; crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO x2); crush.
  unfold osi_pa_to_pg in *; crush.
  destruct H_valid_virtmap with x2 i0 x4 x5 va' ma1; crush.
  assert (madd_owner (smem_mm (st_memory s1)) i0 = Os_PO x2); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va') eqn:?; unfold os_accessible in *; auto.
  rewrite H31 in *; crush.
  destruct v; crush.
  destruct (vadd_eq va' x4); crush.
  (* va' = x4 *)
  destruct H10 with x2 x3 x5; crush.
  destruct H30 with x4 pa ma1; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  elim n0; crush.
  unfold shared_page; crush.
  exists x2; exists osi; exists x3; exists x5; exists x4; crush.
  apply n0.
  eapply shared_page_eq with (s1:=s2) (osi:=osi) (ma1:=ma2); crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  elim n0.
  eapply shared_page_eq with (s1:=s2) (osi:=osi) (ma1:=ma2); crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
Qed.

Lemma new_os_isolation_pt_pages :
  forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd) (pa:padd),
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (New_untrusted osi va pa) s1' ->
    ACT.valid_exec s2 (New_untrusted osi va pa) s2' ->
    osi_pt_pages_eq s1' s2' osi.
Proof.
  Opaque ACT.new.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_new, ACT.post_new, ACT.post_all_new in *;
    crush.
  unfold osi_equivalent, osi_equivalent_smem in *; crush.
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  destruct (padd_eq pt_padd (os_current_page x4)); crush.
  (* pt_padd is the current pt page *)
  rewrite STAC.new_get_mem_pt_addr with (pt_page:=pt_page)
    (pt:=pt) in H22; unfold osi_current_page_table; crush.
  injection H23; intros; subst mp1; clear H23.
  2:assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush;
    unfold madd_owner in *; crush.
  STAC.destruct_valid_state_H H1.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite H4 in *; crush.
  exists (Page (PT_c (PT.set pt0 va x2)) (page_owned_by pt_page0) (page_cacheable pt_page0));
    crush.
  apply STAC.new_get_mem_pt_addr; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists (PT.set pt0 va x2); crush.
  destruct (vadd_eq va va0); crush.
  (* va = va0 *)
  rewrite <- PT.get_set in *; crush.
  assert (pa0 = pa); crush.
  exists x2; crush.
  apply new_os_isolation_pt_pages_page_of_vadd with (s1:=s1) (pt_addr1':=x3)
    (pa:=pa) (ma1:=ma1); crush.
  unfold osi_equivalent, osi_equivalent_smem, osi_equivalent_hyper; crush.
  unfold osi_current_page_table; crush.
  rewrite H4 in *; crush.
  unfold osi_current_page_table; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  (* va <> va0 *)
  rewrite PT.get_set_distinct in *; crush.
  destruct H19 with (os_current_page x4) pt_page pt; crush.
  unfold osi_pa_to_pg; crush.
  destruct H33 with va0 pa0 ma1; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  exists x8; crush.
  destruct (padd_eq pa0 (os_current_page x4)); crush.
  (* pa is the current page *)
  erewrite STAC.new_page_pt_addr with (pt_page:=pt_page) in H25; crush.
  unfold is_rw_page in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 x3 p; crush.
  unfold is_rw_page in *; crush.
  (* pa is not the current page *)
  destruct (padd_eq pa0 pa); crush.
  apply new_os_isolation_pt_pages_page_of_vadd with (s1:=s1) (pt_addr1':=x3)
    (pa:=pa) (ma1:=x5); crush.
  unfold osi_equivalent, osi_equivalent_smem; crush.
  unfold osi_current_page_table; crush.
  rewrite H4; crush.
  unfold osi_current_page_table; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  left; crush.
  exists (os_current_page x4); exists x3; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H28; crush.
  rewrite STAC.new_page_ne; crush.
  rewrite STAC.new_page_ne in H25; crush.
  exists osi; exists (os_current_page x4); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H28; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.  
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists osi; exists (os_current_page x4); exists x3; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; 
    crush.
  rewrite H28; crush.
  exists osi; exists (os_current_page x4); unfold osi_pa_to_ma; crush. 
  unfold madd_type; crush.
  rewrite H30; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists osi; exists (os_current_page x4); exists x0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; 
    crush.
  rewrite <- H32; crush.
  
  (* pt_padd is not the current page *)
  destruct H19 with pt_padd pt_page1 mp1; crush.
  unfold osi_pa_to_pg; crush.
  rewrite STAC.new_get_mem_distinct_pt in H22; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type
    with (osi:=osi) (pt_addr':=x3) (va':=va) (ma':=x5); crush.
  unfold madd_type; crush.
  rewrite <- H23; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists x6; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite STAC.new_get_mem_distinct_pt; crush.
  rewrite H4 in *; crush.
  apply n.
  STAC.destruct_valid_state_H H1.
  rewrite H_injective_hyper0 with osi i1 (os_current_page x4) pt_padd i; crush.
  STAC.destruct_valid_state_H H1.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold madd_type; crush.
  rewrite <- H27; crush.
  exists x7; crush.
  destruct H30 with va0 pa0 ma1; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (padd_eq pa0 pa); crush.
  apply new_os_isolation_pt_pages_page_of_vadd with (s1:=s1) (pt_addr1':=x3)
    (pa:=pa) (ma1:=x5); crush.
  unfold osi_equivalent, osi_equivalent_smem; crush.
  unfold osi_current_page_table; crush.
  unfold osi_current_page_table; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  destruct (vadd_eq va va0); crush.
  left; crush.
  exists pt_padd; exists i0; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite STAC.new_get_mem_distinct_pt in H22; crush.
  rewrite <- H23; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type
    with (osi:=osi) (pt_addr':=x3) (va':=va) (ma':=x5); crush.
  unfold madd_type; crush.
  rewrite <- H23; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  (* pa0 <> pa *)
  destruct (padd_eq pa0 (os_current_page x4)); crush.
  erewrite STAC.new_page_pt_addr with (pt_page:=pt_page) in H36; crush.
  unfold is_rw_page in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 x3 p; crush.
  unfold is_rw_page in *; crush.
  rewrite STAC.new_page_ne; crush.
  rewrite STAC.new_page_ne in H36; crush.
  exists osi; exists (os_current_page x4); unfold osi_pa_to_ma; crush.
  unfold madd_type; crush.
  rewrite H28; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  exists osi; exists pt_padd; exists i0; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite STAC.new_get_mem_distinct_pt in H22; crush.
  rewrite <- H23; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  erewrite <- STAC.new_same_madd_type with (osi:=osi) (pt_addr':=x3) (va':=va) (ma':=x5); crush.
  unfold madd_type; crush.
  rewrite <- H23; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  apply n0.
  STAC.destruct_valid_state_H H1.
  rewrite H_injective_hyper0 with osi i1 pa0 pa x8; crush.
  apply n1.
  rewrite H4 in *; crush.
  STAC.destruct_valid_state_H H1.
  rewrite H_injective_hyper0 with osi i1 pa0 (os_current_page x4) x8; crush.
  exists osi; exists (os_current_page x1); unfold osi_pa_to_ma; crush.
  STAC.destruct_valid_state_H H1.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold madd_type; crush.
  rewrite H42; crush.
  STAC.destruct_valid_state_H H1.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold osi_pa_to_pg in *; crush.
  exists osi; exists pt_padd; exists i; unfold osi_pa_to_ma, madd_of_vadd, madd_page_table; crush.
  rewrite <- H27; crush.
Qed.

Lemma new_os_isolation_hyper :
  forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd) (pa:padd),
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (New_untrusted osi va pa) s1' ->
    ACT.valid_exec s2 (New_untrusted osi va pa) s2' ->
    osi_hyper_eq s1' s2' osi.
Proof.
  Opaque ACT.new.
  unfold ACT.valid_exec in *; crush.
  generalize H2 H5 H3 H4.
  unfold ACT.pre_new, ACT.post_new, ACT.post_all_new;
    crush.
  unfold osi_equivalent, osi_equivalent_hyper in *; crush.
  unfold osi_hyper_eq; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.

  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  rewrite H7 in *; crush.
  destruct (padd_eq pa0 (os_current_page x4)); crush.
  (* pa0 is the current page *)
  rewrite STAC.new_get_mem_pt_addr with (pt_page:=pt_page)
    (pt:=pt) in H29; unfold osi_current_page_table; crush.
  unfold is_rw_page in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  rewrite STAC.new_get_mem_pt_addr with (pt_page:=pt_page)
    (pt:=pt) in H29; unfold osi_current_page_table; crush.
  unfold is_rw_page in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  (* pa0 is not the current page *)
  destruct (padd_eq pa0 pa).
  (* pa0 = pa *)
  subst pa0.
  exists x2.
  match goal with
    | [ H1: ACT.no_va_mapped_to_ma ?osi ?ma1 ?s1',
        H2: ACT.post_new ?osi ?va ?pa ?s1 ?s1'
      |- context[ACT.no_va_mapped_to_ma ?osi ?ma2 ?s2']] =>
      assert (ACT.no_va_mapped_to_ma osi ma2 s2');
        [eapply no_va_mapped_eq with (s:=s1')|]; crush
  end.
  apply new_os_isolation_pt_pages with (s1:=s2) (s2:=s1) (va:=va) (pa:=pa); crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent, osi_equivalent_hyper; crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.

  destruct (madd_of_vadd (smem_mm (st_memory s1)) x3 va) eqn:?; crush. 
  assert (exists p, osi_pa_to_ma s1 osi p = Some m); crush.
  unfold madd_of_vadd, madd_page_table in Heqo0; crush.
  rewrite H34 in *; crush.
  destruct H_va_has_pa with osi x x3 pt_page pt va m; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H30; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (padd_eq x1 pa); crush.
  (* x5 was mapped by va in s1 *)
  destruct H28 with (os_current_page x4) {|
     page_content := PT_c (PT.set pt va x5);
     page_owned_by := page_owned_by pt_page;
     page_cacheable := page_cacheable pt_page |} va
     (PT.set pt va x5); crush.
  unfold osi_pa_to_pg; crush.
  rewrite STAC.new_get_mem_pt_addr with (pt_page:=pt_page)
    (pt:=pt); unfold osi_current_page_table; crush.
  unfold madd_type in H22; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5) eqn:?; crush.
  rewrite <- PT.get_set; crush.

  (* x5 was not mapped by va in s1 *)
  STAC.destruct_valid_state_H H1.
  clear H_valid_currpg; crush.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite <- STAC.new_no_va_mapped_same_mem with (osi:=osi) (pt_page':=pt_page0) (pt':=pt0) (pa:=pa); crush.
  rewrite <- STAC.new_no_va_mapped_same_mem with (osi:=osi) (pt_page':=pt_page) (pt':=pt) (pa:=pa) in H29; crush.
  destruct H24 with pa x5 pg; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (osi':=osi) (va':=va) (pa':=pa) (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  rewrite H30; crush.
  rewrite H7; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (osi':=osi) (va':=va) (pa':=pa) (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  rewrite H7; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (osi':=osi) (va':=va) (pa':=pa) (pt_addr':=x0); crush.
  unfold osi_current_page_table; crush.
  destruct H23; crush.
  edestruct H37 with (pt_padd:=os_current_page x4) (pt_page1:=pt_page0); crush.  
  unfold osi_pa_to_pg; crush.
  destruct H42 with va pa x2; crush.
  unfold madd_of_vadd, madd_page_table in H33; crush.
  rewrite H38 in *; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H38 in *; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite H34 in *; crush. 
  injection H41; crush.

  (* Nothing is mapped from va in s1 *)
  STAC.destruct_valid_state_H H1.
  clear H_valid_currpg; crush.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite <- STAC.new_no_va_mapped_same_mem with (osi:=osi) (pt_page':=pt_page0) (pt':=pt0) (pa:=pa); crush.
  rewrite <- STAC.new_no_va_mapped_same_mem with (osi:=osi) (pt_page':=pt_page) (pt':=pt) (pa:=pa) in H29; crush.
  destruct H24 with pa x5 pg; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (osi':=osi) (va':=va) (pa':=pa) (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  rewrite H30; crush.
  rewrite H7; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) x5 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (osi':=osi) (va':=va) (pa':=pa) (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  rewrite H7; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x2 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (osi':=osi) (va':=va) (pa':=pa) (pt_addr':=x0); crush.
  unfold osi_current_page_table; crush.
  destruct H23; crush.
  edestruct H36 with (pt_padd:=os_current_page x4) (pt_page1:=pt_page0); crush.  
  unfold osi_pa_to_pg; crush.
  destruct H41 with va pa x2; crush.
  unfold madd_of_vadd, madd_page_table in H31; crush.
  rewrite H37 in *; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H37 in *; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite H34 in *; crush. 

  (* pa0 <> pa *)
  assert (ma1 <> x5); crush.
  destruct (madd_of_vadd (smem_mm (st_memory s1)) x3 va) eqn:?; crush. 
  assert (exists p, osi_pa_to_ma s1 osi p = Some m); crush.
  unfold madd_of_vadd, madd_page_table in Heqo0; crush.
  rewrite H34 in *; crush.
  destruct H_va_has_pa with osi x x3 pt_page pt va m; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H30; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (padd_eq x1 pa0); crush.

  destruct H23.
  destruct H23 with (os_current_page x4)pt_page pt; crush.
  unfold osi_pa_to_pg; crush.
  destruct H37 with va pa0 ma1; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H34 in *; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.

  match goal with
    | [ H1: ACT.no_va_mapped_to_ma ?osi ?ma1 ?s1',
        H2: ACT.post_new ?osi ?va ?pa ?s1 ?s1'
      |- context[ACT.no_va_mapped_to_ma ?osi ?ma2 ?s2']] =>
        eapply no_va_mapped_eq with (s:=s1'); crush
  end.
  apply new_os_isolation_pt_pages with s2 s1 va pa; crush.
  unfold osi_equivalent; crush.
  destruct H; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold page_of_vadd_madd_smem in H40; crush.
  assert (p = p0); crush.
  injection (H40 p); crush.
  destruct H_valid_cache with va ma1 p; crush.
  clear H40.
  unfold osi_pa_to_pg in *; crush.
  destruct (shared_page_dec s2 va x2).
  apply STAC.new_get_mem_shared_distinct_ma_some_entry with
    (pt_page':=x1) (pt':=x6); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  STAC.destruct_valid_state_H H1; crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_some_entry with
    (pt_page':=pt_page) (pt':=pt) (pg:=p0) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  apply shared_page_eq with s2 osi pa x2; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  apply STAC.new_get_mem_not_shared_distinct_ma_some_entry with
    (pt_page':=x1) (pt':=x6); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_some_entry with
    (pt_page':=pt_page) (pt':=pt) (pg:=p0) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  elim n1.
  apply shared_page_eq with s1 osi pa x5; crush.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  (* Cache, Mem *)
  destruct (shared_page_dec s2 va x2).
  rewrite  STAC.new_get_mem_shared_distinct_ma_no_entry with
    (pt_page':=x1) (pt':=x6); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_some_entry with
    (pt_page':=pt_page) (pt':=pt) (pg:=p) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  apply shared_page_eq with s2 osi pa x2; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  unfold osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold osi_pa_to_pg in *; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.
  rewrite  STAC.new_get_mem_not_shared_distinct_ma_no_entry with
    (pt_page':=x1) (pt':=x6); crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_some_entry with
    (pt_page':=pt_page) (pt':=pt) (pg:=p) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  elim n1.
  apply shared_page_eq with s1 osi pa x5; crush.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  unfold osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold osi_pa_to_pg in *; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.
  (* Mem, Cache *)
  destruct (shared_page_dec s2 va x2).
  rewrite  STAC.new_get_mem_shared_distinct_ma_some_entry with
    (pt_page':=x1) (pt':=x6) (pg:=p); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_no_entry with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  apply shared_page_eq with s2 osi pa x2; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  unfold osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold osi_pa_to_pg in *; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.
  rewrite  STAC.new_get_mem_not_shared_distinct_ma_some_entry with
    (pt_page':=x1) (pt':=x6) (pg:=p); crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_no_entry with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  elim n1.
  apply shared_page_eq with s1 osi pa x5; crush.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  unfold osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold osi_pa_to_pg in *; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.
  (* Mem, Mem *)
  destruct (shared_page_dec s2 va x2).
  rewrite  STAC.new_get_mem_shared_distinct_ma_no_entry with
    (pt_page':=x1) (pt':=x6); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_no_entry with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  apply shared_page_eq with s2 osi pa x2; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  unfold osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold osi_pa_to_pg in *; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.
  rewrite  STAC.new_get_mem_not_shared_distinct_ma_no_entry with
    (pt_page':=x1) (pt':=x6); crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_no_entry with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  elim n1.
  apply shared_page_eq with s1 osi pa x5; crush.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_smem; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  unfold osi_pa_to_pg in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  STAC.destruct_valid_state_H H1; crush.
  unfold osi_pa_to_pg in *; crush.
  unfold madd_of_vadd, madd_page_table; crush; rewrite <- H36; crush.

  (* x1 <> pa0 *)
  destruct (shared_page_dec s1 va x5); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  destruct H24 with pa0 ma1 pg; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  destruct H_valid_os_idents with osi; crush.
  unfold osi_pa_to_ma in *; crush.
  match goal with
    | [ H1: ACT.no_va_mapped_to_ma ?osi ?ma1 ?s1',
        H2: ACT.post_new ?osi ?va ?pa ?s1 ?s1'
      |- context[ACT.no_va_mapped_to_ma ?osi ?ma2 ?s2']] =>
        eapply no_va_mapped_eq with (s:=s1'); crush
  end.
  apply new_os_isolation_pt_pages with s2 s1 va pa; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  STAC.destruct_valid_state_H H1; crush.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  apply shared_page_eq with s1 osi pa x5; crush.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H42 in *; crush.
  destruct H36 with (os_current_page x4) pt_page0 va pt0; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  (* not shared page *)
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  destruct H24 with pa0 ma1 pg; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  destruct H_valid_os_idents with osi; crush.
  unfold osi_pa_to_ma in *; crush.
  match goal with
    | [ H1: ACT.no_va_mapped_to_ma ?osi ?ma1 ?s1',
        H2: ACT.post_new ?osi ?va ?pa ?s1 ?s1'
      |- context[ACT.no_va_mapped_to_ma ?osi ?ma2 ?s2']] =>
        eapply no_va_mapped_eq with (s:=s1'); crush
  end.
  apply new_os_isolation_pt_pages with s2 s1 va pa; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  STAC.destruct_valid_state_H H1; crush.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  elim n2.
  apply shared_page_eq with s2 osi pa x2; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H42 in *; crush.
  destruct H36 with (os_current_page x4) pt_page0 va pt0; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.

  (* va is not mapped in s1 *)
  destruct (shared_page_dec s1 va x5); crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  destruct H24 with pa0 ma1 pg; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  destruct H_valid_os_idents with osi; crush.
  unfold osi_pa_to_ma in *; crush.
  match goal with
    | [ H1: ACT.no_va_mapped_to_ma ?osi ?ma1 ?s1',
        H2: ACT.post_new ?osi ?va ?pa ?s1 ?s1'
      |- context[ACT.no_va_mapped_to_ma ?osi ?ma2 ?s2']] =>
        eapply no_va_mapped_eq with (s:=s1'); crush
  end.
  apply new_os_isolation_pt_pages with s2 s1 va pa; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  STAC.destruct_valid_state_H H1; crush.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite STAC.new_get_mem_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  apply shared_page_eq with s1 osi pa x5; crush.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H41 in *; crush.
  destruct H33 with (os_current_page x4) pt_page0 va pt0; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
  (* not shared page *)
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page) (pt':=pt) in H29; crush.
  destruct H24 with pa0 ma1 pg; crush.
  unfold osi_pa_to_ma; crush.
  eapply STAC.new_no_va_mapped_old with (pt_addr':=x3); crush.
  unfold osi_current_page_table; crush.
  destruct H_valid_os_idents with osi; crush.
  unfold osi_pa_to_ma in *; crush.
  match goal with
    | [ H1: ACT.no_va_mapped_to_ma ?osi ?ma1 ?s1',
        H2: ACT.post_new ?osi ?va ?pa ?s1 ?s1'
      |- context[ACT.no_va_mapped_to_ma ?osi ?ma2 ?s2']] =>
        eapply no_va_mapped_eq with (s:=s1'); crush
  end.
  apply new_os_isolation_pt_pages with s2 s1 va pa; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  STAC.destruct_valid_state_H H1; crush.
  destruct H_valid_os_idents0 with osi; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite STAC.new_get_mem_not_shared_distinct_ma_not_mapped with
    (pt_page':=pt_page0) (pt':=pt0); crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s2)) x2); crush.
  elim n1.
  apply shared_page_eq with s2 osi pa x2; crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent; crush.
  unfold osi_equivalent_hyper; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H41 in *; crush.
  destruct H33 with (os_current_page x4) pt_page0 va pt0; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_type in *; crush.
  destruct (get_mem (smem_mm (st_memory s1)) x5); crush.
Qed.

Lemma new_os_isolation :
  forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd) (pa:padd),
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (New_untrusted o va pa) s1' ->
    ACT.valid_exec s2 (New_untrusted o va pa) s2' ->
    osi_equivalent s1' s2' osi.
Proof.
  Local Hint Resolve osi_equivalent_symm.
  intros.
  destruct (os_ident_eq o osi); crush.
  unfold osi_equivalent; crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_new, ACT.post_new, ACT.post_all_new, osi_equivalent in *;
    crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_new, ACT.post_new, ACT.post_all_new, osi_equivalent in *;
    crush.
  repeat rewrite <- OSS.get_set; crush.

  unfold osi_equivalent_smem; crush.
  eapply new_os_isolation_pt_pages; crush.
  eapply new_os_isolation_pt_pages with (s1:=s2) (s2:=s1); crush.
  unfold osi_equivalent_hyper; crush.
  eapply new_os_isolation_hyper; crush.
  eapply new_os_isolation_hyper with (s1:=s2) (s2:=s1); crush.

  unfold ACT.valid_exec in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.
Qed.

Lemma del_os_isolation_pt_pages :
  forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd),
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Del_untrusted osi va) s1' ->
    ACT.valid_exec s2 (Del_untrusted osi va) s2' ->
    osi_pt_pages_eq s1' s2' osi.
Proof.
  unfold ACT.valid_exec in *; crush.
  unfold osi_equivalent, osi_equivalent_smem in *; crush.
  clear H8.
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  assert (HYP.get (st_hypervisor s1) osi = Some i); crush.
  unfold ACT.post_del in *; crush.
  clear Heqopt_res.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  destruct (padd_eq pt_padd (os_current_page x)); crush.
  (* pt_padd is the current pt page *)
  rewrite STAC.del_get_mem_pt_page with (s:=s1) (pt_page:=pt_page)
    (pt:=pt) (osi':=osi) (va':=va) in H8; unfold osi_current_page_table; crush.
  injection H10; intros; subst mp1; clear H10.
  replace (st_hypervisor s2') with (st_hypervisor s2); crush.
  2:unfold ACT.post_del in *; crush.
  STAC.destruct_valid_state_H H1.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  rewrite H6 in *; crush.
  exists (Page (PT_c (PT.remove pt0 va)) (page_owned_by pt_page0) (page_cacheable pt_page0));
    crush.
  apply STAC.del_get_mem_pt_page with (s:=s2) (osi':=osi); 
    unfold osi_current_page_table; crush.
  exists (PT.remove pt0 va); crush.
  destruct (vadd_eq va va0); crush.
  rewrite PT.get_remove in *; crush.
  assert (PT.get pt va0 = Some ma1).
  apply PT.get_remove_old with (k':=va); crush.
  clear H9.
  destruct H7 with (os_current_page x) pt_page pt; crush.
  unfold osi_pa_to_pg; crush.
  destruct H24 with va0 pa ma1; crush.
  unfold osi_pa_to_ma; crush.
  exists x3; crush.
  rewrite PT.get_remove_distinct with (inf:=x3); crush.
  unfold osi_pa_to_pg in *; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold osi_pa_to_pg in *; crush.
  destruct (padd_eq pa (os_current_page x)); crush.
  erewrite STAC.del_page_of_vadd_madd_smem_pt_page with
    (s:=s1) (pt_page:=pt_page) in H28; crush.
  subst pg; unfold is_rw_page in *; crush.
  unfold osi_current_page_table; crush.
  rewrite H6; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  (* pa is not the current pt page *)
  generalize H3.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H20 in *; crush.
  injection H23; crush.
  rewrite <- STAC.del_page_of_vadd_madd_smem with
    (s:=s2) (osi':=osi) (pt_addr:=x4) (pt_addr':=x4) (va':=va); crush.
  apply H27; crush.
  generalize H2.
  unfold ACT.pre_del, ACT.pre_all_del; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H16 in *; crush.
  rewrite STAC.del_page_of_vadd_madd_smem with (s':=s1')
    (osi':=osi) (pt_addr:=x2) (pt_addr':=x2) (va':=va); crush.
  unfold osi_current_page_table; crush.
  exists osi; exists (os_current_page x); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H16; crush.
  unfold osi_current_page_table; crush.
  exists osi; exists (os_current_page x); unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H20; crush.
  apply n0.
  unfold osi_pa_to_ma in *; crush.

  (* pt_padd is not the current pt page *)
  erewrite STAC.del_get_mem_distinct_pt with (s:=s1) (pt_addr':=pt_addr) in H8;
    unfold osi_current_page_table; crush.
  2:erewrite <- STAC.del_same_madd_type; crush;
    unfold madd_type; crush;
    rewrite <- H10; crush.
  destruct H7 with pt_padd pt_page1 mp1; crush.
  unfold osi_pa_to_pg; crush.
  unfold osi_pa_to_pg in *; crush.
  replace (st_hypervisor s2') with (st_hypervisor s2); crush.
  2: unfold ACT.post_del in *; crush.
  assert (exists pt_addr0, osi_pa_to_ma s2 osi (os_current_page x) = Some pt_addr0).
  unfold ACT.pre_del, ACT.post_del in H3,H4; crush.
  unfold osi_current_page_table in *; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite H6 in *; crush.
  destruct H20 as [pt_addr0 ?].
  unfold osi_pa_to_ma in *; crush.
  assert (pt_addr0 <> i2); crush.
  STAC.destruct_valid_state_H H1; crush.
  exists x0; crush.
  erewrite STAC.del_get_mem_distinct_pt with (s:=s2) (pt_addr':=pt_addr0);
    unfold osi_current_page_table, madd_type; crush.
  rewrite <- H6; crush.
  rewrite <- H18; crush.
  exists x1; crush.
  destruct H19 with va0 pa ma1; crush.
  destruct (padd_eq pa (os_current_page x)); crush.
  erewrite STAC.del_page_of_vadd_madd_smem_pt_page with
    (s:=s1) (pt_page:=pt_page) in H27; crush.
  subst pg; unfold is_rw_page in *; crush.
  unfold osi_current_page_table; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with va0 ma1 p; crush.
  unfold is_rw_page in *; crush.
  (* pa is not the current pt page *)
  erewrite <- STAC.del_page_of_vadd_madd_smem_ne_pt with (pt_addr':=pt_addr0);
    crush.
  apply H26; crush.
  rewrite STAC.del_page_of_vadd_madd_smem_ne_pt with (s':=s1') (osi':=osi)
    (pt_addr:=i0) (pt_addr':=pt_addr) (va':=va); crush.
  unfold osi_current_page_table; crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H10; crush.
  unfold osi_current_page_table; crush.
  rewrite <- H6; crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H18; crush.
  apply n0.
  STAC.destruct_valid_state_H H1; crush.
Qed.

Lemma del_os_isolation_hyper :
  forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd),
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Del_untrusted osi va) s1' ->
    ACT.valid_exec s2 (Del_untrusted osi va) s2' ->
    osi_hyper_eq s1' s2' osi.
Proof.
  unfold ACT.valid_exec in *; crush.
  generalize H2 H5 H3 H4.
  unfold ACT.pre_del, ACT.pre_all_del,
    ACT.post_del; crush.
  unfold osi_equivalent, osi_equivalent_smem,
    osi_equivalent_hyper in *; crush.
  unfold osi_hyper_eq; crush.
  destruct ACT.no_va_mapped_to_ma_dec with osi ma1 s1; crush.
  (* No va mapped to ma in s1 *)
  destruct H22 with pa ma1 pg; crush.
  erewrite STAC.del_no_va_mapped_same_mem; crush.
  unfold osi_pa_to_ma in *; crush.
  assert (madd_type (smem_mm (delete (st_memory s1) x4 va)) x4 =
    madd_type (smem_mm (st_memory s1)) x4).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type, is_rw_page, madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content pg); crush.
  destruct (page_content p2); crush.
  eapply STAC.del_no_va_mapped_new with (s:=s2); crush.
  erewrite <- STAC.del_no_va_mapped_same_mem; crush.
  assert (madd_type (smem_mm (delete (st_memory s1) x4 va)) x4 =
    madd_type (smem_mm (st_memory s1)) x4).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type, is_rw_page, madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content pg); crush.
  (* Some va maps to ma in s1 *)
  destruct v; crush.
  destruct H21 with x x6 x8; crush.
  unfold osi_pa_to_ma, osi_pa_to_pg in *; crush.
  destruct H34 with x7 pa ma1; try (crush; fail).
  exists x11; split; [crush|].
  match goal with
    | [ H1: ACT.no_va_mapped_to_ma ?osi ?ma1 ?s1',
        H2: ACT.post_del ?osi ?va ?s1 ?s1'
      |- context[ACT.no_va_mapped_to_ma ?osi ?ma2 ?s2']] =>
      assert (ACT.no_va_mapped_to_ma osi ma2 s2');
        [eapply no_va_mapped_eq with (s:=s1')|]; crush
  end.
  apply del_os_isolation_pt_pages with (s1:=s2) (s2:=s1) (va:=va); crush.
  apply osi_equivalent_symm.
  unfold osi_equivalent, osi_equivalent_smem, osi_equivalent_hyper; crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.

  assert (exists pg_old, page_of_vadd_madd_smem (st_memory s1)
    x7 ma1 = Some pg_old).
  unfold page_of_vadd_madd_smem; crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi).
  STAC.destruct_valid_state_H H0; crush.
  unfold madd_owner in H39; crush.
  destruct H39 as [pg_old ?].
  match goal with
    | [H1: ACT.no_va_mapped_to_ma ?osi ?ma ?s2',
       H2: ACT.post_del ?osi ?va ?s2 ?s2'
    |- get_mem (smem_mm (delete (st_memory ?s2) ?pt_addr ?va)) ?ma = Some ?pg] =>
      let s := fresh "s" in
     remember s2' as s; replace (smem_mm (delete (st_memory s2) pt_addr va)) with
       (smem_mm (st_memory s)); [|crush]
  end.
  apply STAC.del_no_va_mapped_smem_then_mem
    with (s:=s2) (osi:=osi) (pt_addr':=x1)
         (va':=va) (pt_padd:=x) (pt_addr:=i0) (va:=x7); crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H33; crush.

  unfold osi_current_page_table, osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  destruct (padd_eq pa (os_current_page x0)); crush.
  rewrite H7 in *; crush.
  assert (madd_type (smem_mm (delete (st_memory s1) x4 va)) x4 =
    madd_type (smem_mm (st_memory s1)) x4).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type, is_rw_page, madd_of_vadd, madd_page_table in *; crush.
  rewrite H25 in *; crush.
  destruct (page_content p2); crush.
  apply H38; crush.
  match goal with
    | [H1: ACT.no_va_mapped_to_ma ?osi ?ma ?s1',
       H2: ACT.post_del ?osi ?va' ?s1 ?s1',
       H3: get_mem (smem_mm (delete (st_memory ?s1) ?pt_addr ?va')) ?ma = Some ?pg
    |- page_of_vadd_madd_smem (st_memory ?s1) ?va ?ma = Some ?pg] => 
      let s := fresh "s" in
     remember s1' as s; replace (smem_mm (delete (st_memory s1) pt_addr va')) with
       (smem_mm (st_memory s)) in H3; [|crush]
  end.
  rewrite STAC.del_no_va_mapped_smem_then_mem 
    with (s:=s1) (osi:=osi) (pt_addr':=x4) (va':=va)
         (pt_padd:=x) (pt_addr:=i2) (va:=x7) (pg:=pg_old)
    in H28; crush.
  unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H30; crush.

  unfold osi_current_page_table, osi_pa_to_ma in *; crush.
  STAC.destruct_valid_state_H H0; crush.
  destruct (padd_eq pa (os_current_page x3)); crush.
  assert (madd_type (smem_mm (delete (st_memory s1) x4 va)) x4 =
    madd_type (smem_mm (st_memory s1)) x4).
  apply STAC.del_hyper_same_madd_type; crush.
  unfold madd_type, is_rw_page, madd_of_vadd, madd_page_table in *; crush.
  rewrite H25 in *; crush.
  destruct (page_content p2); crush.
Qed.  

Lemma del_untrusted_os_isolation :
  forall (s1 s2 s1' s2':state) (osi o:os_ident) (va:vadd),
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Del_untrusted o va) s1' ->
    ACT.valid_exec s2 (Del_untrusted o va) s2' ->
    osi_equivalent s1' s2' osi.
Proof.
  intros.
  destruct (os_ident_eq o osi); crush.
  unfold osi_equivalent; crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_del, ACT.post_del, osi_equivalent in *;
    crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_del, ACT.post_del, osi_equivalent in *;
    crush.
  repeat rewrite <- OSS.get_set; crush.

  unfold osi_equivalent_smem; crush.
  eapply del_os_isolation_pt_pages; crush.
  eapply del_os_isolation_pt_pages with (s1:=s2) (s2:=s1); crush.

  unfold osi_equivalent_hyper; crush.
  eapply del_os_isolation_hyper; crush.
  eapply del_os_isolation_hyper with (s1:=s2) (s2:=s1); crush.
  
  unfold ACT.valid_exec in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.
Qed.

Lemma pin_os_isolation_pt_pages : 
  forall s1 s2 s1' s2' osi pa t,
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Page_pin_untrusted osi pa t) s1' ->
    ACT.valid_exec s2 (Page_pin_untrusted osi pa t) s2' ->
    osi_pt_pages_eq s1' s2' osi.
Proof.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  unfold osi_equivalent, osi_equivalent_smem in *; crush.
  clear H32.
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (padd_eq pa pt_padd); crush.
  (* pa = pt_padd *)
  rewrite <- P2M.get_set in *; crush.
  exists (new_page (Some t) (Os_PO osi)); crush.
  rewrite <- get_add_mem; crush.
  rewrite <- get_add_mem in *; crush.
  Transparent new_page.
  unfold new_page in *; crush.
  Opaque new_page.
  exists PT.empty; crush.
  injection H34; intros; subst mp1.
  rewrite PT.get_empty in *; crush.

  (* pa <> pt_padd *)
  rewrite P2M.get_set_distinct in *; crush.
  destruct (madd_eq x5 i0); crush.
  assert (madd_owner (smem_mm (st_memory s1)) i0 = Os_PO osi); crush.
  STAC.destruct_valid_state_H H0; crush.
  unfold madd_owner in *; crush.
  rewrite get_add_mem_distinct in H32; crush.
  destruct H31 with pt_padd pt_page1 mp1; crush.
  unfold osi_pa_to_pg; crush.
  exists x11; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite get_add_mem_distinct; crush.
  exists x12; crush.
  destruct (padd_eq pa0 pa); crush.
  rewrite <- P2M.get_set in *; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_virtmap with osi i0 pt_page1 mp1 va ma1; crush.
  rewrite <- H_valid_hyper with osi x6 pt_padd i0; crush.
  unfold madd_owner; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *.
  rewrite H39 in *; crush.
  rewrite H40 in *; crush.
  rewrite P2M.get_set_distinct in *; crush.
  destruct H37 with va pa0 ma1; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  destruct (madd_eq x5 ma1); crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi).
  STAC.destruct_valid_state_H H0; crush.
  unfold madd_owner in *; crush.
  destruct (madd_eq x x13); crush.
  assert (madd_owner (smem_mm (st_memory s2)) x13 = Os_PO osi).
  STAC.destruct_valid_state_H H1.
  unfold osi_pa_to_ma in *; crush.
  unfold madd_owner in *; crush.
  unfold page_of_vadd_madd_smem in *; crush;
    try rewrite get_add_mem_distinct in *; crush.
Qed.

Lemma pin_os_isolation_hyper : 
  forall s1 s2 s1' s2' osi pa t,
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Page_pin_untrusted osi pa t) s1' ->
    ACT.valid_exec s2 (Page_pin_untrusted osi pa t) s2' ->
    osi_hyper_eq s1' s2' osi.
Proof.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  unfold osi_equivalent, osi_equivalent_hyper in *; crush.
  unfold osi_hyper_eq; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (padd_eq pa pa0); crush.
  rewrite <- P2M.get_set in *; crush.
  exists x; crush.
  unfold ACT.no_va_mapped_to_ma; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (madd_eq x i0); crush.
  rewrite <- get_add_mem in *; crush.
  Transparent new_page.
  unfold new_page in H38; crush.
  Opaque new_page.
  unfold is_rw_page in *; crush.
  rewrite get_add_mem_distinct in H35; crush.
  destruct (padd_eq pa0 pt_padd); crush.
  rewrite <- P2M.get_set in *; crush.
  rewrite P2M.get_set_distinct in *; crush.
  STAC.destruct_valid_state_H H1; crush.
  destruct H_va_has_pa with osi x0 i0 pt_page mp va x; crush.
  assert (madd_owner (smem_mm (st_memory s2)) i0 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  cut False; crush.
  destruct H_valid_virtmap with osi i0 pt_page mp va x; crush.
  assert (madd_owner (smem_mm (st_memory s2)) i0 = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  unfold os_accessible in *; rewrite H41 in *; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  rewrite <- get_add_mem in *; crush.
  (* pa <> pa0 *)
  rewrite P2M.get_set_distinct in *; crush.
  rewrite get_add_mem_distinct in *; crush.
  destruct H32 with pa0 ma1 pg; crush.
  unfold osi_pa_to_ma; crush.
  unfold ACT.no_va_mapped_to_ma; crush.
  destruct H36 with pt_padd pt_page va mp; crush.
  unfold osi_pa_to_pg; crush.
  rewrite <- HYP.get_set; crush.
  destruct (padd_eq pa pt_padd); crush.
  rewrite <- P2M.get_set in *; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite P2M.get_set_distinct in *; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite get_add_mem_distinct; crush.
  unfold osi_pa_to_ma in *; crush.
  unfold ACT.no_va_mapped_to_ma; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (padd_eq pa pt_padd); crush.
  rewrite <- P2M.get_set in *; crush.
  rewrite <- get_add_mem in *; crush.  
  Transparent new_page.
  unfold new_page in H42; crush.
  Opaque new_page.
  injection H42; intros; subst mp; rewrite PT.get_empty in *; crush.
  rewrite P2M.get_set_distinct in *; crush.
  rewrite get_add_mem_distinct in H41; crush.
  destruct H39 with pt_padd pt_page va mp; crush.
  unfold osi_pa_to_pg; crush.
  rewrite <- get_add_mem in *; crush.
  Transparent new_page.
  unfold new_page in H42; crush.
  Opaque new_page.
  injection H42; intros; subst mp; rewrite PT.get_empty in *; crush.
  rewrite get_add_mem_distinct; crush.
  assert (madd_owner (smem_mm (st_memory s2)) x11 = Os_PO osi).
  STAC.destruct_valid_state_H H1; crush.
  unfold madd_owner in *; crush.
  rewrite <- get_add_mem in *; crush.
  assert (madd_owner (smem_mm (st_memory s1)) ma1 = Os_PO osi).
  STAC.destruct_valid_state_H H0; crush.
  unfold madd_owner in *; crush.
Qed.

Lemma pin_untrusted_os_isolation :
  forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd) (type:ptype),
  valid_state s1 -> valid_state s2 ->
  osi_equivalent s1 s2 osi ->
  ACT.valid_exec s1 (Page_pin_untrusted o pa type) s1' ->
  ACT.valid_exec s2 (Page_pin_untrusted o pa type) s2' ->
  osi_equivalent s1' s2' osi.
Proof.
  intros.
  destruct (os_ident_eq o osi); crush.
  unfold osi_equivalent; crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_pin, ACT.post_pin, osi_equivalent in *;
    crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_pin, ACT.post_pin, osi_equivalent in *;
    crush.
  repeat rewrite <- OSS.get_set; crush.

  unfold osi_equivalent_smem; crush.
  eapply pin_os_isolation_pt_pages; crush.
  eapply pin_os_isolation_pt_pages with (s1:=s2) (s2:=s1); crush.

  unfold osi_equivalent_hyper; crush.
  eapply pin_os_isolation_hyper; crush.
  eapply pin_os_isolation_hyper with (s1:=s2) (s2:=s1); crush.
  
  unfold ACT.valid_exec in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.
Qed.

Lemma unpin_os_isolation_pt_pages : 
  forall s1 s2 s1' s2' osi pa,
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Page_unpin_untrusted osi pa) s1' ->
    ACT.valid_exec s2 (Page_unpin_untrusted osi pa) s2' ->
    osi_pt_pages_eq s1' s2' osi.
Proof.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  unfold osi_equivalent, osi_equivalent_smem in *; crush.
  clear H24.
  unfold osi_pt_pages_eq; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  rewrite <- HYP.get_set in *; crush.
  destruct (padd_eq pa pt_padd); crush.
  (* pa = pt_padd *)
  rewrite P2M.get_remove in *; crush.
  (* pa <> pt_padd *)
  assert (P2M.get x1 pt_padd = Some i0).
  rewrite P2M.get_remove_old with (k':=pa) (inf:=i0); crush.
  clear Heqopt_res0.
  destruct (madd_eq x7 i0); crush.
  elim n.
  STAC.destruct_valid_state_H H0.
  rewrite <- H_injective_hyper with osi x1 pa pt_padd i0; crush.
  rewrite get_add_mem_distinct in *; crush.
  unfold osi_pa_to_pg in *; crush.
  destruct H25 with pt_padd pt_page1 mp1; crush.
  unfold osi_pa_to_pg; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite P2M.get_remove_distinct with (inf:=i); crush.
  exists x0; crush.
  destruct (madd_eq x3 i); crush.
  elim n.
  STAC.destruct_valid_state_H H1.
  rewrite <- H_injective_hyper with osi x pa pt_padd i; crush.
  rewrite get_add_mem_distinct; crush.
  exists x4; crush.
  destruct (padd_eq pa0 pa); crush.
  rewrite P2M.get_remove in *; crush.
  assert (P2M.get x1 pa0 = Some ma1).
  apply P2M.get_remove_old with (k':=pa) (inf:=ma1); crush.
  clear H32.
  destruct H30 with va pa0 ma1; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in *; crush.
  apply P2M.get_remove_distinct; crush.
  destruct (madd_eq x7 ma1); crush.
  elim H23 with pt_padd pt_page1 va mp1; crush.
  unfold osi_pa_to_pg; crush.
  destruct (madd_eq x3 x5); crush.
  destruct H17 with pt_padd x0 va x4; crush.
  unfold osi_pa_to_pg; crush.
  unfold page_of_vadd_madd_smem in *; crush;
    try rewrite get_add_mem_distinct in *; crush.
Qed.

Lemma unpin_os_isolation_hyper : 
  forall s1 s2 s1' s2' osi pa,
    osi_equivalent s1 s2 osi ->
    valid_state s1 -> valid_state s2 ->
    ACT.valid_exec s1 (Page_unpin_untrusted osi pa) s1' ->
    ACT.valid_exec s2 (Page_unpin_untrusted osi pa) s2' ->
    osi_hyper_eq s1' s2' osi.
Proof.
  unfold ACT.valid_exec in *; crush.
  generalize H2 H5 H3 H4.
  unfold ACT.pre_unpin, ACT.post_unpin; crush.
  unfold osi_equivalent, osi_equivalent_hyper in *; crush.
  unfold osi_hyper_eq; crush.
  unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
  rewrite <- HYP.get_set in *; crush.
  assert (ACT.no_va_mapped_to_ma osi ma1 s1).
  eapply STAC.page_unpin_no_va_mapped with (pa:=pa); crush.
  destruct (padd_eq pa pa0); crush.
  rewrite P2M.get_remove in *; crush.
  assert (P2M.get x1 pa0 = Some ma1).
  apply P2M.get_remove_old with (k':=pa) (inf:=ma1); crush.
  clear H33.
  rewrite get_add_mem_distinct in *; crush.
  destruct H30 with pa0 ma1 pg; crush.
  unfold osi_pa_to_ma; crush.
  exists x0; crush.
  unfold osi_pa_to_ma in *; crush.
  rewrite P2M.get_remove_distinct with (inf:=x0); crush.
  eapply STAC.page_unpin_no_va_mapped with (s:=s2) (pa:=pa); crush.
  rewrite get_add_mem_distinct; crush.
  unfold osi_pa_to_ma in *; crush.
  apply n.
  STAC.destruct_valid_state_H H1.
  rewrite H_injective_hyper with osi x pa pa0 x0; crush.
  apply n.
  STAC.destruct_valid_state_H H0.
  rewrite H_injective_hyper with osi x1 pa pa0 ma1; crush.
Qed.

Lemma unpin_untrusted_os_isolation : forall (s1 s2 s1' s2':state) (osi o:os_ident) (pa:padd),
  valid_state s1 -> valid_state s2 ->
  osi_equivalent s1 s2 osi ->
  ACT.valid_exec s1 (Page_unpin_untrusted o pa) s1' ->
  ACT.valid_exec s2 (Page_unpin_untrusted o pa) s2' ->
  osi_equivalent s1' s2' osi.
Proof.
  intros.
  destruct (os_ident_eq o osi); crush.
  unfold osi_equivalent; crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_equivalent in *;
    crush.
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_unpin, ACT.post_unpin, osi_equivalent in *;
    crush.
  repeat rewrite <- OSS.get_set; crush.

  unfold osi_equivalent_smem; crush.
  eapply unpin_os_isolation_pt_pages; crush.
  eapply unpin_os_isolation_pt_pages with (s1:=s2) (s2:=s1); crush.

  unfold osi_equivalent_hyper; crush.
  eapply unpin_os_isolation_hyper; crush.
  eapply unpin_os_isolation_hyper with (s1:=s2) (s2:=s1); crush.

  unfold ACT.valid_exec in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.
Qed.

Theorem os_isolation :
  forall (s1 s2 s1' s2':state) (osi:os_ident) (a:action),
    valid_state s1 ->
    valid_state s2 ->
    ACT.valid_exec s1 a s1' ->
    ACT.valid_exec s2 a s2' ->
    STEQ.osi_equivalent s1 s2 osi ->
    STEQ.osi_equivalent s1' s2' osi.
Proof.  
  Local Hint Constructors osi_action.

  intros.

  STAC.action_induction a; crush.

  (* Silent *)
  unfold ACT.valid_exec in *; crush;
    unfold ACT.post_silent in *; crush.

  (* Read *)
  unfold ACT.valid_exec in *; crush;
    unfold ACT.pre_read, ACT.post_read in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.

  (* Read hyper *)
  unfold ACT.valid_exec in *; crush;
    unfold ACT.pre_read_hyper, ACT.post_read_hyper in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.

  (* Write *)
  eapply write_os_isolation; crush.

  (* Write hyper *)
  eapply write_hyper_os_isolation; crush.

  (* New untrusted *)
  destruct (os_ident_eq o osi); crush.
  eapply new_os_isolation; crush.
  unfold ACT.valid_exec in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.

  (* New hyper *)
  unfold ACT.valid_exec in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.

  (* Del untrusted *)
  eapply del_untrusted_os_isolation; crush.

  (* Del hyper *)
  unfold ACT.valid_exec in *; crush.
  apply osi_equivalent_trans with s1; crush.
  apply osi_equivalent_trans with s2; crush.

  (* Switch *)
  unfold ACT.valid_exec in *; crush;
    unfold ACT.pre_switch, ACT.post_switch in *; crush.
  unfold osi_equivalent; crush.
  destruct (os_ident_eq o osi); [left|right]; crush;
    unfold hyper_running in *; crush.
  unfold osi_equivalent in H3; crush.
  apply osi_equivalent_smem_trans with s1; crush.
  apply osi_equivalent_smem_symm.
  apply ctxt_switch_eq_smem; crush.
  apply osi_equivalent_smem_trans with s2; crush.
  unfold osi_equivalent in *; crush.
  apply ctxt_switch_eq_smem; crush.
  apply osi_equivalent_hyper_trans with s1; crush.
  apply osi_equivalent_hyper_symm.
  apply ctxt_switch_eq_hyper; crush.
  apply osi_equivalent_hyper_trans with s2; crush.
  unfold osi_equivalent in *; crush.
  apply ctxt_switch_eq_hyper; crush.

  (* Lswitch *)
  unfold ACT.valid_exec in *; crush;
    unfold ACT.pre_lswitch, ACT.post_lswitch in *; crush.
  unfold osi_equivalent in *; crush.
  destruct H3; crush.
  destruct (os_ident_eq (st_active_os s1) o); crush;
    [ repeat rewrite <- OSS.get_set | repeat rewrite OSS.get_set_distinct]; crush.
  destruct (os_ident_eq osi o); crush;
    [ repeat rewrite <- OSS.get_set | repeat rewrite OSS.get_set_distinct]; crush.
  apply osi_equivalent_smem_trans with s1; crush.
  apply osi_equivalent_smem_symm.
  apply ctxt_switch_eq_smem; crush.
  apply osi_equivalent_smem_trans with s2; crush.
  apply ctxt_switch_eq_smem; crush.
  apply osi_equivalent_hyper_trans with s1; crush.
  apply osi_equivalent_hyper_symm.
  apply ctxt_switch_eq_hyper; crush.
  apply osi_equivalent_hyper_trans with s2; crush.
  apply ctxt_switch_eq_hyper; crush.

  (* Hcall *)
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_hcall, ACT.post_hcall in *; crush.
  unfold osi_equivalent in *; crush.
  destruct H3; crush.
  destruct H3; crush.
  rewrite H3.  
  repeat rewrite <- OSS.get_set; crush.
  repeat rewrite OSS.get_set_distinct; crush.

  (* Ret ctrl *)
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_ret_ctrl, ACT.post_ret_ctrl in *; crush.
  unfold osi_equivalent in *; crush.
  destruct H3; crush.

  (* Chmod *)
  unfold ACT.valid_exec in *; crush.
  unfold ACT.pre_chmod, ACT.post_chmod in *; crush.
  unfold osi_equivalent in *; crush.
  destruct H3; crush.

  (* Page pin *)
  eapply pin_untrusted_os_isolation with (s1:=s1) (s2:=s2); crush.

  (* Page unpin *)
  eapply unpin_untrusted_os_isolation with (s1:=s1) (s2:=s2); crush.
Qed.

End OS_ISOLATION.

