(***********************************************************************
 * File: Opt_monad.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies the option monad.
 * 
 *************************************************************************)

Require Import Monad.

Set Implicit Arguments.

Definition option_bind X Y (f:X -> option Y) (x:option X) :=
  match x with
    | Some x' => f x'
    | None => None
  end.

Definition option_unit := Some.

Lemma option_bind_bind : forall X Y Z
  (f : X -> option Y) (g : Y -> option Z)
  (x : option X),
  option_bind g (option_bind f x) =
  option_bind (fun u => option_bind g (f u)) x.
Proof.
  destruct x; auto.
Qed.
  
Lemma option_bind_unit : forall X Y (f : X -> option Y) (x : X),
  option_bind f (option_unit x) = f x.
Proof.
  auto.
Qed.

Lemma option_unit_bind : forall X (x : option X),
  option_bind (@option_unit X) x = x.
Proof.
  destruct x; auto.
Qed.

Definition Monad_Option : Monad :=
  Build_Monad option option_bind option_unit option_bind_bind
  option_bind_unit option_unit_bind.

Notation "x <- e1 ; e2" := (@bind Monad_Option _ _ (fun x => e2) e1)
  (right associativity, at level 60).

Ltac simpl_monad :=
  repeat (match goal with
    | [H: Monad.bind Monad_Option ?t ?f ?e = Some ?v |-_] =>
      let opt_res := fresh "opt_res" in
        remember e as opt_res; destruct opt_res;
          [simpl in H | discriminate H; fail]
    | [H: option_bind ?f ?e = Some ?v |-_] =>
      let opt_res := fresh "opt_res" in
        remember e as opt_res; destruct opt_res;
          [simpl in H | discriminate H; fail]
    | [H: Some ?v = option_bind ?f ?e |-_] =>
      let opt_res := fresh "opt_res" in
        remember e as opt_res; destruct opt_res;
          [simpl in H | discriminate H; fail]
    | [H: option_bind ?f ?e = None |-_] =>
      let opt_res := fresh "opt_res" in
        remember e as opt_res; destruct opt_res;
          [discriminate H; fail | simpl in H ]
    | [H: None = option_bind ?f ?e |-_] =>
      let opt_res := fresh "opt_res" in
        remember e as opt_res; destruct opt_res;
          [ discriminate H; fail | simpl in H]
  end).
