(***********************************************************************
 * File: Page.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies the memory pages
 * 
 *************************************************************************)

Require Import Dict_unbound.
Require Export OS.

Module Type PAGE_TABLE := DICT_UNBOUND
  with Definition key := vadd
  with Definition eq_key := vadd_eq
  with Definition info := madd.

Module Type PAGE.
  Declare Module PT : PAGE_TABLE.

  Definition page_table := PT.dict.

  Inductive content : Set :=
    | RW_c : option value -> content
    | PT_c : page_table -> content
    | Other_c : content.

  (* The owner of a page might be either an OS, the hypervisor or
     none *)
  Inductive page_owner : Set :=
    | Hyp_PO   : page_owner
    | Os_PO    : os_ident -> page_owner
    | No_Owner : page_owner.

  Record page : Set :=
    Page {
      page_content : content;
      page_owned_by : page_owner;
      page_cacheable : bool
    }.

  Definition is_rw_page (pg:page) :=
    exists val, page_content pg = RW_c (Some val).

  Definition is_pt_page (pg:page) :=
    exists pt, page_content pg = PT_c pt.

  Definition is_cacheable (pg:page) :=
    page_cacheable pg = true.

  (** Create an empty page of the given type with given owner. *)
  Definition new_page (type:option ptype) (owner:page_owner) : page :=
  match type with
    | Some RW => Page (RW_c None) owner true
    | Some PT => Page (PT_c PT.empty) owner false
    | None    => Page Other_c owner false
  end.

  Definition same_page_type (pg pg':page) : bool :=
    match page_content pg, page_content pg' with
      | RW_c (Some _), RW_c (Some _) => true
      | RW_c None, RW_c None => true
      | PT_c _, PT_c _ => true
      | Other_c, Other_c => true
      | _, _ => false
    end.

  Definition same_page_but_value (pg1 pg2:page) : Prop :=
    page_owned_by pg1 = page_owned_by pg2 /\
    page_cacheable pg1 = page_cacheable pg2 /\
    same_page_type pg1 pg2 = true.

End PAGE.





