(***********************************************************************
 * File: Stealth_action_stream.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Definition infinite action lists.
 *************************************************************************)
Require Import Stealth_actions Stealth_valid Stealth_equivalence
  Stealth_traces Stealth_isolation Stealth_traces_victim
  Stealth_non_stealth_equivalence.

Declare Module TRACES_VIC : STEALTH_TRACES_VICTIM.

Module ACTION_STREAMS.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE STEALTH_EQ
         STEALTH_ISOLATION STEALTH_TRACES STEALTH_NON_STEALTH_EQ.

  Import VALID_SM_STATE.
  Import STEALTH_TRACES.
  Import STEALTH_ISOLATION.
  Import STEALTH_SM_ISOLATION.
  Import STEALTH_EQ.
  Import OBS_LEMMAS.
  Import C.
  Import TRACES_VIC.

  Definition action_stream := Stream sm_action.
  Definition leak_action_stream := Stream leak_action.

  (* (* Actions executed by an OS or by the Hypervisor on behalf of an *)
  (*    OS. *) *)
  (* Definition osi_action (a:sm_action) (osi:os_ident) : Prop := *)
  (*   match a with *)
  (*     | Action Silent => True *)
  (*     | Action (Read _) => True *)
  (*     | Action (Read_hyper _)  => False *)
  (*     | Action (Write _ _) => True *)
  (*     | Action (Write_hyper _ _) => False *)
  (*     | Action (New_untrusted _ _ _) => False *)
  (*     | Action (New_hyper _ _) => False *)
  (*     | Action (Del_untrusted osi' _) => osi = osi' *)
  (*     | Action (Del_hyper _) => False *)
  (*     | Action (Switch _) => False *)
  (*     | Action (Lswitch_untrusted osi' _) => osi = osi' *)
  (*     | Action (Hcall _) => True *)
  (*     | Action (Ret_ctrl) => True *)
  (*     | Action (Chmod) => True *)
  (*     | Action (Page_pin_untrusted osi' _ _) => osi = osi' *)
  (*     | Action (Page_unpin_untrusted osi' _) => osi = osi' *)
  (*     | New_normal osi' _ _ => osi = osi' *)
  (*     | New_sm _ _ => True *)
  (*     | Switch_sm _ => False *)
  (*     | LSwitch_untrusted_sm _ _ => True *)
  (*   end. *)

  (* Definition osi_leak_action (A:leak_action) (osi:os_ident) : Prop := *)
  (*   forall o, osi_action (A o) osi.  *)

  (* CoInductive osi_actions : action_stream -> os_ident -> Prop := *)
  (*   Os_Actions : forall astr a osi, *)
  (*     osi_actions astr osi -> *)
  (*     osi_action a osi -> *)
  (*     osi_actions (Cons a astr) osi. *)

  (* CoInductive osi_leak_actions : leak_action_stream -> os_ident -> Prop := *)
  (*   Os_Leak_Actions : forall astr A osi, *)
  (*     osi_leak_actions astr osi -> *)
  (*     osi_leak_action A osi -> *)
  (*     osi_leak_actions (Cons A astr) osi. *)

  Inductive effect :=
  | EFFRead (addr:vadd)
  | EFFWrite (addr:vadd)
  | EFFSilent.

  Definition effects := Stream effect.

  (* (* Any implementation of our action semantics. *) *)
  (* Parameter exec : state -> sm_action -> option state. *)

  (* Axiom exec_valid : forall s s' a, *)
  (*   exec s a = Some s' -> *)
  (*   valid_sm_exec s a s'. *)

  (* Axiom exec_invalid : forall s a, *)
  (*   exec s a = None -> *)
  (*   ~ defined_action a \/ ~ valid_sm_state s \/ *)
  (*   ~ pre_sm_action a s. *)

  Inductive non_stealth_effect : effect -> Prop :=
  | non_stealth_read: forall va, stealth_va <> va ->
    non_stealth_effect (EFFRead va)
  | non_stealth_write: forall va, stealth_va <> va ->
    non_stealth_effect (EFFWrite va).


  CoInductive same_non_stealth_effects_str :
    effects -> effects -> Prop :=
  | Non_stealth_str : forall astr1 astr2 a,
      non_stealth_effect a ->
      same_non_stealth_effects_str astr1 astr2 ->
      same_non_stealth_effects_str (Cons a astr1) (Cons a astr2)
  | Stealth_str : forall astr1 astr2 a1 a2,
      ~ non_stealth_effect a1 ->
      ~ non_stealth_effect a2 ->
      same_non_stealth_effects_str astr1 astr2 ->
      same_non_stealth_effects_str (Cons a1 astr1) (Cons a2 astr2).

End ACTION_STREAMS.