(***********************************************************************
 * File: Stealth_leakage.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Non-Leakage with StealthMem
 *************************************************************************)
Require Import Stealth_actions Stealth_valid Stealth_non_stealth_equivalence.

Module STEALTH_EQ.
  Import PG C T MM SMEM ST ST_VALID OS_ISO
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_NON_STEALTH_EQ.

Definition same_non_stealth_entries (s1 s2:state) :=
  forall va,
    get_line va <> get_line stealth_va ->
    same_cache_line (smem_c (st_memory s1)) (st_hypervisor s1)
                    (smem_c (st_memory s2)) (st_hypervisor s2)
                    (get_line va).

Hint Resolve same_cache_line_entries_refl
     same_cache_line_entries_symm.

Lemma same_non_stealth_entries_refl :
  forall s, same_non_stealth_entries s s.
Proof.
  unfold same_non_stealth_entries; crush.
Qed.

Lemma same_non_stealth_entries_symm :
  forall s1 s2, same_non_stealth_entries s1 s2 ->
                same_non_stealth_entries s2 s1.
Proof.
  unfold same_non_stealth_entries; crush.
Qed.

Lemma same_non_stealth_entries_trans :
  forall s1 s2 s3,
    same_non_stealth_entries s1 s2 ->
    same_non_stealth_entries s2 s3 ->
    same_non_stealth_entries s1 s3.
Proof.
  intros.
  unfold same_non_stealth_entries; crush.
  apply same_cache_line_entries_trans with
    (smem_c (st_memory s2)) (st_hypervisor s2); crush.
Qed.


Definition osi_stealth_equivalent (s1 s2:state) (osi:os_ident) :=
  STEQ.osi_equivalent s1 s2 osi /\
  osi_non_stealth_equivalent s1 s2 /\
  same_non_stealth_entries s1 s2.

Hint Resolve same_cache_line_entries_refl same_cache_line_entries_symm
     STEQ.osi_equivalent_refl STEQ.osi_equivalent_symm
     osi_non_stealth_equivalent_refl osi_non_stealth_equivalent_symm.

Lemma osi_stealth_equivalent_refl :
  forall s osi, osi_stealth_equivalent s s osi.
Proof.
  unfold osi_stealth_equivalent, same_non_stealth_entries; crush.
Qed.

Lemma osi_stealth_equivalent_symm :
  forall s1 s2 osi, osi_stealth_equivalent s1 s2 osi ->
    osi_stealth_equivalent s2 s1 osi.
Proof.
  unfold osi_stealth_equivalent, same_non_stealth_entries; crush.
Qed.

Lemma osi_stealth_equivalent_trans :
  forall s1 s2 s3 osi, 
    osi_stealth_equivalent s1 s2 osi ->
    osi_stealth_equivalent s2 s3 osi ->
    osi_stealth_equivalent s1 s3 osi.
Proof.
  unfold osi_stealth_equivalent; crush.
  apply STEQ.osi_equivalent_trans with s2; crush.
  apply osi_non_stealth_equivalent_trans with s2; crush.
  apply same_non_stealth_entries_trans with s2; crush.
Qed.


End STEALTH_EQ.
