(***********************************************************************
 * File: Stealth_interleaving.v
 * $Revision: 12473 $
 * $Date: 2014-04-24 14:03:36 -0300 (Thu, 24 Apr 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Interleaving of action streams
 *************************************************************************)
Require Import Sets.Constructive_sets Stealth_actions
  Stealth_valid Stealth_traces Stealth_action_streams
  Stealth_traces_victim Streams.

Module Type INTERLEAVING_OLD.
    Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS
         TRACES_VIC.

  Parameter interleave : state ->
    leak_action_stream -> effects -> leak_state_stream -> Prop.

  Axiom interleave_hd :
    forall astr1 astr2 s0 t,
      interleave s0 astr1 astr2 t ->
      leak_state_state (l_ex_state (hd t)) = s0.

  Axiom interleave_hd_obs :
    forall astr1 astr2 s0 t,
      interleave s0 astr1 astr2 t ->
      leak_state_obs (l_ex_state (hd t)) = Empty_set vadd.

  Axiom interleave_run : forall astr1 astr2 s0 t,
    valid_sm_state s0 ->
    interleave s0 astr1 astr2 t ->
    leak_run t.
  
  Axiom interleave_same_non_stealth_osi_a_actions :
    forall astr1 astr2 astr2' s0 s0' t t',
      same_non_stealth_effects_str astr2 astr2' ->
      interleave s0 astr1 astr2 t ->      
      interleave s0' astr1 astr2' t' ->      
      leak_same_non_stealth_osi_a_actions t t'.

End INTERLEAVING_OLD.

Module PARTIAL_TRACES.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS
         TRACES_VIC STEALTH_EQ STEALTH_ISOLATION.

    Inductive next_sched :=
    | Osi_v_sched : next_sched
    | Osi_a_sched : leak_action -> next_sched.
    Definition partial_trace := list leak_exec_state.

    (* Equivalence of partial traces (same definition as equivalence for
     traces) *)
    Inductive leak_equivalent_partial_traces :
    partial_trace -> partial_trace -> Prop :=
    | eq_partial_trace_nil : leak_equivalent_partial_traces nil nil 
    | eq_partial_trace_osi_a_action: forall (pstr1 pstr2:partial_trace)
      (A1 A2:leak_action) (s1 s2:leak_state),
      leak_equivalent_partial_traces pstr1 pstr2 -> 
      osi_sm_action osi_a (leak_state_state s1) (A1 (leak_state_obs s1)) ->
      osi_sm_action osi_a (leak_state_state s2) (A2 (leak_state_obs s2)) ->
      osi_stealth_equivalent (leak_state_state s1) (leak_state_state s2) osi_a ->
      leak_equivalent_partial_traces (cons (Leak_Ex_State s1 A1 (cache_test osi_a)) pstr1)
                                 (cons (Leak_Ex_State s2 A2 (cache_test osi_a)) pstr2)
    | eq_partial_trace_not_osi_action: forall (pstr1 pstr2:partial_trace)
      (A1 A2:leak_action) (s1 s2:leak_state),
        leak_equivalent_partial_traces pstr1 pstr2 -> 
        osi_sm_action osi_v (leak_state_state s1) (A1 (leak_state_obs s1)) ->
        osi_sm_action osi_v (leak_state_state s2) (A2 (leak_state_obs s2)) ->
        leak_equivalent_partial_traces
          (cons (Leak_Ex_State s1 A1 (cache_test osi_a)) pstr1)
          (cons (Leak_Ex_State s2 A2 (cache_test osi_a)) pstr2).

End PARTIAL_TRACES.

Module Type SCHEDULER.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS
         TRACES_VIC STEALTH_EQ STEALTH_ISOLATION PARTIAL_TRACES.
  
  Definition scheduler := partial_trace -> next_sched.
  Parameter sched : scheduler.

  (* The attacker cannot not distinguish between equivalent partial
     traces, so the scheduler should return the same next action for
     them. *)
  Axiom sched_eq_result : forall pstr1 pstr2,
    leak_equivalent_partial_traces pstr1 pstr2 ->
    sched pstr1 = sched pstr2.
End SCHEDULER.

Module Type INTERLEAVING (S:SCHEDULER).
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS
         TRACES_VIC STEALTH_EQ STEALTH_ISOLATION PARTIAL_TRACES
         S.

  Inductive action_effect : sm_action -> effect -> Prop :=
  | ActEffRead : forall va, action_effect (Action (Read va)) (EFFRead va)
  | ActEffWrite : forall va val, action_effect (Action (Write va val)) (EFFWrite va)
  | ActEffSilent : action_effect (Action Silent) EFFSilent.

  CoInductive interleave' : state -> observation -> 
    partial_trace -> effects -> leak_state_stream -> Prop :=
  | IntVictimCons :
      forall s0 o pstr v_str t a eff,
        interleave' (leak_state_state (l_ex_state (hd t)))
                    (leak_state_obs (l_ex_state (hd t)))
                    (cons (Leak_Ex_State (Leak_State s0 o)
                                           (fun _ => a) (cache_test osi_a)) pstr)
                    v_str t ->
        leak_trace t ->
        action_effect a eff ->
        cache_test osi_a s0 o (leak_state_state (l_ex_state (hd t))) =
          leak_state_obs (l_ex_state (hd t)) ->
        osi_sm_action osi_v s0 a ->
        ~ osi_sm_action osi_a s0 a ->
        valid_sm_exec s0 a (leak_state_state (l_ex_state (hd t))) ->
        sched pstr = Osi_v_sched ->
        interleave' s0 o pstr (Cons eff v_str)
                    (Cons (Leak_Ex_State (Leak_State s0 o)
                                         (fun _ => a) (cache_test osi_a)) t)
  | IntAttackerCons :
      forall s0 o pstr v_str t A,
        interleave' (leak_state_state (l_ex_state (hd t)))
                    (leak_state_obs (l_ex_state (hd t)))
                    (cons (Leak_Ex_State (Leak_State s0 o) A (cache_test osi_a))
                            pstr) v_str t ->
        leak_trace t ->
        cache_test osi_a s0 o (leak_state_state (l_ex_state (hd t))) =
          leak_state_obs (l_ex_state (hd t)) ->
        osi_sm_action osi_a s0 (A o) ->
        valid_sm_exec s0 (A o) (leak_state_state (l_ex_state (hd t))) ->
        sched pstr = Osi_a_sched A ->
        interleave' s0 o pstr v_str
                    (Cons (Leak_Ex_State (Leak_State s0 o) A (cache_test osi_a)) t).
  Definition interleave s0 := interleave' s0 (Empty_set vadd) nil.
End INTERLEAVING.

Module INTERLEAVING_PROOF (S:SCHEDULER) (I:INTERLEAVING S).
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS
         TRACES_VIC STEALTH_EQ STEALTH_ISOLATION
         PARTIAL_TRACES S I.

  Lemma interleave_hd :
    forall astr2 s0 t,
      interleave s0 astr2 t ->
      leak_state_state (l_ex_state (hd t)) = s0.
  Proof.
    intros.
    unfold interleave in *.
    inversion H; crush.
  Qed.

  Lemma interleave_hd_obs :
    forall astr2 s0 t,
      interleave s0 astr2 t ->
      leak_state_obs (l_ex_state (hd t)) = Empty_set vadd.
  Proof.
    intros.
    unfold interleave in *.
    inversion H; crush.
  Qed.

  Lemma interleave_run : forall astr2 s0 t,
    valid_sm_state s0 ->
    interleave s0 astr2 t ->
    leak_run t.
  Proof.
    intros.
    inversion H0; crush;
    repeat (constructor; crush).
  Qed.

  Lemma interleave'_same_non_stealth_osi_a_actions :
    forall pstr pstr' astr2 astr2' s0 s0' o t t',
      valid_sm_state s0 ->
      valid_sm_state s0' ->
      leak_equivalent_partial_traces pstr pstr' ->
      same_non_stealth_effects_str astr2 astr2' ->
      interleave' s0 o pstr astr2 t ->      
      interleave' s0' o pstr' astr2' t' ->      
      osi_stealth_equivalent s0 s0' osi_a ->
      leak_same_non_stealth_osi_a_actions t t'.
  Proof.
    Local Hint Constructors leak_equivalent_partial_traces.
    cofix.
    intros.
    inversion H3; subst; simpl in *;
      inversion H4; subst; simpl in *.

    inversion H2; subst; simpl in *. (* same_non_stealth_effects_str *)
    destruct eff0;
      inversion H16; inversion H8; subst; simpl in *. (* action_effect *)
    (* Read *)
    apply Leak_non_stealth_osi_v_action; crush; (* non_stealth_effect *)
      repeat constructor; inversion H25; crush.
    eapply interleave'_same_non_stealth_osi_a_actions
      with (pstr:=({|
          l_ex_state := {| leak_state_state := s0; leak_state_obs := o |};
          l_ex_action := fun _ : observation => Action (Read addr);
          l_ex_function := cache_test osi_a |} :: pstr))
           (pstr':={|
           l_ex_state := {| leak_state_state := s0'; leak_state_obs := o |};
           l_ex_action := fun _ : observation => Action (Read addr);
           l_ex_function := cache_test osi_a |} :: pstr')
           (astr2:=v_str) (astr2':=v_str0)
           (s0:=leak_state_state (l_ex_state (hd t0)))
           (s0':=leak_state_state (l_ex_state (hd t)))
           (o:=leak_state_obs (l_ex_state (hd t0))); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0'); crush.
    erewrite NON_LEAKAGE.same_obs_two_exec
      with (osi:=osi_a)
             (s1:={| leak_state_state := s0; leak_state_obs := o |})
             (s2:={| leak_state_state := s0'; leak_state_obs := o |})
             (A:=fun _ => Action (Read addr)); crush.
    constructor; crush.
    constructor; crush.
    inversion H18; inversion H10; crush.
    inversion H24; inversion H28; crush.
    apply STEALTH_SM_ISOLATION.stealth_os_isolation
      with s0 s0' (Action (Read addr)); crush.
    inversion H18; inversion H10; crush.
    inversion H24; inversion H28; crush.
    (* Write *)
    apply Leak_non_stealth_osi_v_write with (va:=addr) (val1:=val0) (val2:=val);
      crush;
      try (inversion H25; constructor; crush; fail).
    eapply interleave'_same_non_stealth_osi_a_actions
      with (pstr:=({|
          l_ex_state := {| leak_state_state := s0; leak_state_obs := o |};
          l_ex_action := fun _ : observation => Action (Write addr val0);
          l_ex_function := cache_test osi_a |} :: pstr))
           (pstr':={|
           l_ex_state := {| leak_state_state := s0'; leak_state_obs := o |};
           l_ex_action := fun _ : observation => Action (Write addr val);
           l_ex_function := cache_test osi_a |} :: pstr')
           (astr2:=v_str) (astr2':=v_str0)
           (s0:=leak_state_state (l_ex_state (hd t0)))
           (s0':=leak_state_state (l_ex_state (hd t)))
           (o:=leak_state_obs (l_ex_state (hd t0))); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0'); crush.
    erewrite NON_LEAKAGE.same_obs_two_write
      with (osi:=osi_a)
             (s1:={| leak_state_state := s0; leak_state_obs := o |})
             (s2:={| leak_state_state := s0'; leak_state_obs := o |})
             (A1:=fun _ => Action (Write addr val0))
             (A2:=fun _ => Action (Write addr val)); crush.
    constructor; crush.
    constructor; crush.
    inversion H25; crush.
    inversion H18; inversion H10; crush.
    inversion H24; inversion H28; crush.
    inversion H18; inversion H10; crush.
    inversion H23; inversion H26; crush.
    apply STEALTH_SM_ISOLATION.non_stealth_write_eq
      with s0 s0' addr val0 val; crush.
    inversion H25; crush.
    inversion H18; inversion H10; crush.
    inversion H24; inversion H28; crush.
    inversion H18; inversion H10; crush.
    inversion H23; inversion H26; crush.
    (* Silent *)
    inversion H25.
    (* Stealth effect *)
    apply Leak_stealth_osi_v_action1; crush.
    apply Leak_stealth_osi_v_action2; crush.
    eapply interleave'_same_non_stealth_osi_a_actions
      with (pstr:=({|
          l_ex_state := {| leak_state_state := s0; leak_state_obs := o |};
          l_ex_action := fun _ : observation => a;
          l_ex_function := cache_test osi_a |} :: pstr))
           (pstr':={|
           l_ex_state := {| leak_state_state := s0'; leak_state_obs := o |};
           l_ex_action := fun _ : observation => a0;
           l_ex_function := cache_test osi_a |} :: pstr')
           (astr2:=v_str) (astr2':=v_str0)
           (s0:=leak_state_state (l_ex_state (hd t0)))
           (s0':=leak_state_state (l_ex_state (hd t)))
           (o:=leak_state_obs (l_ex_state (hd t))); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0'); crush.
    erewrite <- NON_LEAKAGE.same_obs_step
      with (osi:=osi_a)
             (s:={| leak_state_state := s0'; leak_state_obs := o |})
             (s':=l_ex_state (hd t))
             (A:=fun _ => a0); crush.
    replace o with
      (leak_state_obs {| leak_state_state := s0; leak_state_obs := o |});
      [|crush].
    erewrite NON_LEAKAGE.same_obs_step
      with (osi:=osi_a)
             (s:={| leak_state_state := s0; leak_state_obs := o |})
             (s':=l_ex_state (hd t0))
             (A:=fun _ => a) at 1; crush.
    constructor; crush.
    inversion H8; crush;
      inversion H22; crush; elim H26;
      constructor; crush.
    constructor; crush.
    inversion H16; crush;
      inversion H22; crush; elim H27;
      constructor; crush.
    apply osi_stealth_equivalent_trans with s0.
    apply osi_stealth_equivalent_symm.
    apply STEALTH_SM_ISOLATION.stealth_isolation_not_osi with a; crush.
    inversion H8; crush;
      inversion H22; crush;
      elim H26; constructor; crush.
    apply osi_stealth_equivalent_trans with s0'; crush.
    apply STEALTH_SM_ISOLATION.stealth_isolation_not_osi with a0; crush.
    inversion H16; crush; 
      inversion H22; crush;
      elim H27; constructor; crush.
    inversion H16; crush; 
      inversion H22; crush;
      elim H27; constructor; crush.
    inversion H8; crush;
      inversion H22; crush;
      elim H26; constructor; crush.

    clear interleave'_same_non_stealth_osi_a_actions.
    rewrite sched_eq_result with pstr pstr' in H13; crush.
    clear interleave'_same_non_stealth_osi_a_actions.
    rewrite sched_eq_result with pstr pstr' in H11; crush.
    rewrite sched_eq_result with pstr pstr' in H11; [|crush].
    rewrite H11 in H17; injection H17; intros; subst A0; clear H17.

    (* Attacker action *)
    apply Leak_osi_a_action; crush.
    eapply interleave'_same_non_stealth_osi_a_actions
      with (pstr:=({|
          l_ex_state := {| leak_state_state := s0; leak_state_obs := o |};
          l_ex_action := A;
          l_ex_function := cache_test osi_a |} :: pstr))
           (pstr':={|
           l_ex_state := {| leak_state_state := s0'; leak_state_obs := o |};
           l_ex_action := A;
           l_ex_function := cache_test osi_a |} :: pstr')
           (astr2:=astr2) (astr2':=astr2')
           (s0:=leak_state_state (l_ex_state (hd t0)))
           (s0':=leak_state_state (l_ex_state (hd t)))
           (o:=leak_state_obs (l_ex_state (hd t0))); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0); crush.
    eapply VALID_SM_STATE_INV.valid_sm_state_inv with (s:=s0'); crush.
    erewrite NON_LEAKAGE.same_obs_two_exec
      with (osi:=osi_a)
             (s1:={| leak_state_state := s0; leak_state_obs := o |})
             (s2:={| leak_state_state := s0'; leak_state_obs := o |})
             (A:=A); crush.
    constructor; crush.
    constructor; crush.
    inversion H5; crush.
    inversion H17; crush.
    destruct H20; crush.
    destruct osi_v_or_osi_a with (st_active_os s0); crush.
    destruct osi_v_or_osi_a with (st_active_os s0'); crush.

    apply STEALTH_SM_ISOLATION.stealth_os_isolation
      with s0 s0' (A o); crush.
    inversion H5; crush.
    inversion H17; crush.
    destruct H20; crush.
    destruct osi_v_or_osi_a with (st_active_os s0); crush.
    destruct osi_v_or_osi_a with (st_active_os s0'); crush.
    Grab Existential Variables. (* ??? *)
    exact (fun _ => Action Silent).
    exact (fun _ => Action Silent).
    exact (fun _ => Action Silent).
    exact (fun _ => Action Silent).
    exact (fun _ => Action Silent).
    exact (fun _ => Action Silent).
    exact (fun _ => Action Silent).
    exact (fun _ => Action Silent).
  Qed.

  Lemma interleave_same_non_stealth_osi_a_actions :
    forall astr2 astr2' s0 s0' t t',
      same_non_stealth_effects_str astr2 astr2' ->
      valid_sm_state s0 ->
      valid_sm_state s0' ->
      interleave s0  astr2 t ->      
      interleave s0' astr2' t' ->      
      osi_stealth_equivalent s0 s0' osi_a ->
      leak_same_non_stealth_osi_a_actions t t'.
  Proof.
    intros.
    eapply interleave'_same_non_stealth_osi_a_actions
      with (s0:=s0) (s0':=s0'); eauto.
  Qed.

End INTERLEAVING_PROOF.


Module Type RR_SCHEDULER <: SCHEDULER.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS
         TRACES_VIC STEALTH_EQ STEALTH_ISOLATION PARTIAL_TRACES.
  
  Definition scheduler := partial_trace -> next_sched.

  (* Timeslice for each OS *)
  Parameter N : nat.
  (* The first OS to execute *)
  Parameter first_os : os_ident.

  (* Function to select the next attacker action, given a partial trace *)
  Parameter attacker : partial_trace -> leak_action.
  (* The attacker function cannot distinguish equivalent partial traces. *)
  Axiom attacker_eq_result : forall pstr1 pstr2,
    leak_equivalent_partial_traces pstr1 pstr2 ->
    attacker pstr1 = attacker pstr2.

  Definition get_next_osi_action (pstr:partial_trace) (osi:os_ident) :=
    if os_ident_eq first_os osi_a then
      Osi_a_sched (attacker pstr)
    else
      Osi_v_sched.

  Lemma osi_sm_action_dec :
    forall osi s a, {osi_sm_action osi s a}+{~osi_sm_action osi s a}.
  Proof.
    Local Hint Constructors osi_sm_action OS_ISO.osi_action.
    sm_action_induction a;
      [clear Case; STAC.action_induction a | | | | ];
    try (destruct (os_ident_eq (st_active_os s) osi); crush;
         [destruct (st_aos_activity s) eqn:?; [|right] |right]; crush;
         inversion H; inversion H1; crush; fail);
    try (destruct (os_ident_eq o osi); crush;
         destruct (os_ident_eq (st_active_os s) osi); crush;
         [destruct (st_aos_activity s) eqn:?; [|right] |right]; crush;
         inversion H; inversion H1; crush; fail);
    try (destruct (os_ident_eq o osi); crush;
         destruct (os_ident_eq (st_active_os s) osi); crush; right; crush;
         inversion H; inversion H1; crush; fail);
    try (destruct (os_ident_eq (st_active_os s) osi); crush; right; crush;
         inversion H; inversion H1; crush; fail).
  Qed.
    
  Fixpoint count_osi_actions (osi:os_ident) (pstr:partial_trace) : nat :=
    match pstr with
      | nil => 0
      | s::pstr' =>
        if osi_sm_action_dec osi
             (leak_state_state (l_ex_state s))
             ((l_ex_action s) (leak_state_obs (l_ex_state s))) then
          1+(count_osi_actions osi pstr')
        else 0
    end.

  Definition sched (pstr:partial_trace) : next_sched :=
    if eq_nat_dec (count_osi_actions osi_v pstr) 0 then
      if eq_nat_dec (count_osi_actions osi_a pstr) 0 then
        get_next_osi_action pstr first_os
      else if ge_dec (count_osi_actions osi_a pstr) N then
        get_next_osi_action pstr osi_v
      else
        get_next_osi_action pstr osi_a
    else if ge_dec (count_osi_actions osi_v pstr) N then
        get_next_osi_action pstr osi_a
      else
        get_next_osi_action pstr osi_v.


  (* The attacker cannot not distinguish between equivalent partial *)
  (*      traces, so the scheduler should return the same next action for *)
  (*      them. *)
  Lemma sched_eq_result : forall pstr1 pstr2,
    leak_equivalent_partial_traces pstr1 pstr2 ->
    sched pstr1 = sched pstr2.
  Proof.
    Local Hint Constructors leak_equivalent_partial_traces.
    Local Hint Resolve attacker_eq_result.
    induction pstr1; induction pstr2; crush;
      inversion_clear H; crush.
    unfold sched; crush;
      unfold get_next_osi_action; crush.
    unfold sched; crush;
      unfold get_next_osi_action; crush.
  Qed.

End RR_SCHEDULER.






