(***********************************************************************
 * File: Stealth_isolation.v
 * $Revision: 12857 $
 * $Date: 2015-07-28 19:11:12 -0300 (Tue, 28 Jul 2015) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Proof of OS Isolation for StealthMem actions.
 *************************************************************************)
Require Import Stealth_actions Stealth_valid Stealth_lemmas.

Module STEALTH_ISOLATION.
  Import PG C T MM SMEM ST OS_ISO ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE STEALTH_LEMMAS.

  Inductive osi_sm_action (osi:os_ident) (s:state) : sm_action -> Prop  :=
  | old_osi_action : forall (a:action), osi_action osi s a -> osi_sm_action osi s (Action a)
  | new_normal_osi  : forall va pa, osi_sm_action osi s (New_normal osi va pa)
  | new_sm_osi  : forall pa, osi_sm_action osi s (New_sm osi pa)
  | lswitch_untrusted_osi  : forall pa, osi_sm_action osi s (LSwitch_untrusted_sm osi pa)
  | switch_from_osi : forall o, st_active_os s = osi -> osi_sm_action osi s (Switch_sm o)
  | switch_to_osi : osi_sm_action osi s (Switch_sm osi).

  Lemma osi_sm_action_old :
    forall osi s a,
      osi_sm_action osi s (Action a) -> osi_action osi s a.
  Proof.
    Local Hint Constructors osi_action osi_sm_action.
    intros; destruct a; inversion H; crush.
  Qed.
  Lemma osi_sm_action_old_back :
    forall osi s a,
      osi_action osi s a -> osi_sm_action osi s (Action a).
  Proof.
    Local Hint Constructors osi_action osi_sm_action.
    intros; destruct a; inversion H; crush.
  Qed.
  Local Hint Resolve osi_sm_action_old osi_sm_action_old_back .

  Lemma new_sm_equivalent_not_osi :
    forall s s' osi osi' pa,
      osi <> osi' ->
      valid_sm_state s ->
      valid_sm_exec s (New_sm osi' pa) s' ->
      STEQ.osi_equivalent s s' osi.
  Proof.
    intros.
    remember (os_ident_eq osi' (st_active_os s)).
    destruct s0; crush; clear Heqs0.

    destruct new_sm_eq_old_new_active with s s' pa; crush.
    apply STEQ.osi_equivalent_trans with x; crush.
    apply os_isolation_not_osi with
      (New_untrusted (st_active_os s) stealth_va pa); crush.
    unfold valid_sm_state in *; crush.
    inversion H4; crush.
    apply os_isolation_not_osi with
      (Read_hyper stealth_va); crush.
    apply EXEC_INV.valid_state_invariant with s
      (New_untrusted (st_active_os s) stealth_va pa); crush.
    unfold valid_sm_state in *; crush.
    inversion H4; crush.
    inversion H2; crush.
    unfold ACT.post_new, ACT.post_all_new in *; crush.

    destruct new_sm_eq_old_new with s s' osi' pa; crush.
    apply os_isolation_not_osi with
      (New_untrusted osi' stealth_va pa); crush.
    unfold ACT.valid_exec; crush.
    unfold valid_sm_state in *; crush.
    inversion H4; crush.
    unfold ACT.pre_new in *; crush.
  Qed.

  Lemma ctxt_switch_equivalent_smem :
    forall s s' osi osi' pt_padd pt_addr pt_page pt,
      valid_sm_state s ->
      osi_pa_to_ma s osi' pt_padd = Some pt_addr ->
      get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
      PT_c pt = page_content pt_page ->
      st_memory s' = ctxt_switch_new_smem s pt_addr ->
      st_hypervisor s' = st_hypervisor s ->
      STEQ.osi_equivalent_smem s s' osi.
  Proof.
    unfold STEQ.osi_equivalent_smem,
      STEQ.osi_pt_pages_eq, osi_pa_to_ma, osi_pa_to_pg; crush;
      rewrite H3, H4 in *; clear dependent s'; crush.
    exists pt_page1; crush.
    rewrite osi_pa_to_pg_ctxt_switch_new_smem with (osi:=osi')
      (pt_padd:=pt_padd) (osi':=osi) (pa':=pt_padd0); crush.
    unfold osi_pa_to_ma; crush.
    unfold osi_pa_to_ma; crush.
    unfold madd_type; crush.
    rewrite <- H6; crush.
    exists mp1; crush.
    rewrite page_of_vadd_madd_smem_ctxt_switch_new_smem
      with (osi:=osi') (osi':=osi) (pt_padd:=pt_padd) 
                       (pt_padd':=pt_padd0) (pt_addr':=i1); crush.
    unfold osi_pa_to_ma; crush.
    unfold osi_pa_to_ma; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H6; crush.

    exists pt_page1; crush.
    rewrite osi_pa_to_pg_ctxt_switch_new_smem with (osi:=osi')
      (pt_padd:=pt_padd) (osi':=osi) (pa':=pt_padd0) in H5; crush.
    unfold osi_pa_to_ma; crush.
    unfold osi_pa_to_ma; crush.
    rewrite <- madd_type_ctxt_switch_new_smem
      with (osi:=osi') (pt_padd:=pt_padd) (pt_addr:=pt_addr) (osi':=osi)
      (pa':=pt_padd0); crush.
    unfold madd_type; crush.
    rewrite <- H6; crush.
    unfold osi_pa_to_ma; crush.
    unfold osi_pa_to_ma; crush.
    exists mp1; crush.
    rewrite page_of_vadd_madd_smem_ctxt_switch_new_smem
      with (osi:=osi') (osi':=osi) (pt_padd:=pt_padd) (pt_padd':=pt_padd0) 
                       (pt_addr':=i1) in H7; crush.
    unfold osi_pa_to_ma; crush.
    unfold osi_pa_to_ma; crush.
    assert (madd_of_vadd (smem_mm (ctxt_switch_new_smem s pt_addr)) i1 va = Some ma1).
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H6; crush.
    rewrite madd_of_vadd_ctxt_switch_new_smem with (osi:=osi') (pt_padd':=pt_padd) in H9; crush.
    unfold valid_sm_state in *; crush.
    unfold osi_pa_to_ma; crush.
  Qed.

Lemma ctxt_switch_equivalent_hyper :
  forall s s' osi osi' pt_padd pt_addr pt_page pt,
    valid_sm_state s ->
    osi_pa_to_ma s osi' pt_padd = Some pt_addr ->
    get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
    PT_c pt = page_content pt_page ->
    st_memory s' = ctxt_switch_new_smem s pt_addr ->
    st_hypervisor s' = st_hypervisor s ->
    STEQ.osi_equivalent_hyper s s' osi.
Proof.
  unfold valid_sm_state, STEQ.osi_equivalent_hyper,
    STEQ.osi_hyper_eq, osi_pa_to_ma, osi_pa_to_pg; crush;
  rewrite H3, H4 in *; intros; simpl in *.
  assert (ACT.no_va_mapped_to_ma osi ma1 s'); crush.
  eapply no_va_mapped_eq with (s:=s) (pa:=pa) (ma1:=ma1); crush.
  destruct ctxt_switch_equivalent_smem
    with (s:=s) (s':=s') (pt_padd:=pt_padd) (pt_addr:=pt_addr)
    (osi:=osi) (osi':=osi') (pt_page:=pt_page) (pt:=pt);  crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H4; crush.
  rewrite <- ctxt_switch_no_va_mapped_same_mem
    with (osi':=osi') (pt_padd':=pt_padd) (osi:=osi) (pa:=pa); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.

  assert (ACT.no_va_mapped_to_ma osi ma1 s); crush.
  eapply no_va_mapped_eq with (s:=s') (pa:=pa) (ma1:=ma1); crush.
  destruct ctxt_switch_equivalent_smem
    with (s:=s) (s':=s') (pt_padd:=pt_padd) (pt_addr:=pt_addr)
    (osi:=osi) (osi':=osi') (pt_page:=pt_page) (pt:=pt);  crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H4; crush.
  unfold osi_pa_to_ma; crush.
  rewrite  ctxt_switch_no_va_mapped_same_mem
    with (osi':=osi') (pt_padd':=pt_padd) (pt_addr':=pt_addr)
         (osi:=osi) (pa:=pa); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma; crush.
Qed.  


Lemma switch_sm_equivalent_not_osi :
  forall s s' osi osi',
    osi <> osi' ->
    osi <> (st_active_os s) ->
    valid_sm_state s ->
    valid_sm_exec s (Switch_sm osi') s' ->
    STEQ.osi_equivalent s s' osi.
Proof.
  intros.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'; crush.
  inversion H2; crush.
  unfold post_sm_switch in *; crush.
  rewrite H17 with pt_addr; crush.
  2:unfold osi_current_page_table; crush.
  unfold STEQ.osi_equivalent; crush.
  apply ctxt_switch_equivalent_smem with osi' (os_current_page x)
    pt_addr pt_page pt; crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.

  apply ctxt_switch_equivalent_hyper with osi' (os_current_page x)
    pt_addr pt_page pt; crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
Qed.

Lemma lswitch_sm_equivalent_not_osi :
  forall s s' osi osi' pt_padd,
    osi <> osi' ->
    valid_sm_state s ->
    valid_sm_exec s (LSwitch_untrusted_sm osi' pt_padd) s' ->
    STEQ.osi_equivalent s s' osi.
Proof.
  intros.
  unfold valid_sm_state in *; crush.
  inversion H1; crush.
  unfold ACT.pre_lswitch in *; crush.
  unfold post_sm_lswitch in *; crush.
  rewrite H10 with x; crush.
  unfold STEQ.osi_equivalent; crush.
  rewrite OSS.get_set_distinct; crush.
  unfold madd_type in *; crush.
  remember (page_content p).
  destruct c; crush.
  unfold valid_sm_state; crush.
  STAC.destruct_valid_state.
  unfold valid_hypervisor in H_valid_hyper.
  unfold osi_pa_to_ma in *; crush.
  
  unfold madd_type in *; crush.
  remember (page_content p).
  destruct c; crush.
  STAC.destruct_valid_state.
  unfold valid_hypervisor in H_valid_hyper.
  STAC.get_aos_data.
  apply ctxt_switch_equivalent_smem with
    (st_active_os s) pt_padd x p p0; crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.

  unfold madd_type in *; crush.
  destruct (page_content p) eqn:?; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  apply ctxt_switch_equivalent_hyper with
    (st_active_os s) pt_padd x p p0; crush.
  unfold valid_sm_state; crush.

  rewrite H10 with x; crush; clear H10.
  unfold STEQ.osi_equivalent; crush.
  destruct (os_ident_eq (st_active_os s) osi); crush.
  rewrite OSS.get_set_distinct; crush.
  unfold STEQ.osi_equivalent_smem, STEQ.osi_pt_pages_eq;
    repeat eexists; crush.
  unfold STEQ.osi_equivalent_hyper, STEQ.osi_pt_pages_eq;
    repeat eexists; crush.
Qed.

Theorem stealth_isolation_not_osi :
  forall (s s':state) (a:sm_action) (osi:os_ident),
  valid_sm_exec s a s' ->
  valid_sm_state s ->
  ~ osi_sm_action osi s a ->
  STEQ.osi_equivalent s s' osi.
Proof.
  intros;
  sm_action_induction a;
    [clear Case; STAC.action_induction a | | | | ]; crush;
      try match goal with
        | [ H1 : valid_sm_exec ?s (Action ?a) ?s',
            H2 : ~ osi_sm_action ?osi ?s (Action ?a) |-
            STEQ.osi_equivalent ?s ?s' ?osi] =>
        apply os_isolation_not_osi with a;
          unfold valid_sm_exec, ACT.valid_exec,
            valid_sm_state in *; crush
      end.

  Local Hint Constructors osi_sm_action.

  (* New normal *)
  destruct (os_ident_eq osi o); crush.
  elim H1; crush.
  apply os_isolation_not_osi with (New_untrusted o v p); crush.
  apply new_normal_eq_old_new; crush.
  unfold valid_sm_state in *; crush.
  inversion H2; crush.
  unfold valid_sm_exec in *; crush;
    unfold pre_new_normal, pre_all_new in *; crush.

  (* New stealth *)
  destruct (os_ident_eq osi o); crush.
  elim H1; crush.
  remember (os_ident_eq o (st_active_os s)).
  destruct s0; crush.
  destruct new_sm_eq_old_new_active with s s' p; crush.
  apply STEQ.osi_equivalent_trans with x; crush.
  apply os_isolation_not_osi with (New_untrusted (st_active_os s) stealth_va p); crush.
  unfold valid_sm_state in *; crush.
  inversion H4; crush.
  apply os_isolation_not_osi with (Read_hyper stealth_va); crush.
  apply EXEC_INV.valid_state_invariant with s
    (New_untrusted (st_active_os s) stealth_va p); crush.
  unfold valid_sm_state in *; crush.
  inversion H4; crush.
  inversion H2; crush.
  unfold ACT.post_new, ACT.post_all_new in *; crush.
  apply os_isolation_not_osi with (New_untrusted o C.stealth_va p); crush.
  apply new_sm_eq_old_new; crush.
  unfold valid_sm_state in *; crush.
  inversion H2; crush.
  unfold valid_sm_exec in *; crush;
    unfold pre_new_sm, pre_all_new in *; crush.

  (* Switch *)
  apply switch_sm_equivalent_not_osi with o; crush.

  (* Lswitch *)
  apply lswitch_sm_equivalent_not_osi with o p; crush.
Qed.

Lemma new_normal_osi_equivalent :
  forall s1 s2 s1' s2' osi osi' va' pa',
      valid_sm_state s1 ->
      valid_sm_state s2 ->
      valid_sm_exec s1 (New_normal osi' va' pa') s1' ->
      valid_sm_exec s2 (New_normal osi' va' pa') s2' ->
      STEQ.osi_equivalent s1 s2 osi ->
      STEQ.osi_equivalent s1' s2' osi.
Proof.
  intros.
  assert (ACT.valid_exec s1 (New_untrusted osi' va' pa') s1').
  apply new_normal_eq_old_new; auto.
  assert (ACT.valid_exec s2 (New_untrusted osi' va' pa') s2').
  apply new_normal_eq_old_new; auto.
  clear H1 H2.
  eapply OS_ISO.os_isolation; crush.
  unfold valid_sm_state in *; crush.
  unfold valid_sm_state in *; crush.
Qed.

Lemma new_sm_osi_equivalent :
  forall s1 s2 s1' s2' osi osi' pa',
      valid_sm_state s1 ->
      valid_sm_state s2 ->
      valid_sm_exec s1 (New_sm osi' pa') s1' ->
      valid_sm_exec s2 (New_sm osi' pa') s2' ->
      STEQ.osi_equivalent s1 s2 osi ->
      STEQ.osi_equivalent s1' s2' osi.
Proof.
  intros.
  assert (valid_state s1).
  unfold valid_sm_state in *; crush.
  assert (valid_state s2).
  unfold valid_sm_state in *; crush.

  destruct (os_ident_eq osi' (st_active_os s1)); crush.
  destruct (os_ident_eq (st_active_os s1) (st_active_os s2)); crush.

  (* osi' active in both states *)
  destruct new_sm_eq_old_new_active with s1 s1' pa'; crush.
  destruct new_sm_eq_old_new_active with s2 s2' pa'; crush.
  apply STEQ.osi_equivalent_trans with x.
  apply STEQ.osi_equivalent_symm.
  apply OS_ISO.os_isolation_not_osi with (Read_hyper stealth_va); crush.
  apply EXEC_INV.valid_state_invariant
    with s1 (New_untrusted (st_active_os s1) stealth_va pa'); crush.
  inversion H10; crush.
  inversion H10; crush.
  inversion H7; crush; unfold ACT.pre_read_hyper in *; crush.
  apply STEQ.osi_equivalent_trans with x0 ;crush.
  apply OS_ISO.os_isolation with s1 s2
    (New_untrusted (st_active_os s1) stealth_va pa'); crush.
  apply OS_ISO.os_isolation_not_osi with (Read_hyper stealth_va); crush.
  apply EXEC_INV.valid_state_invariant
    with s2 (New_untrusted (st_active_os s1) stealth_va pa'); crush.
  inversion H10; crush.
  inversion H10; crush.
  inversion H9; crush; unfold ACT.pre_read_hyper in *; crush.

  (* osi' active in s1 only *)
  destruct new_sm_eq_old_new_active with s1 s1' pa'; crush.
  destruct new_sm_eq_old_new with s2 s2' (st_active_os s1) pa'; crush.
  apply STEQ.osi_equivalent_trans with x.
  apply STEQ.osi_equivalent_symm.
  apply OS_ISO.os_isolation_not_osi with (Read_hyper stealth_va); crush.
  apply EXEC_INV.valid_state_invariant
    with s1 (New_untrusted (st_active_os s1) stealth_va pa'); crush.
  inversion H10; crush.
  inversion H10; crush.
  inversion H7; crush; unfold ACT.pre_read_hyper in *; crush.
  apply OS_ISO.os_isolation with s1 s2
    (New_untrusted (st_active_os s1) stealth_va pa'); crush.
  unfold ACT.valid_exec; crush.
  
  destruct (os_ident_eq osi' (st_active_os s2)); crush.
  (* osi' active in s2 only *)
  destruct new_sm_eq_old_new with s1 s1' (st_active_os s2) pa'; crush.
  destruct new_sm_eq_old_new_active with s2 s2' pa'; crush.
  apply STEQ.osi_equivalent_trans with x.
  apply OS_ISO.os_isolation with s1 s2
    (New_untrusted (st_active_os s2) stealth_va pa'); crush.
  unfold ACT.valid_exec; crush.
  apply OS_ISO.os_isolation_not_osi with (Read_hyper stealth_va); crush.
  apply EXEC_INV.valid_state_invariant
    with s2 (New_untrusted (st_active_os s2) stealth_va pa'); crush.
  inversion H10; crush.
  inversion H10; crush.
  inversion H9; crush; unfold ACT.pre_read_hyper in *; crush.

  (* osi' is not active *)
  destruct new_sm_eq_old_new with s1 s1' osi' pa'; crush.
  destruct new_sm_eq_old_new with s2 s2' osi' pa'; crush.
  apply OS_ISO.os_isolation with s1 s2
    (New_untrusted osi' stealth_va pa'); crush.
  unfold ACT.valid_exec; crush.
  unfold ACT.valid_exec; crush.
Qed.  

Lemma switch_sm_osi_equivalent :
  forall s1 s2 s1' s2' osi osi',
      valid_sm_state s1 ->
      valid_sm_state s2 ->
      valid_sm_exec s1 (Switch_sm osi') s1' ->
      valid_sm_exec s2 (Switch_sm osi') s2' ->
      STEQ.osi_equivalent s1 s2 osi ->
      STEQ.osi_equivalent s1' s2' osi.
Proof.
  Local Hint Resolve ctxt_switch_equivalent_smem.
  intros.
  generalize H, H0; unfold valid_sm_state; crush.
  unfold valid_sm_exec in *; crush.
  unfold ACT.pre_switch, post_sm_switch in *; crush.
  STAC.destruct_valid_state_H H4.
  destruct H_valid_os_idents with osi'; crush.
  STAC.get_osi_current_pt_page osi'.
  rewrite H21 with pt_addr in *;
    unfold osi_current_page_table; crush; clear H21.
  clear H_valid_currpg.
  STAC.destruct_valid_state_H H5.
  destruct H_valid_os_idents0 with osi'; crush.
  STAC.get_osi_current_pt_page osi'.
  rewrite H19 with pt_addr0 in *;
    unfold osi_current_page_table; crush; clear H19.

  unfold STEQ.osi_equivalent in *; crush.
  destruct H3; crush.
  destruct (os_ident_eq osi' (st_active_os s1)); crush.
  destruct (os_ident_eq osi' osi); crush.
  unfold hyper_running in *; crush.
  left; crush.

  apply STEQ.osi_equivalent_smem_trans with s1.
  apply STEQ.osi_equivalent_smem_symm.
  apply ctxt_switch_equivalent_smem
    with osi' (os_current_page x) pt_addr pt_page pt; crush.
  unfold osi_pa_to_ma; crush.
  apply STEQ.osi_equivalent_smem_trans with s2; auto.
  apply ctxt_switch_equivalent_smem
    with osi' (os_current_page x1) pt_addr0 pt_page0 pt0; crush.
  unfold osi_pa_to_ma; crush.

  apply STEQ.osi_equivalent_hyper_trans with s1.
  apply STEQ.osi_equivalent_hyper_symm.
  apply ctxt_switch_equivalent_hyper
    with osi' (os_current_page x) pt_addr pt_page pt; crush.
  unfold osi_pa_to_ma; crush.
  apply STEQ.osi_equivalent_hyper_trans with s2; auto.
  apply ctxt_switch_equivalent_hyper
    with osi' (os_current_page x1) pt_addr0 pt_page0 pt0; crush.
  unfold osi_pa_to_ma; crush.
Qed.

Lemma lswitch_sm_osi_equivalent :
  forall s1 s2 s1' s2' osi osi' pa,
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    valid_sm_exec s1 (LSwitch_untrusted_sm osi' pa) s1' ->
    valid_sm_exec s2 (LSwitch_untrusted_sm osi' pa) s2' ->
    STEQ.osi_equivalent s1 s2 osi ->
    STEQ.osi_equivalent s1' s2' osi.
Proof.
  Local Hint Resolve ctxt_switch_equivalent_hyper
        STEQ.osi_equivalent_smem_trans
        STEQ.osi_equivalent_hyper_trans.

  intros.
  generalize H, H0; unfold valid_sm_state; crush.
  unfold valid_sm_exec in *; crush.
  unfold ACT.pre_lswitch, post_sm_lswitch in *; crush.
  rewrite H21 with x0 in *; crush; clear H21.
  rewrite H19 with x in *; crush; clear H19.
  
  unfold madd_type in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (page_content p) eqn:?; crush.

  unfold STEQ.osi_equivalent in *; crush.
  destruct (os_ident_eq osi (st_active_os s2)); crush.
  repeat rewrite <- OSS.get_set; crush.
  repeat rewrite OSS.get_set_distinct; crush.

  apply STEQ.osi_equivalent_smem_trans with s1; crush.
  apply STEQ.osi_equivalent_smem_symm; crush.
  apply STEQ.osi_equivalent_hyper_trans with s1; crush.
  apply STEQ.osi_equivalent_hyper_symm; crush.

  rewrite H21 with x0 in *; crush; clear H21.
  rewrite H19 with x in *; crush; clear H19.
  unfold madd_type in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (page_content p) eqn:?; crush.

  unfold STEQ.osi_equivalent in *; crush.
  destruct (os_ident_eq osi (st_active_os s2)); crush.
  repeat rewrite <- OSS.get_set; crush.
  repeat rewrite OSS.get_set_distinct; crush.

  rewrite H21 with x0 in *; crush; clear H21.
  rewrite H19 with x in *; crush; clear H19.
  unfold madd_type in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (page_content p) eqn:?; crush.

  unfold STEQ.osi_equivalent in *; crush.
  destruct (os_ident_eq osi (st_active_os s1)); crush.
  repeat rewrite <- OSS.get_set; crush.
  repeat rewrite OSS.get_set_distinct; crush.

  apply STEQ.osi_equivalent_smem_trans with s1; crush.
  apply STEQ.osi_equivalent_smem_symm; crush.
  apply STEQ.osi_equivalent_hyper_trans with s1; crush.
  apply STEQ.osi_equivalent_hyper_symm; crush.

  rewrite H21 with x0 in *; crush; clear H21.
  rewrite H19 with x in *; crush; clear H19.
  unfold madd_type in *; crush.
  destruct (page_content p0) eqn:?; crush.
  destruct (page_content p) eqn:?; crush.

  unfold STEQ.osi_equivalent in *; crush.
  destruct (os_ident_eq osi osi'); crush.
  repeat rewrite <- OSS.get_set; crush.
  repeat rewrite OSS.get_set_distinct; crush.
Qed.

Theorem stealth_os_isolation :
  forall (s1 s1' s2 s2':state) (a:sm_action) (osi:os_ident),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    valid_sm_exec s1 a s1' ->
    valid_sm_exec s2 a s2' ->
    STEQ.osi_equivalent s1 s2 osi ->
    STEQ.osi_equivalent s1' s2' osi.
Proof.
  Local Hint Resolve
       new_normal_osi_equivalent
       new_sm_osi_equivalent
       switch_sm_osi_equivalent 
       lswitch_sm_osi_equivalent.
    intros;
    sm_action_induction a;
      [clear Case; STAC.action_induction a | | | | ]; crush;
        try match goal with
          | [ H1 : valid_sm_exec ?s1 (Action ?a) ?s1',
              H2 : valid_sm_exec ?s2 (Action ?a) ?s2' |-
              STEQ.osi_equivalent ?s1' ?s2' ?osi] =>
          apply os_isolation with s1 s2 a;
            unfold valid_sm_exec, ACT.valid_exec,
              valid_sm_state in *; crush
        end.
  Qed.

End STEALTH_ISOLATION.


