(***********************************************************************
 * File: Stealth_leakage.v
 * $Revision: 12920 $
 * $Date: 2016-09-22 18:44:11 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Non-Leakage with StealthMem
 *************************************************************************)
Require Import Sets.Constructive_sets
  Stealth_actions Stealth_valid Stealth_obs_lemmas Stealth_lemmas
  Stealth_non_stealth_equivalence Stealth_equivalence Stealth_isolation
  Stealth_sm_isolation Stealth_traces Streams.

Module NON_LEAKAGE.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES STEALTH_ISOLATION
         STEALTH_SM_ISOLATION STEALTH_EQ STEALTH_LEMMAS
         OBS_LEMMAS OS_ISO STEALTH_NON_STEALTH_EQ.

  Lemma same_obs_step :
    forall (osi:os_ident) (s s':leak_state) (A A':leak_action),
      leak_system_transition (Leak_Ex_State s A (cache_test osi)) 
                             (Leak_Ex_State s' A' (cache_test osi)) ->
      valid_sm_state (leak_state_state s) ->
      ~ osi_sm_action osi (leak_state_state s) (A (leak_state_obs s)) ->
      ~ non_stealth_access (A (leak_state_obs s)) ->
      leak_state_obs s = leak_state_obs s'.
  Proof.
    intros. 
    inversion H; crush.
    subst s0; subst o0; subst s; subst.
    subst o. subst s'0. subst a.

    rewrite <- H4; symmetry.
    remember (A (leak_state_obs ls)) as a.
    clear Heqa A A' H H4.

    Local Hint Constructors osi_action osi_sm_action non_stealth_access.
    Local Hint Resolve write_stealth_same_obs
      new_normal_same_obs new_stealth_same_obs.

    sm_action_induction a;
      [clear Case; STAC.action_induction a | | | | ];
      generalize H3; unfold valid_sm_exec in H3; intro; crush;
        unfold ACT.post_silent, ACT.pre_read, ACT.pre_read_hyper,
          ACT.pre_write, ACT.pre_write_hyper, ACT.post_hcall,
          ACT.post_ret_ctrl, ACT.post_chmod, ACT.post_pin, ACT.post_unpin in *;
          crush;
          try (match goal with
            | [ H: _ = leak_state_state ?s' |- context[?s']] =>
              rewrite H; apply same_state_same_obs
            | [ H: leak_state_state ?s' = _ |- context[?s']] =>
              rewrite H; apply same_state_same_obs
          end; crush; fail);
          try match goal with
            | [ H: ~ non_stealth_access ?a |- _] =>
              elim H; crush; fail
          end.

    (* Read *)
    destruct (vadd_eq v stealth_va);
      [eapply read_stealth_same_obs | elim H2]; crush.

    (* Read hyper *)
    destruct (vadd_eq v stealth_va);
      [eapply read_stealth_same_obs | elim H2]; crush.

    (* Write *)
    destruct (vadd_eq v stealth_va);
      [ eapply write_stealth_same_obs | elim H2];
      crush.

    (* Del untrusted *)
    destruct (vadd_eq v stealth_va);
      [ apply del_same_obs with o v | elim H2];
      crush.

    (* New sm *)
    eapply new_stealth_same_obs; crush.
    crush.

    (* Switch *)
    apply switch_same_obs with o; crush.
  Qed.

Lemma stealth_cached_eq : forall (osi:os_ident) (s1 s2:state) (ma1:madd) (pg:page),
  valid_sm_state s1 -> valid_sm_state s2 ->
  osi_stealth_equivalent s1 s2 osi ->
  get_cache (smem_c (st_memory s1)) (stealth_va,ma1) = Some pg ->
  page_owned_by pg = Os_PO osi -> 
  exists ma2, get_cache (smem_c (st_memory s2)) (stealth_va,ma2) = Some pg.
Proof.
  intros.
  unfold osi_stealth_equivalent, STEQ.osi_equivalent,
  STEQ.osi_equivalent_smem in *;
    crush.
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state_H H.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  rename pt_addr into pt_addr1, pt_page into pt_page1, pt into pt1,
    p2m into p2m1.
  destruct stealth_cached_is_aos with s1 pt_addr1 ma1 pg; crush.
  unfold valid_sm_state; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  rewrite H21 in H3; crush.
  injection H3; intros; subst osi; clear H3.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H19 in *; crush.

  unfold STEQ.osi_pt_pages_eq in H7; crush.
  destruct H7 with (os_current_page aos) pt_page1 pt1; crush; clear H7.
  unfold osi_pa_to_pg; crush.
  destruct H_va_has_pa with (st_active_os s1) p2m1 pt_addr1
    pt_page1 pt1 stealth_va ma1; crush.
  apply stealth_va_accessible.
  rename x1 into pa.

  destruct H23 with stealth_va pa ma1; crush; clear H23.
  unfold osi_pa_to_ma; crush.
  rename x1 into ma2.
  assert (page_of_vadd_madd_smem (st_memory s2) stealth_va ma2 = Some pg).
  apply H26; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H_valid_cache with stealth_va ma1 pg; crush.

  clear H26.
  assert (st_active_os s1 = st_active_os s2).
  destruct H1; crush.
    
  unfold page_of_vadd_madd_smem in H23; crush.
  unfold osi_pa_to_pg in H3; crush.
  rewrite H6 in H_oss.
  rewrite H26 in H25, Heqopt_res, H_oss.
  clear H26.
  clear dependent s1.
  unfold valid_sm_state in H0; crush.
  destruct H0 with i0 ma2 pg; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H22; crush.
Qed.

  Lemma ctxt_switch_obs_two_exec :
    forall s1 s1' s2 s2' osi' pt_padd1 pt_padd2 pt_addr1 pt_addr2
           ma1 pg1 ma2 pg2,
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    (osi' <> st_active_os s1 \/ pt_padd1 = pt_padd2) ->
    osi_stealth_equivalent s1 s2 (st_active_os s1) ->
    osi_pa_to_ma s1 osi' pt_padd1 = Some pt_addr1 ->
    osi_pa_to_ma s2 osi' pt_padd2 = Some pt_addr2 ->
    madd_type (smem_mm (st_memory s1)) pt_addr1 = Some PT ->
    madd_type (smem_mm (st_memory s2)) pt_addr2 = Some PT ->
    st_memory s1' = ctxt_switch_new_smem s1 pt_addr1 ->
    st_memory s2' = ctxt_switch_new_smem s2 pt_addr2 ->
    get_cache (smem_c (st_memory s1)) (stealth_va,ma1) = Some pg1 ->
    get_cache (smem_c (st_memory s1')) (stealth_va,ma1) = None ->
    get_cache (smem_c (st_memory s2)) (stealth_va,ma2) = Some pg2 ->
    get_cache (smem_c (st_memory s2')) (stealth_va,ma2) = None.
  Proof.    
    intros; crush.
    rewrite H8; rewrite H7 in H10; clear dependent s1';
      clear dependent s2'.
    unfold ctxt_switch_new_smem; crush.
    Focus 2.
    (* aos_current_page_table s2 = None, ABSURD *)
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rename m into pt_addr_aos2.

    (* The aos has a current PT *)
    remember (madd_of_vadd (smem_mm (st_memory s2)) pt_addr2 stealth_va).
    destruct o; crush.
    (* The new PT has a stealth page *)
    unfold valid_sm_state in H0; crush.
    remember (page_of_vadd_madd_smem (st_memory s2) stealth_va m).
    destruct o; crush.
    destruct H14 with pt_addr2 m p; crush.
    unfold is_rw_page in H16; crush.
    rename m into new_stealth_ma, p into new_stealth_pg, x into new_stealth_val.
    rewrite get_new_cache with (ma:=new_stealth_ma)
      (pg:=new_stealth_pg) (val:=new_stealth_val); crush.

    destruct (madd_eq new_stealth_ma ma2); crush.
    (* The new stealth page is the same as the old one *)
    assert (osi' = st_active_os s2).

    STAC.destruct_valid_state.
    unfold madd_of_vadd, madd_page_table in Heqo; crush.
    remember (page_content p0).
    destruct c; crush.
    assert (madd_owner (smem_mm (st_memory s2)) ma2 = Os_PO osi').
    destruct H_valid_virtmap with osi' pt_addr2 p0 p1 stealth_va ma2; crush.
    unfold osi_pa_to_ma in H4; crush.
    unfold osi_pa_to_ma in H4; crush.
    generalize (H_valid_hyper osi' i pt_padd2 pt_addr2); intros.
    unfold madd_owner in H18; crush.
    specialize (H19 stealth_va_accessible).
    unfold madd_owner; crush.
    unfold madd_owner in H18; crush.
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s2 pt_addr_aos2 ma2 pg2; crush.
    unfold valid_sm_state; crush.
    destruct H_valid_cache with stealth_va ma2 pg2; crush.
    subst osi'.
    destruct H1; crush.
    unfold osi_stealth_equivalent in H2; crush.
    unfold STEQ.osi_equivalent in H2; crush.
    destruct H2; crush.
    rename pt_padd2 into pt_padd.

    assert (exists pa, osi_pa_to_ma s2 (st_active_os s2) pa = Some ma2).
    unfold valid_sm_state in H0.
    STAC.destruct_valid_state.
    unfold osi_pa_to_ma in H4; crush.
    unfold madd_of_vadd, madd_page_table in Heqo; crush.
    remember (page_content p0).
    destruct c; crush.
    destruct H_va_has_pa with (st_active_os s2) i pt_addr2 p0 p1 stealth_va ma2; crush.
    generalize (H_valid_hyper (st_active_os s2) i pt_padd pt_addr2); intros.
    unfold madd_owner in H1; crush.
    apply stealth_va_accessible.
    exists x; unfold osi_pa_to_ma; crush.

    destruct H1 as [pa ?].
    unfold osi_stealth_equivalent in *; crush.
    unfold STEQ.osi_equivalent, STEQ.osi_equivalent_smem in *; crush.
    unfold madd_of_vadd, madd_page_table in Heqo; crush.
    remember (page_content p0).
    destruct c; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.
    destruct H23 with (os_current_page aos) pt_page pt; crush.
    unfold osi_pa_to_pg; crush.
    unfold osi_pa_to_ma in H4; crush.
    destruct H2; crush.
    rewrite <- H26; crush.
    destruct H30 with stealth_va pa ma2; crush.
    destruct H2; crush.
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s2 pt_addr_aos2 ma2 pg2; crush.
    unfold valid_sm_state; crush.
    unfold madd_of_vadd, madd_page_table in H33; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite H27 in Heqopt_res; crush.
    unfold osi_pa_to_ma; crush.
    destruct H2; crush.
    rewrite <- H31; crush.
    unfold osi_pa_to_pg in H26; crush.
    rename i0 into pt_addr_aos1.
    assert (page_of_vadd_madd_smem (st_memory s1) stealth_va x1 = Some pg2); crush.
    apply H33; crush.
    unfold page_of_vadd_madd_smem; crush.
    destruct H_valid_cache with stealth_va ma2 pg2; crush.
    clear H33.
    destruct (madd_eq x1 ma1); crush.
    Focus 2.
    (* x1 <> ma1 ABSURD *)
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s1 pt_addr_aos1 ma1 pg1; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
    destruct H2; crush.
    rewrite H33 in H_oss.
    rewrite <- H20 in H_oss; crush.
    unfold madd_of_vadd, madd_page_table in H33; crush.
    rewrite <- H29 in *; crush.

    clear H30 H34.
    destruct H23 with pt_padd p0 p1; crush.
    unfold osi_pa_to_pg; crush.
    destruct H2; crush.
    rewrite <- H30; crush.
    unfold osi_pa_to_ma in H4; crush.
    destruct H34 with stealth_va pa ma2; crush.
    unfold osi_pa_to_ma; crush.
    destruct H2; crush.
    rewrite <- H35; crush.
    clear H34 H37.
    destruct H2; crush.
    rewrite H34 in H_oss.
    rewrite <- H20 in H_oss.
    rewrite H34 in H3.

    cut False; crush.
    clear H34; clear dependent s2.

    generalize H10; clear H10; intros H10.
    unfold ctxt_switch_new_smem in H10; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    unfold valid_sm_state in *; crush.
    assert (is_rw_page pg1).
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va ma1 pg1; crush.
    unfold is_rw_page in H11; crush.
    rename x3 into val.

    clear dependent pt_page.
    rename x1 into pt_page.
    unfold osi_pa_to_pg in H30; crush.
    assert (i0 = pt_addr1).
    unfold osi_pa_to_ma in H3; crush.
    subst i0.

    rewrite get_new_cache with (ma:=ma1) (pg:=pg1) (val:=val) in H10; crush.
    rewrite <- get_add_cache_cacheable in H10; crush.
    destruct H7 with pt_addr1 ma1 pg1; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H33; crush.
    unfold page_of_vadd_madd_smem; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite <- H29; crush.
    rewrite madd_of_vadd_smem_flush_ma; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    unfold madd_type in *; crush.
    rewrite <- H33; crush.
    rewrite STEALTH_LEMMAS.madd_of_vadd_smem_ctxt_switch; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H33; crush.
      
    rewrite get_page_result with (pg:=pg1); crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite <- H29; crush.
    unfold page_of_vadd_smem; crush.
    rewrite madd_of_vadd_smem_flush_ma; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H33; crush.
    rewrite STEALTH_LEMMAS.page_of_vadd_madd_smem_flush_ma
      with (pt_addr:=pt_addr_aos1); crush.
    unfold page_of_vadd_madd_smem; crush.
    exists (st_active_os s1); exists (os_current_page aos);
      unfold osi_pa_to_ma; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H29; crush.
    rewrite ctxt_switch_mem in *; crush.

  (* new_stealth_ma <> ma2 *)
    match goal with
      | [|- context[add_cache ?c (stealth_va,?stealth_ma) ?stealth_pg]] =>
        assert (get_cache c (stealth_va,ma2) = None)
    end.
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s2 pt_addr_aos2 ma2 pg2; crush.
    unfold valid_sm_state; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.
    remember (page_content p).
    destruct c; crush.
    rewrite flush_ma_cache; crush.
    rewrite get_remove_ma_cache_eq; crush.
    rewrite ctxt_switch_mem in Heqval0; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.

    match goal with
        [|- get_cache ?c (?va,?ma) = None] =>
        remember (get_cache c (va, ma))
    end.
    destruct o; crush.
    generalize (get_add_cache_distinct_not_new
      (smem_c (remove_osi_stealth_page (ctxt_switch (st_memory s2)) pt_addr_aos2))
      (stealth_va,new_stealth_ma) (stealth_va,ma2) new_stealth_pg p); crush.
    rewrite <- H19 in H18; crush.
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in Heqval0; crush.
    rewrite H21; crush.
    rewrite madd_of_vadd_smem_flush_ma; crush.
    rewrite STEALTH_LEMMAS.madd_of_vadd_smem_ctxt_switch; crush.
    rewrite STEALTH_LEMMAS.madd_of_vadd_smem_ctxt_switch; crush.
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite get_page_result with (pg:=new_stealth_pg); crush.
    unfold page_of_vadd_smem; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite H21; crush.
    rewrite madd_of_vadd_smem_flush_ma; crush.

    rewrite STEALTH_LEMMAS.page_of_vadd_madd_smem_flush_ma
      with (pt_addr:=pt_addr2); crush.
    rewrite STEALTH_LEMMAS.madd_of_vadd_smem_ctxt_switch; crush.
    rewrite STEALTH_LEMMAS.page_of_vadd_madd_smem_ctxt_switch; crush.
    rewrite STEALTH_LEMMAS.madd_of_vadd_smem_ctxt_switch; crush.
    rewrite STEALTH_LEMMAS.page_of_vadd_madd_smem_ctxt_switch; crush.

    (* Absurd case *)
    unfold madd_of_vadd, madd_page_table in Heqo; crush.
    remember (page_content p0).
    destruct c; crush.
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state.
    destruct H_valid_virtmap with osi' pt_addr2 p0 p1 stealth_va m; crush.
    unfold osi_pa_to_ma in H4; crush.
    unfold osi_pa_to_ma in H4; crush.
    generalize (H_valid_hyper osi' i pt_padd2 pt_addr2); intros.
    unfold madd_owner in H16; crush.
    unfold page_of_vadd_madd_smem in Heqo0; crush.

    (* The new PT has no stealth page *)
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite get_invalid; crush.
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s2 pt_addr_aos2 ma2 pg2; crush.
    unfold valid_sm_state; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite H19; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.
    rewrite H19 in Heqopt_res; crush.
    rewrite flush_ma_cache; crush.
    rewrite get_remove_ma_cache_eq; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.
    rewrite H19 in Heqopt_res; crush.
    rewrite ctxt_switch_mem in *; crush.

    unfold page_of_vadd_smem; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite H19; crush.
    rewrite madd_of_vadd_smem_flush_ma; crush.
    rewrite <- Heqo; crush.
    rewrite STEALTH_LEMMAS.madd_of_vadd_smem_ctxt_switch; crush.
    rewrite <- Heqo; crush.
    rewrite STEALTH_LEMMAS.madd_of_vadd_smem_ctxt_switch; crush.
    rewrite <- Heqo; crush.
  Qed.

  Lemma included_obs_two_exec :
    forall (osi:os_ident) (s1 s1' s2 s2':state) (a:sm_action) (o:observation),
      valid_sm_state s1 ->
      valid_sm_state s2 ->
      st_active_os s1 = st_active_os s2 ->
      valid_sm_exec s1 a s1' ->
      valid_sm_exec s2 a s2' ->
      osi_stealth_equivalent s1 s2 osi ->
      Included vadd (get_cache_observation osi (smem_c (st_memory s1))
        (smem_c (st_memory s1')) o)
      (get_cache_observation osi (smem_c (st_memory s2))
        (smem_c (st_memory s2')) o).
  Proof.
    unfold Included; crush.

    destruct get_cache_observation_spec with osi (smem_c (st_memory s1))
      (smem_c (st_memory s1')) o; crush.
    clear H7 H8.
    rename x into v.
    destruct H6 with v; crush; clear H5 H6.
    destruct get_cache_observation_spec with osi (smem_c (st_memory s2))
      (smem_c (st_memory s2')) o; crush.

    rename x into m, x0 into p.
    destruct get_cache_observation_spec with osi (smem_c (st_memory s2))
      (smem_c (st_memory s2')) o; crush.
    clear H5 H10.
    generalize H4; intro H_eq.
    unfold osi_stealth_equivalent in H4; crush.
    destruct get_cache_then_pa with s1 osi v m p; crush.
    unfold valid_sm_state in *; crush.

    rename x into pa.
    destruct (vadd_eq v stealth_va); crush.

    Focus 2.
    (* v <> stealth_va *)
    destruct H5 with osi; clear H5; crush.
    assert (exists pt_padd, exists pt_addr,
      osi_pa_to_ma s1 osi pt_padd = Some pt_addr /\
      madd_of_vadd (smem_mm (st_memory s1)) pt_addr v = Some m).
    unfold valid_sm_state in H; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache_va_ma with v m p; crush.
    unfold osi_pa_to_pg in *; crush.
    assert (x = osi); crush.
    assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO x).
    destruct H_valid_virtmap with x i0 x1 x2 v m; crush.
    generalize (H_valid_hyper x i x0 i0); intros.
    unfold madd_owner in H23; crush.
    unfold madd_owner; crush.
    apply H24; crush.
    destruct H_valid_cache with v m p; crush.
    destruct (ctxt_vadd_accessible ctxt v) eqn:?;
      [unfold os_accessible | cut False]; crush.
    rewrite H25 in H29; crush.
    assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO osi); crush.
    unfold osi_pa_to_ma in H11; crush.
    exists x0; exists i0; unfold madd_of_vadd, madd_page_table; crush.
    unfold osi_pa_to_ma; crush.
    rewrite H21; crush.
    destruct H14; crush.
    rename x into pt_padd.
    rename x0 into pt_addr.
    destruct H13.
    unfold madd_of_vadd, madd_page_table in H15; crush.
    destruct (page_content p1) eqn:?; crush.
    destruct H13 with pt_padd p1 p2 v m; crush.
    unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
    destruct H21 with p; crush.
    unfold page_of_vadd_madd_smem; crush.

    assert (get_cache (smem_c (st_memory s2)) (v,x1) = Some x2).
    destruct (same_cache_line_entries_some_pg
      (smem_c (st_memory s1)) (smem_c (st_memory s2))
      (st_hypervisor s1) (st_hypervisor s2) osi v v pa m x1 p); crush.
    apply H10; crush.
    unfold valid_sm_state in H; crush.    
    unfold page_of_vadd_madd_smem in *; crush.

Lemma get_cache_some_none_abs :
  forall s va ma,
    (exists pg, get_cache (smem_c (st_memory s)) (va, ma) = Some pg) ->
    get_cache (smem_c (st_memory s)) (va, ma) = None ->
    False.
Proof.
  crush.
Qed.
    destruct get_cache_some_none_abs with s2 v x1; crush.
    apply H6 with x1 x2; crush; clear H6.
    
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with v x1 x2; crush.
    specialize (H20 pa H11).
    unfold osi_pa_to_ma in H20; crush.
    generalize (H_valid_hyper osi i pa x1); intros.
    unfold madd_owner in H36.
    rewrite <- H32 in H36.
    rewrite H35.
    apply H36; auto.
    specialize (H20 pa H11).
    unfold osi_pa_to_ma in H20; crush.
    
    apply (same_cache_line_entries_no_pg 
      (smem_c (st_memory s1')) (smem_c (st_memory s2'))
      (st_hypervisor s1) (st_hypervisor s2) v pa m); crush.
    apply same_cache_line_entries_trans with (smem_c (st_memory s1'))
      (st_hypervisor s1'); crush.
    apply same_cache_line_entries_trans with (smem_c (st_memory s2'))
      (st_hypervisor s2'); crush.

    destruct stealth_os_isolation with s1 s1' s2 s2' a osi; crush.
    unfold same_non_stealth_entries in *; crush.
    apply H26; crush.
    unfold valid_sm_state in H0; crush.
    exists osi; crush.
    unfold osi_pa_to_ma_hyp, osi_pa_to_pg,
      osi_pa_to_ma in *; crush.

    (* v = stealth_va *)
    unfold valid_sm_state in H; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s1 pt_addr m p; crush.
    unfold valid_sm_state; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
    rewrite H8 in H23.
    injection H23; intros; subst osi.

    destruct (deleted_stealth_cache_action s1 s1' a m p); crush.
    unfold valid_sm_state; crush.
    (* Delete untrusted *)
    unfold STEQ.osi_equivalent, STEQ.osi_equivalent_smem in *; crush.
    destruct H25 with (os_current_page aos) pt_page pt; crush.
    unfold osi_pa_to_pg; crush.
    destruct H30 with stealth_va pa m; crush.
    unfold madd_of_vadd, madd_page_table in H20; crush.
    rewrite H21 in Heqopt_res; crush.

    apply H6 with x1 p; crush.
    rewrite <- H33 with p; crush.
    unfold page_of_vadd_madd_smem; crush.
    unfold valid_sm_state in H0; crush.
    unfold osi_pa_to_pg in *; crush.
    destruct H34 with i0 x1 p; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
    destruct H4; crush.
    rewrite H40; crush.
    rewrite <- H24; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H29; crush.
    rewrite <- H33; crush.
    unfold page_of_vadd_madd_smem; crush.
    unfold page_of_vadd_madd_smem; crush.
    destruct H_valid_cache with stealth_va m p; crush.
    unfold page_of_vadd_madd_smem; crush.
    destruct H_valid_cache with stealth_va m p; crush.

    unfold valid_sm_exec in H3; crush; unfold ACT.post_del in *; crush.
    rewrite delete_cache with (ma:=x1); crush.
    rewrite get_remove_cache; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    unfold osi_current_page_table in *; crush.
    unfold osi_pa_to_pg in *; crush.

    unfold valid_sm_state in *; crush.
    STAC.destruct_valid_state.
    destruct H4; crush.
    destruct H_valid_currpg0 with (st_active_os s2) x3 i; crush.
    rewrite H43; crush.
    rewrite H43; crush.
    unfold is_pt_page in *; crush.
    rewrite H46; crush.
    rewrite H24 in *; crush.
    crush.
    rewrite <- H29 in H46; crush.
    injection H46; crush.
    (* Done delete *)

    (* Ctxt switch *)
    destruct stealth_cached_eq with (st_active_os s1) s1 s2 m p; crush.
    unfold valid_sm_state; crush.
    apply H6 with x0 p; crush.
    destruct H24; crush.

    destruct H_valid_os_idents with x; crush.
    STAC.get_osi_current_pt_page x.
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state_H H0.
    destruct H_valid_os_idents0 with x; crush.
    STAC.get_osi_current_pt_page x.
    
    apply ctxt_switch_obs_two_exec with s1 s1' s2 x (os_current_page x1)
      (os_current_page x3) pt_addr0 pt_addr1 m p p; crush.
    unfold valid_sm_state; crush.
    unfold valid_sm_state; crush.
    destruct (os_ident_eq x (st_active_os s1)); crush.
    right.
    unfold STEQ.osi_equivalent in H4; crush.
    unfold osi_pa_to_ma; crush.
    unfold osi_pa_to_ma; crush.
    unfold madd_type; crush.
    rewrite H30; crush.
    unfold madd_type; crush.
    rewrite H42; crush.
    inversion H2; crush.
    unfold post_sm_switch in H45; crush.
    rewrite H45 with pt_addr0; crush.
    unfold osi_current_page_table; crush.
    inversion H3; crush.
    unfold post_sm_switch in H45; crush.
    rewrite H45 with pt_addr1; crush.
    unfold osi_current_page_table; crush.

    (* LSwitch *)
    clear dependent x.
    inversion H2; crush.
    unfold ACT.pre_lswitch in H26; crush.
    unfold madd_type in H31; crush.
    remember (page_content p0).
    destruct c; crush.
    unfold post_sm_lswitch in H27; crush.
    inversion H3; crush.
    unfold ACT.pre_lswitch in H32; crush.
    unfold madd_type in H37; crush.
    remember (page_content p2).
    destruct c; crush.
    unfold post_sm_lswitch in H33; crush.
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state_H H0.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.

    apply ctxt_switch_obs_two_exec with s1 s1' s2 (st_active_os s1) x2
      x2 x x1 m p p; crush.
    unfold valid_sm_state; crush.
    unfold valid_sm_state; crush.
    unfold madd_type; crush.
    rewrite <- Heqc; crush.
    unfold madd_type; crush.
    rewrite <- Heqc0; crush.
    rewrite H27 with x; crush.
    rewrite H33 with x1; crush.

    elim n. 
    unfold STEQ.osi_equivalent in H4; crush.
    destruct H4; crush.
    rewrite H27 with x in H9; crush.
Qed.    

Lemma same_obs_two_exec :
    forall (osi:os_ident) (s1 s1' s2 s2':leak_state)
      (A A1' A2':leak_action),
    leak_system_transition (Leak_Ex_State s1 A (cache_test osi)) 
                           (Leak_Ex_State s1' A1' (cache_test osi)) ->
    leak_system_transition (Leak_Ex_State s2 A (cache_test osi)) 
                           (Leak_Ex_State s2' A2' (cache_test osi)) ->
                           
    st_active_os (leak_state_state s1) = st_active_os (leak_state_state s2) ->
    valid_sm_state (leak_state_state s1) ->
    valid_sm_state (leak_state_state s2) ->
    osi_stealth_equivalent (leak_state_state s1) (leak_state_state s2) osi ->
    (leak_state_obs s1) = (leak_state_obs s2) ->

    (leak_state_obs s1') = (leak_state_obs s2').
  Proof.
    intros.
    inversion H; crush.
    subst s; subst o0; subst s1; subst.
    subst o; subst s'; subst a.
    clear H.

    inversion H0; crush.
    subst s; subst o0; subst s2; subst.
    subst o; subst s'; subst a.
    clear H0.

    rewrite <- H5 in *.
    rename ls into s1, ls0 into s2.
    remember (A (leak_state_obs s1)) as a.

    rewrite <- H7, <- H8.
    apply Extensionality_Ensembles.
    split; apply included_obs_two_exec with (a:=A (leak_state_obs s1)); crush.
    apply osi_stealth_equivalent_symm; crush.
  Qed.

  Lemma included_obs_two_write :
    forall (osi:os_ident) (s1 s1' s2 s2':state) (va:vadd) (val1 val2:value)
           (o:observation),
      valid_sm_state s1 ->
      valid_sm_state s2 ->
      st_active_os s1 = st_active_os s2 ->
      va <> stealth_va -> 
      valid_sm_exec s1 (Action (Write va val1)) s1' ->
      valid_sm_exec s2 (Action (Write va val2)) s2' ->
      osi <> st_active_os s1 ->
      osi_stealth_equivalent s1 s2 osi ->
      Included vadd (get_cache_observation osi (smem_c (st_memory s1))
        (smem_c (st_memory s1')) o)
      (get_cache_observation osi (smem_c (st_memory s2))
        (smem_c (st_memory s2')) o).
  Proof.
    unfold Included; crush.

    destruct get_cache_observation_spec with osi (smem_c (st_memory s1))
      (smem_c (st_memory s1')) o; crush.
    clear H9 H10.
    rename x into v.
    destruct H8 with v; crush; clear H7 H8.
    destruct get_cache_observation_spec with osi (smem_c (st_memory s2))
      (smem_c (st_memory s2')) o; crush.

    rename x into m, x0 into p.
    destruct get_cache_observation_spec with osi (smem_c (st_memory s2))
      (smem_c (st_memory s2')) o; crush.
    clear H7 H12.
    generalize H6; intro H_eq.
    unfold osi_stealth_equivalent in H6; crush.
    destruct get_cache_then_pa with s1 osi v m p; crush.
    unfold valid_sm_state in *; crush.

    rename x into pa.
    destruct (vadd_eq v stealth_va); crush.

    Focus 2.
    (* v <> stealth_va *)
    destruct H7 with osi; clear H7; crush.
    assert (exists pt_padd, exists pt_addr,
      osi_pa_to_ma s1 osi pt_padd = Some pt_addr /\
      madd_of_vadd (smem_mm (st_memory s1)) pt_addr v = Some m).
    unfold valid_sm_state in H; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache_va_ma with v m p; crush.
    unfold osi_pa_to_pg in *; crush.
    assert (x = osi); crush.
    assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO x).
    destruct H_valid_virtmap with x i0 x1 x2 v m; crush.
    generalize (H_valid_hyper x i x0 i0); intros.
    unfold madd_owner in H25; crush.
    unfold madd_owner; crush.
    apply H26; crush.
    destruct H_valid_cache with v m p; crush.
    destruct (ctxt_vadd_accessible ctxt v) eqn:?;
      [unfold os_accessible | cut False]; crush.
    rewrite H27 in H31; crush.
    assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO osi); crush.
    unfold osi_pa_to_ma in H13; crush.
    exists x0; exists i0; unfold madd_of_vadd, madd_page_table; crush.
    unfold osi_pa_to_ma; crush.
    rewrite H23; crush.
    destruct H16; crush.
    rename x into pt_padd.
    rename x0 into pt_addr.
    destruct H15.
    unfold madd_of_vadd, madd_page_table in H17; crush.
    destruct (page_content p1) eqn:?; crush.
    destruct H15 with pt_padd p1 p2 v m; crush.
    unfold osi_pa_to_pg, osi_pa_to_ma in *; crush.
    destruct H23 with p; crush.
    unfold page_of_vadd_madd_smem; crush.

    assert (get_cache (smem_c (st_memory s2)) (v,x1) = Some x2).
    destruct (same_cache_line_entries_some_pg
      (smem_c (st_memory s1)) (smem_c (st_memory s2))
      (st_hypervisor s1) (st_hypervisor s2) osi v v pa m x1 p); crush.
    apply H12; crush.
    unfold valid_sm_state in H; crush.    
    unfold page_of_vadd_madd_smem in *; crush.
    destruct get_cache_some_none_abs with s2 v x1; crush.
    apply H8 with x1 x2; crush; clear H8.
    
    unfold valid_sm_state in H0; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with v x1 x2; crush.
    specialize (H22 pa H13).
    unfold osi_pa_to_ma in H22; crush.
    generalize (H_valid_hyper osi i pa x1); intros.
    unfold madd_owner in H38.
    rewrite <- H34 in H38.
    rewrite H37.
    apply H38; auto.
    specialize (H22 pa H13).
    unfold osi_pa_to_ma in H22; crush.

    apply (same_cache_line_entries_no_pg 
      (smem_c (st_memory s1')) (smem_c (st_memory s2'))
      (st_hypervisor s1) (st_hypervisor s2) v pa m); crush.
    apply same_cache_line_entries_trans with (smem_c (st_memory s1'))
      (st_hypervisor s1'); crush.
    apply same_cache_line_entries_trans with (smem_c (st_memory s2'))
      (st_hypervisor s2'); crush.

    destruct non_stealth_write_eq with s1 s2 s1' s2' osi va val1 val2; crush.
    unfold same_non_stealth_entries in *; crush.
    apply H28; crush.
    unfold valid_sm_state in H0; crush.
    exists osi; crush.
    unfold osi_pa_to_ma_hyp, osi_pa_to_pg,
      osi_pa_to_ma in *; crush.

    (* v = stealth_va *)
    unfold valid_sm_state in H; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s1 pt_addr m p; crush.
    unfold valid_sm_state; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
Qed.    


Lemma same_obs_two_write :
    forall (osi:os_ident) (s1 s1' s2 s2':leak_state)
      (A1 A2 A1' A2':leak_action) (va:vadd) (val1 val2:value),
    leak_system_transition (Leak_Ex_State s1 A1 (cache_test osi)) 
                           (Leak_Ex_State s1' A1' (cache_test osi)) ->
    leak_system_transition (Leak_Ex_State s2 A2 (cache_test osi)) 
                           (Leak_Ex_State s2' A2' (cache_test osi)) ->

    va <> stealth_va ->
    osi <> st_active_os (leak_state_state s1) ->
    A1 (leak_state_obs s1) = Action (Write va val1) ->
    A2 (leak_state_obs s2) = Action (Write va val2) ->
                           
    st_active_os (leak_state_state s1) = st_active_os (leak_state_state s2) ->
    valid_sm_state (leak_state_state s1) ->
    valid_sm_state (leak_state_state s2) ->
    osi_stealth_equivalent (leak_state_state s1) (leak_state_state s2) osi ->
    (leak_state_obs s1) = (leak_state_obs s2) ->

    (leak_state_obs s1') = (leak_state_obs s2').
  Proof.
    intros.
    inversion H; crush.
    subst s; subst o0; subst s1; subst.
    subst o; subst s'; subst a.
    clear H.

    inversion H0; crush.
    subst s; subst o0; subst s2; subst.
    subst o; subst s'; subst a.
    clear H0.

    rewrite <- H9 in *.
    rename ls into s1, ls0 into s2.

    rewrite <- H11, <- H12.
    apply Extensionality_Ensembles.
    split.
    apply included_obs_two_write with va val1 val2; crush.
    apply included_obs_two_write with va val2 val1; crush.
    apply osi_stealth_equivalent_symm; crush.
  Qed.
  
End NON_LEAKAGE.


