(***********************************************************************
 * File: Stealth_lemmas.v
 * $Revision: 12919 $
 * $Date: 2016-09-22 11:57:12 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * StealthMem properties and tactics.
 *************************************************************************)

Require Import Sets.Constructive_sets
  Stealth_actions Stealth_valid.

Module STEALTH_LEMMAS.
  Import PG C T MM SMEM ST ST_VALID OS_ISO
         STEALTH_ACTIONS VALID_SM_STATE. 

Lemma stealth_cached_is_aos : forall s pt_addr ma pg,
  valid_sm_state s ->
  get_cache (smem_c (st_memory s)) (stealth_va,ma) = Some pg ->
  aos_current_page_table s = Some pt_addr ->
  madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = Some ma /\
  page_owned_by pg = Os_PO (st_active_os s).
Proof.
  intros.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = Some ma);
    crush.
  unfold valid_sm_state in *; crush; STAC.destruct_valid_state.
  destruct (page_of_vadd_smem (st_memory s) pt_addr stealth_va) eqn:?; crush.
  unfold page_of_vadd_smem in Heqo; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  generalize (H5 ma m pg p); crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  destruct H2 with pt_addr m p; crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  rewrite H3 with (pt_addr:=pt_addr) in H0; crush.
  unfold page_of_vadd_smem in Heqo; crush.
  destruct (madd_of_vadd_smem (st_memory s) pt_addr stealth_va) eqn:?; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = Some m).
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold madd_of_vadd, madd_page_table in H8; crush.
  destruct (page_content p0) eqn:?; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct H_valid_virtmap with (st_active_os s) pt_addr p0 p1 stealth_va m; crush.
  rewrite <- H_valid_hyper with (st_active_os s) i0 (os_current_page i) pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem in Heqo; crush.
  unfold madd_of_vadd_smem in Heqo0; crush.

  unfold valid_sm_state in *; crush; STAC.destruct_valid_state.
  STAC.get_aos_data; STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct H_valid_virtmap with (st_active_os s) pt_addr pt_page pt stealth_va ma; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H12 in Heqopt_res; crush.
  specialize (H11 stealth_va_accessible); crush.
  destruct H_valid_cache with stealth_va ma pg; crush.
Qed.    

(* New lemmas *)
  Lemma new_normal_eq_old_new : forall s s' o v p,
    valid_sm_exec s (New_normal o v p) s' ->
    ACT.valid_exec s (New_untrusted o v p) s'.
  Proof.
    crush.
    destruct H; crush.
    unfold ACT.valid_exec; crush.
    unfold pre_new_normal, pre_all_new, ACT.pre_new in *; crush.
    exists x; crush.
  Qed.

  Lemma new_sm_eq_old_new_active : forall s s'' p,
    valid_sm_state s ->
    valid_sm_exec s (New_sm (st_active_os s) p) s'' ->
    exists s',
      (ACT.valid_exec s  (New_untrusted (st_active_os s) C.stealth_va p) s' /\
       ACT.valid_exec s' (Read_hyper C.stealth_va) s'').
  Proof.
    crush.
    destruct H0; crush.
    unfold post_new_sm in H2; crush.
    rename x into ma, x0 into pt_addr, x1 into aos.   
    assert (madd_type (smem_mm (st_memory s)) pt_addr = Some PT).
    unfold madd_type, valid_sm_state in *; crush.
    STAC.destruct_valid_state.
    unfold osi_current_page_table in *; crush.
    destruct H_valid_currpg with (st_active_os s) aos i; crush.
    unfold is_pt_page in *; crush.
    rewrite H12; crush.

    exists (State (st_active_os s) (OSS.set (st_oss s) (st_active_os s)
                                            (OS None (os_current_page aos)))
                  (st_exec_mode s) (st_aos_activity s) (st_hypervisor s)
                  (SMem (smem_c (new_not_shared (st_memory s) pt_addr C.stealth_va ma))
                        (remove_tlb (smem_t (new_not_shared (st_memory s) pt_addr C.stealth_va ma))
                                    C.stealth_va)
                        (smem_mm (new_not_shared (st_memory s) pt_addr C.stealth_va ma))));

      crush.

    unfold ACT.valid_exec; crush.
    unfold pre_new_sm, pre_all_new in H1; crush.
    unfold ACT.pre_new; crush.
    exists x0; crush.
    unfold madd_type, is_rw_page in *; crush.
    rewrite H12 in H14; crush.

    unfold ACT.post_new, ACT.post_all_new; crush.
    exists ma; exists pt_addr; exists aos; crush.
    unfold ACT.new; crush.
    unfold pre_new_sm in H1; crush.

    unfold ACT.valid_exec; crush.
    unfold ACT.pre_read_hyper, ACT.pre_all_read; crush.

    unfold pre_new_sm in H1; crush.
    exists x1; crush.
    unfold aos_page_of_vadd_smem; crush.
    unfold aos_current_page_table; crush.    
    unfold osi_current_page_table in *; crush.    
    rewrite <- OSS.get_set; crush.

    unfold page_of_vadd_smem; crush.
    unfold madd_of_vadd_smem; crush.
    rewrite get_remove_tlb in Heqval; crush.
    rewrite STAC.madd_of_vadd_new_not_shared; crush.
    unfold page_of_vadd_madd_smem; crush.

    rewrite new_cache_not_shared in Heqval0; crush.
    unfold remove_new_va_cache in *; crush.
    unfold valid_sm_state in H; crush.
    unfold no_stealth_page in H13; crush.
    rewrite H13 with (pt_addr:=x) in Heqval0; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.

    remember (add_page_table_entry (smem_mm (st_memory s)) x C.stealth_va x0).
    destruct o; crush.
    rewrite new_mem_not_shared with (mm':=m); crush.
    unfold remove_new_va_mem; crush.
    unfold add_page_table_entry in Heqo; crush.
    remember (page_content p0).
    destruct c; crush.
    rewrite get_add_mem_distinct; crush.
    unfold is_rw_page in *; crush.
    remember (page_content x1).
    destruct c; crush.
    unfold add_page_table_entry in Heqo; crush.
    unfold valid_sm_state in *; crush.
    STAC.destruct_valid_state.
    destruct H_valid_currpg with (st_active_os s) aos i; crush.
    unfold is_pt_page in *; crush.
    rewrite H19 in Heqo; crush.
    unfold is_rw_page in *; crush.
    unfold is_rw_page in *; crush.
    unfold valid_sm_state in *; crush.

    unfold hyper_running; crush.
    unfold pre_new_sm in *; crush.
    unfold pre_all_new in *; crush.

    unfold ACT.post_read_hyper, ACT.post_all_read; crush.
    unfold new_sm; crush.
    unfold page_of_vadd_madd_smem in Heqval0; crush.
    unfold valid_sm_state in *; crush.
    unfold no_stealth_page in H8; crush.
    rewrite H8 with (pt_addr:=pt_addr) in Heqval1; crush.
    unfold pre_new_sm in *; crush.
    unfold pre_new_sm in *; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite <- OSS.get_set in Heqopt_res; crush.

    match goal with
        [|- State _ _ _ _ _ ?mem = State _ _ _ _ _ ?mem'] =>
          assert (mem = mem'); crush
    end.

    match goal with
        [|- SMem ?cache ?tlb ?mm = ?smem'] =>
        assert (cache = smem_c smem' /\
                tlb = smem_t smem' /\
                mm = smem_mm smem'); crush
    end.

    (* Same Cache *)
    unfold is_rw_page in H11; crush.
    rewrite get_new_cache with (ma:=x0) (pg:=x1) (val:=x2); crush.
    rewrite new_cache_not_shared.
    unfold remove_new_va_cache; crush.
    unfold madd_of_vadd_smem; crush.
    rewrite  get_remove_tlb in Heqval0; crush.
    apply STAC.madd_of_vadd_new_not_shared; crush.
    unfold valid_sm_state in *; crush.

    rewrite get_page_result with (pg:=x1); crush.
    unfold page_of_vadd_smem; crush.
    unfold madd_of_vadd_smem; crush.
    rewrite get_remove_tlb in Heqval0; crush.
    rewrite STAC.madd_of_vadd_new_not_shared; crush.

    unfold page_of_vadd_madd_smem; crush.
    rewrite new_cache_not_shared in Heqval2.
    unfold remove_new_va_cache in Heqval2; crush.
    remember (add_page_table_entry (smem_mm (st_memory s)) x C.stealth_va x0).
    destruct o; crush.
    rewrite new_mem_not_shared with (mm':=m); crush.
    unfold remove_new_va_mem; crush.
    unfold add_page_table_entry in Heqo; crush.
    unfold madd_type in *; crush.
    remember (page_content p0).
    destruct c; crush.
    rewrite get_add_mem_distinct; crush.
    unfold add_page_table_entry in Heqo; crush.
    unfold madd_type in *; crush.
    remember (page_content p0).
    destruct c; crush.
    unfold valid_sm_state in *; crush.

    (* Same Tlb *)
    rewrite get_new_tlb with (ma:=x0); crush.
    unfold madd_of_vadd_smem; crush.
    rewrite get_remove_tlb in Heqval0; crush.
    rewrite STAC.madd_of_vadd_new_not_shared; crush.
    unfold valid_sm_state in *; crush.

    (* Same mem *)
    rewrite get_mem_unchanged with (ma:=x0) (pg:=x1); crush. 
    remember (snd (add_cache (smem_c (new_not_shared (st_memory s) x C.stealth_va x0))
     (C.stealth_va,x0) x1)).
    destruct y; crush.
    destruct p0.
    destruct c. 
    assert (C.get_line C.stealth_va = C.get_line v).
    apply C.cache_inertia with (smem_c (new_not_shared (st_memory s) x C.stealth_va x0))
      x0 m x1 p0; crush.
    unfold valid_sm_state in *; crush.
    unfold stealth_page_reserved_line in *.
    assert (get_cache (smem_c (st_memory s)) (v,m) = Some p0).
    STAC.destruct_add_cache (smem_c (new_not_shared (st_memory s) x C.stealth_va x0)).
    STAC.simpl_some_cache_entry.
    rewrite new_cache_not_shared in H20; crush.
    unfold remove_new_va_cache in H20; crush.
    assert (v = C.stealth_va).
    apply H15 with m p0; crush.
    crush.
    unfold no_stealth_page in H14.
    rewrite H14 with (pt_addr:=x) in H19; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.

    match goal with
      [|- _ = page_smem_smem (get ?smem ?va ?ma)] =>
      remember (smem)
    end.
    match goal with
      [|- _ = ?smem'] =>
      destruct smem'
    end; crush.

    unfold page_of_vadd_madd_smem in Heqval0; crush.
    unfold valid_sm_state in *; crush.
    STAC.destruct_valid_state.
    unfold valid_hypervisor in H_valid_hyper.
    assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO (st_active_os s)).
    unfold osi_pa_to_ma in *; crush.
    unfold madd_owner in *; crush.
  Qed.

  Lemma new_sm_eq_old_new : forall s s' o p,
    o <> st_active_os s ->
    valid_sm_exec s (New_sm o p) s' ->
    ACT.valid_exec s (New_untrusted o C.stealth_va p) s'.
  Proof.
    crush.
    destruct H0; crush.
    unfold ACT.valid_exec; crush.
    unfold pre_new_sm, pre_all_new, ACT.pre_new in *; crush.
    exists x0; crush.
    unfold madd_type, is_rw_page in *; crush.
    rewrite H11 in H13; crush.

    unfold post_new_sm, ACT.post_new, ACT.post_all_new in *; crush.
    exists x; exists x0; exists x1; crush.
    unfold new_sm; crush.
    unfold ACT.new; crush.
    unfold pre_new_sm in *; crush.
  Qed.

  Ltac inv_crush :=
    repeat (match goal with
        (* OSS *)                
        | [ |- context[OSS.get (OSS.set ?oss ?osi ?val) ?osi]] =>
          rewrite <- OSS.get_set
        | [ H: context[OSS.get (OSS.set ?oss ?osi ?val) ?osi] |-_] =>
          rewrite <- OSS.get_set in H; crush
        | [ H: ?osi1 <> ?osi2
          |- context[OSS.get (OSS.set ?oss ?osi1 ?val) ?osi2]] =>
        rewrite OSS.get_set_distinct
        | [ H1: ?osi1 <> ?osi2,
            H2: context[OSS.get (OSS.set ?oss ?osi1 ?val) ?osi2] |-_] =>
        rewrite OSS.get_set_distinct
        | [ H: ?osi2 <> ?osi1
          |- context[OSS.get (OSS.set ?oss ?osi1 ?val) ?osi2]] =>
        rewrite OSS.get_set_distinct
        | [ H1: ?osi2 <> ?osi1,
            H2: context[OSS.get (OSS.set ?oss ?osi1 ?val) ?osi2] |-_] =>

        (* HYP *) 
        rewrite HYP.get_set_distinct
        | [ |- context[HYP.get (HYP.set ?oss ?osi ?val) ?osi]] =>
          rewrite <- HYP.get_set
        | [ H: context[HYP.get (HYP.set ?oss ?osi ?val) ?osi] |-_] =>
          rewrite <- HYP.get_set in H; crush
        | [ H: ?osi1 <> ?osi2
          |- context[HYP.get (HYP.set ?oss ?osi1 ?val) ?osi2]] =>
        rewrite HYP.get_set_distinct
        | [ H1: ?osi1 <> ?osi2,
            H2: context[HYP.get (HYP.set ?oss ?osi1 ?val) ?osi2] |-_] =>
        rewrite HYP.get_set_distinct
        | [ H: ?osi2 <> ?osi1
          |- context[HYP.get (HYP.set ?oss ?osi1 ?val) ?osi2]] =>
        rewrite HYP.get_set_distinct
        | [ H1: ?osi2 <> ?osi1,
            H2: context[HYP.get (HYP.set ?oss ?osi1 ?val) ?osi2] |-_] =>
        rewrite HYP.get_set_distinct

        | [|- ex (fun x => (Some ?e = Some x /\ _)) ] =>
          exists e; split
        | [|- ex (fun x => (Some x = Some ?e /\ _)) ] =>
          exists e; split

      end; crush); crush.


  Lemma remove_osi_stealth_page_valid_state : forall s s' o oss pt_addr,
    valid_sm_state s ->
    aos_current_page_table s = Some pt_addr ->
    running_os_not_hypercall o (st_aos_activity s') oss ->
    valid_os_idents oss (st_hypervisor s') ->
    valid_current_page (st_hypervisor s') (smem_mm (st_memory s')) oss ->
    s' = State o oss (st_exec_mode s)
           (st_aos_activity s) (st_hypervisor s)
           (remove_osi_stealth_page (ctxt_switch (st_memory s)) pt_addr) ->
    valid_state s' /\ stealth_page_rw_cacheable s'.
  Proof.
    crush.
    match goal with [H: valid_sm_state _ |-_] => unfold valid_sm_state in H end;
      crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in Heqval; crush.
    match goal with
      | [H: page_content ?pt_page = PT_c _ |-
        context[page_content ?pt_page]] =>
      rewrite H; crush
    end.
    unfold valid_sm_state, valid_state; crush.

    (* valid_hypervisor *)
    rewrite flush_ma_mem.
    unfold update_cache_entry; crush.
    rewrite ctxt_switch_cache in Heqval0; crush.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; crush.
    rewrite ctxt_switch_mem.
    unfold valid_hypervisor; crush.
    unfold madd_owner; crush.
    destruct (madd_eq m ma); crush.
    rewrite <- get_add_mem in Heqval1; crush.
    rewrite <- (H_valid_hyper osi p2m0 pa ma); crush.
    unfold madd_owner; crush.
    destruct H_valid_cache with v ma p0; crush.
    destruct H_valid_cache with v ma p0; crush.
    rewrite get_add_mem_distinct in Heqval1; crush.
    rewrite <- (H_valid_hyper osi p2m0 pa ma); crush.
    unfold madd_owner; crush.
    destruct (madd_eq m ma); crush.
    rewrite <- get_add_mem in Heqval1; crush.
    rewrite get_add_mem_distinct in Heqval1; crush.
    rewrite <- (H_valid_hyper osi p2m0 pa ma); crush.
    unfold madd_owner; crush.
    rewrite ctxt_switch_mem; crush.

    (* Valid pt non cacheable *)
    unfold valid_pt_non_cacheable; crush.
    rewrite flush_ma_mem in *; crush.
    unfold update_cache_entry in *; crush.
    edestruct get_remove_ma_cache_some_entry; crush.
    rewrite ctxt_switch_cache in *; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite get_add_mem_distinct in H10; crush.
    destruct H_valid_pt_non_cacheable with pt_addr0 pt_page0; crush.
    rewrite <- get_add_mem in *; crush.
    destruct H_valid_cache with v pt_addr0 pt_page0; crush.
    unfold is_rw_page, is_pt_page in *; crush.
    rewrite ctxt_switch_mem in *; crush.
    destruct H_valid_pt_non_cacheable with pt_addr0 pt_page0; crush.

    (* Non cacheable synonyms *)
    rewrite flush_ma_mem.
    unfold update_cache_entry; crush.
    rewrite ctxt_switch_cache in Heqval0; crush.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; crush.
    rewrite ctxt_switch_mem.
    unfold non_cacheable_synonyms; crush.
    destruct H_valid_cache with v m p0; crush.
    destruct (madd_eq m pt_addr0); crush.
    unfold is_rw_page in *; crush.
    rewrite <- get_add_mem in *; crush.
    destruct (madd_eq m pt_addr'); crush.
    unfold is_rw_page in *; crush.
    rewrite <- get_add_mem in *; crush.
    destruct (madd_eq m ma); crush.
    rewrite <- get_add_mem in *; crush.
    rewrite get_add_mem_distinct in *; crush.
    destruct H_syn with pt_addr0 pt_page0 mp va ma x5 pt_addr' pt_page' mp' va';
      crush.
    rewrite get_add_mem_distinct in *; crush.
    destruct H_syn with pt_addr0 pt_page0 mp va ma pg pt_addr' pt_page' mp' va';
      crush.
    rewrite ctxt_switch_mem; crush.

    (* Valid virtual mapping *)
    rewrite flush_ma_mem.
    unfold update_cache_entry; crush.
    rewrite ctxt_switch_cache in Heqval0; crush.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; crush.
    rewrite ctxt_switch_mem.
    unfold valid_virtual_mapping; crush.
    destruct H_valid_cache with v m p0; crush.
    destruct (madd_eq m pt_addr0); crush.
    unfold is_rw_page in *; crush.
    rewrite <- get_add_mem in *; crush.
    destruct (madd_eq m ma); crush.
    rewrite <- get_add_mem in *; crush.
    rewrite get_add_mem_distinct in *; crush.
    destruct H_valid_virtmap with osi pt_addr0 pt_page0 pt0 va ma;
      crush.
    exists p0; crush.
    rewrite get_add_mem_distinct in *; crush.
    rewrite ctxt_switch_mem; crush.

    (* Valid current page *)
    unfold remove_osi_stealth_page in *; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite H13 in H3; crush.
    rewrite ctxt_switch_mem in *; crush.

    (* Va has valid pa *)
    rewrite flush_ma_mem.
    unfold update_cache_entry; crush.
    rewrite ctxt_switch_cache in Heqval0; crush.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; crush.
    rewrite ctxt_switch_mem.
    unfold va_has_valid_pa; crush.
    destruct H_valid_cache with v m p0; crush.
    destruct (madd_eq m pt_addr0); crush.
    unfold is_rw_page in *; crush.
    rewrite <- get_add_mem in *; crush.
    rewrite get_add_mem_distinct in *; crush.
    rewrite ctxt_switch_mem; crush.

    (* Valid cache *)
    unfold valid_cache; crush.
    rewrite flush_ma_cache in *; crush.
    rewrite ctxt_switch_cache in *; crush.
    destruct (madd_eq i ma); crush.
    rewrite get_remove_ma_cache_eq in *; crush.
    rewrite get_remove_ma_cache_ne in *; crush.
    destruct H_valid_cache with va ma pg_cache; crush.
    rewrite flush_ma_cache in *; crush.
    rewrite ctxt_switch_cache in *; crush.
    destruct (madd_eq i ma); crush.
    rewrite get_remove_ma_cache_eq in *; crush.
    rewrite get_remove_ma_cache_ne in *; crush.
    destruct H_valid_cache with va ma pg_cache; crush.
    rewrite flush_ma_cache in *; crush.
    rewrite ctxt_switch_cache in *; crush.
    destruct (madd_eq i ma); crush.
    rewrite get_remove_ma_cache_eq in *; crush.
    rewrite get_remove_ma_cache_ne in *; crush.
    destruct H_valid_cache with va ma pg_cache; crush.
    rewrite flush_ma_mem; crush.
    unfold update_cache_entry; crush.
    rewrite ctxt_switch_cache in Heqval0; crush.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; crush.
    rewrite ctxt_switch_mem.
    rewrite get_add_mem_distinct; crush.
    rewrite ctxt_switch_mem; crush.

    (* Valid cache va ma *)
    unfold valid_cache_va_ma; crush.
    rewrite flush_ma_cache in *; crush.
    rewrite ctxt_switch_cache in *; crush.
    destruct (madd_eq i ma); crush.
    rewrite get_remove_ma_cache_eq in *; crush.
    rewrite get_remove_ma_cache_ne in *; crush.
    unfold osi_pa_to_pg; crush.
    destruct H_valid_cache with va ma pg_cache; crush.
    destruct H_valid_cache_va_ma with va ma pg_cache; crush.
    exists x0; exists x1; exists x2; exists x3; crush.
    unfold osi_pa_to_pg in *; crush.
    rewrite flush_ma_mem; crush.
    unfold update_cache_entry; crush.
    rewrite ctxt_switch_cache in Heqval0; crush.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; crush.
    rewrite ctxt_switch_mem.

    destruct (madd_eq m i1); crush.
    destruct H_valid_cache with v i1 p0; crush.
    unfold is_rw_page in *; crush.
    rewrite get_add_mem_distinct; crush.
    rewrite ctxt_switch_mem; crush.

    (* Valid tlb *)
    unfold valid_tlb; crush.
    rewrite flush_ma_tlb in *; crush.
    rewrite ctxt_switch_tlb in *; crush.
    rewrite get_empty_tlb in *; crush.

    unfold valid_state; crush;
      try rewrite ctxt_switch_mem;
      try rewrite ctxt_switch_cache;
      try rewrite ctxt_switch_tlb; crush.

    unfold remove_osi_stealth_page in *; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite H13 in H3; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite ctxt_switch_mem in *; crush.
    unfold valid_cache; intros; simpl in *.
    rewrite ctxt_switch_cache in *.
    rewrite ctxt_switch_mem.
    destruct H_valid_cache with va ma pg_cache; crush.
    unfold valid_cache_va_ma, osi_pa_to_pg; crush.
    rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
    unfold valid_tlb; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite ctxt_switch_tlb in *; crush.
    rewrite get_empty_tlb in *; crush.
    rewrite ctxt_switch_mem in *; crush.

    (* Stealth pg rw cacheable *)
    unfold valid_sm_state in *; crush.
    clear H3.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.    
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite H12; crush.
    unfold stealth_page_rw_cacheable; intros; simpl in *.
    unfold page_of_vadd_madd_smem in *.
    remember (get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) i))
      (C.stealth_va, ma)).
    destruct o0; simpl in *.
    injection H11; intro; subst; clear H11.
    rewrite flush_ma_cache in *.
    rewrite ctxt_switch_cache in *.
    destruct (madd_eq i ma).
    crush; rewrite get_remove_ma_cache_eq in Heqo0; crush.
    rewrite get_remove_ma_cache_ne in Heqo0; auto.
    destruct H_valid_cache with C.stealth_va ma pg; crush.
    rewrite flush_ma_cache, flush_ma_mem in *.
    rewrite ctxt_switch_cache, ctxt_switch_mem in *.
    unfold update_cache_entry in *.
    remember (snd (remove_ma_cache (smem_c (st_memory s)) i)) as o0.
    destruct o0; simpl in *.
    destruct p0; destruct c.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; auto; subst.
    destruct (madd_eq m ma).
    subst.
    rewrite <- get_add_mem in *.
    injection H11; intro; subst; clear H11.
    destruct H_valid_cache with v ma pg; crush.
    rewrite get_add_mem_distinct in *; auto.
    destruct H8 with pt_addr0 ma pg; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.
    rewrite get_add_mem_distinct in Heqopt_res0; crush.
    rewrite <- get_add_mem in *; crush.
    destruct H_valid_cache with v pt_addr0 p0; crush.
    unfold is_rw_page in *; crush.
    rewrite H14 in *; crush.
    unfold page_of_vadd_madd_smem; crush.
    rewrite get_remove_ma_cache_ne in Heqo0; crush.

    destruct (madd_eq i ma).
    subst.
    destruct H8 with pt_addr0 ma pg; crush.
    unfold page_of_vadd_madd_smem; crush.
    rewrite get_remove_ma_cache_no_entry in Heqval1; crush.
    rewrite get_remove_ma_cache_ne in Heqo0; auto.
    destruct H8 with pt_addr0 ma pg; crush.
    unfold page_of_vadd_madd_smem; crush.

    unfold stealth_page_rw_cacheable; intros; simpl in *.
    destruct H8 with pt_addr0 ma pg; crush.
    rewrite ctxt_switch_mem in *; crush.

    unfold page_of_vadd_madd_smem in *; crush;
      rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.

    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.    
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite ctxt_switch_mem in *; crush.
  Qed.

  Lemma valid_currpg_remove_osi_stealh_page :
    forall s pt_addr,
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    valid_current_page (st_hypervisor s)
     (smem_mm (remove_osi_stealth_page (ctxt_switch (st_memory s)) pt_addr))
     (st_oss s).
  Proof.
    intros.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.
    unfold valid_current_page; crush.
    unfold remove_osi_stealth_page; crush.
    rewrite ctxt_switch_mem in *; crush.
    rewrite H4; crush.
    rewrite flush_ma_mem; crush.
    unfold update_cache_entry; crush.
    rewrite ctxt_switch_cache in *; crush.
    destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
      v i m p0; crush.
    rewrite ctxt_switch_mem; crush.
    destruct H_valid_currpg with osi system p2m0; crush.
    exists x; exists x0; crush.
    destruct (madd_eq m x); crush.
    destruct H_valid_cache with v x p0; crush.
    unfold is_rw_page, is_pt_page in *; crush.
    rewrite get_add_mem_distinct; crush.
    rewrite ctxt_switch_mem; crush.
    rewrite ctxt_switch_mem; crush.
    rewrite ctxt_switch_mem; crush.
  Qed.    

(* Ctxt switch lemmas *)
Lemma madd_of_vadd_smem_ctxt_switch :
  forall s pt_addr va,
    madd_of_vadd_smem (ctxt_switch (st_memory s)) pt_addr va =
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va.
Proof.
  intros.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *;
    rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
Qed.

Lemma madd_of_vadd_flush_ma :
  forall s ma_flushed pt_addr va,
    valid_state s ->
    madd_of_vadd (smem_mm (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
                 pt_addr va =
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va.
Proof.    
  intros; rewrite flush_ma_mem; unfold madd_of_vadd, madd_page_table; crush.
  unfold update_cache_entry; crush; 
    try rewrite ctxt_switch_cache, ctxt_switch_mem in *; crush.
  remember (madd_type (smem_mm (st_memory s)) pt_addr).
  destruct o; crush; try (destruct p; crush);
    unfold madd_type in *; crush.

  remember (page_content p).
  destruct c; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite <- Heqc; crush.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
    v ma_flushed pt_addr p0; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p0; crush.
  unfold madd_type, is_rw_page in *; crush.

  remember (page_content p).
  destruct c; crush.

  destruct (madd_eq m pt_addr); crush.
  (* m = pt_addr *)
  rewrite <- get_add_mem; crush.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
     v ma_flushed pt_addr p0; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H0; crush.
  (* m <> pt_addr *)
  rewrite get_add_mem_distinct; crush.
  rewrite <- Heqc; crush.

  remember (get_mem (smem_mm (st_memory s)) pt_addr).
  destruct o; crush.

  remember (page_content p).
  destruct c; crush.
  
  destruct (madd_eq m pt_addr); crush.
  (* m = pt_addr *)
  rewrite <- get_add_mem; crush.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
     v ma_flushed pt_addr p0; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p0; crush.
  unfold is_rw_page in *; crush.
  (* m <> pt_addr *)
  rewrite get_add_mem_distinct; crush.
  rewrite <- Heqc; crush.

  destruct (madd_eq m pt_addr); crush.
  (* m = pt_addr *)
  rewrite <- get_add_mem; crush.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
     v ma_flushed pt_addr p0; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p0; crush.
  (* m <> pt_addr *)
  rewrite get_add_mem_distinct; crush.
  rewrite <- Heqo0; crush.
Qed.

Lemma page_of_vadd_madd_smem_flush_ma :
  forall s ma_flushed pt_addr va ma,
    valid_state s ->
    (exists osi pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
    page_of_vadd_madd_smem
      (flush_ma (ctxt_switch (st_memory s)) ma_flushed) va ma =
    page_of_vadd_madd_smem (st_memory s) va ma.
Proof.
  intros.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_cache in Heqval; crush.
  destruct (madd_eq ma ma_flushed); crush.
  rewrite get_remove_ma_cache_eq in Heqval; crush.
  rewrite get_remove_ma_cache_ne in Heqval; crush.
  rewrite ctxt_switch_cache in *; crush.

  rewrite flush_ma_cache in Heqval; crush.
  destruct (madd_eq ma ma_flushed); crush.
  rewrite get_remove_ma_cache_eq in Heqval; crush.
  rewrite get_remove_ma_cache_ne in Heqval; crush.
  rewrite ctxt_switch_cache in *; crush.

  rewrite flush_ma_cache in Heqval; crush.
  destruct (madd_eq ma ma_flushed); crush.
  rewrite get_remove_ma_cache_eq in Heqval; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
    v ma_flushed m p1; crush.
  rewrite <- get_add_mem; crush.
  destruct (vadd_eq v va); crush.
  destruct STAC.no_cache_alias with s va v m p p1; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache in *; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval0; crush.

  rewrite get_remove_ma_cache_ne in Heqval; crush.
  rewrite ctxt_switch_cache in *; crush.

  rewrite flush_ma_cache in Heqval; crush.
  destruct (madd_eq ma ma_flushed); crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
    v ma_flushed m p0; crush.
  rewrite <- get_add_mem.
  rewrite get_remove_ma_cache_eq in Heqval.
  destruct (vadd_eq va v); crush.
  STAC.destruct_valid_state.  
  STAC.synonym_cached.
  rewrite ctxt_switch_mem; crush.

  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
    v ma_flushed m p0; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
Qed.

Lemma page_of_vadd_madd_smem_ctxt_switch :
  forall s va ma,
    page_of_vadd_madd_smem
      (ctxt_switch (st_memory s)) va ma =
    page_of_vadd_madd_smem (st_memory s) va ma.
Proof.
  intros.
  unfold page_of_vadd_madd_smem; crush;
  try rewrite ctxt_switch_cache in *;
  try rewrite ctxt_switch_mem in *; crush.
Qed.

Lemma page_of_vadd_smem_ctxt_switch :
  forall s ma_flushed pt_addr va,
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    page_of_vadd_smem (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
                      pt_addr va =
    page_of_vadd_smem (st_memory s) pt_addr va.
Proof.
  intros.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush;
    try (rewrite flush_ma_tlb in *;
       rewrite ctxt_switch_tlb in *;
       rewrite get_empty_tlb in *; crush).

  rewrite madd_of_vadd_flush_ma; crush.
  remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr va).
  destruct o; crush.

  unfold madd_of_vadd, madd_page_table in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.

  generalize H0.
  unfold aos_current_page_table, osi_current_page_table in H0; crush.
  rewrite H4 in *; crush.
  unfold valid_tlb in H_valid_tlb.

  generalize (H_valid_tlb va m pt_addr Heqval0 H3); intros.
  unfold madd_of_vadd, madd_page_table in H1; crush.
  rewrite H4 in *; crush.
  apply page_of_vadd_madd_smem_flush_ma with pt_addr; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s); exists (os_current_page aos); crush.

  STAC.destruct_valid_state.
  rewrite H_valid_tlb with (ma:=m) in Heqo; crush.
  
  rewrite madd_of_vadd_flush_ma; crush.
  remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr va).
  destruct o; crush.
  apply page_of_vadd_madd_smem_flush_ma with pt_addr; crush.
  unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
  exists (st_active_os s); exists (os_current_page i); crush.
Qed.

Lemma madd_of_vadd_smem_flush_ma :
  forall s pt_addr va ma',
    valid_state s ->
    madd_type (smem_mm (st_memory s)) pt_addr = Some PT ->
    madd_of_vadd_smem (flush_ma (ctxt_switch (st_memory s)) ma') pt_addr va =
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va.
Proof.
  intros.
  unfold madd_type in *; crush.
  remember (page_content p).
  destruct c; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in Heqval.
  rewrite ctxt_switch_tlb in Heqval.
  rewrite get_empty_tlb in Heqval; crush.
  clear Heqval.
  rewrite flush_ma_mem.
  unfold update_cache_entry; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_cache in Heqval.
  destruct get_remove_ma_cache_some_entry with
  (smem_c (st_memory s)) v ma' pt_addr p2; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p2; crush.
  unfold is_rw_page in *; crush.
  rewrite ctxt_switch_mem; crush.
Qed.

Lemma madd_of_vadd_get_ctxt_switch : 
  forall s osi pt_padd' pt_addr pt_addr' va va',
    valid_state s ->
    osi_pa_to_ma s osi pt_padd' = Some pt_addr' ->
    madd_of_vadd (smem_mm
      (page_smem_smem (get (ctxt_switch (st_memory s))
                           va' pt_addr'))) pt_addr va =
       madd_of_vadd (smem_mm (st_memory s)) pt_addr va.
Proof.
  intros.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?; crush.
  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) va' m).
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  destruct H_valid_virtmap with osi pt_addr' p0 p1 va' m; crush.
  rewrite <- H_valid_hyper with osi i pt_padd' pt_addr'; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H1.
  rename x into pg.
  destruct (add_cache (smem_c (ctxt_switch (st_memory s)))
    (va',m) pg) eqn:?; crush.
  rewrite ctxt_switch_cache in Heqp.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    
  unfold madd_of_vadd, madd_page_table; crush.
  destruct (madd_eq pt_addr m0); crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=m) (pg:=pg) (pg':=x0);
    crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m0 x0; crush.
  unfold is_rw_page in *; crush.
  rewrite H7, H2; crush.
  rewrite ctxt_switch_cache; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=pg) (pg':=x0);
    crush.
  rewrite ctxt_switch_mem.
  unfold update_cache_entry; crush.
  rewrite ctxt_switch_cache; crush.

  rewrite get_mem_unchanged with (ma:=m) (pg:=pg); crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_cache; crush.
  rewrite Heqp; crush.
  rewrite get_invalid; crush.
  rewrite ctxt_switch_mem; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
Qed.

Lemma madd_of_vadd_get_flush_ma :
  forall s osi ma_flushed pt_padd' pt_addr pt_addr' va va',
    valid_state s ->
    osi_pa_to_ma s osi pt_padd' = Some pt_addr' ->
    madd_of_vadd (smem_mm
      (page_smem_smem (get (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
                           va' pt_addr'))) pt_addr va =
       madd_of_vadd (smem_mm (st_memory s)) pt_addr va.
Proof.
  intros.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr' va') eqn:?; crush.
  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) va' m).
  unfold madd_of_vadd, madd_page_table in *; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  destruct H_valid_virtmap with osi pt_addr' p0 p1 va' m; crush.
  rewrite <- H_valid_hyper with osi i pt_padd' pt_addr'; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H1.
  rename x into pg.
  destruct (add_cache
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (va',m) pg) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    
  rewrite flush_ma_cache in H3.
  destruct (madd_eq m0 ma_flushed); crush;
    [rewrite get_remove_ma_cache_eq in * |
     rewrite get_remove_ma_cache_ne in *]; crush.
  rewrite ctxt_switch_cache in H3; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  destruct (madd_eq pt_addr m0); crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=m) (pg:=pg) (pg':=x0);
    crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m0 x0; crush.
  unfold is_rw_page in *; crush.
  rewrite H7, H2; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=m) (ma':=m0) (pg:=pg) (pg':=x0);
    crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_mem.
  unfold update_cache_entry; crush.
  destruct (madd_eq pt_addr m1); crush.
  rewrite <- MM.get_add_mem; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in H5; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v0 m1 p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H8, H2; crush.
  rewrite MM.get_add_mem_distinct; crush.

  rewrite get_mem_unchanged with (ma:=m) (pg:=pg); crush.
  apply madd_of_vadd_flush_ma; crush.
  rewrite Heqp; crush.
  rewrite get_invalid; crush.
  apply madd_of_vadd_flush_ma; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite Heqo; crush.
Qed.

Lemma madd_of_vadd_ctxt_switch_new_smem :
  forall s osi pt_padd' pt_addr pt_addr' va,
    valid_state s ->
    osi_pa_to_ma s osi pt_padd' = Some pt_addr' ->
    madd_of_vadd (smem_mm (ctxt_switch_new_smem s pt_addr')) pt_addr va =
    madd_of_vadd (smem_mm (st_memory s)) pt_addr va.
Proof.
  Local Hint Resolve madd_of_vadd_get_flush_ma
    madd_of_vadd_get_ctxt_switch.
  intros.
  unfold ctxt_switch_new_smem; crush.
  unfold remove_osi_stealth_page; crush.
  destruct (page_content p) eqn:?; crush.
Qed.

Lemma page_of_vadd_madd_smem_get_flush_ma:
  forall s ma_flushed osi osi' pt_padd pt_padd' pt_addr pt_addr' va ma,
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pt_padd' = Some pt_addr' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va = Some ma ->
    page_of_vadd_madd_smem (page_smem_smem
      (get (flush_ma (ctxt_switch (st_memory s)) ma_flushed) stealth_va pt_addr))
      va ma = page_of_vadd_madd_smem (st_memory s) va ma.
Proof.
  unfold valid_sm_state; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?; crush.
  rename m into x.

  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) stealth_va x).
  unfold madd_of_vadd, madd_page_table in Heqo; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  destruct H_valid_virtmap with osi pt_addr p0 p1 stealth_va x; crush.
  rewrite <- H_valid_hyper with osi i0 pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H9 as [pg ?].
  destruct (add_cache
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) pg) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    
  rewrite flush_ma_cache in H11.
  destruct (madd_eq m ma_flushed); crush;
    [rewrite get_remove_ma_cache_eq in * |
     rewrite get_remove_ma_cache_ne in *]; crush.
  rewrite ctxt_switch_cache in *; crush.
  assert (is_rw_page pg).
  destruct H7 with pt_addr x pg; crush.
  unfold is_rw_page in *; crush.
  rename x0 into val.
  unfold page_of_vadd_madd_smem; crush.
  generalize (get_new_cache (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
    pt_addr stealth_va x pg val); crush.
  rewrite H13 in Heqval0; clear H13; crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in Heqval0; [|crush].
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,ma) pg)) in Heqval0; [|crush].
  rewrite <- get_add_cache_cacheable in Heqval0; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H7 with pt_addr ma pg; crush.
  assert (Some p0 = get_cache
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)).
  generalize (get_add_cache_distinct_not_new
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p); intros.
  rewrite <- Heqp in H13.
  rewrite flush_ma_cache in *.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq ma_flushed ma); crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in Heqval0; [|crush].  
  generalize (get_add_cache_distinct_not_new
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p); intros.
  rewrite <- Heqp in H14.
  rewrite <- H14 in H13; auto.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in *;
    auto.
  rewrite <- Heqp; auto.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.  
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.

  rewrite get_page_result with (pg:=pg); crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  clear Heqval2.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr); crush.

  generalize (get_new_cache (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
    pt_addr stealth_va x pg val); crush.
  rewrite H13 in Heqval0; clear H13; crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in Heqval0; [|crush].
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,ma) pg)) in Heqval0; [|crush].
  rewrite <- get_add_cache_cacheable in Heqval0; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H7 with pt_addr' ma pg; crush.
  assert (Some p = get_cache
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)).
  generalize (get_add_cache_distinct_not_new
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p); intros.
  rewrite <- Heqp in H13.
  apply H13; crush.
  rewrite flush_ma_cache in *.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq ma_flushed ma); crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.  
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite get_page_result with (pg:=pg); crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  clear Heqval2.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr); crush.

  generalize (get_new_cache (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
    pt_addr stealth_va x pg val); crush.
  rewrite H13 in Heqval0; clear H13; crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in Heqval0; [|crush].
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,ma) pg)) in Heqval0; [|crush].
  rewrite <- get_add_cache_cacheable in Heqval0; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H7 with pt_addr' ma p0; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct (madd_eq ma_flushed ma); crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_cache, ctxt_switch_mem.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct (vadd_eq v0 va); crush.
  rewrite <- MM.get_add_mem; crush.
  destruct STAC.no_cache_alias with s v0 va m0 p1 p; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval1; crush.
  destruct (madd_eq ma m); crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  destruct (vadd_eq v va); crush.
  destruct STAC.no_cache_alias with s v va m x1 p; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_cache, ctxt_switch_mem.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) m0)) (stealth_va,x) pg)) in Heqval0; [|crush].
  assert (None = get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) m)) (va,ma)).
  destruct (get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) m)) (va,ma)) eqn:?; crush.
  generalize (get_add_cache_distinct
    (smem_c (flush_ma (ctxt_switch (st_memory s)) m0))
    (stealth_va,x) (va,ma) pg p0); intros.
  rewrite <- Heqp in H13.
  rewrite <- H13 in Heqval0; auto; clear H13.
  right.
  exists (v,m); exists x1; crush.
  rewrite flush_ma_cache in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  rewrite flush_ma_cache in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  rewrite ctxt_switch_cache in *; crush.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in Heqval0; [|crush].
  assert (None = get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)).
  destruct (get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)) eqn:?; crush.
  generalize (get_add_cache_distinct
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p0); intros.
  rewrite <- Heqp in H13.
  rewrite <- H13 in Heqval0; crush; clear H13.
  right.
  exists (v,m); exists x1; crush.
  rewrite flush_ma_cache in *; crush.
  rewrite get_remove_ma_cache_ne in H13; crush.
  rewrite ctxt_switch_cache in *; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.  
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite get_page_result with (pg:=pg); crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  clear Heqval2.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr); crush.

  destruct (madd_eq m ma); crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=x) (ma':=ma) (pg:=pg) (pg':=x1);
    crush.
  destruct (vadd_eq va v); crush.  
  STAC.destruct_valid_state.
  STAC.synonym_cached.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct (madd_eq m0 ma); crush.
  rewrite ctxt_switch_cache in *; crush.
  destruct (vadd_eq va v0); crush.
  STAC.destruct_valid_state.
  STAC.synonym_cached.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.

  unfold page_of_vadd_madd_smem; crush.
  assert (is_rw_page pg).
  destruct H7 with pt_addr x pg; crush.
  unfold is_rw_page in H10; crush.
  rename x0 into val.
  generalize (get_new_cache (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
    pt_addr stealth_va x pg val); crush.
  rewrite H11 in Heqval; clear H11; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H7 with pt_addr' ma p0; crush.
  unfold page_of_vadd_madd_smem; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  assert (Some p = get_cache
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)).
  generalize (get_add_cache_distinct_not_new
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p); intros.
  rewrite Heqp in H11.
  apply H11; auto.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)); auto.
  rewrite Heqp; auto.
  rewrite flush_ma_cache in *.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq ma_flushed ma); crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.  
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite get_page_result with (pg:=pg); crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  clear Heqval1.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr); crush.
  assert (is_rw_page pg).
  destruct H7 with pt_addr x pg; crush.
  unfold is_rw_page in H10; crush.
  rename x0 into val.
  generalize (get_new_cache (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
    pt_addr stealth_va x pg val); crush.
  rewrite H11 in Heqval; clear H11; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H7 with pt_addr' ma pg; crush.
  assert (Some p = get_cache
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)).
  generalize (get_add_cache_distinct_not_new
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p); intros.
  rewrite Heqp in H11.
  apply H11; auto.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)); auto.
  rewrite Heqp; auto.
  rewrite flush_ma_cache in *.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq ma_flushed ma); crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.  
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite get_page_result with (pg:=pg); crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  clear Heqval1.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr); crush.
  assert (is_rw_page pg).
  destruct H7 with  pt_addr x pg; crush.
  unfold is_rw_page in H10; crush.
  rename x0 into val.
  generalize (get_new_cache (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
    pt_addr stealth_va x pg val); crush.
  rewrite H11 in Heqval; clear H11; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H7 with pt_addr' ma p0; crush.
  unfold page_of_vadd_madd_smem; crush.
  assert (None = get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)).
  destruct (get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)) eqn:?; crush.
  generalize (get_add_cache_distinct
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p0); intros.
  rewrite Heqp in H11.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in H11; auto.
  rewrite <- H11 in Heqval; crush; clear H11.
  rewrite Heqp; auto.
  destruct (madd_eq ma_flushed ma); crush.
  rewrite flush_ma_cache in H11.
  rewrite ctxt_switch_cache in H11.
  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush.
  2: rewrite Heqp; crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_cache; rewrite ctxt_switch_mem.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite <- MM.get_add_mem; crush.
  destruct (vadd_eq va v); crush.
  destruct STAC.no_cache_alias with s v va m p1 p; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval0; crush.
  rewrite flush_ma_cache in H11.
  rewrite ctxt_switch_cache in H11.
  rewrite get_remove_ma_cache_ne in *; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.  
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite get_page_result with (pg:=pg); crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  clear Heqval1.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr); crush.
  assert (is_rw_page pg).
  destruct H7 with pt_addr x pg; crush.
  unfold is_rw_page in H10; crush.
  rename x0 into val.
  generalize (get_new_cache (flush_ma (ctxt_switch (st_memory s)) ma_flushed)
    pt_addr stealth_va x pg val); crush.
  rewrite H11 in Heqval; clear H11; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct H7 with pt_addr' ma pg; crush.
  unfold page_of_vadd_madd_smem; crush.
  assert (None = get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)).
  destruct (get_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (va,ma)) eqn:?; crush.
  generalize (get_add_cache_distinct
    (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    (stealth_va,x) (va,ma) pg p); intros.
  rewrite Heqp in H11.
  replace c with (fst (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg)) in H11; auto.
  rewrite <- H11 in Heqval; crush; clear H11.
  rewrite Heqp; auto.
  destruct (madd_eq ma_flushed ma); crush.
  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush.
  2: rewrite Heqp; crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_cache; rewrite ctxt_switch_mem.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  destruct (vadd_eq va v); crush.
  STAC.destruct_valid_state.
  STAC.synonym_cached.
  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush.
  2: rewrite Heqp; crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_cache; rewrite ctxt_switch_mem.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite MM.get_add_mem_distinct; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.  
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite get_page_result with (pg:=pg); crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  clear Heqval1.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr); crush.

  (* No old stealth page *)
  rewrite get_invalid; crush.
  rewrite page_of_vadd_madd_smem_flush_ma with (pt_addr:=pt_addr'); crush.
  unfold page_of_vadd_smem; crush.
  destruct (madd_type (smem_mm (st_memory s)) pt_addr) eqn:?.
  destruct p; crush.
  rewrite madd_of_vadd_smem_flush_ma; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  assert (madd_of_vadd (smem_mm (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    pt_addr stealth_va = None); crush.
  2:rewrite H9; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  destruct (madd_eq m pt_addr); crush.
  rewrite <- MM.get_add_mem; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H9; crush.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  assert (madd_of_vadd (smem_mm (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    pt_addr stealth_va = None); crush.
  2:rewrite H9; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  destruct (madd_eq m pt_addr); crush.
  rewrite <- MM.get_add_mem; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v pt_addr p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H9; crush.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
Qed.

Lemma page_of_vadd_madd_smem_get_ctxt_switch:
  forall s osi osi' pt_padd pt_padd' pt_addr pt_addr' va ma,
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pt_padd' = Some pt_addr' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va = Some ma ->
    page_of_vadd_madd_smem (page_smem_smem
      (get (ctxt_switch (st_memory s)) stealth_va pt_addr))
      va ma = page_of_vadd_madd_smem (st_memory s) va ma.
Proof.
  unfold valid_sm_state; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?; crush.
  rename m into x.

  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) stealth_va x).
  unfold madd_of_vadd, madd_page_table in Heqo; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in *; crush.
  destruct H_valid_virtmap with osi pt_addr p0 p1 stealth_va x; crush.
  rewrite <- H_valid_hyper with osi i0 pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H9 as [pg ?].

  assert (madd_of_vadd_smem (ctxt_switch (st_memory s)) pt_addr stealth_va = Some x).
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  assert (page_of_vadd_smem (ctxt_switch (st_memory s)) pt_addr stealth_va = Some pg).
  unfold page_of_vadd_smem; crush.
  rewrite page_of_vadd_madd_smem_ctxt_switch; crush.

  assert (is_rw_page pg).
  destruct H7 with pt_addr x pg; crush.
  unfold is_rw_page in H12; crush.
  rename x0 into val.

  destruct (add_cache (smem_c (st_memory s)) (stealth_va,x) pg) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    

  unfold page_of_vadd_madd_smem at 1; crush.
  rewrite get_new_cache with (ma:=x) (pg:=pg) (val:=val) in Heqval0;
    try erewrite get_page_result; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval0; crush.
  destruct H7 with pt_addr' ma pg; crush.
  generalize (get_add_cache_distinct_not_new
    (smem_c (st_memory s)) (stealth_va,x) (va,ma) pg p); intros.
  rewrite <- Heqp in H13.
  rewrite ctxt_switch_cache in *.
  replace c with (fst (add_cache (smem_c (st_memory s)) (stealth_va,x) pg)) in H13.
  unfold page_of_vadd_madd_smem; crush.
  rewrite <- Heqp; crush.

  rewrite get_new_cache with (ma:=x) (pg:=pg) (val:=val) in Heqval0;
    try erewrite get_page_result; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval0; crush.
  destruct H7 with pt_addr' ma pg; crush.
  rewrite ctxt_switch_cache in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct (eq_cache_key (v,m) (va,ma)); crush.
  rewrite get_cache_entry1 with (va':=va) (ma:=x) (ma':=ma) (pg:=pg) (pg':=x1);
    crush.
  rewrite ctxt_switch_cache.
  rewrite <- Heqp; crush.
  generalize (get_add_cache_distinct
    (smem_c (st_memory s)) (stealth_va,x) (va,ma) pg p); intros.
  rewrite <- Heqp in H13.
  rewrite <- H13 in Heqval0; crush.
  destruct (madd_eq m ma); crush.
  destruct (vadd_eq va v); crush.
  STAC.destruct_valid_state.
  STAC.synonym_cached.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_cache.
  rewrite <- Heqp; crush.
  
  unfold page_of_vadd_madd_smem at 1; crush.
  rewrite get_new_cache with (ma:=x) (pg:=pg) (val:=val) in Heqval0;
    try erewrite get_page_result; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval0; crush.
  destruct H7 with pt_addr' ma pg; crush.
  generalize (get_add_cache_distinct_not_new
    (smem_c (st_memory s)) (stealth_va,x) (va,ma) pg p); intros.
  rewrite Heqp in H13.
  rewrite ctxt_switch_cache in *.
  replace c with (fst (add_cache (smem_c (st_memory s)) (stealth_va,x) pg)) in H13.
  unfold page_of_vadd_madd_smem; crush.
  rewrite Heqp; crush.

  rewrite get_new_cache with (ma:=x) (pg:=pg) (val:=val) in Heqval0;
    try erewrite get_page_result; crush.
  destruct (eq_cache_key (stealth_va,x) (va,ma)); crush.
  rewrite <- get_add_cache_cacheable in Heqval0; crush.
  destruct H7 with pt_addr' ma pg; crush.
  rewrite ctxt_switch_cache in *; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush. 
  rewrite ctxt_switch_mem; crush.
  generalize (get_add_cache_distinct
    (smem_c (st_memory s)) (stealth_va,x) (va,ma) pg p); intros.
  rewrite Heqp in H13.
  replace c with (fst (add_cache (smem_c (st_memory s)) (stealth_va,x) pg)) in H13.
  rewrite <- H13 in Heqval0; crush.
  rewrite Heqp; crush.
  rewrite ctxt_switch_cache.
  rewrite Heqp; crush.
  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush. 
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_cache.
  rewrite Heqp; crush.

  (* No old stealth page *)
  rewrite get_invalid; crush.
  unfold page_of_vadd_madd_smem;
    try rewrite ctxt_switch_cache, ctxt_switch_mem in *; crush.
  unfold page_of_vadd_smem; crush.
  destruct (madd_type (smem_mm (st_memory s)) pt_addr) eqn:?.
  destruct p; crush.
  rewrite madd_of_vadd_smem_ctxt_switch; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  assert (madd_of_vadd (smem_mm (ctxt_switch (st_memory s)))
    pt_addr stealth_va = None); crush.
  2:rewrite H9; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite madd_of_vadd_smem_ctxt_switch; crush.
  rewrite Heqo; crush.
Qed.

Lemma page_of_vadd_madd_smem_ctxt_switch_new_smem:
  forall s osi osi' pt_padd pt_padd' pt_addr pt_addr' va ma,
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pt_padd' = Some pt_addr' ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr' va = Some ma ->
    page_of_vadd_madd_smem (ctxt_switch_new_smem s pt_addr) va ma =
    page_of_vadd_madd_smem (st_memory s) va ma.
Proof.
  Local Hint Resolve page_of_vadd_madd_smem_get_ctxt_switch
    page_of_vadd_madd_smem_get_flush_ma.
  intros.
  generalize H.
  unfold valid_sm_state; crush.
  unfold ctxt_switch_new_smem; crush.
  unfold remove_osi_stealth_page; crush.
  rewrite ctxt_switch_mem in *.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite H13; crush.
Qed.

Lemma osi_pa_to_pg_get_flush_ma :
  forall s ma_flushed osi pt_padd pt_addr osi' pa' ma',
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pa' = Some ma' ->
    madd_type (smem_mm (st_memory s)) ma' = Some PT ->
    get_mem (smem_mm (page_smem_smem
      (get (flush_ma (ctxt_switch (st_memory s)) ma_flushed) stealth_va pt_addr))) ma' =
    get_mem (smem_mm (st_memory s)) ma'.
Proof.
  unfold valid_sm_state; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?; crush.
  rename m into x.

  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) stealth_va x).
  unfold madd_of_vadd, madd_page_table in Heqo; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in H0; crush.
  destruct H_valid_virtmap with osi pt_addr p0 p1 stealth_va x; crush.
  rewrite <- H_valid_hyper with osi i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H9 as [pg ?].

  destruct (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    
  
  destruct (madd_eq m ma'); crush.
  rewrite flush_ma_cache in H11; crush.
  rewrite get_remove_ma_cache_ne in H11; crush.
  rewrite ctxt_switch_cache in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' x1; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H14 in *; crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq m0 ma'); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v0 ma' p0; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H15 in *; crush.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.

  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush. 
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq m ma'); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' p0; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H13 in *; crush.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqp; crush.

  rewrite get_invalid; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq m ma'); crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' p0; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H12 in *; crush.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
  unfold page_of_vadd_smem; crush.
  destruct (madd_type (smem_mm (st_memory s)) pt_addr) eqn:?.
  destruct p; crush.
  rewrite madd_of_vadd_smem_flush_ma; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  assert (madd_of_vadd (smem_mm (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    pt_addr stealth_va = None); crush.
  2:rewrite H9; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite Heqo; crush.
Qed.

Lemma osi_pa_to_pg_get_ctxt_switch :
  forall s osi pt_padd pt_addr osi' pa' ma',
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pa' = Some ma' ->
    madd_type (smem_mm (st_memory s)) ma' = Some PT ->
    get_mem (smem_mm (page_smem_smem
      (get (ctxt_switch (st_memory s)) stealth_va pt_addr))) ma' =
    get_mem (smem_mm (st_memory s)) ma'.
Proof.
  unfold valid_sm_state; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?; crush.
  rename m into x.

  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) stealth_va x).
  unfold madd_of_vadd, madd_page_table in Heqo; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in H0; crush.
  destruct H_valid_virtmap with osi pt_addr p0 p1 stealth_va x; crush.
  rewrite <- H_valid_hyper with osi i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H9 as [pg ?].

  destruct (add_cache (smem_c (ctxt_switch (st_memory s))) (stealth_va,x) pg) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    
  
  destruct (madd_eq m ma'); crush.
  rewrite ctxt_switch_cache in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' x1; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H14 in *; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite ctxt_switch_mem; crush.

  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush. 
  rewrite ctxt_switch_mem; crush.
  rewrite Heqp; crush.

  rewrite get_invalid; crush.
  rewrite ctxt_switch_mem; crush.
  unfold page_of_vadd_smem; crush.
  destruct (madd_type (smem_mm (st_memory s)) pt_addr) eqn:?.
  destruct p; crush.
  rewrite madd_of_vadd_smem_ctxt_switch; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
Qed.  

Lemma osi_pa_to_pg_ctxt_switch_new_smem:
  forall s osi pt_padd pt_addr osi' pa' ma',
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pa' = Some ma' ->
    madd_type (smem_mm (st_memory s)) ma' = Some PT ->
    get_mem (smem_mm (ctxt_switch_new_smem s pt_addr)) ma' =
    get_mem (smem_mm (st_memory s)) ma'.
Proof.
  Local Hint Resolve osi_pa_to_pg_get_ctxt_switch
       osi_pa_to_pg_get_flush_ma.
  intros.
  generalize H.
  unfold valid_sm_state; crush.
  unfold ctxt_switch_new_smem; crush.
  unfold remove_osi_stealth_page; crush.
  unfold valid_sm_state; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H13; crush.
Qed.

Lemma madd_type_get_flush_ma :
  forall s ma_flushed osi pt_padd pt_addr osi' pa' ma',
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pa' = Some ma' ->
    madd_type (smem_mm (page_smem_smem
      (get (flush_ma (ctxt_switch (st_memory s)) ma_flushed) stealth_va pt_addr))) ma' =
    madd_type (smem_mm (st_memory s)) ma'.
Proof.
  unfold valid_sm_state; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?; crush.
  rename m into x.

  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) stealth_va x).
  unfold madd_of_vadd, madd_page_table in Heqo; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in H0; crush.
  destruct H_valid_virtmap with osi pt_addr p0 p1 stealth_va x; crush.
  rewrite <- H_valid_hyper with osi i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H8 as [pg ?].

  destruct (add_cache (smem_c (flush_ma (ctxt_switch (st_memory s)) ma_flushed)) (stealth_va,x) pg) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    
  
  destruct (madd_eq m ma'); crush.
  unfold madd_type; crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=x) (ma':=ma') (pg:=pg) (pg':=x1);
    crush.
  rewrite flush_ma_cache in H10; crush.
  rewrite get_remove_ma_cache_ne in H10; crush.
  rewrite ctxt_switch_cache in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' x1; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H9, H14; crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  unfold madd_type; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq m0 ma'); crush.
  rewrite <- MM.get_add_mem; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v0 ma' p0; crush.
  unfold is_rw_page in *; crush.
  rewrite H9, H15; crush.
  rewrite MM.get_add_mem_distinct; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.

  unfold madd_type; crush.
  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush. 
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq m ma'); crush.
  rewrite <- MM.get_add_mem; crush.  
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' p0; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H9, H13; crush.
  rewrite MM.get_add_mem_distinct; crush.  
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqp; crush.

  unfold madd_type; crush.
  rewrite get_invalid; crush.
  rewrite flush_ma_mem; crush.
  unfold update_cache_entry; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_cache in *.
  destruct (madd_eq m ma'); crush.
  rewrite <- MM.get_add_mem; crush.  
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' p0; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H8, H12; crush.
  rewrite MM.get_add_mem_distinct; crush.  
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
  
  unfold page_of_vadd_smem; crush.
  destruct (madd_type (smem_mm (st_memory s)) pt_addr) eqn:?.
  destruct p; crush.
  rewrite madd_of_vadd_smem_flush_ma; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  assert (madd_of_vadd (smem_mm (flush_ma (ctxt_switch (st_memory s)) ma_flushed))
    pt_addr stealth_va = None); crush.
  2:rewrite H8; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb, ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite madd_of_vadd_flush_ma; crush.
  rewrite Heqo; crush.
Qed.

Lemma madd_type_get_ctxt_switch :
  forall s osi pt_padd pt_addr osi' pa' ma',
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pa' = Some ma' ->
    madd_type (smem_mm (page_smem_smem
      (get (ctxt_switch (st_memory s)) stealth_va pt_addr))) ma' =
    madd_type (smem_mm (st_memory s)) ma'.
Proof.
  unfold valid_sm_state; crush.
  destruct (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va) eqn:?; crush.
  rename m into x.

  assert (exists pg, Some pg = page_of_vadd_madd_smem (st_memory s) stealth_va x).
  unfold madd_of_vadd, madd_page_table in Heqo; crush.
  destruct (page_content p0) eqn:?; crush.
  STAC.destruct_valid_state.
  unfold osi_pa_to_ma in H0; crush.
  destruct H_valid_virtmap with osi pt_addr p0 p1 stealth_va x; crush.
  rewrite <- H_valid_hyper with osi i pt_padd pt_addr; crush.
  unfold madd_owner; crush.
  unfold page_of_vadd_madd_smem; crush.
  destruct H8 as [pg ?].

  destruct (add_cache (smem_c (ctxt_switch (st_memory s))) (stealth_va,x) pg) eqn:?; crush.
  destruct o. destruct p. destruct c0.
  symmetry in Heqp.
  STAC.simpl_some_cache_entry.    
  
  destruct (madd_eq m ma'); crush.
  unfold madd_type; crush.
  rewrite get_cache_entry1 with (va':=v) (ma:=x) (ma':=ma') (pg:=pg) (pg':=x1);
    crush.
  rewrite ctxt_switch_cache in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v ma' x1; unfold madd_type in *; crush.
  unfold is_rw_page in *; crush.
  rewrite H9, H14; crush.
  unfold madd_type; crush.
  rewrite get_cache_entry2 with (va':=v) (ma:=x) (ma':=m) (pg:=pg) (pg':=x1);
    crush.
  rewrite ctxt_switch_mem; crush.

  unfold madd_type; crush.
  rewrite get_mem_unchanged with (ma:=x) (pg:=pg); crush. 
  rewrite ctxt_switch_mem; crush.
  rewrite Heqp; crush.

  unfold madd_type; crush.
  rewrite get_invalid; crush.
  rewrite ctxt_switch_mem; crush.
  
  unfold page_of_vadd_smem; crush.
  destruct (madd_type (smem_mm (st_memory s)) pt_addr) eqn:?.
  destruct p; crush.
  rewrite madd_of_vadd_smem_ctxt_switch; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in *; rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite Heqo; crush.
Qed.

Lemma madd_type_ctxt_switch_new_smem:
  forall s osi pt_padd pt_addr osi' pa' ma',
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    osi_pa_to_ma s osi' pa' = Some ma' ->
    madd_type (smem_mm (ctxt_switch_new_smem s pt_addr)) ma' =
    madd_type (smem_mm (st_memory s)) ma'.
Proof.
  Local Hint Resolve madd_type_get_ctxt_switch
       madd_type_get_flush_ma.
  intros.
  generalize H.
  unfold valid_sm_state; crush.
  unfold ctxt_switch_new_smem; crush.
  unfold remove_osi_stealth_page; crush.
  unfold valid_sm_state; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H12; crush.
Qed.

Lemma ctxt_switch_no_va_mapped_same_mem :
  forall s osi' pt_padd' pt_addr' osi pa ma,
    valid_sm_state s ->
    osi_pa_to_ma s osi' pt_padd' = Some pt_addr' ->
    osi_pa_to_ma s osi pa = Some ma ->
    ACT.no_va_mapped_to_ma osi ma s ->
    get_mem (smem_mm (st_memory s)) ma =
    get_mem (smem_mm (ctxt_switch_new_smem s pt_addr')) ma.
Proof.
  unfold osi_pa_to_ma, valid_sm_state; crush.
  unfold ctxt_switch_new_smem; crush.

  assert (forall va, get_cache (smem_c (st_memory s)) (va,ma) = None).
  intros.
  destruct (get_cache (smem_c (st_memory s)) (va,ma)) eqn:?; crush.
  STAC.destruct_valid_state; crush.
  destruct H_valid_cache_va_ma with va ma p; crush.
  destruct H2 with x0 x1 va x2; crush.
  unfold osi_pa_to_pg in *; crush.  
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO osi); crush.
  assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO x); crush.
  destruct H_valid_virtmap with x i2 x1 x2 va ma; crush.
  assert (madd_owner (smem_mm (st_memory s)) i2 = Os_PO x); crush.
  unfold madd_owner in *; crush.
  unfold madd_owner in *; crush.
  destruct (ctxt_vadd_accessible ctxt va) eqn:?; unfold os_accessible in *; auto.
  rewrite H15 in *; crush.
  rewrite H12 in H13; injection H13; crush.

  transitivity (get_mem (smem_mm (remove_osi_stealth_page
    (ctxt_switch (st_memory s)) m)) ma); crush.
  unfold remove_osi_stealth_page; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rewrite H13; crush.
  rewrite flush_ma_mem; crush.
  rewrite ctxt_switch_mem, ctxt_switch_cache; crush.
  unfold update_cache_entry; crush.
  rewrite get_add_mem_distinct; crush.
  edestruct get_remove_ma_cache_some_entry; crush.
  rewrite ctxt_switch_mem; crush.
  rewrite ctxt_switch_mem; crush.
  destruct (page_of_vadd_smem (remove_osi_stealth_page
    (ctxt_switch (st_memory s)) m) pt_addr' stealth_va) eqn:?; crush.
  unfold page_of_vadd_smem in *; crush.
  destruct (add_cache (smem_c (remove_osi_stealth_page
    (ctxt_switch (st_memory s)) m)) (stealth_va,m0) p) eqn:?; crush.
  destruct o.
  destruct p0; destruct c0.
  symmetry in Heqp0.
  STAC.simpl_some_cache_entry.
  rewrite get_cache_entry2 with (va':=v) (ma':=m1) (pg':=x0)
    (ma:=m0) (pg:=p); crush.
  unfold remove_osi_stealth_page in H11; crush.
  rewrite ctxt_switch_mem in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite H15 in H11; crush.
  rewrite flush_ma_cache in H11; crush.
  destruct (madd_eq ma i1); crush.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in H11; crush.
  rewrite ctxt_switch_cache in *; crush.
  rewrite ctxt_switch_cache in *; crush.
  rewrite ctxt_switch_cache in *; crush.
  rewrite get_mem_unchanged with (ma:=m0) (pg:=p); crush.
  rewrite Heqp0; crush.
  rewrite get_invalid; crush.
Qed.  

End STEALTH_LEMMAS.
