(***********************************************************************
 * File: Stealth.v
 * $Revision: 12919 $
 * $Date: 2016-09-22 11:57:12 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Non-Leakage with StealthMem
 *************************************************************************)

Require Import Sets.Constructive_sets
  Stealth_actions Stealth_valid Stealth_lemmas.

Module OBS_LEMMAS.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE.

  (* Lemmas *)
  Lemma get_cache_then_pa : forall s osi va ma pg,
    valid_state s ->
    get_cache (smem_c (st_memory s)) (va,ma) = Some pg ->
    page_owned_by pg = Os_PO osi ->
    exists pa, osi_pa_to_ma s osi pa = Some ma.
  Proof.
    crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache_va_ma with va ma pg; crush.
    unfold osi_pa_to_pg in *; crush.
    rename x into osi', x0 into pt_padd,
      x1 into pt_page, x2 into pt, i into p2m, i0 into pt_addr.

    assert (madd_owner (smem_mm (st_memory s))
      pt_addr = Os_PO osi'); crush.
    unfold madd_owner in *; crush.
    rename p into pt_page.
    assert (os_accessible va); crush.
    destruct H_valid_virtmap with osi' pt_addr pt_page pt va ma;
      crush.
    destruct H_valid_cache with va ma pg; crush.
    unfold os_accessible.
    remember (ctxt_vadd_accessible ctxt va).
    destruct b; auto.
    rewrite H7 in H12; unfold os_accessible.
    rewrite H12 in H1; discriminate H1.
    intro; rewrite <- Heqb in H13; discriminate H13.
    assert (osi' = osi); crush.
    destruct H_valid_virtmap with osi' pt_addr pt_page pt va ma;
      crush.
    destruct H_valid_cache with va ma pg; crush.

    destruct H_va_has_pa with osi p2m pt_addr pt_page pt va ma;
      crush.
    unfold osi_pa_to_ma; crush.
    rewrite <- Heqopt_res; crush.
  Qed.

  Lemma same_osi_pa_to_ma : forall s s' a osi pa va ma pg,
    valid_sm_state s ->
    valid_sm_exec s a s' ->
    osi_pa_to_ma s osi pa = Some ma ->
    get_cache (smem_c (st_memory s)) (va,ma) = Some pg ->
    osi_pa_to_ma s' osi pa = Some ma.
  Proof.    
    crush.
    unfold osi_pa_to_ma in *; crush.
    sm_action_induction a; destruct_sm_and_normal_action_sem; crush.

    (* Page pin *)
    unfold osi_pa_to_ma in *; crush.
    destruct (os_ident_eq osi o); crush.
    rewrite <- HYP.get_set; crush.
    destruct (padd_eq pa p); crush.
    rewrite P2M.get_set_distinct; crush.
    rewrite HYP.get_set_distinct; crush.

    (* Page unpin *)
    unfold osi_pa_to_ma in *; crush.
    destruct (os_ident_eq osi o); crush.
    rewrite <- HYP.get_set; crush.
    destruct (padd_eq pa p); crush.
    unfold valid_sm_state in *; crush;
      STAC.destruct_valid_state.
    destruct H_valid_cache_va_ma with va x3 pg; crush.
    unfold ACT.no_va_mapped_to_ma in *; crush.
    unfold osi_pa_to_pg in *; crush.
    destruct H12 with x1 x4 va x5; crush.
    assert (madd_owner (smem_mm (st_memory s)) x3 = Os_PO o).
    apply H_valid_hyper with x p; crush.
    unfold madd_owner in *; crush.
    destruct H_valid_virtmap with x0 i0 x4 x5 va x3; crush.
    rewrite <- H_valid_hyper with x0 i x1 i0; crush.
    unfold madd_owner; crush.
    rewrite H21 in H20; crush.
    injection H20; crush.
    unfold os_accessible in *.
    destruct (ctxt_vadd_accessible ctxt va) eqn:?.
    rewrite H21 in H20; auto.
    rewrite H22 in H20; auto.
    discriminate.

    rewrite P2M.get_remove_distinct with (inf:=ma); crush.
    rewrite HYP.get_set_distinct; crush.

    (* Switch *)
    unfold valid_sm_state in *; crush.
    STAC.destruct_valid_state.
    destruct H_valid_os_idents with o; crush.
    destruct H_valid_currpg with o x x0; crush.
    rewrite Hpost with x1; crush.
    unfold osi_current_page_table; crush.

    (* Lswitch untrusted *)
    unfold ACT.pre_lswitch in *; crush.
    rewrite Hpost with x; crush.

    unfold ACT.pre_lswitch in *; crush.
    rewrite Hpost with x; crush.
  Qed.

Local Hint Resolve STAC.madd_of_vadd_smem_then_mmem.

  Lemma deleted_stealth_cache_action : 
    forall s s' a ma pg,
      valid_sm_state s ->
      valid_sm_exec s a s' ->
      get_cache (smem_c (st_memory s)) (stealth_va, ma) = Some pg ->
      get_cache (smem_c (st_memory s')) (stealth_va, ma) = None ->
      (a = Action (Del_untrusted (st_active_os s) stealth_va) \/
       exists o, a = Switch_sm o \/
       exists o, exists pa, a = LSwitch_untrusted_sm o pa).
  Proof.
    crush.
    sm_action_induction a; destruct_sm_and_normal_action_sem; crush.

    (* Read *)
    unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
    unfold is_rw_page in *; crush.
    rewrite get_new_cache with (ma:=m) (pg:=x) (val:=x0) in H2; crush.
    2: rewrite get_page_result with (pg:=x); unfold page_of_vadd_smem; crush.
    destruct (vadd_eq v stealth_va); crush.
    unfold valid_sm_state in *; crush.
    unfold page_of_vadd_madd_smem in *; crush.
    assert (m=ma); crush.
    rewrite <- get_add_cache_cacheable in H2; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va ma pg; crush.
    destruct H7 with pt_addr0 m x; crush.

    generalize (get_add_cache_distinct (smem_c (st_memory s)) (v,m)
      (stealth_va,ma) x pg); crush.
    rewrite <- H7 in H2; crush.
    destruct o; crush.
    destruct p; destruct c0; crush.
    clear H7.
    
    generalize (STAC.cache_deleted_entry (smem_c (st_memory s))
      v stealth_va m ma x pg); intro.
    assert (snd (add_cache (smem_c (st_memory s)) (v,m) x) = Some (v0,m0,p)).
    crush.
    rewrite H7 in H8; crush.
    2:STAC.some_entry_add_cache.
    clear H7.
    replace c with (fst (add_cache (smem_c (st_memory s)) (v,m) x)) in H2.
    2:crush.

    assert (get_line v = get_line stealth_va); crush.
    apply cache_inertia with (smem_c (st_memory s))
      m m0 x p; crush.
    assert (v = stealth_va); crush.
    unfold valid_sm_state in *; crush.

    (* Read hyper *)
    unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
    unfold is_rw_page in *; crush.
    rewrite get_new_cache with (ma:=m) (pg:=x) (val:=x0) in H2; crush.
    2: rewrite get_page_result with (pg:=x); unfold page_of_vadd_smem; crush.
    destruct (vadd_eq v stealth_va); crush.
    unfold valid_sm_state in *; crush.
    unfold page_of_vadd_madd_smem in *; crush.
    assert (m=ma); crush.
    rewrite <- get_add_cache_cacheable in H2; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va ma pg; crush.
    destruct H6 with pt_addr0 m x; crush.

    generalize (get_add_cache_distinct (smem_c (st_memory s)) (v,m)
      (stealth_va,ma) x pg); crush.
    rewrite <- H6 in H2; crush.
    destruct o; crush.
    destruct p; destruct c0; crush.
    clear H6.
    
    generalize (STAC.cache_deleted_entry (smem_c (st_memory s))
      v stealth_va m ma x pg); intro.
    assert (snd (add_cache (smem_c (st_memory s)) (v,m) x) = Some (v0,m0,p)).
    crush.
    rewrite H6 in H7; crush.
    2:STAC.some_entry_add_cache.
    clear H6.
    replace c with (fst (add_cache (smem_c (st_memory s)) (v,m) x)) in H2.
    2:crush.

    assert (get_line v = get_line stealth_va); crush.
    apply cache_inertia with (smem_c (st_memory s))
      m m0 x p; crush.
    assert (v = stealth_va); crush.
    unfold valid_sm_state in *; crush.

    (* Write *)
    unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
    unfold is_rw_page in *; crush.
    rewrite set_new_cache with (ma:=m) (pg:=x) (old_val:=x0) in H2; crush.
    2: unfold page_of_vadd_smem; crush.
    destruct (vadd_eq v stealth_va); crush.
    unfold valid_sm_state in *; crush.
    unfold page_of_vadd_madd_smem in *; crush.
    assert (m=ma); crush.
    rewrite <- get_add_cache_cacheable in H2; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va ma pg; crush.
    destruct H7 with pt_addr0 m x; crush.

    generalize (get_add_cache_distinct (smem_c (st_memory s)) (v,m)
      (stealth_va,ma)
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) pg); crush.
    rewrite <- H7 in H2; crush.
    destruct o; crush.
    destruct p; destruct c0; crush.
    clear H7.
    
    generalize (STAC.cache_deleted_entry (smem_c (st_memory s))
      v stealth_va m ma
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) pg); intro.
    assert (snd (add_cache (smem_c (st_memory s)) (v,m)
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x))) =
    Some (v1,m0,p)).
    crush.
    rewrite H7 in H8; crush.
    2:STAC.some_entry_add_cache.
    clear H7.
    replace c with (fst (add_cache (smem_c (st_memory s)) (v,m)
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)))) in H2.
    2:crush.

    assert (get_line v = get_line stealth_va); crush.
    apply cache_inertia with (smem_c (st_memory s))
      m m0 (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p; crush.
    assert (v = stealth_va); crush.
    unfold valid_sm_state in *; crush.
    unfold no_stealth_page_line in *.

    (* Write hyper *)
    unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
    unfold is_rw_page in *; crush.
    rewrite set_new_cache with (ma:=m) (pg:=x) (old_val:=x0) in H2; crush.
    2: unfold page_of_vadd_smem; crush.
    destruct (vadd_eq v stealth_va); crush.
    unfold valid_sm_state in *; crush.
    unfold page_of_vadd_madd_smem in *; crush.
    assert (m=ma); crush.
    rewrite <- get_add_cache_cacheable in H2; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va ma pg; crush.
    destruct H7 with pt_addr0 m x; crush.

    generalize (get_add_cache_distinct (smem_c (st_memory s)) (v,m)
      (stealth_va,ma) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x))
      pg); crush.
    rewrite <- H7 in H2; crush.
    destruct o; crush.
    destruct p; destruct c0; crush.
    clear H7.
    
    generalize (STAC.cache_deleted_entry (smem_c (st_memory s))
      v stealth_va m ma
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) pg); intro.
    assert (snd (add_cache (smem_c (st_memory s)) (v,m)
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x))) =
      Some (v1,m0,p)).
    crush.
    rewrite H7 in H8; crush.
    2:STAC.some_entry_add_cache.
    clear H7.
    replace c with (fst (add_cache (smem_c (st_memory s)) (v,m)
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)))) in H2.
    2:crush.

    assert (get_line v = get_line stealth_va); crush.
    apply cache_inertia with (smem_c (st_memory s))
      m m0 (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p; crush.
    assert (v = stealth_va); crush.
    unfold valid_sm_state in *; crush.
    unfold no_stealth_page_line in *.

    (* Del untrusted *)
    rewrite delete_cache with (ma:=x2) in H2; crush.
    destruct (eq_cache_key (stealth_va,ma) (v,x2)); crush.
    2:rewrite get_remove_cache_distinct with (pg:=pg) in H2; crush.
    assert (o = st_active_os s); crush.
    unfold valid_sm_state in *; crush.
    assert (page_owned_by pg = Os_PO o); crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va x2 pg; crush.
    unfold osi_current_page_table in *; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.
    remember (page_content p0).
    destruct c; crush.
    destruct H_valid_virtmap with o x1 p0 p1 stealth_va x2; crush.
    rewrite <- H_valid_hyper with o i (os_current_page x0) x1;
      unfold madd_owner; crush.

    STAC.destruct_valid_state.
    STAC.get_aos_data; STAC.get_aos_current_pt_page; crush.
    
    remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va).
    destruct o0; crush.
    destruct H_valid_virtmap with (st_active_os s) pt_addr pt_page pt
      stealth_va m; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.
    rewrite H19 in Heqopt_res0; crush.
    specialize (H21 stealth_va_accessible); clear H22.
    destruct H7 with pt_addr m x; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
    assert (m = x2); crush.
    unfold no_stealth_page in H10.
    rewrite H10 with (pt_addr:=pt_addr) in H1; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.

    (* New normal *)
    unfold ACT.new in *; crush.
    (* shared page *)
    rewrite new_cache_shared in H2; crush.
    unfold valid_sm_state in *; crush.
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.
    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s pt_addr ma pg; crush.
    unfold valid_sm_state; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
    destruct (madd_eq x2 ma); crush.
    unfold osi_pa_to_ma in *; crush.
    assert (madd_owner (smem_mm (st_memory s)) ma = Os_PO o).
    crush.
    unfold madd_owner in *; crush.
    destruct H_valid_cache with stealth_va ma pg; crush.
    rewrite <- H29 in H24.
    rewrite H23 in H24.
    injection H24; intros; subst o; clear H24.
    destruct H11 with (os_current_page aos) pt_addr; crush.
    unfold osi_pa_to_ma; crush.

    rewrite get_remove_ma_cache_ne in H2; crush.
    unfold remove_new_va_cache in H2; crush.
    destruct (vadd_eq v stealth_va); crush.
    rewrite get_remove_cache_distinct with (pg:=pg) in H2; crush.

    (* non shared page *)
    rewrite new_cache_not_shared in H2; crush.
    unfold remove_new_va_cache in H2; crush.
    rewrite get_remove_cache_distinct with (pg:=pg) in H2; crush.

    (* New stealth *)
    unfold valid_sm_state in *; crush.
    rewrite H15 with (pt_addr:=x2) in H1; crush.
    rewrite new_cache_not_shared in H2; crush.
    unfold remove_new_va_cache in H2; crush.
    rewrite new_cache_not_shared in H2; crush.
    unfold remove_new_va_cache in H2; crush.

(* Lswitch untrusted *)
    unfold ACT.pre_lswitch in *; crush.
    specialize (Hpost x H0); crush.
    right.
    exists (st_active_os s).
    right; crush.
    unfold ACT.pre_lswitch in *; crush.
    specialize (Hpost x H0); crush.
  Qed.

  Lemma same_state_same_obs : forall c osi o,
    get_cache_observation osi c c o = o.
  Proof.    
    crush.
    destruct get_cache_observation_spec with osi c c o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    destruct H with x; crush.
  Qed.

  Lemma read_stealth_same_obs :
    forall s s' o osi,
      valid_sm_state s ->
      ACT.pre_all_read stealth_va s ->
      ACT.post_all_read stealth_va s s' ->
      get_cache_observation osi (smem_c (st_memory s))
        (smem_c (st_memory s')) o = o.
  Proof.
    unfold valid_sm_state; crush; clear H7.

    unfold ACT.pre_all_read in H0; crush.
    unfold ACT.post_all_read in H1; crush.
    STAC.unify_read_post; crush.


    generalize H8.
    unfold aos_page_of_vadd_smem, page_of_vadd_smem in H8; crush.
    assert (is_cacheable x).
    unfold page_of_vadd_madd_smem in H8; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va m x; crush.
    destruct H2 with pt_addr0 m x; crush.

    unfold is_rw_page in H7; crush.
    rewrite get_new_cache with (ma:=m) (pg:=x) (val:=x0); crush.
    2: apply get_page_result; crush;
         unfold aos_page_of_vadd_smem in H11; crush.

    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (fst (add_cache (smem_c (st_memory s)) (stealth_va, m) x)) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    rename x1 into va.
    destruct H10 with va; crush.
    rename x1 into ma.

    assert (get_line stealth_va = get_line va).
    apply cache_inertia with (smem_c (st_memory s)) m ma x x2.
    apply STAC.cache_deleted_entry; crush.
    assert (stealth_va = va); crush.
    rewrite H4 with va ma x2; crush.

    unfold page_of_vadd_madd_smem in H8; crush.
    assert (m = ma); crush.
    rewrite <- get_add_cache_cacheable in H16; crush.

    destruct H2 with pt_addr0 m x; crush.
  Qed.

  Lemma write_stealth_same_obs :
    forall s s' o osi val,
      valid_sm_state s ->
      ACT.pre_all_write stealth_va val s ->
      ACT.post_all_write stealth_va val s s' ->
      get_cache_observation osi (smem_c (st_memory s))
        (smem_c (st_memory s')) o = o.
  Proof.
    unfold valid_sm_state; crush; clear H7.

    unfold ACT.pre_all_write in H0; crush.
    unfold ACT.post_all_write in H1; crush.
    STAC.unify_write_post; crush.

    generalize H8.
    unfold aos_page_of_vadd_smem, page_of_vadd_smem in H8; crush.
    assert (is_cacheable x).
    unfold page_of_vadd_madd_smem in H8; crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va m x; crush.
    destruct H2 with pt_addr0 m x; crush.

    unfold is_rw_page in H7; crush.
    rewrite set_new_cache with (ma:=m) (pg:=x) (old_val:=x0); crush.

    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (fst (add_cache (smem_c (st_memory s)) (stealth_va, m)
        (Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x)))) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    rename x1 into va.
    destruct H10 with va; crush.
    rename x1 into ma.

    assert (get_line stealth_va = get_line va).
    apply cache_inertia with (smem_c (st_memory s)) m ma
      (Page (RW_c (Some val)) (page_owned_by x) (page_cacheable x))
      x2.
    apply STAC.cache_deleted_entry; crush.
    assert (stealth_va = va); crush.
    rewrite H4 with va ma x2; crush.

    unfold page_of_vadd_madd_smem in H8; crush.
    assert (m = ma); crush.
    rewrite <- get_add_cache_cacheable in H16; crush.
    destruct H2 with pt_addr0 m x; crush.
  Qed.

  Lemma new_normal_same_obs : forall osi osi' o s s' va pa,
    osi <> osi' ->
    valid_sm_state s ->
    valid_sm_exec s (New_normal osi' va pa) s' ->
    get_cache_observation osi (smem_c (st_memory s))
      (smem_c (st_memory s')) o = o.
  Proof.
    unfold valid_sm_state; crush; clear H7.
    unfold valid_sm_exec in H1; crush.
    unfold ACT.post_new, ACT.post_all_new in H8; crush.

    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (smem_c (ACT.new s osi x0 va x)) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    
    destruct (H11 x2 H14); crush.
    clear H11 H12 H13 H14 Heqval0 Heqval.
    unfold pre_new_normal, pre_all_new in H7; crush.
    rename x3 into ma, x2 into va', x4 into pg; crush.
    destruct (madd_eq x5 ma); crush.
    assert (page_owned_by pg = Os_PO osi'); crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with va' ma pg; crush.
    unfold osi_pa_to_ma in H11; crush.
    rewrite <- H_valid_hyper with osi' i pa ma; crush.
    unfold madd_owner; crush.

    (* ma <> m *)
    rewrite new_cache_shared in H17; crush.
    rewrite get_remove_ma_cache_ne in H17; crush.
    unfold remove_new_va_cache in H17; crush.
    remember (eq_cache_key (va,m) (va',ma)).
    destruct s2; crush; clear Heqs2.
    injection e; crush; clear e.
    2:rewrite get_remove_cache_distinct with (pg:=pg) in H17; crush.
    clear s1 H17.
    assert (madd_owner (smem_mm (st_memory s)) x0 = Os_PO osi').
    STAC.destruct_valid_state.
    unfold madd_owner; crush.
    unfold valid_current_page in H_valid_currpg.
    unfold osi_current_page_table in *; crush.
    destruct H_valid_currpg with osi' x1 i; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.    
    unfold madd_owner in H17; crush.
    rename p into pt_page.
    STAC.destruct_valid_state.
    unfold madd_of_vadd, madd_page_table in *; crush.
    remember (page_content pt_page).
    destruct c; crush.
    destruct H_valid_virtmap with osi' x0 pt_page p0 va' ma; crush.
    unfold osi_current_page_table in *; crush.
    destruct H_valid_cache with va' ma pg; crush.

    rewrite new_cache_not_shared; crush.
    unfold remove_new_va_cache; crush;
      try apply same_state_same_obs.
    
    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (remove_cache (smem_c (st_memory s)) (va,m)) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    destruct (H11 x2 H14); crush.
    clear H11 H12 H13 H14 Heqval.
    remember (eq_cache_key (va,m) (x2,x3)).
    destruct s0; crush; clear Heqs0.
    injection e; crush; clear e.
    2:rewrite get_remove_cache_distinct with (pg:=x4) in H17; crush.
    clear H17.
    assert (madd_owner (smem_mm (st_memory s)) x0 = Os_PO osi').
    STAC.destruct_valid_state.
    unfold madd_owner; crush.
    unfold valid_current_page in H_valid_currpg.
    unfold osi_current_page_table in *; crush.
    destruct H_valid_currpg with osi' x1 i; crush.
    unfold madd_of_vadd, madd_page_table in *; crush.    
    unfold madd_owner in H11; crush.
    rename p into pt_page.
    STAC.destruct_valid_state.
    unfold madd_of_vadd, madd_page_table in *; crush.
    remember (page_content pt_page).
    destruct c; crush.
    destruct H_valid_virtmap with osi' x0 pt_page p0 x2 x3; crush.
    unfold osi_current_page_table in *; crush.
    destruct H_valid_cache with x2 x3 x4; crush.
    elim H; crush.
    rewrite H16 in H21.
    unfold os_accessible in *; crush.
    destruct (ctxt_vadd_accessible ctxt x2); crush.
    rewrite H13 in H21; crush.
    rewrite H14 in H21; crush.
  Qed.

  Lemma new_stealth_same_obs : forall osi osi' o s s' pa,
    osi <> osi' ->
    valid_sm_state s ->
    valid_sm_exec s (New_sm osi' pa) s' ->
    get_cache_observation osi (smem_c (st_memory s))
      (smem_c (st_memory s')) o = o.
  Proof.
    intros; unfold valid_sm_state, valid_sm_exec in *; crush;
      unfold ACT.post_all_new, post_new_sm in H3; crush.

    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (fst (add_cache (smem_c (st_memory s)) (stealth_va,x) p)) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    destruct H12 with x2; crush.
    destruct (eq_cache_key (stealth_va, x) (x2,x3)); crush.
    remember (page_cacheable p).
    destruct b; crush.
    rewrite <- get_add_cache_cacheable in H18; crush.
    rewrite <- get_add_cache_non_cacheable in H18; crush.
    
    generalize (get_add_cache_distinct (smem_c (st_memory s)) 
      (stealth_va,x) (x2,x3) p x4); intros.
    remember (add_cache (smem_c (st_memory s)) (stealth_va,x) p).
    destruct p0; simpl in *.
    rewrite <- H19 in H18; auto.
    inversion H18.
    clear H19.
    destruct o0; [destruct p0|]; simpl in *.
    right; crush.
    exists c0; exists p0; crush.

    assert (C.get_line C.stealth_va = C.get_line x2).
    apply C.cache_inertia with (smem_c (st_memory s))
      x x3 p p0; crush.
    assert (x2 = C.stealth_va).
    apply H6 with x3 x4; crush.
    crush.
    elim H.
    STAC.destruct_valid_state.
    destruct H_valid_cache with stealth_va x3 x4; crush.


    destruct STEALTH_LEMMAS.stealth_cached_is_aos with s x0 x3 x4; crush.
    unfold valid_sm_state; crush.
    left; crush.

    unfold pre_new_sm in *; crush.
    unfold page_of_vadd_madd_smem in Heqval0; crush.

    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (smem_c (new_not_shared (st_memory s) x0 stealth_va x)) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    destruct H12 with x2; crush.

    rewrite new_cache_not_shared in H18; crush.
    unfold remove_new_va_cache in H18; crush.
    rewrite get_remove_cache_distinct with (pg:=x4) in H18; crush.
    unfold pre_new_sm in *; crush.
  Qed.

  Lemma del_same_obs : forall osi osi' o s s' va,
    osi <> osi' ->
    valid_sm_state s ->
    valid_sm_exec s (Action (Del_untrusted osi' va)) s' ->
    get_cache_observation osi (smem_c (st_memory s))
      (smem_c (st_memory s')) o = o.
  Proof.
    unfold valid_sm_state in *; crush; clear H7.

    unfold valid_sm_exec in *; crush;
      unfold ACT.post_del, ACT.pre_del, ACT.pre_all_del in *; crush.

    unfold valid_sm_exec in *; crush;
      unfold ACT.post_del, ACT.pre_del, ACT.pre_all_del in *; crush.
    rename x2 into ma.
    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (smem_c (delete (st_memory s) x1 va)) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    destruct H8 with x; crush.
    rename x into v, x2 into m, x3 into pg. 

    destruct (madd_eq ma m); crush.
    assert (page_owned_by pg = Os_PO osi'); crush.
    STAC.destruct_valid_state.
    destruct H_valid_os_idents with osi'; crush.
    STAC.get_osi_current_pt_page osi'.
    unfold osi_current_page_table in *; crush.
    destruct H_valid_virtmap with osi' x1 pt_page pt va m; crush.
    unfold madd_of_vadd in *; crush.
    unfold madd_page_table in Heqopt_res; crush.
    rewrite H25 in Heqopt_res; crush.
    destruct H_valid_cache with v m pg; crush.

    (* ma <> m *)
    rewrite delete_cache with (ma:=ma) in H20; crush.
    rewrite get_remove_cache_distinct with (pg:=pg) in H20; crush.
  Qed.

  Ltac switch_crush :=
    crush; repeat (match goal with
      | [ H: context[remove_osi_stealth_page _ _]|-_] =>
        unfold remove_osi_stealth_page in H
      | [ |- context[remove_osi_stealth_page _ _]] =>
        unfold remove_osi_stealth_page
      | [ H: context[update_cache_entry _ _ _]|-_] =>
        unfold update_cache_entry in H
      | [ |- context[update_cache_entry _ _]] =>
        unfold update_cache_entry
          
      | [ H: context[smem_mm (ctxt_switch _)]|-_] =>
        rewrite ctxt_switch_mem in H
      | [ H: context[smem_c (ctxt_switch _)]|-_] =>
        rewrite ctxt_switch_cache in H
      | [ H: context[smem_t (ctxt_switch _)] |-_] =>
        rewrite ctxt_switch_tlb in H
      | [ H: context[get_tlb empty_tlb _] |-_] =>
        rewrite get_empty_tlb in H
      | [ |- context[smem_mm (ctxt_switch _)]] =>
        rewrite ctxt_switch_mem
      | [ |- context[smem_c (ctxt_switch _)]] =>
        rewrite ctxt_switch_cache
      | [ |- context[smem_t (ctxt_switch _)]] =>
        rewrite ctxt_switch_tlb

      | [ H: context[smem_mm (flush_ma _ _)]|-_] =>
        rewrite flush_ma_mem in H
      | [ H: context[smem_c (flush_ma _ _)]|-_] =>
        rewrite flush_ma_cache in H
      | [ H: context[smem_t (flush_ma _ _)] |-_] =>
        rewrite flush_ma_tlb in H
      | [ |- context[smem_mm (flush_ma _ _)]] =>
        rewrite flush_ma_mem
      | [ |- context[smem_c (flush_ma _ _)]] =>
        rewrite flush_ma_cache
      | [ |- context[smem_t (flush_ma _ _)] ] =>
        rewrite flush_ma_tlb

      | [ H: context[get_mem (add_mem _ ?ma _) ?ma] |-_]=>
        rewrite <- get_add_mem in H
      | [ |- context[get_mem (add_mem _ ?ma _) ?ma]]=>
        rewrite <- get_add_mem

      | [ H1: context[get_mem (add_mem _ ?ma _) ?ma'] |-_]=>
        match goal with
          | [ H2: ?ma <> ?ma'|-_]=>
            rewrite get_add_mem_distinct in H1
          | [ H2: ?ma' <> ?ma|-_]=>
            rewrite get_add_mem_distinct in H1
        end
      | [ |- context[get_mem (add_mem _ ?ma _) ?ma']]=>
        match goal with
          | [ H: ?ma <> ?ma'|-_]=>
            rewrite get_add_mem_distinct
          | [ H: ?ma <> ?ma'|-_]=>
            rewrite get_add_mem_distinct
        end

      | [ H: context[get_cache (fst (remove_ma_cache _ ?ma)) (_,?ma)] |- _] =>
        rewrite get_remove_ma_cache_eq in H
      | [ H1: context[get_cache (fst (remove_ma_cache _ ?ma)) (_,?ma')] |- _ ] =>
        match goal with
          | [ H2: ?ma <> ?ma' |- _] =>
            rewrite get_remove_ma_cache_ne in H1
          | [ H2: ?ma' <> ?ma |- _] =>
            rewrite get_remove_ma_cache_ne in H1
        end

      | [ H: context[match page_content ?pt_page with
                      | RW_c _ => _ | PT_c _ => _ | Other_c => _
                     end] |- _ ] =>
        let content := fresh "content" in
        remember (page_content pt_page) as content; destruct content
      | [ H1: context[match page_content ?pt_page with
                       | RW_c _ => _ | PT_c _ => _ | Other_c => _
                      end],
          H2: page_content ?pt_page = PT_c _ |- _ ] =>
        rewrite H2 in H1
      | [ |- context[match page_content ?pt_page with
                      | RW_c _ => _ | PT_c _ => _ | Other_c => _
                     end]] =>
        let content := fresh "content" in
        remember (page_content pt_page) as content; destruct content
      | [ H: page_content ?pt_page = PT_c _ |-
          context[match page_content ?pt_page with
                    | RW_c _ => _ | PT_c _ => _ | Other_c => _
                  end]] =>
        rewrite H

      | [ H: Some (?va, ?ma, ?pg) =
               snd (remove_ma_cache ?c ?ma') |-_] =>
        notHyp (get_cache c (va, ma) = Some pg);
          destruct get_remove_ma_cache_some_entry with c va ma' ma pg
      | [ H1: None = snd (remove_ma_cache ?c ?ma),
          H2: Some ?pg = get_cache ?c (_,?ma) |-_]=>
      rewrite get_remove_ma_cache_no_entry in H2
    end; crush).

  Lemma page_of_vadd_madd_smem_flush_ctxt_switch :
    forall s pt_addr pt_addr' va ma,
      valid_state s ->
      aos_current_page_table s = Some pt_addr ->
      (exists osi', exists pt_padd',
        osi_pa_to_ma s osi' pt_padd' = Some pt_addr') ->
      madd_of_vadd (smem_mm (st_memory s)) pt_addr' va = Some ma ->
      page_of_vadd_madd_smem (st_memory s) va ma =
      page_of_vadd_madd_smem (remove_osi_stealth_page
        (ctxt_switch (st_memory s)) pt_addr) va ma.
  Proof.
    intros.
    STAC.destruct_valid_state;
      STAC.get_aos_data;
      STAC.get_aos_current_pt_page.
    unfold aos_current_page_table, osi_current_page_table in *;
      crush.
    unfold page_of_vadd_madd_smem; switch_crush.

    destruct (madd_eq ma i); switch_crush.
    destruct (madd_eq ma m); switch_crush.
    destruct (vadd_eq va v); switch_crush.
    STAC.synonym_cached.
    destruct (madd_eq ma i); switch_crush.
    destruct (madd_eq ma i); switch_crush.
    destruct (madd_eq ma m); switch_crush.
    destruct (vadd_eq va v); switch_crush.
    STAC.synonym_cached.
  Qed.

  Lemma madd_of_vadd_flush_ctxt_switch_not_new : forall s osi pt_padd
    pt_addr pt_addr',
    valid_state s ->
    aos_current_page_table s = Some pt_addr ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr' ->
    madd_type (smem_mm (st_memory s)) pt_addr' = Some PT ->
    madd_of_vadd (smem_mm (st_memory s)) pt_addr' stealth_va = None ->
    madd_of_vadd (smem_mm (remove_osi_stealth_page (ctxt_switch (st_memory s))
      pt_addr)) pt_addr' stealth_va = None.
  Proof.
    intros.
    unfold madd_of_vadd, madd_type, madd_page_table in *; switch_crush.
    rewrite get_add_mem_distinct; switch_crush.
    STAC.destruct_valid_state.
    destruct H_valid_cache with v pt_addr' p4; crush.
    unfold is_rw_page in *; crush.
  Qed.

  Lemma ctxt_switch_same_obs :
    forall s s' o osi osi' pt_padd pt_addr pt_page pt,
      osi <> (st_active_os s) ->
      osi <> osi' ->
      valid_sm_state s ->
      osi_pa_to_ma s osi' pt_padd = Some pt_addr ->
      get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
      PT_c pt = page_content pt_page ->
      st_memory s' = ctxt_switch_new_smem s pt_addr ->
      get_cache_observation osi (smem_c (st_memory s))
        (smem_c (st_memory s')) o = o.
  Proof.
    unfold valid_sm_state; crush; clear H11.

    STAC.destruct_valid_state.

    rewrite H5; clear H5 s'.
    destruct get_cache_observation_spec with osi (smem_c (st_memory s))
      (smem_c (ctxt_switch_new_smem s pt_addr)) o; crush.
    apply Extensionality_Ensembles.
    split; unfold Included; crush.
    destruct H5 with x; crush.
    rename x into va, x0 into ma, x1 into pg.
    clear H5 H11 H12 H13.

    unfold ctxt_switch_new_smem in H16; crush.
    rename m into aos_pt_addr.
    assert (get_cache (smem_c
        (remove_osi_stealth_page (ctxt_switch (st_memory s)) aos_pt_addr))
        (va, ma) = Some pg).
    switch_crush.
    unfold aos_current_page_table, osi_current_page_table in Heqval; crush.
    destruct (madd_eq ma i); crush.
    destruct H_valid_virtmap with (st_active_os s) aos_pt_addr p0 p
      stealth_va i; crush.
    destruct H_valid_currpg with (st_active_os s) i0 i1; crush.
    specialize (H11 stealth_va_accessible); crush.
    destruct H_valid_cache with va i pg; crush.
    rewrite get_remove_ma_cache_ne; crush.

    remember (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va).
    destruct o0; crush.
    (* stealth_va is used by osi'. *)
    unfold madd_of_vadd, madd_page_table in Heqo0; crush.
    rewrite <- H4 in Heqopt_res; crush.
    unfold osi_pa_to_ma in *; crush.
    destruct H_valid_virtmap with osi' pt_addr pt_page pt stealth_va m; crush.

    rewrite <- H_valid_hyper with (pa:=pt_padd) (p2m:=i) (ma:=pt_addr); 
      unfold madd_owner; crush. 
    specialize (H12 stealth_va_accessible); clear H13; crush.
    destruct (madd_eq m ma); crush.
    destruct H_valid_cache with va ma pg; crush.

    assert (is_rw_page x).
    remember (get_cache (smem_c (st_memory s)) (stealth_va, m)).
    destruct o0; crush.
    destruct H_valid_cache with stealth_va m p; crush.
    destruct H10 with pt_addr m x; crush.
    exists osi', pt_padd; unfold osi_pa_to_ma; crush.

    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H4; crush.
    unfold page_of_vadd_madd_smem; crush.
    unfold is_rw_page in *; crush.
    rename x into stealth_pg; rename x0 into stealth_val.
    
    assert (madd_of_vadd_smem (remove_osi_stealth_page
      (ctxt_switch (st_memory s)) aos_pt_addr) pt_addr stealth_va = Some m).
    assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = Some m);
      crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H4; crush.
    clear H5 H16.
    unfold madd_of_vadd_smem; switch_crush.
    unfold madd_of_vadd, madd_page_table in H17; crush.
    rewrite <- H4 in Heqopt_res0; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite get_add_mem_distinct; crush.
    rewrite <- H4; crush.
    destruct H_valid_cache with v pt_addr p2; crush.
    unfold is_rw_page in H19; crush.

    remember (get_cache (smem_c (st_memory s)) (stealth_va, m)).
    destruct o0; crush.
    rename p into cache_stealth_pg.
    destruct H_valid_cache with stealth_va m cache_stealth_pg; crush.
    unfold is_rw_page in H18; crush.
    rename x into cache_stealth_val.
    rewrite get_new_cache with (ma:=m) (pg:=cache_stealth_pg)
      (val:=cache_stealth_val) in H16; crush.
    Focus 2.
    rewrite get_page_result with (pg:=cache_stealth_pg); crush.
    unfold page_of_vadd_smem; crush.
    rewrite <- page_of_vadd_madd_smem_flush_ctxt_switch
      with (pt_addr':=pt_addr); crush.
    unfold page_of_vadd_madd_smem; crush.
    exists osi', pt_padd; unfold osi_pa_to_ma; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H4; crush.
    remember (remove_osi_stealth_page (ctxt_switch (st_memory s))
      aos_pt_addr) as tmp_smem.
    clear Heqtmp_smem.

    generalize (get_add_cache_distinct (smem_c tmp_smem) (stealth_va,m) (va,ma) cache_stealth_pg pg);
      crush.
    rewrite <- H23 in H16; crush.
    destruct o0; [right | left]; crush.
    destruct p; destruct c0.
    exists (v,m0); exists p; crush.
    assert (get_line stealth_va = get_line v); crush.
    apply cache_inertia with (smem_c tmp_smem) m m0 cache_stealth_pg p; crush.
    assert (v = stealth_va); crush.

    rewrite get_new_cache with (ma:=m) (pg:=stealth_pg)
      (val:=stealth_val) in H16; crush.
    Focus 2.
    rewrite get_page_result with (pg:=stealth_pg); crush.
    unfold page_of_vadd_smem; crush.
    rewrite <- page_of_vadd_madd_smem_flush_ctxt_switch
      with (pt_addr':=pt_addr); crush.
    unfold page_of_vadd_madd_smem; crush.
    exists osi', pt_padd; unfold osi_pa_to_ma; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite <- H4; crush.
    remember (remove_osi_stealth_page (ctxt_switch (st_memory s))
      aos_pt_addr) as tmp_smem.
    clear Heqtmp_smem.

    generalize (get_add_cache_distinct (smem_c tmp_smem) (stealth_va,m) (va,ma) stealth_pg pg);
      crush.
    rewrite <- H18 in H16; crush.
    destruct o0; [right | left]; crush.
    destruct p; destruct c0.
    exists (v,m0); exists p; crush.
    assert (get_line stealth_va = get_line v); crush.
    apply cache_inertia with (smem_c tmp_smem) m m0 stealth_pg p; crush.
    assert (v = stealth_va); crush.
    clear H18 H19.
    remember (madd_of_vadd (smem_mm (st_memory s)) aos_pt_addr stealth_va).
    destruct o0; crush.
    (* stealth_va was used by the AOS *)
    unfold madd_of_vadd, madd_page_table in *; crush.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page; crush.
    unfold aos_current_page_table, osi_current_page_table in *; crush.
    rewrite H21 in Heqopt_res0; crush.
    destruct H_valid_virtmap with (st_active_os s) aos_pt_addr p1 pt0 stealth_va m1; crush.
    specialize (H19 stealth_va_accessible); crush.
    destruct H6 with aos_pt_addr m1 x; crush.
    unfold aos_current_page_table, osi_current_page_table; crush.
    unfold madd_of_vadd, madd_page_table; crush.
    rewrite H21; crush.
    assert (m0 = m1); crush.
    (* stealth_va was not used by the AOS *)
    unfold no_stealth_page in H6.
    rewrite H7 with (pt_addr:=aos_pt_addr) in H14; crush.

    (* stealth_va is not used by osi' *)
    rewrite get_invalid in H16; crush.
    unfold page_of_vadd_smem; crush.
    unfold madd_of_vadd_smem; crush.
    rewrite H_valid_tlb with (ma:=m) in Heqo0; switch_crush.
    clear H16 Heqval0.
    assert (madd_of_vadd (smem_mm (remove_osi_stealth_page (ctxt_switch (st_memory s))
      aos_pt_addr)) pt_addr stealth_va = None); crush.
    2: rewrite H11; crush.
    switch_crush.
    unfold madd_of_vadd, madd_page_table in *; crush.
    destruct (madd_eq m pt_addr); crush.
    destruct H_valid_cache with v pt_addr p2; crush.
    unfold is_rw_page in *; crush.    
    rewrite get_add_mem_distinct; crush.
  Qed.

  Lemma switch_same_obs : forall osi osi' o s s',
    osi <> osi' ->
    osi <> (st_active_os s) ->
    valid_sm_state s ->
    valid_sm_exec s (Switch_sm osi') s' ->
    get_cache_observation osi (smem_c (st_memory s))
      (smem_c (st_memory s')) o = o.
  Proof.
    unfold valid_sm_state in *; crush.
    STAC.destruct_valid_state.
    destruct H_valid_os_idents with osi'; crush.
    STAC.get_osi_current_pt_page osi'.
    apply ctxt_switch_same_obs with osi' (os_current_page x) 
      pt_addr pt_page pt; crush.
    unfold valid_sm_state; crush.
    unfold osi_pa_to_ma; crush.
    unfold valid_sm_exec in *; crush;
      unfold post_sm_switch in *; crush.
    rewrite H16 with pt_addr; crush.
    unfold osi_current_page_table; crush.
  Qed.

  Lemma lswitch_same_obs : forall osi osi' o s s' pa,
    osi <> osi' ->
    valid_sm_state s ->
    valid_sm_exec s (LSwitch_untrusted_sm osi' pa) s' ->
    get_cache_observation osi (smem_c (st_memory s))
      (smem_c (st_memory s')) o = o.
  Proof.
    unfold valid_sm_state in *; crush.
    unfold valid_sm_exec in *; crush;
      unfold ACT.pre_lswitch, post_sm_lswitch in *; crush.

    unfold madd_type in *; crush.
    remember (page_content p).
    destruct c; crush.
    apply ctxt_switch_same_obs with (st_active_os s) pa x p p0;
      crush.
    unfold valid_sm_state; crush.
    rewrite H9 with x; crush.
    rewrite H9 with x; crush.
    apply same_state_same_obs.
  Qed.
    
End OBS_LEMMAS.
