(***********************************************************************
 * File: Stealth_secure_program.v
 * $Revision: 12471 $
 * $Date: 2014-04-23 13:48:43 -0300 (Wed, 23 Apr 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Definition of sufficient conditions for a program to be secure wrt
 * cache leakage.
 *************************************************************************)
Require Import Sets.Constructive_sets Stealth_actions
  Stealth_valid Stealth_traces Stealth_action_streams
  Stealth_traces_victim Stealth_obs_lemmas
  Stealth_interleaving Streams.

Module Type SECURE_P (S:SCHEDULER) (INT:INTERLEAVING S).
    Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS STEALTH_EQ
         TRACES_VIC OBS_LEMMAS INT.

  (* L is the type of the language *)
  Parameter L : Type.
  (* Lstate is the type of states of the language *)
  Parameter Lstate : Type.

  (* Given a program, and an initial state, return the action stream generated
     by the execution of that program. *)
  Parameter get_trace : L -> Lstate -> effects -> Prop.

  (* Actions generated by the program execution are osi_actions *)
  (* Axiom get_trace_os_actions : forall P s0 str, *)
  (*   get_trace P s0 str -> *)
  (*   osi_actions str osi_v. *)

  (* In program P, states s and s' differ only in values of high
     variables. *)
  Parameter low_var_eq : L -> Lstate -> Lstate -> Prop.

  (* A program is secure if starting with the same low variables, the
     sequence of non stealth actions are the same. *)
  Definition secure (P:L) :=
    forall s0 s0' str str',
      low_var_eq P s0 s0' ->
      get_trace P s0 str -> get_trace P s0' str' ->
      same_non_stealth_effects_str str str'.

End SECURE_P.


Module SECURE_NON_LEAKAGE (S:SCHEDULER) (INT:INTERLEAVING S) (SP:SECURE_P S INT).
    Module INT_PROOF := INTERLEAVING_PROOF S INT.
    Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE
         STEALTH_TRACES ACTION_STREAMS STEALTH_EQ
         TRACES_VIC OBS_LEMMAS INT SP INT_PROOF.

  Theorem secure_P_non_leakage : forall P s0 s0' Ls0 Ls0' trace trace' str str',
    valid_sm_state s0 ->
    valid_sm_state s0' ->
    (* osi_leak_actions str_attacker osi_a -> *)
    secure P ->
    low_var_eq P Ls0 Ls0' ->
    osi_stealth_equivalent s0 s0' osi_a ->
    get_trace P Ls0 trace -> get_trace P Ls0' trace' ->
    interleave s0 trace str ->
    interleave s0' trace' str' ->
    leak_osi_equivalent_traces osi_a str str'.
  Proof.
    Local Hint Resolve interleave_run.
    intros.
    apply Non_leakage; crush.
    eapply interleave_same_non_stealth_osi_a_actions
      with (s0:=s0) (s0':=s0'); crush.
    erewrite interleave_hd; crush.
    erewrite interleave_hd; crush.
  Qed.

End SECURE_NON_LEAKAGE.
  

