(***********************************************************************
 * File: Stealth_sm_iso_lemmas.v
 * $Revision: 12920 $
 * $Date: 2016-09-22 18:44:11 -0300 (Thu, 22 Sep 2016) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Proofs of same_non_stealth_entries invariance.
 *************************************************************************)


Require Import Stealth_actions Stealth_valid Stealth_equivalence
  Stealth_traces Stealth_isolation Stealth_non_stealth_equivalence
  Stealth_lemmas.

Module STEALTH_SM_ISO_LEMMAS.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE STEALTH_EQ
         STEALTH_ISOLATION STEALTH_NON_STEALTH_EQ STEALTH_LEMMAS.
  

  
Lemma read_stealth_same_cache_entires :
  forall s s',
    valid_sm_state s ->
    ACT.pre_all_read stealth_va s ->
    ACT.post_all_read stealth_va s s' ->
    
    same_non_stealth_entries s s'.
Proof.
  unfold valid_sm_state; crush.

  unfold same_non_stealth_entries; crush.
  unfold ACT.pre_all_read, ACT.post_all_read in *; crush.
  STAC.unify_read_post; crush.

  unfold page_of_vadd_smem in H0; crush.
  unfold is_rw_page in *; crush.
  rewrite get_new_cache with (ma:=m) (pg:=x) (val:=x0); crush.
  apply same_cache_line_add_cache_other_line; auto.

  rewrite get_page_result with (pg:=x); crush.
  unfold page_of_vadd_smem; crush.
Qed.

Lemma write_stealth_same_cache_entires :
  forall s s' val,
    valid_sm_state s ->
    ACT.pre_all_write stealth_va val s ->
    ACT.post_all_write stealth_va val s s' ->
    
    same_non_stealth_entries s s'.
Proof.
  unfold valid_sm_state; crush.

  unfold same_non_stealth_entries; crush.
  unfold ACT.pre_all_write, ACT.post_all_write in *; crush.
  STAC.unify_write_post; crush.

  unfold page_of_vadd_smem in H0; crush.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=m) (pg:=x) (old_val:=x0); crush.
  apply same_cache_line_add_cache_other_line; auto.
  
  unfold page_of_vadd_smem; crush.
Qed.

Lemma del_stealth_same_cache_entires :
  forall s s' o,
    valid_sm_state s ->
    ACT.pre_del o stealth_va s ->
    ACT.post_del o stealth_va s s' ->
    same_non_stealth_entries s s'.
Proof.
  unfold valid_sm_state; crush.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  unfold same_non_stealth_entries, osi_pa_to_ma; crush.
  rewrite delete_cache with (ma:=x2); crush.
  apply same_cache_line_remove_cache_other_line; crush.
Qed.

Lemma pin_stealth_same_cache_entries :
  forall s s' o pa ptype,
    valid_sm_state s ->
    ACT.pre_pin o pa ptype s ->
    ACT.post_pin o pa ptype s s' ->
    same_non_stealth_entries s s'.
Proof.
  unfold valid_sm_state; crush.
  unfold ACT.pre_pin, ACT.post_pin in *; crush.
  unfold same_non_stealth_entries; crush.
  unfold same_non_stealth_entries; crush.
Qed.

Lemma unpin_stealth_same_cache_entries :
  forall s s' o pa,
    valid_sm_state s ->
    ACT.pre_unpin o pa s ->
    ACT.post_unpin o pa s s' ->
    same_non_stealth_entries s s'.
Proof.
  unfold valid_sm_state; crush.
  unfold ACT.pre_unpin, ACT.post_unpin in *; crush.
  unfold same_non_stealth_entries, osi_pa_to_ma; crush.
Qed.

Lemma new_sm_stealth_same_cache_entries :
  forall s s' o pa,
    valid_sm_state s ->
    pre_new_sm o pa s ->
    post_new_sm o pa s s' ->
    same_non_stealth_entries s s'.
Proof.
  unfold valid_sm_state; crush.
  unfold pre_new_sm, pre_all_new, post_new_sm in *; crush.
  unfold same_non_stealth_entries; crush.
  apply same_cache_line_add_cache_other_line; crush.

  rewrite new_cache_not_shared; crush.
  unfold remove_new_va_cache; crush.
  rewrite new_cache_not_shared; crush.
  unfold remove_new_va_cache; crush.
Qed.

(* Lemmas for Switch *)
Lemma stealth_va_mem_rw : forall s pt_addr ma pg,
  valid_state s ->
  stealth_page_rw_cacheable s ->
  (exists osi, exists pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->
  madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = Some ma ->
  get_mem (smem_mm (st_memory s)) ma = Some pg ->
  is_rw_page pg.
Proof.
  intros.
  remember (get_cache (smem_c (st_memory s)) (stealth_va,ma)).
  destruct o.
  STAC.destruct_valid_state.
  destruct H_valid_cache with stealth_va ma p; crush.
  destruct H0 with pt_addr ma pg; crush.
  unfold page_of_vadd_madd_smem; crush.
Qed.

Lemma get_add_cache_non_stealth :
  forall s va ma stealth_ma stealth_pg,
    va <> stealth_va ->
    stealth_page_reserved_line s ->
    get_cache (fst (add_cache (smem_c (st_memory s))
                              (stealth_va, stealth_ma) stealth_pg)) (va,ma) =
    get_cache (smem_c (st_memory s)) (va,ma).
Proof.
  intros.
  remember (get_cache (smem_c (st_memory s)) (va,ma)).
  destruct o.
  rename p into pg.
  generalize (get_add_cache_distinct
                (smem_c (st_memory s))
                (stealth_va,stealth_ma) (va,ma) stealth_pg pg); crush.
  rewrite H1; crush.
  destruct o; crush.
  destruct p.
  right; exists c0; exists p; crush; clear H1.
  STAC.simpl_some_cache_entry.
  assert (get_line stealth_va = get_line va); crush.
  apply cache_inertia with (smem_c (st_memory s)) stealth_ma ma stealth_pg x0; crush.

  match goal with
    | [|- ?e = _] =>
      remember e
  end.
  destruct o; crush.
  generalize (get_add_cache_distinct_not_new (smem_c (st_memory s))
                                             (stealth_va,stealth_ma) (va,ma) stealth_pg p); crush.
  rewrite <- H1 in Heqo; crush.
Qed.

Lemma reserved_line_flush : forall s s' ma,
  valid_sm_state s ->
  st_memory s' = flush_ma (ctxt_switch (st_memory s)) ma ->
  stealth_page_reserved_line s'.
Proof.
  intros.
  unfold valid_sm_state in H; crush.
  unfold stealth_page_reserved_line in *; crush.
  rewrite H0 in *; clear dependent s'.
  apply H3 with ma0 pg; crush; clear H3.
  rewrite flush_ma_cache in H8; crush.
  remember (madd_eq ma ma0).
  destruct s0; crush; clear Heqs0.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  rewrite ctxt_switch_cache in *; crush.
Qed.

Lemma get_tlb_flush_ctxt_switch :
  forall s ma va,
    get_tlb (smem_t (flush_ma (ctxt_switch (st_memory s)) ma)) va = None.
Proof.
  intros.
  rewrite flush_ma_tlb.
  rewrite ctxt_switch_tlb.
  rewrite get_empty_tlb.
  reflexivity.
Qed.

Lemma get_cache_some_none_abs :
  forall c va ma,
    (exists pg, get_cache c (va, ma) = Some pg) ->
    get_cache c (va, ma) = None ->
    False.
Proof.
  crush.
Qed.


Lemma ctxt_switch_stealth_same_cache_entries :
  forall s s' osi pt_padd pt_addr pt_page pt,
    valid_sm_state s ->
    osi_pa_to_ma s osi pt_padd = Some pt_addr ->
    get_mem (smem_mm (st_memory s)) pt_addr = Some pt_page ->
    PT_c pt = page_content pt_page ->
    st_memory s' = ctxt_switch_new_smem s pt_addr ->
    st_hypervisor s' = st_hypervisor s ->
    same_non_stealth_entries s s'.
Proof.
  unfold valid_sm_state, osi_stealth_equivalent; crush.

  unfold same_non_stealth_entries; crush.
  rewrite H3.
  rewrite H4.
  clear dependent s'.
  unfold osi_pa_to_ma in *; crush.
  unfold ctxt_switch_new_smem; crush.

  rename m into pt_addr_aos.
  unfold remove_osi_stealth_page; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite ctxt_switch_mem in *; crush.
  rename p into pt_page_aos.
  rewrite H13; crush.
  
  remember (PT.get pt stealth_va).
  destruct o; crush.

  (* The new PT has a stealth page *)
  destruct H_valid_virtmap with osi pt_addr pt_page pt stealth_va i1; crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  specialize (H4 stealth_va_accessible); clear H12.
  assert (exists pg,
            page_of_vadd_madd_smem (st_memory s) stealth_va i1 = Some pg).
  unfold page_of_vadd_madd_smem; crush.
  destruct H12.
  rename x0 into new_stealth_pg.
  assert (is_rw_page new_stealth_pg).
  unfold page_of_vadd_madd_smem in H12; crush.
  destruct H_valid_cache with stealth_va i1 new_stealth_pg; crush.
  destruct H9 with pt_addr i1 new_stealth_pg; crush.
  exists osi; exists pt_padd;
    unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.
  unfold page_of_vadd_madd_smem; crush.
  unfold is_rw_page in *; crush.
  rename x0 into new_stealth_val, i1 into new_stealth_ma.

  rewrite get_new_cache with (ma:=new_stealth_ma)
    (pg:=new_stealth_pg) (val:=new_stealth_val); crush.

  rewrite flush_ma_cache.
  rewrite ctxt_switch_cache.
  apply same_cache_line_entries_trans with
        (fst (remove_ma_cache (smem_c (st_memory s)) i0)) (st_hypervisor s);
    crush.
  destruct H_valid_virtmap with (st_active_os s) pt_addr_aos pt_page_aos pt0
    stealth_va i0; crush.
  specialize (H17 stealth_va_accessible); clear H18.
  destruct H5 with pt_addr_aos i0 x0; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.

  eapply same_cache_line_remove_ma_cache_other_line with stealth_va
    (Page (RW_c (Some x1)) (page_owned_by x0) true); crush.
  remember (snd (remove_ma_cache (smem_c (st_memory s)) i0)).
  destruct y.
  destruct p; destruct c.
  destruct get_remove_ma_cache_some_entry with (smem_c (st_memory s))
    v i0 m p; crush.
  remember (vadd_eq stealth_va v).
  destruct s0; crush; clear Heqs0.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr_aos stealth_va = Some m);
    crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  STAC.synonym_cached.
  rewrite get_remove_ma_cache_no_entry in H20; crush.
  apply same_cache_line_add_cache_other_line; crush.

  rewrite madd_of_vadd_smem_flush_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.
  unfold madd_type; crush.
  rewrite <- H2; crush.
  
  rewrite get_page_result with (pg:=new_stealth_pg); crush.
  unfold page_of_vadd_smem.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = Some new_stealth_ma).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.
  rewrite madd_of_vadd_smem_flush_ma; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite flush_ma_cache in Heqval2; crush.
  remember (madd_eq new_stealth_ma i0).
  destruct s0; crush; clear Heqs0.
  rewrite get_remove_ma_cache_eq in *; crush.
  rewrite get_remove_ma_cache_ne in *; crush.
  rewrite ctxt_switch_cache in *; crush.
  unfold stealth_page_unique in H8.

  destruct H_valid_virtmap with (st_active_os s)
    pt_addr_aos pt_page_aos pt0 stealth_va i0; crush.
  specialize (H18 stealth_va_accessible); clear H19.
  destruct H5 with pt_addr_aos i0 x0; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  elim n; apply H8 with p
    (Page (RW_c (Some x1)) (page_owned_by x0) true); crush.

  rewrite flush_ma_cache in Heqval2; crush.
  remember (madd_eq i0 new_stealth_ma).
  destruct s0; crush; clear Heqs0.
  clear Heqval2.
  rewrite flush_ma_mem.
  unfold update_cache_entry; crush.
  rewrite ctxt_switch_cache in Heqval2.
  destruct get_remove_ma_cache_some_entry with
    (smem_c (st_memory s)) v new_stealth_ma m p0; crush.
  destruct H5 with pt_addr_aos m x; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  remember (vadd_eq stealth_va v).
  destruct s0; crush; clear Heqs0.
  rewrite ctxt_switch_mem.
  rewrite <- get_add_mem; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr_aos stealth_va = Some m).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  STAC.synonym_cached.

  rewrite ctxt_switch_mem.
  rewrite ctxt_switch_cache in Heqval2.
  unfold page_of_vadd_madd_smem in *; crush.
  rewrite get_remove_ma_cache_no_entry in Heqval3; crush.

  unfold page_of_vadd_madd_smem in *; crush.
  rewrite get_remove_ma_cache_ne in Heqval2; crush.
  rewrite ctxt_switch_cache in *; crush.
  rewrite flush_ma_mem.
  rewrite ctxt_switch_mem, ctxt_switch_cache.
  unfold update_cache_entry; crush.
  destruct get_remove_ma_cache_some_entry with
    (smem_c (st_memory s)) v i0 m p0; crush.
  rewrite get_add_mem_distinct; crush.
  unfold madd_type; crush; rewrite <- H2; crush.
  
  (* The new PT does not have a stealth page *)
  rewrite get_invalid.
  rewrite flush_ma_cache.
  rewrite ctxt_switch_cache; crush.
  remember (snd (remove_ma_cache (smem_c (st_memory s)) i0)).
  destruct y; crush.
  (* Some entry is removed from the cache *)
  destruct p. destruct c.
  destruct get_remove_ma_cache_some_entry with
    (smem_c (st_memory s)) v i0 m p; crush.
  remember (vadd_eq v stealth_va).
  destruct s0; crush; clear Heqs0.
  apply same_cache_line_remove_ma_cache_other_line with stealth_va p; crush.
  assert (madd_of_vadd (smem_mm (st_memory s)) pt_addr_aos stealth_va = Some m).
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  assert (stealth_va <> v); crush.
  STAC.synonym_cached.
  (* No entry is removed from the cache, ABSURD *)
  assert (get_cache (smem_c (st_memory s)) (stealth_va,i0) = None).
  apply get_remove_ma_cache_no_entry; auto.
  destruct H_valid_virtmap with (st_active_os s) pt_addr_aos
    pt_page_aos pt0 stealth_va i0; crush.
  specialize (H12 stealth_va_accessible); clear H15.
  destruct H5 with pt_addr_aos i0 x; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  elim get_cache_some_none_abs with (smem_c (st_memory s)) stealth_va i0; crush.

  unfold page_of_vadd_smem, madd_of_vadd_smem; crush.
  rewrite flush_ma_tlb in *; crush.
  rewrite ctxt_switch_tlb in *; crush.
  rewrite get_empty_tlb in *; crush.
  clear Heqval2.
  
  rewrite madd_of_vadd_flush_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.
  rewrite <- Heqo; crush.

  (* The AOS does not have a stealth page *)
  assert (forall ma, get_cache (smem_c (st_memory s)) (stealth_va, ma) = None).
  intros.
  unfold no_stealth_page in H6.
  apply H6 with pt_addr_aos; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite H13; crush.
  remember (PT.get pt stealth_va).
  destruct o; crush.
  destruct H_valid_virtmap with osi pt_addr pt_page pt stealth_va i0;
    crush.
  assert (madd_owner (smem_mm (st_memory s)) pt_addr = Os_PO osi); crush.
  unfold madd_owner in *; crush.
  specialize (H12 stealth_va_accessible); clear H15.
  destruct H9 with pt_addr i0 x; crush.
  exists osi; exists pt_padd; unfold osi_pa_to_ma; crush.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.
  unfold page_of_vadd_madd_smem; crush.
  unfold is_rw_page in H15; crush.

  rewrite get_new_cache with (ma:=i0) (pg:=x) (val:=x0); crush.
  rewrite ctxt_switch_cache.
  apply same_cache_line_add_cache_other_line; crush.
  rewrite madd_of_vadd_smem_ctxt_switch;
    unfold madd_of_vadd_smem, madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.

  rewrite get_page_result with (pg:=x); crush.
  unfold page_of_vadd_smem.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in Heqval2; rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.
  unfold page_of_vadd_madd_smem; crush.
  rewrite ctxt_switch_cache in *; crush.
  rewrite ctxt_switch_mem; crush.

  (* The new PT does not have a stealth page *)
  rewrite get_invalid.
  rewrite ctxt_switch_cache; crush.
  unfold page_of_vadd_smem; crush.
  unfold madd_of_vadd_smem; crush.
  rewrite ctxt_switch_tlb in Heqval2; rewrite get_empty_tlb in *; crush.
  rewrite ctxt_switch_mem.
  unfold madd_of_vadd, madd_page_table; crush.
  rewrite <- H2; crush.
  rewrite <- Heqo; crush.

  unfold aos_current_page_table, osi_current_page_table in *; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  destruct H_valid_currpg with (st_active_os s) aos p2m; crush.
  rewrite ctxt_switch_mem in *; crush.
Qed.

Lemma switch_stealth_same_cache_entries :
  forall s s' o,
    valid_sm_state s ->
    ACT.pre_switch o s ->
    post_sm_switch o s s' ->
    same_non_stealth_entries s s'.
Proof.
  intros.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with o; crush.
  STAC.get_osi_current_pt_page o; crush.
  unfold post_sm_switch in *; crush.
  rewrite H1 with pt_addr; crush.
  2: unfold osi_current_page_table; crush.

  rewrite <- H1; crush.
  apply (ctxt_switch_stealth_same_cache_entries s s' o 
    (os_current_page x) pt_addr pt_page pt); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma; crush.
  rewrite H1 with pt_addr; crush.
  unfold osi_current_page_table; crush.
  rewrite H1 with pt_addr; crush.
  unfold osi_current_page_table; crush.
  unfold osi_current_page_table; crush.
Qed.

Lemma lswitch_stealth_same_cache_entries :
  forall s s' pt_padd o,
    valid_sm_state s ->
    ACT.pre_lswitch pt_padd o s ->
    post_sm_lswitch pt_padd o s s' ->
    same_non_stealth_entries s s'.
Proof.
  intros.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with o; crush.
  STAC.get_osi_current_pt_page o; crush.
  unfold ACT.pre_lswitch, post_sm_lswitch in *; crush.

  unfold madd_type in H17; crush.
  destruct (page_content p) eqn:?; crush.

  apply (ctxt_switch_stealth_same_cache_entries s s' (st_active_os s) 
    pt_padd x1 p p0); crush.
  unfold valid_sm_state; crush.
  unfold osi_pa_to_ma in H0; crush.
  generalize (H_valid_hyper (st_active_os s) x0 pt_padd x1); intros.
  unfold madd_owner in *; crush.
  rewrite H1 with x1; crush.
  unfold osi_pa_to_ma; crush.

  rewrite H1 with x1; crush.
  rewrite H1 with x1; crush.
  unfold same_non_stealth_entries; crush.
Qed.  

Theorem same_non_stealth_entries_iso_not_osi :
  forall (s s':state) (a:sm_action),
    valid_sm_exec s a s' ->
    valid_sm_state s ->
    ~ non_stealth_access a ->
    same_non_stealth_entries s s'.
Proof.
  Local Hint Resolve same_non_stealth_entries_refl.
  Local Hint Constructors non_stealth_access osi_sm_action.

  sm_action_induction a; [clear Case; STAC.action_induction a | | | | ];
    unfold valid_sm_exec in *; crush;
    unfold ACT.post_silent, ACT.pre_read, ACT.pre_read_hyper,
      ACT.post_read, ACT.post_read_hyper in *;
    crush.

  (* Read *)
  apply read_stealth_same_cache_entires; crush;
    destruct (vadd_eq v stealth_va); crush;
    elim H1; crush.

  (* Read hyper *)
  destruct (vadd_eq v stealth_va); crush;
    try (elim H1; crush; fail).
  apply read_stealth_same_cache_entires; crush.

  (* Write *)
  destruct (vadd_eq v stealth_va); crush;
    try (elim H1; crush; fail).
  apply write_stealth_same_cache_entires with v0; crush.
  elim H2; crush.
  
  (* Write hyper *)
  destruct (vadd_eq v stealth_va); crush;
    try (elim H1; crush; fail).

  (* Del untrusted *)
  destruct (vadd_eq v stealth_va); crush;
    try (elim H1; crush; fail).
  apply del_stealth_same_cache_entires with o; crush.

  (* Hcall *)
  unfold same_non_stealth_entries, ACT.post_hcall in *; crush.

  (* Ret_ctrl *)
  unfold same_non_stealth_entries, ACT.post_ret_ctrl in *; crush.

  (* Chmod *)
  unfold same_non_stealth_entries, ACT.post_chmod in *; crush.

  (* Pin *)
  apply pin_stealth_same_cache_entries with o p p0; crush.

  (* Unpin *)
  elim H1; crush.

  (* new normal *)
  elim H1; crush.

  (* new stealth *)
  apply new_sm_stealth_same_cache_entries with o p; crush.

  (* Switch *)
  apply switch_stealth_same_cache_entries with o; crush.

  (* Lswitch *)
  elim H1; crush.
Qed.


(**************************************
 * Two step executions
 **************************************)

Lemma read_non_stealth_eq_two_exec :
  forall s1 s2 s1' s2' va,
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    va <> stealth_va ->
    st_active_os s1 = st_active_os s2 ->
    ACT.pre_all_read va s1 ->
    ACT.post_all_read va s1 s1' ->
    ACT.pre_all_read va s2 ->
    ACT.post_all_read va s2 s2' ->

    osi_non_stealth_equivalent s1 s2 ->
    same_non_stealth_entries s1 s2 ->
    same_non_stealth_entries s1' s2'.
Proof.
  intros.

  unfold same_non_stealth_entries; crush.

  unfold ACT.pre_all_read, ACT.post_all_read in H3, H4; crush.
  STAC.unify_read_post; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  unfold is_rw_page in *; crush.
  rewrite get_new_cache with (ma:=m) (pg:=x) (val:=x0); crush.
  2: rewrite get_page_result with (pg:=x); crush;
     unfold page_of_vadd_smem; crush.

  unfold ACT.pre_all_read, ACT.post_all_read in H5, H6; crush.
  STAC.unify_read_post; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  unfold is_rw_page in *; crush.
  rewrite get_new_cache with (ma:=m0) (pg:=x1) (val:=x2); crush.
  2: rewrite get_page_result with (pg:=x1); crush;
     unfold page_of_vadd_smem; crush.

  remember (line_dec (get_line va0) (get_line va)).
  destruct s; crush; clear Heqs.
  (* va0 and va are in the same line *)
  apply same_cache_line_add_cache_inv; auto.
  
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.

  unfold osi_non_stealth_equivalent in *; crush.
  destruct H7 with (st_active_os s1); crush.
  destruct H24.
  STAC.get_aos_data; crush.
  STAC.get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct H24 with (pt_padd:=(os_current_page aos)) (pt_page1:=pt_page)
    (mp1:=pt0) (va:=va) (ma1:=m); crush.
  unfold osi_pa_to_pg; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) pt_addr0 va = Some m).
  unfold madd_of_vadd_smem in Heqopt_res; crush.
  rewrite H_valid_tlb with (ma:=m1); crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in H18; crush.
  rewrite H29 in *; crush.
  destruct H31 with x; crush.
  assert (m0 = x5); crush.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state_H H0.
  assert (madd_of_vadd (smem_mm (st_memory s2)) pt_addr1 va = Some m0).
  unfold madd_of_vadd_smem in Heqopt_res0; crush.
  rewrite H_valid_tlb0 with (ma:=m1); crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in H40; crush.
  clear H_valid_currpg.
  STAC.get_aos_data; crush.
  STAC.get_aos_current_pt_page; crush.
  rewrite H44 in *; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite H2 in *; crush.
  destruct H20.
  destruct H20 with aos; crush.

  (* va0 and va are in different lines *)
  apply same_cache_line_entries_trans with (smem_c (st_memory s1))
    (st_hypervisor s1); auto.
  apply same_cache_line_entries_symm; auto.
  apply same_cache_line_add_cache_other_line; auto.
  apply same_cache_line_entries_trans with (smem_c (st_memory s2))
    (st_hypervisor s2); auto.
  apply same_cache_line_add_cache_other_line; auto.
Qed.

Lemma write_non_stealth_eq_two_exec :
  forall s1 s2 s1' s2' va val,
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    va <> stealth_va ->
    st_active_os s1 = st_active_os s2 ->
    ACT.pre_all_write va val s1 ->
    ACT.post_all_write va val s1 s1' ->
    ACT.pre_all_write va val s2 ->
    ACT.post_all_write va val s2 s2' ->

    osi_non_stealth_equivalent s1 s2 ->
    same_non_stealth_entries s1 s2 ->
    same_non_stealth_entries s1' s2'.
Proof.
  intros.

  unfold same_non_stealth_entries; crush.

  unfold ACT.pre_all_write, ACT.post_all_write in H3, H4; crush.
  STAC.unify_write_post; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=m) (pg:=x) (old_val:=x0); crush.
  2: unfold page_of_vadd_smem; crush.

  unfold ACT.pre_all_write, ACT.post_all_write in H5, H6; crush.
  STAC.unify_write_post; crush.
  unfold aos_page_of_vadd_smem, page_of_vadd_smem in *; crush.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=m0) (pg:=x1) (old_val:=x2); crush.
  2: unfold page_of_vadd_smem; crush.

  destruct (line_dec (get_line va0) (get_line va)); crush.
  (* va0 and va are in the same line *)
  apply same_cache_line_add_cache_inv; auto.
  
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.

  unfold osi_non_stealth_equivalent in *; crush.
  STAC.get_aos_data; crush.
  STAC.get_aos_current_pt_page; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct H7 with (st_active_os s1); crush.
  destruct H23.
  destruct H23 with (pt_padd:=(os_current_page aos)) (pt_page1:=pt_page) (mp1:=pt0)
    (va:=va) (ma1:=m); crush.
  unfold osi_pa_to_pg; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) pt_addr0 va = Some m); crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H25 in *; crush.
  destruct H31 with x; crush.

  assert (m0 = x5); crush.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state_H H0.
  assert (madd_of_vadd (smem_mm (st_memory s2)) pt_addr1 va = Some m0).
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in H40; crush.
  clear H_valid_currpg.
  STAC.get_aos_data; crush.
  STAC.get_aos_current_pt_page; crush.
  rewrite H44 in *; crush.
  unfold osi_pa_to_pg in *; crush.
  rewrite H2 in *; crush.
  destruct H18.
  destruct H18 with aos; crush.
  unfold same_page_but_value in *; crush.

  (* va0 and va are in different lines *)
  apply same_cache_line_entries_trans with (smem_c (st_memory s1))
    (st_hypervisor s1); auto.
  apply same_cache_line_entries_symm; auto.
  apply same_cache_line_add_cache_other_line; auto.
  apply same_cache_line_entries_trans with (smem_c (st_memory s2))
    (st_hypervisor s2); auto.
  apply same_cache_line_add_cache_other_line; auto.
Qed.

Lemma del_non_stealth_eq_two_exec :
  forall s1 s2 s1' s2' o va,
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    va <> stealth_va ->
    ACT.pre_del o va s1 ->
    ACT.pre_del o va s2 ->
    ACT.post_del o va s1 s1' ->
    ACT.post_del o va s2 s2' ->
    osi_non_stealth_equivalent s1 s2 ->
    same_non_stealth_entries s1 s2 ->
    same_non_stealth_entries s1' s2'.
Proof.
  unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del; crush.
  unfold same_non_stealth_entries; crush.
  rewrite delete_cache with (mem:=st_memory s1) (ma:=x5); crush.
  rewrite delete_cache with (mem:=st_memory s2) (ma:=x4); crush.
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  destruct H_valid_os_idents with o; crush.
  STAC.get_osi_current_pt_page o; crush.
  unfold osi_current_page_table in H5; crush.
  unfold madd_of_vadd, madd_page_table in H15; crush.
  rewrite H29 in *; crush.
  destruct H_va_has_pa with o x1 x pt_page pt va x5; crush.

  apply same_cache_line_remove_cache_inv with o x6; crush.
  unfold osi_pa_to_ma_hyp; crush.

  destruct H6 with o; crush.
  destruct H32.
  destruct H32 with (os_current_page x2) pt_page pt va x5; crush.
  unfold osi_pa_to_pg; crush.
  assert (osi_pa_to_ma s2 o x6 = Some x9).
  apply H37; crush.
  unfold osi_pa_to_ma; crush.
  unfold osi_pa_to_ma in H39; crush.
  unfold osi_pa_to_ma_hyp; crush.

  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with o; crush.
  clear H_valid_currpg.
  STAC.get_osi_current_pt_page o; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H50 in Heqopt_res; crush.
  unfold osi_pa_to_pg in H34; crush.
  unfold same_current_pt_page in *; crush.
  destruct H28 with x2; crush.
  rewrite H52 in *; crush.
  rewrite <- H35 in H50.
  injection H50; intros; subst x8; crush.
Qed.

Lemma same_cache_line_remove_new_va_cache :
  forall s1 s2 osi pt_addr1 pt_addr2 va va',
    valid_state s1 ->
    valid_state s2 ->

    osi_non_stealth_equivalent s1 s2 ->
    same_non_stealth_entries s1 s2 ->

    get_line va' <> get_line stealth_va ->
    get_line va <> get_line stealth_va ->
    os_accessible va ->
    osi_current_page_table s1 osi = Some pt_addr1 ->
    osi_current_page_table s2 osi = Some pt_addr2 ->

    same_cache_line (remove_new_va_cache (st_memory s1) pt_addr1 va)
      (st_hypervisor s1) (remove_new_va_cache (st_memory s2) pt_addr2 va)
      (st_hypervisor s2) (get_line va').
Proof.
  intros.
  generalize (H2 va H4); generalize (H2 va' H3); intros; clear H2.

  unfold remove_new_va_cache; crush.
  (* entry in both caches *)
  assert (exists pa, osi_pa_to_ma s1 osi pa = Some m).
  STAC.destruct_valid_state_H H.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H6; crush.
  destruct H_va_has_pa with osi x0 pt_addr1 pt_page  pt va m; crush.
  unfold madd_of_vadd, madd_page_table in Heqval; crush.
  rewrite H14 in *; crush.
  exists x1; unfold osi_pa_to_ma; crush.
  destruct H2.
  apply same_cache_line_remove_cache_inv with osi x; crush.
  STAC.destruct_valid_state_H H.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H6; crush.
  destruct H1 with osi; crush.
  destruct H17.
  destruct H17 with (os_current_page x0) pt_page pt va m; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H15 in *; crush.
  assert (osi_pa_to_ma s2 osi x = Some x4).
  apply H22; crush.
  unfold osi_pa_to_ma_hyp; unfold osi_pa_to_ma in H24; crush.
  clear H_valid_currpg.
  unfold valid_sm_state in *; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in Heqval1; crush.
  rewrite H30 in *; crush.
  unfold osi_pa_to_pg in H19; crush.
  unfold same_current_pt_page in *; crush.
  destruct H12 with x0; crush.
  rewrite H32 in *; crush.
  rewrite <- H20 in H30.
  injection H30; intros; subst x3; crush.

  (* No cache entry in s2 *)
  assert (exists pa, osi_pa_to_ma s1 osi pa = Some m).
  STAC.destruct_valid_state_H H.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H6; crush.
  destruct H_va_has_pa with osi x0 pt_addr1 pt_page  pt va m; crush.
  unfold madd_of_vadd, madd_page_table in Heqval; crush.
  rewrite H14 in *; crush.
  exists x1; unfold osi_pa_to_ma; crush.
  destruct H2.
  destruct same_cache_line_entries_some_pg with
    (smem_c (st_memory s1)) (smem_c (st_memory s2))
    (st_hypervisor s1) (st_hypervisor s2) osi va va x m m0 p; crush.
  STAC.destruct_valid_state_H H.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H6; crush.
  destruct H1 with osi; crush.
  destruct H17.
  destruct H17 with (os_current_page x0) pt_page pt va m; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H15 in *; crush.
  assert (osi_pa_to_ma s2 osi x = Some x4); crush.
  unfold osi_pa_to_ma_hyp; unfold osi_pa_to_ma in H24; crush.
  clear H_valid_currpg.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in Heqval1; crush.
  rewrite H30 in *; crush.
  unfold osi_pa_to_pg in H19; crush.
  unfold same_current_pt_page in *; crush.
  destruct H12 with x0; crush.
  rewrite H32 in *; crush.
  rewrite <- H20 in H30.
  injection H30; intros; subst x3; crush.
  unfold osi_pa_to_ma in H2; crush.
  STAC.destruct_valid_state_H H.
  generalize (H_valid_hyper osi i x m); intros.
  unfold madd_owner in H10; crush.
  destruct H_valid_cache with va m p; crush.  
  elim get_cache_some_none_abs with (smem_c (st_memory s2)) va m0; crush.

  (* Entry not mapped to va in s2 *)
  STAC.destruct_valid_state_H H.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H6; crush.
  destruct H1 with osi; crush.
  destruct H16.
  destruct H16 with (os_current_page x) pt_page pt va m; crush.
  unfold osi_pa_to_pg; crush.
  assert (exists pa, osi_pa_to_ma s1 osi pa = Some m).
  destruct H_va_has_pa with osi x0 pt_addr1 pt_page pt va m; crush.
  unfold madd_of_vadd, madd_page_table in Heqval; crush.
  rewrite H14 in *; crush.
  exists x1; unfold osi_pa_to_ma; crush.
  destruct H18.
  unfold madd_of_vadd, madd_page_table in Heqval; crush.
  rewrite H14 in *; crush.
  STAC.destruct_valid_state_H H0.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold osi_pa_to_pg in H18; crush.
  unfold madd_of_vadd, madd_page_table in Heqval1; crush.
  rewrite H28 in *; crush.
  unfold same_current_pt_page in H11; crush.
  destruct H11 with x; crush.

  (* No cache entry in s1 *)
  assert (exists pa, osi_pa_to_ma s2 osi pa = Some m0).
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H7; crush.
  destruct H_va_has_pa with osi x0 pt_addr2 pt_page pt va m0; crush.
  unfold madd_of_vadd, madd_page_table in Heqval1; crush.
  rewrite H14 in *; crush.
  exists x1; unfold osi_pa_to_ma; crush.
  destruct H2.
  destruct same_cache_line_entries_some_pg with
    (smem_c (st_memory s2)) (smem_c (st_memory s1))
    (st_hypervisor s2) (st_hypervisor s1) osi va va x m0 m p; crush.
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H7; crush.
  destruct H1 with osi; crush.
  destruct H17.
  destruct H18 with (os_current_page x0) pt_page pt va m0; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H15 in *; crush.
  assert (osi_pa_to_ma s1 osi x = Some x4); crush.
  unfold osi_pa_to_ma_hyp; unfold osi_pa_to_ma in H24; crush.
  clear H_valid_currpg.
  STAC.destruct_valid_state_H H.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in Heqval; crush.
  rewrite H30 in *; crush.
  unfold osi_pa_to_pg in H19; crush.
  unfold same_current_pt_page in *; crush.
  destruct H27 with x0; crush.
  rewrite H32 in *; crush.
  rewrite <- H20 in H30.
  injection H30; intros; subst x3; crush.
  unfold osi_pa_to_ma in H2; crush.
  STAC.destruct_valid_state_H H0.
  generalize (H_valid_hyper osi i x m0); intros.
  unfold madd_owner in H10; crush.
  destruct H_valid_cache with va m0 p; crush.  
  elim get_cache_some_none_abs with (smem_c (st_memory s1)) va m; crush.

  (* Entry not mapped to va in s1 *)
  STAC.destruct_valid_state_H H0.
  destruct H_valid_os_idents with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in H7; crush.
  destruct H1 with osi; crush.
  destruct H16.
  destruct H17 with (os_current_page x) pt_page pt va m; crush.
  unfold osi_pa_to_pg; crush.
  assert (exists pa, osi_pa_to_ma s2 osi pa = Some m).
  destruct H_va_has_pa with osi x0 pt_addr2 pt_page  pt va m; crush.
  unfold madd_of_vadd, madd_page_table in Heqval0; crush.
  rewrite H14 in *; crush.
  exists x1; unfold osi_pa_to_ma; crush.
  destruct H18.
  unfold madd_of_vadd, madd_page_table in Heqval0; crush.
  rewrite H14 in *; crush.
  STAC.destruct_valid_state_H H.
  clear H_valid_currpg.
  destruct H_valid_os_idents0 with osi; crush.
  STAC.get_osi_current_pt_page osi; crush.
  unfold osi_current_page_table in *; crush.
  unfold osi_pa_to_pg in H18; crush.
  unfold madd_of_vadd, madd_page_table in Heqval; crush.
  rewrite H28 in *; crush.
  unfold same_current_pt_page in H11; crush.
  destruct H25 with x; crush.
Qed.

Lemma new_non_stealth_eq_two_exec :
  forall s1 s2 s1' s2' o va pa,
    valid_sm_state s1 ->
    valid_sm_state s2 ->

    pre_new_normal o va pa s1 ->
    pre_new_normal o va pa s2 ->
    ACT.post_new o va pa s1 s1' ->
    ACT.post_new o va pa s2 s2' ->
    
    osi_non_stealth_equivalent s1 s2 ->
    same_non_stealth_entries s1 s2 ->
    same_non_stealth_entries s1' s2'.
Proof.
  intros.
  unfold same_non_stealth_entries; crush.
  unfold pre_new_normal, pre_all_new,
    ACT.post_new, ACT.post_all_new in *; crush.
  repeat rewrite new_cache_shared.

  assert (same_cache_line
     (remove_new_va_cache (st_memory s1) x3 va)
     (st_hypervisor s1)
     (remove_new_va_cache (st_memory s2) x0 va)
     (st_hypervisor s2) (get_line va0)).
  apply same_cache_line_remove_new_va_cache with o; crush;
    unfold valid_sm_state in *; crush.

  destruct (snd (remove_ma_cache (remove_new_va_cache (st_memory s1) x3 va) x))
    eqn:?; crush.
  (* Some entry is removed from s1 *)
  destruct p. destruct c.
  destruct get_remove_ma_cache_some_entry with
    (remove_new_va_cache (st_memory s1) x3 va)
     v x m p; crush.
  assert (get_cache (smem_c (st_memory s1)) (v,m) = Some p).
  unfold remove_new_va_cache in H26; crush.
  destruct (eq_cache_key (va,m0) (v,m)); crush.
  rewrite get_remove_cache in H26; crush.
  apply get_remove_cache_old with (va,m0); crush.
  
  assert (get_line v <> get_line stealth_va).
  destruct (vadd_eq v stealth_va); crush.
  (* v = stealth_va *)
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache_va_ma with stealth_va m p; crush.  
  unfold osi_pa_to_pg in H34; crush.
  destruct H23 with x2 i0; crush.
  unfold osi_pa_to_ma; crush.
  assert (madd_owner (smem_mm (st_memory s1)) m = Os_PO o); crush.
  unfold osi_pa_to_ma in H9; crush.
  destruct H_valid_virtmap with x i0 x6 x7 stealth_va m; crush.
  generalize (H_valid_hyper x i x2 i0); intros.
  unfold madd_owner in H38; crush.
  specialize (H39 stealth_va_accessible); clear H40; crush.
  unfold madd_owner in H37; crush.
  rewrite H37 in H39; injection H39; crush.
  unfold madd_of_vadd, madd_page_table; crush. 
  rewrite H35; crush.
  (* v <> stealth_va *)
  apply n; crush.
  unfold valid_sm_state in H; crush.

  destruct (snd (remove_ma_cache (remove_new_va_cache (st_memory s2) x0 va) x5))
    eqn:?; crush.
  destruct p0. destruct c.
  destruct get_remove_ma_cache_some_entry with
    (remove_new_va_cache (st_memory s2) x0 va)
     v0 x5 m0 p0; crush.
  destruct same_cache_line_entries_some_pg with
    (remove_new_va_cache (st_memory s1) x3 va)
    (remove_new_va_cache (st_memory s2) x0 va)
    (st_hypervisor s1) (st_hypervisor s2) o v v pa m m0 p; crush.
  apply same_cache_line_remove_new_va_cache with o; crush;
    try (unfold valid_sm_state in *; crush; fail).
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  unfold osi_pa_to_ma in H9; crush.
  rewrite H40; crush.
  generalize (H_valid_hyper o i pa m); intros.
  unfold madd_owner in H41; crush.

  assert (get_cache (smem_c (st_memory s2)) (v0,m0) = Some p0).
  unfold remove_new_va_cache in H29; crush.
  destruct (eq_cache_key (va,m1) (v0,m0)); crush.
  rewrite get_remove_cache in H29; crush.
  apply get_remove_cache_old with (va,m1); crush.
  assert (get_cache (smem_c (st_memory s2)) (v,m0) = Some x).
  unfold remove_new_va_cache in H28; crush.
  destruct (eq_cache_key (va,m1) (v,m0)); crush.
  rewrite get_remove_cache in H28; crush.
  apply get_remove_cache_old with (va,m1); crush.

  destruct (vadd_eq v v0); crush.
  2:destruct STAC.no_cache_alias with s2 v0 v m0 p0 x; crush;
    unfold valid_sm_state in *; crush.
  destruct (line_dec (get_line v0) (get_line va0)).
  apply same_cache_line_remove_ma_cache_inv with v0 p p0; crush.
  apply same_cache_line_entries_trans with
    (remove_new_va_cache (st_memory s1) x3 va)
    (st_hypervisor s1).
  apply same_cache_line_entries_symm.
  apply same_cache_line_remove_ma_cache_other_line with (va':=v0) (pg:=p); crush. 
  apply same_cache_line_entries_trans with
    (remove_new_va_cache (st_memory s2) x0 va)
    (st_hypervisor s2); crush.
  apply same_cache_line_remove_ma_cache_other_line with (va':=v0) (pg:=p0); crush. 

  generalize (get_remove_ma_cache_no_entry
    (remove_new_va_cache (st_memory s2) x0 va) x5 Heqo1 v); intros.
  destruct same_cache_line_entries_some_pg with
    (remove_new_va_cache (st_memory s1) x3 va)
    (remove_new_va_cache (st_memory s2) x0 va)
    (st_hypervisor s1) (st_hypervisor s2) o v v pa m x5 p; crush.
  apply same_cache_line_remove_new_va_cache with o; crush;
    try (unfold valid_sm_state in *; crush; fail).
  unfold valid_sm_state in H; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  unfold osi_pa_to_ma in H9; crush.
  rewrite H40; crush.
  generalize (H_valid_hyper o i pa m); intros.
  unfold madd_owner in H41; crush.
  elim get_cache_some_none_abs with (remove_new_va_cache (st_memory s2) x0 va)
    v x5; crush.
  
  (* No entry is removed from s1 *)
  destruct (snd (remove_ma_cache (remove_new_va_cache (st_memory s2) x0 va) x5))
    eqn:?; crush.
  destruct p. destruct c.
  destruct get_remove_ma_cache_some_entry with
    (remove_new_va_cache (st_memory s2) x0 va)
     v x5 m p; crush.
  assert (get_cache (smem_c (st_memory s2)) (v,m) = Some p).
  unfold remove_new_va_cache in H26; crush.
  destruct (eq_cache_key (va,m0) (v,m)); crush.
  rewrite get_remove_cache in H26; crush.
  apply get_remove_cache_old with (va,m0); crush.
  
  assert (get_line v <> get_line stealth_va).
  destruct (vadd_eq v stealth_va); crush.
  (* v = stealth_va *)
  unfold valid_sm_state in H0; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache_va_ma with stealth_va m p; crush.  
  unfold osi_pa_to_pg in H34; crush.
  destruct H18 with x5 i0; crush.
  unfold osi_pa_to_ma; crush.
  assert (madd_owner (smem_mm (st_memory s2)) m = Os_PO o); crush.
  unfold osi_pa_to_ma in H8; crush.
  destruct H_valid_virtmap with x2 i0 x6 x7 stealth_va m; crush.
  generalize (H_valid_hyper x2 i x5 i0); intros.
  unfold madd_owner in H38; crush.
  specialize (H39 stealth_va_accessible); clear H40; crush.
  unfold madd_owner in H37; crush.
  rewrite H37 in H39; injection H39; crush.
  unfold madd_of_vadd, madd_page_table; crush. 
  rewrite H35; crush.
  (* v <> stealth_va *)
  apply n; crush.
  unfold valid_sm_state in H0; crush.

  generalize (get_remove_ma_cache_no_entry
    (remove_new_va_cache (st_memory s1) x3 va) x Heqo0 v); intros.
  destruct same_cache_line_entries_some_pg with
    (remove_new_va_cache (st_memory s2) x0 va)
    (remove_new_va_cache (st_memory s1) x3 va)
    (st_hypervisor s2) (st_hypervisor s1) o v v pa m x p; crush.
  apply same_cache_line_entries_symm.
  apply same_cache_line_remove_new_va_cache with o; crush;
    try (unfold valid_sm_state in *; crush; fail).
  unfold valid_sm_state in H0; crush.
  STAC.destruct_valid_state.
  destruct H_valid_cache with v m p; crush.
  unfold osi_pa_to_ma in H8; crush.
  rewrite H40; crush.
  generalize (H_valid_hyper o i pa m); intros.
  unfold madd_owner in H41; crush.
  elim get_cache_some_none_abs with (remove_new_va_cache (st_memory s1) x3 va)
    v x; crush.

  apply same_cache_line_remove_ma_cache_no_entry_inv; crush.

  (* shared in s1, not shared in s2 *)
  elim n; apply shared_page_non_stealth_eq with s1 o pa x; crush.
  elim n; apply shared_page_non_stealth_eq with s2 o pa x5; crush.

  repeat rewrite new_cache_not_shared.
  apply same_cache_line_remove_new_va_cache with o; crush;
    unfold valid_sm_state in *; crush.
Qed.

Ltac eq_one_step :=
  match goal with
    | [H1: same_non_stealth_entries ?s1 ?s2,
       H2: valid_sm_exec ?s1 ?a ?s1',
       H3: valid_sm_exec ?s2 ?a ?s2'
       |-  same_non_stealth_entries ?s1' ?s2'] =>
      apply same_non_stealth_entries_trans with s1;
        [apply same_non_stealth_entries_symm |
         apply same_non_stealth_entries_trans with s2]; crush;
      try eapply same_non_stealth_entries_iso_not_osi; crush;
      try match goal with
        | [H4: non_stealth_access a |- _ ] => inversion H4; crush; fail
      end
  end.

Lemma write_hyper_stealth_invalid_exec : forall s s' val,
  valid_sm_state s ->
  valid_sm_exec s (Action (Write_hyper stealth_va val)) s' ->
  False.
Proof.
  intros.
  unfold valid_sm_state, valid_sm_exec in *; crush;
  unfold ACT.pre_write_hyper in *; crush.
  unfold aos_page_owner in *; crush.
  unfold aos_page_of_vadd_smem in *; crush.
  rename m into pt_addr.
  unfold page_of_vadd_smem in *; crush.
  unfold page_of_vadd_madd_smem in *; crush.
  destruct stealth_cached_is_aos with s pt_addr m p0; crush.
  unfold valid_sm_state; crush.
  destruct H3 with pt_addr m p; crush.
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
Qed.    

Theorem same_non_stealth_entries_iso: 
  forall (s1 s2 s1' s2':state) (a:sm_action) (osi:os_ident),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    st_active_os s1 = st_active_os s2 ->
    valid_sm_exec s1 a s1' ->
    valid_sm_exec s2 a s2' ->
    osi_stealth_equivalent s1 s2 osi ->
    same_non_stealth_entries s1' s2'.
Proof.  
  Local Hint Resolve same_non_stealth_entries_refl.
  Local Hint Constructors non_stealth_access osi_sm_action.
  unfold osi_stealth_equivalent; crush.
  
  sm_action_induction a; [clear Case; STAC.action_induction a | | | | ]; 
    try (unfold valid_sm_exec in *; crush;
    unfold ACT.post_silent in *; crush; fail); crush.

  (* Read *)
  destruct (vadd_eq v stealth_va); crush;
    [eq_one_step | apply read_non_stealth_eq_two_exec with s1 s2 v; crush];
  try (unfold valid_sm_exec in *; crush; unfold ACT.pre_read, ACT.post_read in *; crush; fail).

  (* Read_hyper *)
  destruct (vadd_eq v stealth_va); crush;
    [eq_one_step | apply read_non_stealth_eq_two_exec with s1 s2 v; crush];
  try (unfold valid_sm_exec in *; crush;
       unfold ACT.pre_read_hyper, ACT.post_read_hyper in *; crush; fail).

  (* Write *)
  destruct (vadd_eq v stealth_va); crush;
    [eq_one_step | apply write_non_stealth_eq_two_exec with s1 s2 v v0; crush];
    try (unfold valid_sm_exec in *; crush;
         unfold ACT.pre_write, ACT.post_write in *; crush; fail).

  (* Write hyper *)
  destruct (vadd_eq v stealth_va); crush;
    [| apply write_non_stealth_eq_two_exec with s1 s2 v v0; crush];
    try (unfold valid_sm_exec in *; crush;
         unfold ACT.pre_write_hyper, ACT.post_write_hyper in *; crush; fail).
  apply same_non_stealth_entries_trans with s1;
    [apply same_non_stealth_entries_symm |
     apply same_non_stealth_entries_trans with s2]; crush;
    apply write_stealth_same_cache_entires with (val:=v0); crush;
    unfold valid_sm_exec in *; crush;
    unfold ACT.pre_write_hyper, ACT.post_write_hyper in *; crush.
    
  (* Del untrusted *)
  destruct (vadd_eq v stealth_va); crush.
  apply same_non_stealth_entries_trans with s1.
  apply same_non_stealth_entries_symm.
  apply del_stealth_same_cache_entires with (o:=o); crush;
    unfold valid_sm_exec in *; crush;
    unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  apply same_non_stealth_entries_trans with s2; crush.
  apply del_stealth_same_cache_entires with (o:=o); crush;
    unfold valid_sm_exec in *; crush;
    unfold ACT.pre_del, ACT.pre_all_del, ACT.post_del in *; crush.
  apply del_non_stealth_eq_two_exec with s1 s2 o v; crush;
  unfold valid_sm_exec in *; crush.

  (* Hcall *)
  unfold valid_sm_exec in *; crush;
    unfold same_non_stealth_entries, ACT.post_hcall in *; crush.

  (* Ret_ctrl *)
  unfold valid_sm_exec in *; crush;
    unfold same_non_stealth_entries, ACT.post_ret_ctrl in *; crush.

  (* Chmod *)
  unfold valid_sm_exec in *; crush;
    unfold same_non_stealth_entries, ACT.post_chmod in *; crush.

  (* Pin *)
  apply same_non_stealth_entries_trans with s1;
    [apply same_non_stealth_entries_symm |
     apply same_non_stealth_entries_trans with s2]; crush.
  apply pin_stealth_same_cache_entries with o p p0; crush;
    unfold valid_sm_exec in *; crush.
  apply pin_stealth_same_cache_entries with o p p0; crush;
    unfold valid_sm_exec in *; crush.
  
  (* Unpin *)
  apply same_non_stealth_entries_trans with s1;
    [apply same_non_stealth_entries_symm |
     apply same_non_stealth_entries_trans with s2]; crush.
  apply unpin_stealth_same_cache_entries with o p; crush;
    unfold valid_sm_exec in *; crush.
  apply unpin_stealth_same_cache_entries with o p; crush;
    unfold valid_sm_exec in *; crush.

  (* New normal *)
  apply new_non_stealth_eq_two_exec with s1 s2 o v p; crush;
    unfold valid_sm_exec in *; crush;
    unfold pre_new_normal in *; crush.

  (* New_sm *)
  apply same_non_stealth_entries_trans with s1;
    [apply same_non_stealth_entries_symm |
     apply same_non_stealth_entries_trans with s2]; crush.
  apply new_sm_stealth_same_cache_entries with o p; crush;
    unfold valid_sm_exec in *; crush.
  apply new_sm_stealth_same_cache_entries with o p; crush;
    unfold valid_sm_exec in *; crush.

  (* Switch *) 
  eq_one_step.
  
  (* Lswitch *)
  unfold valid_sm_exec in *; crush.
  apply same_non_stealth_entries_trans with s1;
    [apply same_non_stealth_entries_symm |
     apply same_non_stealth_entries_trans with s2]; crush.
  apply lswitch_stealth_same_cache_entries with p o; crush.
  apply lswitch_stealth_same_cache_entries with p o; crush.
Qed.

End STEALTH_SM_ISO_LEMMAS.
