(***********************************************************************
 * File: Stealth_sm_isolation.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Proof of OS Stealth Isolation for StealthMem actions.
 *************************************************************************)
Require Import Stealth_actions Stealth_valid Stealth_equivalence
  Stealth_traces Stealth_isolation Stealth_non_stealth_equivalence
  Stealth_sm_iso_lemmas Stealth_lemmas.

Module STEALTH_SM_ISOLATION.
  Import PG C T MM SMEM ST ST_VALID OS_ISO
         STEALTH_ACTIONS VALID_SM_STATE STEALTH_EQ
         STEALTH_ISOLATION STEALTH_TRACES STEALTH_NON_STEALTH_EQ
         STEALTH_SM_ISO_LEMMAS STEALTH_LEMMAS.

Theorem stealth_isolation_not_osi :
  forall (s s':state) (a:sm_action) (osi:os_ident),
    valid_sm_exec s a s' ->
    valid_sm_state s ->
    ~ osi_sm_action osi s a ->
    ~ non_stealth_access a ->
    osi_stealth_equivalent s s' osi.
Proof.
  Local Hint Resolve stealth_isolation_not_osi
       same_non_stealth_entries_iso_not_osi
       os_non_stealth_isolation_not_osi.
  unfold osi_stealth_equivalent; crush.
Qed.


(* TWO STEP EXECUTIONS *)  

(* Ltac eq_one_step := *)
(*   match goal with *)
(*     | [H1: osi_stealth_equivalent ?s1 ?s2 ?osi, *)
(*        H2: valid_sm_exec ?s1 ?a ?s1', *)
(*        H3: valid_sm_exec ?s2 ?a' ?s2' *)
(*        |- osi_stealth_equivalent ?s1' ?s2' ?osi] => *)
(*       apply osi_stealth_equivalent_trans with s1; *)
(*         [apply osi_stealth_equivalent_symm | *)
(*          apply osi_stealth_equivalent_trans with s2]; crush *)
(*   end. *)

Lemma eq_osi_ne_active_s2 :
  forall s1 s2 osi,
    osi <> st_active_os s1 ->
    osi_stealth_equivalent s1 s2 osi ->
    st_active_os s2 <> osi.
Proof.
  intros.
  unfold osi_stealth_equivalent in *; crush.
  unfold OS_ISO.STEQ.osi_equivalent in *; crush.
  destruct H0; crush.
Qed.

Lemma non_stealth_write_eq :
  forall s1 s2 s1' s2' osi va val1 val2,
    va <> stealth_va ->
    osi <> st_active_os s1 ->
    st_active_os s1 = st_active_os s2 ->
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    osi_stealth_equivalent s1 s2 osi ->
    valid_sm_exec s1 (Action (Write va val1)) s1' ->
    valid_sm_exec s2 (Action (Write va val2)) s2' ->
    osi_stealth_equivalent s1' s2' osi.
Proof.
  unfold osi_stealth_equivalent; crush.
  
  assert (st_active_os s2 <> osi);
    [eapply eq_osi_ne_active_s2; unfold osi_stealth_equivalent; crush|].
  apply OS_ISO.STEQ.osi_equivalent_trans with s1; crush;
    [apply OS_ISO.STEQ.osi_equivalent_symm |
     apply OS_ISO.STEQ.osi_equivalent_trans with s2]; crush.
  apply os_isolation_not_osi with (a:=Write va val1);
    unfold ACT.valid_exec, valid_sm_exec, valid_sm_state in *; crush;
    try inversion H10; crush.
  apply os_isolation_not_osi with (a:=Write va val2);
    unfold ACT.valid_exec, valid_sm_exec, valid_sm_state in *; crush;
    try inversion H10; crush.

  apply osi_non_stealth_equivalent_trans with s1; crush;
    [apply osi_non_stealth_equivalent_symm |
     apply osi_non_stealth_equivalent_trans with s2]; crush.
  apply write_stealth_os_non_stealth_isolation with va val1; crush;
    unfold valid_sm_exec in *; crush; unfold ACT.pre_write in *; crush.
  apply write_stealth_os_non_stealth_isolation with va val2; crush;
    unfold valid_sm_exec in *; crush; unfold ACT.pre_write in *; crush.

  unfold same_non_stealth_entries; crush.
  unfold valid_sm_exec in *; crush;
    unfold ACT.pre_write, ACT.pre_all_write,
    ACT.post_write, ACT.post_all_write in *; crush.
  STAC.unify_write_post; crush.
  STAC.unify_write_post; crush.

  unfold page_of_vadd_smem in *; crush.
  unfold is_rw_page in *; crush.
  rewrite set_new_cache with (ma:=m0) (pg:=x0) (old_val:=x2);
    unfold page_of_vadd_smem; crush.
  rewrite set_new_cache with (ma:=m) (pg:=x) (old_val:=x1);
    unfold page_of_vadd_smem; crush.
  apply same_cache_line_add_cache_inv; crush.
  assert (same_page_but_value x0 x); crush;
    [|destruct H18; destruct H20;  unfold same_page_but_value in *; auto].

  unfold osi_non_stealth_equivalent in *; crush.
  generalize (H7 (st_active_os s1)); crush.
  unfold same_current_pt_page, same_current_pt_page_one in *; crush.
  unfold valid_sm_state in H2; crush.
  STAC.destruct_valid_state.
  STAC.get_aos_data.
  STAC.get_aos_current_pt_page.
  unfold aos_current_page_table, osi_current_page_table in Heqopt_pt_addr0; crush.
  destruct H19 with aos; crush.

  destruct H_va_has_pa with (st_active_os s1) p2m pt_addr1 pt_page pt va m0; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold madd_of_vadd_smem in Heqopt_res0; crush.  
  unfold valid_tlb, madd_of_vadd, madd_page_table in *; crush.
  generalize (H_valid_tlb va m1 pt_addr1); intros.
  rewrite <- H31 in H35; simpl in *.
  rewrite H33 in H35; simpl in *.
  apply H35; auto.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in Heqopt_res0; crush.
  rewrite H33 in *; crush.
  
  destruct H22.
  destruct H22 with (os_current_page aos) pt_page pt va m0; crush.
  unfold osi_pa_to_pg; crush.
  unfold madd_of_vadd_smem in Heqopt_res0; crush.
  generalize (H_valid_tlb va m1 pt_addr1 Heqval); intros.
  unfold madd_of_vadd, madd_page_table in H37; simpl in *.
  rewrite <- H31 in H37; simpl in *.
  rewrite H33 in H37; simpl in *.
  apply H37; auto.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H33 in *; crush.
  assert (osi_pa_to_ma s2 (st_active_os s1) x4 = Some x7); crush.
  apply H40; unfold osi_pa_to_ma; crush.

  destruct H22 with (os_current_page aos) pt_page pt va m0; crush.
  unfold osi_pa_to_pg; crush.
  assert (madd_of_vadd (smem_mm (st_memory s1)) pt_addr1 va = Some m0).
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  unfold madd_of_vadd, madd_page_table in *; crush.
  rewrite H33 in *; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  destruct H46 with x0; crush.
  unfold osi_pa_to_ma in H42; crush.
  rewrite H1 in *; crush.
  rewrite <- H32 in Heqopt_pt_addr; crush.
  unfold osi_pa_to_pg in *; crush.
  unfold aos_page_of_vadd_smem in H20; crush.
  unfold page_of_vadd_smem in *; crush.
  unfold aos_current_page_table, osi_current_page_table in *; crush.
  rewrite <- H32 in *; crush.
  assert (madd_of_vadd (smem_mm (st_memory s2)) m1 va = Some m2).
  apply STAC.madd_of_vadd_smem_then_mmem; crush.
  unfold valid_sm_state in *; crush.
  unfold aos_current_page_table, osi_current_page_table; crush.
  unfold madd_of_vadd, madd_page_table in H49; crush.
  rewrite <- H43 in *; crush.

  simpl.
  unfold same_page_type in *; crush.  
Qed.


Local Hint Resolve stealth_os_isolation
  same_non_stealth_entries_iso
  os_non_stealth_isolation.

Theorem stealth_os_isolation :
  forall (s1 s1' s2 s2':state) (a:sm_action) (osi:os_ident),
    valid_sm_state s1 ->
    valid_sm_state s2 ->
    st_active_os s1 = st_active_os s2 ->
    valid_sm_exec s1 a s1' ->
    valid_sm_exec s2 a s2' ->
    osi_stealth_equivalent s1 s2 osi ->
    osi_stealth_equivalent s1' s2' osi.
Proof.
  intros.
  generalize H4.
  unfold osi_stealth_equivalent; crush.
Qed.

End STEALTH_SM_ISOLATION.

