(***********************************************************************
 * File: Stealth_traces_victim.v
 * $Revision: 12408 $
 * $Date: 2014-03-29 18:21:33 -0300 (Sat, 29 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * Stealth traces for the victim OS with StealthMem. Assumes only two
 * OSs (victim and attacker).
 *************************************************************************)
Require Export Sets.Constructive_sets Stealth_actions
  Stealth_valid Stealth_valid_inv Stealth_obs_lemmas Stealth_sm_isolation 
  Stealth_traces Stealth_isolation Stealth_equivalence Stealth_leakage
  Stealth_non_stealth_equivalence Streams.

Module Type STEALTH_TRACES_VICTIM.
  Import PG C T MM SMEM ST ST_VALID
         STEALTH_ACTIONS VALID_SM_STATE VALID_SM_STATE_INV
         STEALTH_TRACES STEALTH_ISOLATION
         STEALTH_SM_ISOLATION STEALTH_EQ OBS_LEMMAS
         OS_ISO NON_LEAKAGE STEALTH_NON_STEALTH_EQ.

  (* Returns the union of observations regarding all OS but osi *)
  Parameter osi_v osi_a : os_ident.
  Axiom osi_v_ne_osi_a : osi_v <> osi_a.
  Axiom osi_v_or_osi_a : forall osi, osi = osi_v \/ osi = osi_a.

(* The actions of the attacker and non stealth actions of the victim
   are the same in both traces.
   This predicate implies that the traces only have actions of osi_a
   and osi_v (not of other OSs or the hypervisor). *)
CoInductive leak_same_non_stealth_osi_a_actions :
    leak_state_stream -> leak_state_stream -> Prop :=
  | Leak_osi_a_action: forall (str1 str2:leak_state_stream)
      (A:leak_action) (s1 s2:leak_state),
        osi_sm_action osi_a (leak_state_state s1) (A (leak_state_obs s1)) ->
        osi_sm_action osi_a (leak_state_state s2) (A (leak_state_obs s2)) ->
        leak_same_non_stealth_osi_a_actions str1 str2 -> 
        leak_same_non_stealth_osi_a_actions 
          (Cons (Leak_Ex_State s1 A (cache_test osi_a)) str1)
          (Cons (Leak_Ex_State s2 A (cache_test osi_a)) str2)
  | Leak_non_stealth_osi_v_action : forall (str1 str2:leak_state_stream)
      (A:leak_action) (s1 s2:leak_state),
        osi_sm_action osi_v (leak_state_state s1) (A (leak_state_obs s1)) ->
        non_stealth_access (A (leak_state_obs s1)) ->
        osi_sm_action osi_v (leak_state_state s2) (A (leak_state_obs s2)) ->
        non_stealth_access (A (leak_state_obs s2)) ->        
        leak_same_non_stealth_osi_a_actions str1 str2 -> 
        leak_same_non_stealth_osi_a_actions
          (Cons (Leak_Ex_State s1 A (cache_test osi_a)) str1)
          (Cons (Leak_Ex_State s2 A (cache_test osi_a)) str2)
  | Leak_non_stealth_osi_v_write : forall (str1 str2:leak_state_stream)
      (A A':leak_action) (va:vadd) (val1 val2:value) (s1 s2:leak_state),
        osi_sm_action osi_v (leak_state_state s1) (A (leak_state_obs s1)) ->
        osi_sm_action osi_v (leak_state_state s2) (A' (leak_state_obs s2)) ->
        (A (leak_state_obs s1)) = (Action (Write va val1)) ->
        (A' (leak_state_obs s2)) = (Action (Write va val2)) ->
        non_stealth_access (A (leak_state_obs s1)) -> 
        non_stealth_access (A (leak_state_obs s2)) ->        
        leak_same_non_stealth_osi_a_actions str1 str2 -> 
        leak_same_non_stealth_osi_a_actions
          (Cons (Leak_Ex_State s1 A (cache_test osi_a)) str1)
          (Cons (Leak_Ex_State s2 A' (cache_test osi_a)) str2)
  | Leak_stealth_osi_v_action1: forall (str1 str2:leak_state_stream)
      (A:leak_action) (s1:leak_state),
        leak_same_non_stealth_osi_a_actions str1 str2 -> 
        ~(osi_sm_action osi_a (leak_state_state s1) (A (leak_state_obs s1))) ->
        osi_sm_action osi_v (leak_state_state s1) (A (leak_state_obs s1)) ->
        ~(non_stealth_access (A (leak_state_obs s1))) ->
        leak_same_non_stealth_osi_a_actions (Cons (Leak_Ex_State s1 A (cache_test osi_a)) str1) str2
  | Leak_stealth_osi_v_action2: forall (str1 str2:leak_state_stream)
      (A:leak_action) (s2:leak_state),
        leak_same_non_stealth_osi_a_actions str1 str2 -> 
        ~(osi_sm_action osi_a (leak_state_state s2) (A (leak_state_obs s2))) ->
        osi_sm_action osi_v (leak_state_state s2) (A (leak_state_obs s2)) ->
        ~(non_stealth_access (A (leak_state_obs s2))) ->
        leak_same_non_stealth_osi_a_actions str1 (Cons (Leak_Ex_State s2 A (cache_test osi_a)) str2).

Lemma osi_v_non_stealth_action_not_osi_a : forall s A,
  osi_sm_action osi_v (leak_state_state s) (A (leak_state_obs s)) ->
  non_stealth_access (A (leak_state_obs s)) ->
  pre_sm_action (A (leak_state_obs s)) (leak_state_state s) ->
  ~(osi_sm_action osi_a (leak_state_state s) (A (leak_state_obs s))).
Proof.
  intros.
  remember (A (leak_state_obs s)) as a.
  clear Heqa A.
  
  sm_action_induction a; crush;
    try match goal with | [Case := "Action"%string : string|-_] => clear Case;
      STAC.action_induction a end; crush;
    unfold ACT.pre_del, ACT.pre_pin, ACT.pre_unpin,
      pre_new_sm, pre_new_normal, pre_all_new, ACT.pre_lswitch, hyper_running in *;
    inversion H; inversion H2; crush;
    try inversion H0; crush;
    try (inversion H4; inversion H6; crush;
         apply osi_v_ne_osi_a; crush).
Qed.  

Theorem Non_leakage : forall (str1 str2:leak_state_stream),
  leak_run str1 -> leak_run str2 ->
  leak_same_non_stealth_osi_a_actions str1 str2 ->
  osi_stealth_equivalent (leak_state_state (l_ex_state (hd str1)))
    (leak_state_state (l_ex_state (hd str2))) osi_a ->
  leak_osi_equivalent_traces osi_a str1 str2.
Proof.
  intros.
  inversion H; subst; clear H.
  inversion H0; subst; clear H0.
  destruct H3.
  destruct H.
  assert (leak_state_obs (l_ex_state (hd str1)) =
    leak_state_obs (l_ex_state (hd str2))); crush.

  clear H3 H6.
  generalize dependent str2.
  generalize dependent str1.

  cofix str1.
  intros str2 H_valid2 H_trace2 str3 H_R H_first_osi_eq
    H_valid3 H_trace3 H_first_obs_eq.

  inversion H_R; subst; simpl in *.
  (* Same osi_a action in both traces *)
  assert (valid_sm_exec (leak_state_state s1)
    (A (leak_state_obs s1))
    (leak_state_state (l_ex_state (hd str0)))).
  inversion H_trace2; crush.
  inversion H5; crush.
  subst s. subst s'. subst o0. subst s1.
  crush.

  assert (valid_sm_exec (leak_state_state s2)
    (A (leak_state_obs s2))
    (leak_state_state (l_ex_state (hd str4)))).
  inversion H_trace3; crush.
  inversion H6; crush.
  subst s. subst s'. subst o0. subst s2.
  crush.

  constructor; crush.
  
  apply str1; auto.
  unfold leak_valid_state.
  apply valid_sm_state_inv with (leak_state_state s1) (A (leak_state_obs s1)); crush.
  inversion H_trace2; crush.
  
  apply stealth_os_isolation with (leak_state_state s1) (leak_state_state s2) (A (leak_state_obs s1)); crush.
  unfold osi_stealth_equivalent in *; crush.
  unfold STEQ.osi_equivalent in *; crush.
  destruct H4; crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s1))); crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s2))); crush.
  apply valid_sm_state_inv with (leak_state_state s2) (A (leak_state_obs s2)); crush.
  inversion H_trace3; crush.
  apply same_obs_two_exec with osi_a s1 s2 A (l_ex_action (hd str0)) (l_ex_action (hd str4)) ; crush.
  inversion H_trace2; crush.
  inversion H_trace3; crush.
  unfold osi_stealth_equivalent in *; crush.
  unfold STEQ.osi_equivalent in *; crush.
  destruct H4; crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s1))); crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s2))); crush.

  (* Same non stealth osi_v action *) 
  assert (valid_sm_exec (leak_state_state s1)
    (A (leak_state_obs s1))
    (leak_state_state (l_ex_state (hd str0)))).
  inversion H_trace2; crush.
  inversion H7; crush.
  subst s. subst s'. subst o0. subst s1.
  crush.

  assert (valid_sm_exec (leak_state_state s2)
    (A (leak_state_obs s2))
    (leak_state_state (l_ex_state (hd str4)))).
  inversion H_trace3; crush.
  inversion H8; crush.
  subst s. subst s'. subst o0. subst s2.
  crush.

  apply eq_trace_not_osi_action1; crush.
  apply eq_trace_not_osi_action2; crush.
    
  apply str1; auto.
  unfold leak_valid_state.
  apply valid_sm_state_inv with (leak_state_state s1) (A (leak_state_obs s1)); crush.
  inversion H_trace2; crush.
  apply stealth_os_isolation with (leak_state_state s1) (leak_state_state s2) (A (leak_state_obs s1)); crush.
  unfold osi_stealth_equivalent in *; crush.
  unfold STEQ.osi_equivalent in *; crush.
  destruct H6; crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s1))); crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s2))); crush.
  
  apply valid_sm_state_inv with (leak_state_state s2) (A (leak_state_obs s2)); crush.
  inversion H_trace3; crush.
  apply same_obs_two_exec with osi_a s1 s2 A (l_ex_action (hd str0)) (l_ex_action (hd str4)) ; crush.
  inversion H_trace2; crush.
  inversion H_trace3; crush.
  unfold osi_stealth_equivalent in *; crush.
  unfold STEQ.osi_equivalent in *; crush.
  destruct H6; crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s1))); crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s2))); crush.

  apply osi_v_non_stealth_action_not_osi_a with s2 A; crush.
  unfold valid_sm_exec in *; crush.
  apply osi_v_non_stealth_action_not_osi_a with s1 A; crush.
  unfold valid_sm_exec in *; crush.

  (* non stealth write *)
  assert (valid_sm_exec (leak_state_state s1)
    (A (leak_state_obs s1))
    (leak_state_state (l_ex_state (hd str0)))).
  inversion H_trace2; crush.
  inversion H9; crush.
  subst s. subst s'. subst o0. subst s1.
  crush.

  assert (valid_sm_exec (leak_state_state s2)
    (A' (leak_state_obs s2))
    (leak_state_state (l_ex_state (hd str4)))).
  inversion H_trace3; crush.
  inversion H10; crush.
  subst s. subst s'. subst o0. subst s2.
  crush.

  apply eq_trace_not_osi_action1; crush.
  apply eq_trace_not_osi_action2; crush.

  apply str1; auto.
  unfold leak_valid_state.
  apply valid_sm_state_inv with (leak_state_state s1) (A (leak_state_obs s1)); crush.
  inversion H_trace2; crush.
  apply non_stealth_write_eq with (leak_state_state s1) (leak_state_state s2) va val1 val2; crush.
  rewrite H1 in H3; inversion H3; crush.
  elim osi_v_non_stealth_action_not_osi_a with s1 A; crush.
  inversion H6; crush.
  rewrite H1.
  apply osi_sm_action_old_back.
  constructor; crush.
  rewrite H1 in H6; inversion H6; crush.
  unfold ACT.pre_write in H10; crush.
  inversion H; inversion H0; crush.
  inversion H9; inversion H11; crush.

  apply valid_sm_state_inv with (leak_state_state s2) (A' (leak_state_obs s2)); crush.
  inversion H_trace3; crush.
  apply same_obs_two_write with osi_a s1 s2 A A'
    (l_ex_action (hd str0)) (l_ex_action (hd str4)) va val1 val2; crush.
  inversion H_trace2; crush.
  inversion H_trace3; crush.
  inversion H3; crush.

  inversion H; crush.
  rewrite H1 in *; crush.
  injection H9; intros; subst a; clear H9; crush.
  inversion H10; crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s1))); crush;
    elim osi_v_ne_osi_a; crush.

  unfold osi_stealth_equivalent in *; crush.
  unfold STEQ.osi_equivalent in *; crush.
  destruct H8; crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s1))); crush.
  destruct (osi_v_or_osi_a (st_active_os (leak_state_state s2))); crush.

  rewrite H2 in *; crush.
  inversion H8; crush.
  inversion H10; crush.
  inversion H0; crush.
  inversion H13; crush.
  elim osi_v_ne_osi_a; crush.

  rewrite H1 in *; crush.
  inversion H8; crush.
  inversion H10; crush.
  inversion H; crush.
  inversion H13; crush.
  elim osi_v_ne_osi_a; crush.

  (* Arbitrary osi_v stealth action in str1 *)
  assert (valid_sm_exec (leak_state_state s1)
    (A (leak_state_obs s1))
    (leak_state_state (l_ex_state (hd str0)))).
  inversion H_trace2; crush.
  inversion H6; crush.
  subst s. subst s'. subst o0. subst s1.
  crush.

  apply eq_trace_not_osi_action1; auto.
  apply str1; crush.
  unfold leak_valid_state.
  apply valid_sm_state_inv with (leak_state_state s1) (A (leak_state_obs s1)); crush.
  inversion H_trace2; crush.
  apply osi_stealth_equivalent_trans
    with (leak_state_state s1); crush.
  apply osi_stealth_equivalent_symm; crush.
  apply stealth_isolation_not_osi with (A (leak_state_obs s1)); crush.
  rewrite <- same_obs_step with (osi:=osi_a)
    (s:=s1) (A:=A) (A':=(l_ex_action (hd str0))); crush.
  inversion H_trace2; crush.

  (* Arbitrary osi_v stealth action in str2 *)
  assert (valid_sm_exec (leak_state_state s2)
    (A (leak_state_obs s2))
    (leak_state_state (l_ex_state (hd str4)))).
  inversion H_trace3; crush.
  inversion H6; crush.
  subst s. subst s'. subst o0. subst s2.
  crush.

  apply eq_trace_not_osi_action2; auto.
  apply str1; crush.
  apply osi_stealth_equivalent_trans
    with (leak_state_state s2); crush.
  apply stealth_isolation_not_osi with (A (leak_state_obs s2));
    crush.
  unfold leak_valid_state.
  apply valid_sm_state_inv with (leak_state_state s2) (A (leak_state_obs s2)); crush.
  inversion H_trace3; crush.
  transitivity (leak_state_obs s2); crush.
  apply same_obs_step with (osi:=osi_a)
    (s:=s2) (A:=A) (A':=(l_ex_action (hd str4))); crush.
  inversion H_trace3; crush.
Qed.    

End STEALTH_TRACES_VICTIM.
