(***********************************************************************
 * File: Stealth_valid.v
 * $Revision: 12857 $
 * $Date: 2015-07-28 19:11:12 -0300 (Tue, 28 Jul 2015) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * State validity of StealthMem, and new invariants.
 *************************************************************************)
Require Import State_tactics OS_Isolation
  State Stealth_actions.

Module VALID_SM_STATE.
  Import PG.
  Import C.
  Import T.
  Import MM.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  Import STEALTH_ACTIONS.

  (* New invariants *)
  (* The current stealth page of the AOS is always cached *)
  Definition stealth_page_cached (s:state) :=
    forall pt_addr ma pg,
      aos_current_page_table s = Some pt_addr ->
      madd_of_vadd (smem_mm (st_memory s)) pt_addr C.stealth_va = Some ma ->
      get_mem (smem_mm (st_memory s)) ma = Some pg ->
      is_cacheable pg /\ is_rw_page pg /\
      exists val, get_cache (smem_c (st_memory s)) (C.stealth_va,ma) =
        Some (Page (RW_c (Some val)) (page_owned_by pg) true).

  Definition no_stealth_page (s:state) :=
    forall pt_addr ma,
      aos_current_page_table s = Some pt_addr ->
      madd_of_vadd (smem_mm (st_memory s)) pt_addr C.stealth_va = None ->
      get_cache (smem_c (st_memory s)) (C.stealth_va,ma) = None.

  (* If an entry is cached in the stealth page line,
     it must be the stealth page of the AOS. *)
  Definition stealth_page_reserved_line (s:state) :=
    forall va ma pg,
      C.get_line va = C.get_line C.stealth_va ->
      get_cache (smem_c (st_memory s)) (va,ma) = Some pg ->
      va = C.stealth_va.

  Definition stealth_page_unique (s:state) :=
    forall ma ma' pg pg',
      get_cache (smem_c (st_memory s)) (C.stealth_va,ma) = Some pg ->
      get_cache (smem_c (st_memory s)) (C.stealth_va,ma') = Some pg' ->
      ma = ma'.

  Definition stealth_page_rw_cacheable (s:state) :=
    forall pt_addr ma pg,
      (exists osi, exists pt_padd, osi_pa_to_ma s osi pt_padd = Some pt_addr) ->
      madd_of_vadd (smem_mm (st_memory s)) pt_addr stealth_va = Some ma ->
      page_of_vadd_madd_smem (st_memory s) C.stealth_va ma = Some pg ->
      is_rw_page pg /\ is_cacheable pg.


  Definition no_stealth_page_line (s:state) :=
    forall va pt_addr ma,
      madd_of_vadd (smem_mm (st_memory s)) pt_addr va =
        Some ma ->
      C.get_line va = C.get_line C.stealth_va ->
      va = C.stealth_va.


  Definition valid_sm_state (s:state) :=
    valid_state s /\ stealth_page_cached s /\
    no_stealth_page s /\ stealth_page_reserved_line s /\
    stealth_page_unique s /\ stealth_page_rw_cacheable s /\
    no_stealth_page_line s.

End VALID_SM_STATE.
