(***********************************************************************
 * File: Tlb.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 *************************************************************************
 * This file specifies a VIPT Tlb
 * 
 *************************************************************************)

Require Import Dict_bound Addressing Page.

Module Type TLB.
(* Page is a module *)
  Parameter tlb : Set.
  
  Definition tlb_key : Set := vadd.
  Axiom eq_tlb_key : forall (k k':tlb_key), {k = k'} + {k <> k'}.

  Parameter get_tlb : tlb -> tlb_key -> option madd.
  Parameter add_tlb : tlb -> tlb_key -> madd -> tlb*option (tlb_key*madd).
  Parameter remove_tlb : tlb -> tlb_key -> tlb.
  Parameter empty_tlb : tlb.
  Parameter is_empty_tlb : tlb -> Prop.
  Parameter is_full_tlb : tlb -> Prop.

  Axiom is_empty_tlb_empty : is_empty_tlb empty_tlb.
  Axiom not_empty_add_tlb : forall (t:tlb) (k:tlb_key) (ma:madd),
    ~ is_empty_tlb (fst (add_tlb t k ma)).

  Axiom is_full_tlb_dec : forall (t:tlb), {is_full_tlb t}+{~(is_full_tlb t)}.

  Axiom get_empty_tlb : forall (k:tlb_key),
    get_tlb empty_tlb k = None.
  Axiom get_add_tlb : forall (t:tlb) (k:tlb_key) (ma:madd),
    Some ma = get_tlb (fst (add_tlb t k ma)) k.

  Axiom get_add_tlb_distinct : forall (t:tlb) (k k':tlb_key) (ma ma':madd),
    let (t',entry) := add_tlb t k ma in
     k <> k' -> (entry = None \/ (exists k'', exists ma'', entry = Some (k'',ma'') /\ k' <> k'')) ->
      Some ma' = get_tlb t k' ->
      Some ma' = get_tlb t' k'.

  Axiom get_add_tlb_distinct_not_new : forall (t:tlb) (k k':tlb_key) (ma ma':madd),
    let (t',entry) := add_tlb t k ma in
      k <> k' -> 
      Some ma' = get_tlb t' k' ->
      Some ma' = get_tlb t k'.

  Axiom add_tlb_entry_behavior_not_full : forall (t:tlb) (k:tlb_key) (ma:madd),
    let (c',entry) := add_tlb t k ma in
      ~ is_full_tlb t -> entry = None.

  Axiom add_tlb_entry_behavior_update : forall (t:tlb) (k:tlb_key) (ma:madd),
  let (c',entry) := add_tlb t k ma in
    get_tlb t k <> None -> entry = None.

  Axiom add_tlb_entry_behavior_full : forall (t:tlb) (k:tlb_key) (ma:madd),
    let (c',entry) := add_tlb t k ma in
      is_full_tlb t -> get_tlb t k = None ->
      exists k', exists ma',
        entry = Some (k',ma') /\
        get_tlb t k' = Some ma' /\
        get_tlb c' k' = None.

  Axiom get_remove_tlb : forall (t:tlb) (k:tlb_key),
    get_tlb (remove_tlb t k) k = None.

  Axiom get_remove_tlb_distinct : forall (t:tlb) (k k':tlb_key) (ma:madd),
    k <> k' -> get_tlb t k = Some ma -> get_tlb (remove_tlb t k') k = Some ma.

  Axiom get_remove_tlb_old : forall (t:tlb) (k k':tlb_key) (ma:madd),
    get_tlb (remove_tlb t k') k = Some ma -> get_tlb t k = Some ma.
End TLB.

Module TLB_OPS : TLB.
(* Madd is a module *)
  Declare Module D : DICT_BOUND
    with Definition key := vadd
    with Definition info := madd.

  Definition tlb := D.dict.

  Definition tlb_key : Set := vadd.
  Definition eq_tlb_key := vadd_eq.

  Definition get_tlb := D.get.
  Definition add_tlb := D.set.
  Definition remove_tlb := D.remove.
  Definition empty_tlb := D.empty.
  Definition is_empty_tlb := D.is_empty.
  Definition is_full_tlb := D.is_full.
  Definition is_full_tlb_dec := D.is_full_dec.

  Definition is_empty_tlb_empty := D.is_empty_empty.
  Definition not_empty_add_tlb := D.not_empty_set.
  Definition get_empty_tlb := D.get_empty.
  Definition get_add_tlb := D.get_set.
  Definition get_add_tlb_distinct := D.get_set_distinct.
  Definition get_add_tlb_distinct_not_new := D.get_set_distinct_not_new.
  Definition add_tlb_entry_behavior_not_full := D.set_entry_behavior_not_full.
  Definition add_tlb_entry_behavior_update := D.set_entry_behavior_update.
  Definition add_tlb_entry_behavior_full := D.set_entry_behavior_full.
  Definition get_remove_tlb := D.get_remove.
  Definition get_remove_tlb_distinct := D.get_remove_distinct.
  Definition get_remove_tlb_old := D.get_remove_old.
End TLB_OPS.
