(***********************************************************************
 * File: Write_Isolation.v
 * $Revision: 12365 $
 * $Date: 2014-03-12 18:50:32 -0300 (Wed, 12 Mar 2014) $
 * $LastChangedBy: jdcampo $
 ***********************************************************************
 * This files proves write isolation property on valid executions. 
 **********************************************************************)
Require Import State State_valid State_tactics.


Module WRITE_ISOLATION (C:CACHE PG) (SMEM:SYSTEM_MEM PG C T MM)
  (ST:STATE C SMEM) (ACT:ACTIONS C SMEM ST) (ST_VALID:ST_VALID C SMEM ST).

  Import PG.
  Import C.
  Import T.
  Import SMEM.
  Import ST.
  Import ST_VALID.
  
  Module STAC := STAC C SMEM ST ACT ST_VALID.  
   
  Definition write_isolation (s s':state) (a:action) : Prop :=
    valid_state s ->
    ~(hyper_running s) ->
    ACT.valid_exec s a s' ->     
    forall (va:vadd) (pt_padd:padd) (ma pt_addr:madd) (pg:page) (osi:os_ident), 
      osi_pa_to_ma s osi pt_padd = Some pt_addr ->
      madd_of_vadd (smem_mm (st_memory s)) pt_addr va = Some ma ->
      page_of_vadd_madd_smem (st_memory s) va ma = Some pg ->
      ((page_owned_by pg = Os_PO osi /\ osi <> (st_active_os s)) \/
        page_owned_by pg = Hyp_PO) ->
      page_of_vadd_madd_smem (st_memory s') va ma = Some pg.


  Theorem write_isolation_proof :
    forall s s' a, write_isolation s s' a.
  Proof.
    unfold write_isolation, aos_page_owner; intros;
      STAC.action_induction a; STAC.destruct_action_sem;
      crush.

    (* Read action *)
    destruct H8.
    unfold page_of_vadd_smem in H1; crush.
    rewrite <- STAC.read_same_mem with s va ma v m pt_addr pt_addr0 x x0; crush.
    unfold aos_current_page_table, osi_current_page_table, osi_pa_to_ma in *; crush.
    exists (st_active_os s); crush.
    exists (os_current_page i0); crush.

    (* Write *)
    STAC.destruct_valid_state.
    STAC.get_aos_data.
    STAC.get_aos_current_pt_page.

    STAC.get_aos_current_pt_page_f.

    (* Obtain the madd corresponding to v *)
    STAC.simpl_aos_page_of_vadd_smem.
    unfold page_of_vadd_smem in *; crush.
  
    assert (ma <> m).
    intro; subst.
    assert (page_owned_by x = Os_PO (st_active_os s)).
    erewrite STAC.valid_state_aos_page_owner; crush.
    unfold aos_page_of_vadd_smem, page_of_vadd_smem; crush.
    generalize (STAC.page_same_owner_cache_mem s v va m x pg); crush.
    rewrite H15 in H12; crush; clear H15.
    destruct H5; crush.
    unfold valid_state; crush.

    unfold page_of_vadd_madd_smem.
    unfold is_rw_page in H8; destruct H8.
    unfold page_of_vadd_madd_smem in H4.
    rewrite set_new_cache with (st_memory s) pt_addr0 v m x x0 v0; crush;
      try (unfold page_of_vadd_smem; crush).

  (* Cache, Cache *)
    generalize (get_add_cache_distinct_not_new (smem_c (st_memory s))
      (v,m) (va,ma) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p); crush.
    rewrite <- H4 in Heqval; crush.

  (* Cache, Memory *)
    STAC.destruct_add_cache (smem_c (st_memory s)).
    (* Cache add returned an entry *)
    case (eq_cache_key (v1,m0) (va,ma)); crush.
    (* The entry is (va,ma) *)
    apply set_cache_entry1 with va m (Page (RW_c (Some v0)) (page_owned_by x)
      (page_cacheable x)); crush.
    rewrite STAC.cache_deleted_entry_was_there with (smem_c (st_memory s))
      v m (va,ma) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p
        in Heqval; crush.
    (* The entry is not (va,ma) *)
    generalize (get_add_cache_distinct (smem_c (st_memory s))
      (v,m) (va,ma) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) pg); crush.
    rewrite <- H4 in Heqval0; crush.
    (* Cache add returned no entry *)
    generalize (get_add_cache_distinct (smem_c (st_memory s))
      (v,m) (va,ma) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) pg); crush.
    rewrite <- H4 in Heqval0; crush.

  (* Memory, Cache *)
    (* Absurd: (va,ma) was not the written page *)
    generalize (get_add_cache_distinct_not_new (smem_c (st_memory s))
      (v,m) (va,ma) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p); crush.
    rewrite <- H15 in Heqval; crush.

  (* Memory, Memory *)
    STAC.destruct_add_cache (smem_c (st_memory s)).
    (* Cache add returned an entry *)
    case (eq_cache_key (v1,m0) (va,ma)); crush.
    (* The entry is (va,ma) *)
    rewrite STAC.cache_deleted_entry_was_there with (smem_c (st_memory s))
      v m (va,ma) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p in Heqval; crush.
    (* The entry is not (va,ma) *)
    generalize (set_mem_unchanged1 (st_memory s) pt_addr0 v v1 m m0 ma v0
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p); crush.
    rewrite H15; crush.
    
    (* Alias between the entry (in cache) and (va,ma) *)
    clear H15.
    assert (va <> v1).
    intro; subst; apply n; reflexivity.
    generalize (STAC.cache_deleted_entry_was_there (smem_c (st_memory s))
      v m (v1,m0) (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x)) p); crush.
    assert (Some (v1,m0,p)=Some (v1,m0,p)); try reflexivity.
    apply H16 in H17; clear H16.
    STAC.synonym_cached.

    rewrite  (set_mem_unchanged2 (st_memory s) pt_addr0 v m ma v0
      (Page (RW_c (Some v0)) (page_owned_by x) (page_cacheable x))); crush.
  Qed.

End WRITE_ISOLATION.
