#ifdef little_endian   /* Eg: Intel */
#include <dos.h>
#include <graphics.h>
#include <io.h>
#endif

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

#ifdef little_endian   /* Eg: Intel */
#include <alloc.h>
#endif

#include <ctype.h>

#ifdef little_endian   /* Eg: Intel */
#include <dir.h>
#include <bios.h>
#endif

#ifdef big_endian
#include <Types.h>
#endif

#include "blowfish.h"
#include "test.h"

#define N               16
#define noErr            0
#define DATAERROR         -1
#define KEYBYTES         8
#define subkeyfilename   "blowfish.dat"

unsigned long P[N + 2];
unsigned long S[4][256];
FILE*         SubkeyFile;

// ADDED FOR THE STEALTHMEM ANALYSIS (BEGIN)
static FILE* fopen_mine(char* s1, char* s2)
{
  return NULL; 
}
static short fread_mine(unsigned long* target, int x, int y, FILE* f)
{
  return 0; 
}
// ADDED FOR THE STEALTHMEM ANALYSIS (END)

short opensubkeyfile(void) /* read only */
{
  short error;

  error = noErr;

  if((SubkeyFile = fopen_mine(subkeyfilename,"rb")) == NULL) {
    error = DATAERROR;
  }
 
  return error;
}

unsigned long F(unsigned long x)
{
  unsigned short a;
  unsigned short b;
  unsigned short c;
  unsigned short d;
  unsigned long  y;

  d = x & 0x00FF;
  x >>= 8;
  c = x & 0x00FF;
  x >>= 8;
  b = x & 0x00FF;
  x >>= 8;
  a = x & 0x00FF;
  //y = ((S[0][a] + S[1][b]) ^ S[2][c]) + S[3][d];
  y = S[0][a] + S[1][b];
  y = y ^ S[2][c];
  y = y + S[3][d];

  return y;
}

void Blowfish_encipher(unsigned long *xl, unsigned long *xr)
{
  unsigned long  Xl;
  unsigned long  Xr;
  unsigned long  temp;
  short          i;

  Xl = *xl;
  Xr = *xr;

  for (i = 0; i < N; ++i) {
    Xl = Xl ^ P[i];
    Xr = F(Xl) ^ Xr;

    temp = Xl;
    Xl = Xr;
    Xr = temp;
  }

  temp = Xl;
  Xl = Xr;
  Xr = temp;

  Xr = Xr ^ P[N];
  Xl = Xl ^ P[N + 1];
  
  *xl = Xl;
  *xr = Xr;
}

void Blowfish_decipher(unsigned long *xl, unsigned long *xr)
{
  unsigned long  Xl;
  unsigned long  Xr;
  unsigned long  temp;
  short          i;

  Xl = *xl;
  Xr = *xr;

  for (i = N + 1; i > 1; --i) {
    Xl = Xl ^ P[i];
    Xr = F(Xl) ^ Xr;

    /* Exchange Xl and Xr */
    temp = Xl;
    Xl = Xr;
    Xr = temp;
  }

  /* Exchange Xl and Xr */
  temp = Xl;
  Xl = Xr;
  Xr = temp;

  Xr = Xr ^ P[1];
  Xl = Xl ^ P[0];

  *xl = Xl;
  *xr = Xr;
}

short InitializeBlowfish(char key[], short keybytes)
{
  short          i;
  short          j;
  short          k;
  short          error;
  short          numread;
  unsigned long  data;
  unsigned long  datal;
  unsigned long  datar;

  /* First, open the file containing the array initialization data */
  error = opensubkeyfile();
  if (error == noErr) {
    for (i = 0; i < N + 2; ++i) {
      numread = fread_mine(&data, 4, 1, SubkeyFile);
#ifdef little_endian      /* Eg: Intel   We want to process things in byte   */
      /*   order, not as rearranged in a longword          */
      data = ((data & 0xFF000000) >> 24) |
	((data & 0x00FF0000) >>  8) |
	((data & 0x0000FF00) <<  8) |
	((data & 0x000000FF) << 24);
#endif

      if (numread != 1) {
	return DATAERROR;
      } else {
	P[i] = data;
      }
    }

    for (i = 0; i < 4; ++i) {
      for (j = 0; j < 256; ++j) {
	numread = fread(&data, 4, 1, SubkeyFile);

#ifdef little_endian      /* Eg: Intel   We want to process things in byte   */
	/*   order, not as rearranged in a longword          */
	data = ((data & 0xFF000000) >> 24) |
	  ((data & 0x00FF0000) >>  8) |
	  ((data & 0x0000FF00) <<  8) |
	  ((data & 0x000000FF) << 24);
#endif

	if (numread != 1) {
	  return DATAERROR;
	} else {
	  S[i][j] = data;
	}
      }
    }

    fclose(SubkeyFile);

    j = 0;
    for (i = 0; i < N + 2; ++i) {
      data = 0x00000000;
      for (k = 0; k < 4; ++k) {
	data = (data << 8) | key[j];
	j = j + 1;
	if (j >= keybytes) {
	  j = 0;
	}
      }
      P[i] = P[i] ^ data;
    }

    datal = 0x00000000;
    datar = 0x00000000;

    for (i = 0; i < N + 2; i += 2) {
      Blowfish_encipher(&datal, &datar);

      P[i] = datal;
      P[i + 1] = datar;
    }

    for (i = 0; i < 4; ++i) {
      for (j = 0; j < 256; j += 2) {

	Blowfish_encipher(&datal, &datar);
   
	S[i][j] = datal;
	S[i][j + 1] = datar;
      }
    }
  } else {
    //    printf("Unable to open subkey initialization file : %d\n", error); MODIFIED FOR THE STEALTHMEM ANALYSIS
  }

  return error;
}



void
print_variable_key (int i)
{
  int j;
  for (j = 0; j < 8; j++)
    printf ("%.2X", variable_key[i][j] & 0xff);
}

int
main (int argc, char *argv[])
{
  short error;
  int i;

  //  printf ("Key bytes         Clear bytes       Cypher bytes      Result\n"); MODIFIED FOR THE STEALTHMEM ANALYSIS
  for (i = 0; i < NUM_VARIABLE_KEY_TESTS; i++)
    {
      error = InitializeBlowfish(variable_key[i], 8);
      if (error != 0)
	{
	  //	  fprintf (stderr, "Error %d\n", error); MODIFIED FOR THE STEALTHMEM ANALYSIS
	  return error; // exit (error); MODIFIED FOR THE STEALTHMEM ANALYSIS
	}

      unsigned long plain_l = plaintext_l[i];
      unsigned long plain_r = plaintext_r[i];
      Blowfish_encipher(&plain_l, &plain_r);

      //      print_variable_key (i);
      //      printf ("  %.8X%.8X  %.8X%.8X  %.8X%.8X\n",
      //	      plaintext_l[i], plaintext_r[i],
      //	      ciphertext_l[i], ciphertext_r[i],
      //	      plain_l, plain_r);
    }
      //  printf ("\n"); MODIFIED FOR THE STEALTHMEM ANALYSIS

  return 0;
}
