/* snow-2.0.c */

/* 
 * SNOW 2.0.  
 *
 * ECRYPT wrapper around SNOW 2.0 code written by Patrik Ekdahl &
 * Thomas Johansson.
 * see http://www.it.lth.se/cryptology/snow/
 *
 * Author: Christophe De Canni\`ere, K.U.Leuven.
 */

/* ------------------------------------------------------------------------- */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "snow.h"
#include "snow2tab.h"

#define ECRYPT_KEYSIZE 128
#define ECRYPT_IVSIZE  128
#define ECRYPT_BLOCKLENGTH 64
#define LONG_TEST_STREAM_SIZEB 0x20000
#define TEST_CHUNK 64

#define NONZEROSIZE(s) (((s) <= 0) ? 4 : (s))

#define MAXKEYSIZEB NONZEROSIZE((ECRYPT_KEYSIZE + 7) / 8)
#define MAXIVSIZEB NONZEROSIZE((ECRYPT_IVSIZE + 7) / 8)

/* ECRYPT_ctx struct */
typedef struct {
  u32 s15, s14, s13, s12, s11, s10, s9, s8, s7, s6, s5, s4, s3, s2, s1, s0;
  u32 r1, r2;
} CTX;

u32 keysize;
u8 key[32];
CTX ctx;

u8 iv[MAXIVSIZEB];
u8 plaintext[LONG_TEST_STREAM_SIZEB];
u8 ciphertext[LONG_TEST_STREAM_SIZEB];
u8 xored[TEST_CHUNK];
int msglen;
int vector;


#define byte(n,w)   (((w)>>(n*8)) & 0xff)
#define ainv_mul(w) (((w)>>8)^(snow_alphainv_mul[w&0xff]))
#define a_mul(w)    (((w)<<8)^(snow_alpha_mul[w>>24]))

void snow_loadkey_fast(u32 IV3,u32 IV2,u32 IV1,u32 IV0)
{
 int i;
 
    ctx.s15=(((u32)*(key+0))<<24) | (((u32)*(key+1))<<16) |
      (((u32)*(key+2))<<8) | (((u32)*(key+3)));
    ctx.s14=(((u32)*(key+4))<<24) | (((u32)*(key+5))<<16) |
      (((u32)*(key+6))<<8) | (((u32)*(key+7)));
    ctx.s13=(((u32)*(key+8))<<24) | (((u32)*(key+9))<<16) |
      (((u32)*(key+10))<<8) | (((u32)*(key+11)));
    ctx.s12=(((u32)*(key+12))<<24) | (((u32)*(key+13))<<16) |
      (((u32)*(key+14))<<8) | (((u32)*(key+15)));
    ctx.s11 =~ctx.s15; /* bitwise inverse */
    ctx.s10 =~ctx.s14;
    ctx.s9  =~ctx.s13;
    ctx.s8  =~ctx.s12;
    ctx.s7  = ctx.s15; /* just copy */ 
    ctx.s6  = ctx.s14;
    ctx.s5  = ctx.s13;
    ctx.s4  = ctx.s12;
    ctx.s3  =~ctx.s15; /* bitwise inverse */
    ctx.s2  =~ctx.s14;
    ctx.s1  =~ctx.s13;
    ctx.s0  =~ctx.s12;
 
  /* XOR IV values */
    ctx.s15^=IV0;
    ctx.s12^=IV1;
    ctx.s10^=IV2;
    ctx.s9^=IV3;
 

  ctx.r1=0;
  ctx.r2=0;

  /* Do 32 initial clockings */
 for(i=0;i<2;i++)
    {
      u32 outfrom_fsm, fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s15 )^ctx.r2;
      ctx.s0 =a_mul(ctx.s0 )^ ctx.s2 ^ainv_mul(ctx.s11 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s5 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s0 )^ctx.r2;
      ctx.s1 =a_mul(ctx.s1 )^ ctx.s3 ^ainv_mul(ctx.s12 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s6 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s1 )^ctx.r2;
      ctx.s2 =a_mul(ctx.s2 )^ ctx.s4 ^ainv_mul(ctx.s13 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s7 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s2 )^ctx.r2;
      ctx.s3 =a_mul(ctx.s3 )^ ctx.s5 ^ainv_mul(ctx.s14 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s8 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s3 )^ctx.r2;
      ctx.s4 =a_mul(ctx.s4 )^ ctx.s6 ^ainv_mul(ctx.s15 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s9 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s4 )^ctx.r2;
      ctx.s5 =a_mul(ctx.s5 )^ ctx.s7 ^ainv_mul(ctx.s0 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s10 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s5 )^ctx.r2;
      ctx.s6 =a_mul(ctx.s6 )^ ctx.s8 ^ainv_mul(ctx.s1 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s11 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s6 )^ctx.r2;
      ctx.s7 =a_mul(ctx.s7 )^ ctx.s9 ^ainv_mul(ctx.s2 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s12 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s7 )^ctx.r2;
      ctx.s8 =a_mul(ctx.s8 )^ ctx.s10 ^ainv_mul(ctx.s3 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s13 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s8 )^ctx.r2;
      ctx.s9 =a_mul(ctx.s9 )^ ctx.s11 ^ainv_mul(ctx.s4 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s14 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s9 )^ctx.r2;
      ctx.s10 =a_mul(ctx.s10 )^ ctx.s12 ^ainv_mul(ctx.s5 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s15 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s10 )^ctx.r2;
      ctx.s11 =a_mul(ctx.s11 )^ ctx.s13 ^ainv_mul(ctx.s6 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s0 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s11 )^ctx.r2;
      ctx.s12 =a_mul(ctx.s12 )^ ctx.s14 ^ainv_mul(ctx.s7 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s1 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s12 )^ctx.r2;
      ctx.s13 =a_mul(ctx.s13 )^ ctx.s15 ^ainv_mul(ctx.s8 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s2 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s13 )^ctx.r2;
      ctx.s14 =a_mul(ctx.s14 )^ ctx.s0 ^ainv_mul(ctx.s9 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s3 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;

      outfrom_fsm=(ctx.r1+ ctx.s14 )^ctx.r2;
      ctx.s15 =a_mul(ctx.s15 )^ ctx.s1 ^ainv_mul(ctx.s10 )^outfrom_fsm;
      fsmtmp=ctx.r2+ ctx.s4 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
    }
}

/*
 * Function: snow_keystream_fast
 * 
 * Synopsis:
 *   Clocks the cipher 16 times and returns 16 words of keystream symbols
 *   in keystream_block.
 * 
 * Returns: void
 * 
 */

void snow_keystream_fast(u32 *keystream_block) 
{
      u32 fsmtmp;

      ctx.s0 =a_mul(ctx.s0 )^ ctx.s2 ^ainv_mul(ctx.s11 );
      fsmtmp=ctx.r2+ ctx.s5 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[0]=(ctx.r1+ ctx.s0 )^ctx.r2^ ctx.s1 ;

      ctx.s1 =a_mul(ctx.s1 )^ ctx.s3 ^ainv_mul(ctx.s12 );
      fsmtmp=ctx.r2+ ctx.s6 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[1]=(ctx.r1+ ctx.s1 )^ctx.r2^ ctx.s2 ;

      ctx.s2 =a_mul(ctx.s2 )^ ctx.s4 ^ainv_mul(ctx.s13 );
      fsmtmp=ctx.r2+ ctx.s7 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[2]=(ctx.r1+ ctx.s2 )^ctx.r2^ ctx.s3 ;

      ctx.s3 =a_mul(ctx.s3 )^ ctx.s5 ^ainv_mul(ctx.s14 );
      fsmtmp=ctx.r2+ ctx.s8 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[3]=(ctx.r1+ ctx.s3 )^ctx.r2^ ctx.s4 ;

      ctx.s4 =a_mul(ctx.s4 )^ ctx.s6 ^ainv_mul(ctx.s15 );
      fsmtmp=ctx.r2+ ctx.s9 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[4]=(ctx.r1+ ctx.s4 )^ctx.r2^ ctx.s5 ;

      ctx.s5 =a_mul(ctx.s5 )^ ctx.s7 ^ainv_mul(ctx.s0 );
      fsmtmp=ctx.r2+ ctx.s10 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[5]=(ctx.r1+ ctx.s5 )^ctx.r2^ ctx.s6 ;

      ctx.s6 =a_mul(ctx.s6 )^ ctx.s8 ^ainv_mul(ctx.s1 );
      fsmtmp=ctx.r2+ ctx.s11 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[6]=(ctx.r1+ ctx.s6 )^ctx.r2^ ctx.s7 ;

      ctx.s7 =a_mul(ctx.s7 )^ ctx.s9 ^ainv_mul(ctx.s2 );
      fsmtmp=ctx.r2+ ctx.s12 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[7]=(ctx.r1+ ctx.s7 )^ctx.r2^ ctx.s8 ;

      ctx.s8 =a_mul(ctx.s8 )^ ctx.s10 ^ainv_mul(ctx.s3 );
      fsmtmp=ctx.r2+ ctx.s13 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp; 
      keystream_block[8]=(ctx.r1+ ctx.s8 )^ctx.r2^ ctx.s9 ;

      ctx.s9 =a_mul(ctx.s9 )^ ctx.s11 ^ainv_mul(ctx.s4 );
      fsmtmp=ctx.r2+ ctx.s14 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[9]=(ctx.r1+ ctx.s9 )^ctx.r2^ ctx.s10 ;

      ctx.s10 =a_mul(ctx.s10 )^ ctx.s12 ^ainv_mul(ctx.s5 );
      fsmtmp=ctx.r2+ ctx.s15 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[10]=(ctx.r1+ ctx.s10 )^ctx.r2^ ctx.s11 ;

      ctx.s11 =a_mul(ctx.s11 )^ ctx.s13 ^ainv_mul(ctx.s6 );
      fsmtmp=ctx.r2+ ctx.s0 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;    
      keystream_block[11]=(ctx.r1+ ctx.s11 )^ctx.r2^ ctx.s12 ;

      ctx.s12 =a_mul(ctx.s12 )^ ctx.s14 ^ainv_mul(ctx.s7 );
      fsmtmp=ctx.r2+ ctx.s1 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;
      keystream_block[12]=(ctx.r1+ ctx.s12 )^ctx.r2^ ctx.s13 ;

      ctx.s13 =a_mul(ctx.s13 )^ ctx.s15 ^ainv_mul(ctx.s8 );
      fsmtmp=ctx.r2+ ctx.s2 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;   
      keystream_block[13]=(ctx.r1+ ctx.s13 )^ctx.r2^ ctx.s14 ;

      ctx.s14 =a_mul(ctx.s14 )^ ctx.s0 ^ainv_mul(ctx.s9 );
      fsmtmp=ctx.r2+ ctx.s3 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp; 
      keystream_block[14]=(ctx.r1+ ctx.s14 )^ctx.r2^ ctx.s15 ;

      ctx.s15 =a_mul(ctx.s15 )^ ctx.s1 ^ainv_mul(ctx.s10 );
      fsmtmp=ctx.r2+ ctx.s4 ;
      ctx.r2=snow_T0[byte(0,ctx.r1)]^snow_T1[byte(1,ctx.r1)]^snow_T2[byte(2,ctx.r1)]^snow_T3[byte(3,ctx.r1)];
      ctx.r1=fsmtmp;    
      keystream_block[15]=(ctx.r1+ ctx.s15 )^ctx.r2^ ctx.s0 ;
}

void ECRYPT_ivsetup()
{ 
  snow_loadkey_fast(
    U8TO32_LITTLE(iv),
    U8TO32_LITTLE(iv + 4),
    U8TO32_LITTLE(iv + 8),
    U8TO32_LITTLE(iv + 12));
}

u32 keystream[16];

void ECRYPT_process_bytes(
  const u8* input, 
  u8* output, 
  u32 msglen)
{ 
  u32 i;

  for ( ; msglen >= 64; msglen -= 64, input += 64, output += 64)
    {
      snow_keystream_fast(keystream);

      for (i = 0; i < 16; ++i)
	((u32*)output)[i] = ((u32*)input)[i] ^ U32TO32_LITTLE(keystream[i]);
    }

  if (msglen > 0)
    {
      snow_keystream_fast(keystream);

      for (i = 0; i < msglen; i ++)
	output[i] = input[i] ^ ((u8*)keystream)[i];
    }
}

/* ------------------------------------------------------------------------- */

void print_data(const char *str, const u8 *val, int len)
{
  //  int i;

  //  static const char hex[] = "0123456789ABCDEF";

  //  printf("%28s = ", str); MODIFIED FOR THE STEALTHMEM ANALYSIS

  //  for (i = 0; i < len; i++)
  //    {
      //      if (i > 0 && (i & 0xF) == 0 && (len > 24))
	//        printf("\n%28s   ", ""); MODIFIED FOR THE STEALTHMEM ANALYSIS

	//      putc(hex[(val[i] >> 4) & 0xF], stdout); MODIFIED FOR THE STEALTHMEM ANALYSIS
	//      putc(hex[(val[i]     ) & 0xF], stdout); MODIFIED FOR THE STEALTHMEM ANALYSIS
  //    }

  //  printf("\n"); MODIFIED FOR THE STEALTHMEM ANALYSIS
}

void print_chunk(const char *str, const u8 *val, int start, int len)
{
  char indexed[80];

  //  sprintf(indexed, "%s[%d..%d]", str, start, start + len - 1); MODIFIED FOR THE STEALTHMEM ANALYSIS
  print_data(indexed, val + start, len);
}

// ADDED FOR THE STEALTHMEM ANALYSIS (BEGIN)
static void memset_uchar(unsigned char* tab, int val, int size)
{
  int i;
  for (i=0; i < size; i++ ) *(tab+i) = val;
}
// ADDED FOR THE STEALTHMEM ANALYSIS (END)

void xor_digest(const u8 *stream, int size, u8 *out, int outsize)
{
  int i;
  memset_uchar(out, 0, outsize);
  for (i = 0; i < size; i++)
    out[i % outsize] ^= stream[i];
}

void print_primitive()
{
  //  printf("\n");
  //  printf("Primitive Name: SNOW-2.0\n");
  //  printf("================%.*s\n", (int)strlen("SNOW-2.0"),
  //  	  "==========================================");
  //  printf("Profile: bench\n");
  //  printf("Key size: %d bits\n", ECRYPT_KEYSIZE);
  //  printf("IV size: %d bits\n", ECRYPT_IVSIZE);
  //  printf("\n"); MODIFIED FOR THE STEALTHMEM ANALYSIS
}

void print_stream()
{
  const int chunk = TEST_CHUNK;
  
  print_data("key", key, (ECRYPT_KEYSIZE + 7) / 8); 
  print_data("IV", iv, (ECRYPT_IVSIZE + 7) / 8);    
  
  print_chunk("stream", ciphertext, 0, chunk);
  print_chunk("stream", ciphertext, msglen/2-chunk, chunk);
  print_chunk("stream", ciphertext, msglen/2, chunk);
  print_chunk("stream", ciphertext, msglen-chunk, chunk);
      
  xor_digest(ciphertext, msglen, xored, chunk);
  print_data("xor-digest", xored, chunk);  
}

/* ------------------------------------------------------------------------- */

void encrypt_and_check()
{
  unsigned int i;

  memset_uchar(ciphertext, 0, sizeof(ciphertext));
  ECRYPT_ivsetup();
  ECRYPT_process_bytes(plaintext, ciphertext, msglen);

  print_stream();
  //  printf("\n");
}

int
main ()
{
  int i, v;

  print_primitive();

  memset_uchar(plaintext, 0, sizeof(plaintext));
  memset_uchar(ciphertext, 0, sizeof(ciphertext));

  /* check key stream */

  msglen = LONG_TEST_STREAM_SIZEB;

  //  printf("Test vectors -- set 6\n");
  //  printf("=====================\n\n");

  for (v = 0; v < 4; v++)
  {
    for (i = 0; i < sizeof(key); i++)
      key[i] = (i * 0x53 + v * 5) & 0xFF;

    for (i = 0; i < sizeof(iv); i++)
      iv[i] = (i * 0x67 + v * 9 + 13) & 0xFF;

    //    printf("Set %d, vector#%3d:\n", 6, vector = v);
    encrypt_and_check();
  }

  //  printf("\n\nEnd of test vectors\n");
  //  fflush(stdout);

  return 0;
}
