#ifndef SNOW_H
#define SNOW_H

#include <limits.h>

#define I8T char
#define U8C(v) (v##U)
#define I16T short
#define U16C(v) (v##U)
#define I32T int
#define U32C(v) (v##U)
#define I64T long
#define U64C(v) (v##UL)

#define MAXT long


typedef signed I8T s8;
typedef unsigned I8T u8;
typedef signed I16T s16;
typedef unsigned I16T u16;
typedef signed I32T s32;
typedef unsigned I32T u32;
typedef signed I64T s64;
typedef unsigned I64T u64;

#define U8V(v) ((u8)(v) & U8C(0xFF))
#define U16V(v) ((u16)(v) & U16C(0xFFFF))
#define U32V(v) ((u32)(v) & U32C(0xFFFFFFFF))
#define U64V(v) ((u64)(v) & U64C(0xFFFFFFFFFFFFFFFF))

#define U16TO16_LITTLE(v) (v)
#define U32TO32_LITTLE(v) (v)
#define U64TO64_LITTLE(v) (v)

#define U8TO16_LITTLE(p) \
  (((u16)((p)[0])      ) | \
   ((u16)((p)[1]) <<  8))

#define U8TO32_LITTLE(p) \
  (((u32)((p)[0])      ) | \
   ((u32)((p)[1]) <<  8) | \
   ((u32)((p)[2]) << 16) | \
   ((u32)((p)[3]) << 24))

#define U8TO64_LITTLE(p) \
  (((u64)((p)[0])      ) | \
   ((u64)((p)[1]) <<  8) | \
   ((u64)((p)[2]) << 16) | \
   ((u64)((p)[3]) << 24) | \
   ((u64)((p)[4]) << 32) | \
   ((u64)((p)[5]) << 40) | \
   ((u64)((p)[6]) << 48) | \
   ((u64)((p)[7]) << 56))

#define U16TO8_LITTLE(p, v) \
    (p)[0] = U8V((v)      ); \
    (p)[1] = U8V((v) >>  8);

#define U32TO8_LITTLE(p, v) \
    (p)[0] = U8V((v)      ); \
    (p)[1] = U8V((v) >>  8); \
    (p)[2] = U8V((v) >> 16); \
    (p)[3] = U8V((v) >> 24);

#define U64TO8_LITTLE(p, v) \
    (p)[0] = U8V((v)      ); \
    (p)[1] = U8V((v) >>  8); \
    (p)[2] = U8V((v) >> 16); \
    (p)[3] = U8V((v) >> 24); \
    (p)[4] = U8V((v) >> 32); \
    (p)[5] = U8V((v) >> 40); \
    (p)[6] = U8V((v) >> 48); \
    (p)[7] = U8V((v) >> 56);

#endif
