#include <stdint.h>
#include <stdio.h>

/* Taken from http://en.wikipedia.org/wiki/Tiny_Encryption_Algorithm */
 
void
encrypt (uint32_t* v, uint32_t* k)
{
  uint32_t v0=v[0], v1=v[1], sum=0, i;           /* set up */
  uint32_t delta=0x9e3779b9;                     /* a key schedule constant */
  uint32_t k0=k[0], k1=k[1], k2=k[2], k3=k[3];   /* cache key */
  for (i=0; i < 32; i++) {                       /* basic cycle start */
    sum += delta;
    v0 += ((v1<<4) + k0) ^ (v1 + sum) ^ ((v1>>5) + k1);
    v1 += ((v0<<4) + k2) ^ (v0 + sum) ^ ((v0>>5) + k3);  
  }                                              /* end cycle */
  v[0]=v0; v[1]=v1;
}
 
void
decrypt (uint32_t* v, uint32_t* k)
{
  uint32_t v0=v[0], v1=v[1], sum=0xC6EF3720, i;  /* set up */
  uint32_t delta=0x9e3779b9;                     /* a key schedule constant */
  uint32_t k0=k[0], k1=k[1], k2=k[2], k3=k[3];   /* cache key */
  for (i=0; i<32; i++) {                         /* basic cycle start */
    v1 -= ((v0<<4) + k2) ^ (v0 + sum) ^ ((v0>>5) + k3);
    v0 -= ((v1<<4) + k0) ^ (v1 + sum) ^ ((v1>>5) + k1);
    sum -= delta;                                   
  }                                              /* end cycle */
  v[0]=v0; v[1]=v1;
}

uint32_t pz[1024];

/* Test case take from http://www.cix.co.uk/~klockstone/teavect.htm */
int
main (int argc, char *argv[])
{
  uint32_t n;

 repeat:

  for (n = 1; n < 65; n++)
    {
      encrypt (pz + n, pz + n + 2);

      if (n == (n & -n))                         /* if n power of 2 */
	//        printf ("\n%3d %8x %8x %8x %8x  %8x %8x",
	//	       n, pz[n], pz[n + 1], pz[n + 2], 
	//	       pz[n + 3], pz[n + 4], pz[n + 5]); MODIFIED FOR THE STEALTHMEM ANALYSIS

        pz[n + 6] = pz[n];
    }

    for (n = 64; n > 0; n--)
    {
      pz[n] = pz[n + 6];
      decrypt (pz + n, pz + n + 2);
    }

    n = 1;

    //    printf ("\n%3d %8x %8x %8x %8x  %8x %8x\n",
    //            n, pz[n], pz[n + 1], pz[n + 2], 
    //            pz[n + 3], pz[n + 4], pz[n + 5]); MODIFIED FOR THE STEALTHMEM ANALYSIS

    return 0;
}

